Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import List.
Require Import Setoid.
Require Import Morphisms.


Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomSet.
Require Import DomMap.

Require Import Syntax.

Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.

Require Import SemAbsBlock.
Require Import SemAccSen.
Require Import SemAccInsen.

Require Import SemSenLocal.
Require Import SemInsenLocal.

Set Implicit Arguments.

Section AccMemorySpecs.
  Variables (prog:progT) (acc:accT) (state:stateT)
    (Hacc: SemAccInsen.acc_correct prog acc state).

  Lemma memory_update_mrestrict_join_commutative_subset :
    forall (tlocs: Set! [locT]) (locs: \powerset locT)
           (Hloc: le (set_to_powerset tlocs) locs)
           (val:valueT) (mem mem':memT),
      mrestrict_join (E:=valueT) locs
                     (memory_update acc (Elt tlocs) val mem) mem'
      ≡
      memory_update acc (Elt tlocs) val
      (mrestrict_join (E:=valueT) locs mem mem').
  Proof.
    move=> tlocs locs Hloc val [mem|] [mem'|]; try done.
    apply mem_eq_spec; try done.
    move=> loc.
    rewrite memory_update_spec mem_mrestrict_join_spec.
    rewrite {1}/memory_update.
    rewrite {1}/mrestrict_join/mrestrict/memT_restricting_maplike
            /withtop_restricting_maplike.
    case Htlocsin: (loc !\in tlocs).
    { case Hlocsin: (loc !\in locs).
      { rewrite memory_update_spec Htlocsin.
        case (_ && _); first by exact:eq_refl.
        rewrite [join (Elt _) (Elt _)]/join/join_/memT_hasJoin/withtop_hasJoin
                /withtop_join.
        rewrite mem_mrestrict_join_spec Hlocsin.
        exact:eq_refl.
      }
      { elimtype False.
        rewrite/le/le_/morphism_poset/le/le_/bool_poset in Hloc.
        pose (Hloc loc) as Hlocin.
        simpl in Hlocsin.
        rewrite -> Hlocsin in Hlocin.
        rewrite/set_to_powerset in Hlocin.
        simpl in Hlocin.
        simpl in Htlocsin.
        rewrite Htlocsin in Hlocin.
        move/implyP in Hlocin.
        inversion Hlocin.
      }
    }
    { case Hlocsin: (loc !\in locs).
      { rewrite [join (Elt _) (Elt _)]/join/join_/memT_hasJoin/withtop_hasJoin
                /withtop_join.
        rewrite mem_mrestrict_join_spec Hlocsin.
        rewrite memory_update_spec Htlocsin.
        exact:eq_refl.
      }
      { rewrite [join (Elt _) (Elt _)]/join/join_/memT_hasJoin/withtop_hasJoin
                /withtop_join.
        rewrite mem_mrestrict_join_spec Hlocsin.
        exact:eq_refl.
      }
    }
  Qed.

  Lemma memory_lookup_mrestrict_join_spec :
    forall locs rlocs mem mem'
           (Hsubset : SetInterface.subset locs rlocs),
      memory_lookup (Elt locs)
                    (mrestrict_join (set_to_powerset rlocs)
                                    (Elt mem) (Elt mem'))
      ≡
      memory_lookup (Elt locs) (Elt mem).
  Proof.
    intros.
    apply le_antisym.
    { apply memory_lookup_lub=> loc Hloc.
      erewrite <- memory_lookup_ub; last by eauto.
      dom_tac; rewrite/memT_hasJoin/withtop_hasJoin/withtop_join.
      rewrite/mlookup/memT_maplike/withtop_maplike.
      dom_tac.
      move: (Hloc)=> Hloc'.
      rewrite/= in Hloc'.
      apply SetFacts.mem_iff in Hloc'.
      apply SetFacts.subset_iff in Hsubset.
      apply Hsubset in Hloc'.
      apply SetFacts.mem_iff in Hloc'.
      rewrite/smem/set_setlike Hloc' /negb.
      rewrite join_bot.
      exact/DomType.eq_le/eq_refl.
    }
    { apply memory_lookup_lub=> loc Hloc.
      erewrite <- memory_lookup_ub; last by eauto.
      dom_tac; rewrite/memT_hasJoin/withtop_hasJoin/withtop_join.
      rewrite/mlookup/memT_maplike/withtop_maplike.
      dom_tac.
      move: (Hloc)=> Hloc'.
      rewrite/= in Hloc'.
      apply SetFacts.mem_iff in Hloc'.
      apply SetFacts.subset_iff in Hsubset.
      apply Hsubset in Hloc'.
      apply SetFacts.mem_iff in Hloc'.
      rewrite/smem/set_setlike Hloc' /negb.
      rewrite join_bot.
      exact/DomType.eq_le/eq_refl.
    }
  Qed.      

  Lemma memory_lookup_mrestrict_join_valid_1 :
    forall fid le mem mem' locs
           (Hargs : acc_correct_acclocs_libcall_args acc fid le mem)
           (Hlocs : locs = locs_of
                             (sem_expr fid (hd (expr_const lit_bot) le) mem)),
      memory_lookup locs (mrestrict_join
                            (set_to_powerset (coerce (acclocs acc) [fid])) mem
                            (Elt mem'))
      ≡
      memory_lookup locs mem.
  Proof.
    intros; subst.
    case: le Hargs => [|hd tl] Hargs /=.
    { rewrite/bot/bot_/set_hasBot.
      repeat rewrite SetProperties.fold_1b; try apply SetInterface.empty_1.
      reflexivity.
    }
    { simpl in Hargs; move/andP in Hargs;
      inversion Hargs as [Hhd Htl]; clear Hargs Htl.
      case: (locs_of (sem_expr fid hd mem)) Hhd => [locs|] Hhd;
      last by reflexivity.
      case: mem => [mem|].
      { apply memory_lookup_mrestrict_join_spec; done. }
      { simpl; reflexivity. }
    }
  Qed.

  Lemma memory_lookup_mrestrict_join_valid_2 :
    forall fid le mem mem' locs
           (Hargs : acc_correct_acclocs_libcall_args acc fid le mem)
           (Hlocs : locs = locs_of
                             (sem_expr fid (hd (expr_const lit_bot)
                                               (tl le)) mem)),
      memory_lookup locs (mrestrict_join
                            (set_to_powerset (coerce (acclocs acc) [fid])) mem
                            (Elt mem'))
      ≡
      memory_lookup locs mem.
  Proof.
    intros; subst.
    case: le Hargs => [|hd [|thd ttl]] Hargs /=.
    { rewrite/bot/bot_/set_hasBot.
      repeat rewrite SetProperties.fold_1b; try apply SetInterface.empty_1.
      reflexivity.
    }
    { rewrite/bot/bot_/set_hasBot.
      repeat rewrite SetProperties.fold_1b; try apply SetInterface.empty_1.
      reflexivity.
    }
    { simpl in Hargs.
      move/andP in Hargs; inversion Hargs as [Hhd Htl]; clear Hargs Hhd.
      move/andP in Htl; inversion Htl as [Hhd Httl]; clear Htl Httl.
      case: (locs_of (sem_expr fid thd mem)) Hhd => [locs|] Hhd;
      last by reflexivity.
      case: mem => [mem|].
      { apply memory_lookup_mrestrict_join_spec; done. }
      { simpl; reflexivity. }
    }
  Qed.
End AccMemorySpecs.
