Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import List.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomSet.
Require Import DomMap.

Require Import Syntax.

Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomPidxSen.
Require Import DomPidxInsen.

Require Import SemAbsBlock.
Require Import SemAccSen.
Require Import SemAccInsen.

Require Import SemSenLocal.
Require Import SemInsenLocal.

Require Import TrsSen.

Require Import CorAccExpr.
Require Import CorConPruning.

Set Implicit Arguments.

Section Accs.
  Variables (prog:progT) (acc:accT) (state:stateT)
    (Hacc: SemAccInsen.acc_correct prog acc state).

  Definition tstate' := tstate prog state.

  Inductive subexp (sub_e:exprT) : exprT -> Prop :=
  | subexp_eq : subexp sub_e sub_e
  | subexp_op2_l : forall e o e' (Hsub:subexp sub_e e), 
    subexp sub_e (expr_op2 o e e')
  | subexp_op2_r : forall e o e' (Hsub:subexp sub_e e), 
    subexp sub_e (expr_op2 o e' e)
  | subexp_op1 : forall e o (Hsub:subexp sub_e e), 
    subexp sub_e (expr_op1 o e)
  | subexp_value_of : forall l (Hsub:subexp_lv sub_e l), 
    subexp sub_e (expr_value_of l)
  | subexp_addr_of : forall l (Hsub:subexp_lv sub_e l), 
    subexp sub_e (expr_addr_of l)
  | subexp_ctype : forall e ct (Hsub:subexp sub_e e), 
    subexp sub_e (expr_cast ct e)
  | subexp_size_of : forall e (Hsub:subexp sub_e e), 
    subexp sub_e (expr_size_of e)
  with subexp_lv (sub_e:exprT) : lv -> Prop :=
  | subexp_deref : forall e (Hsub:subexp sub_e e),
    subexp_lv sub_e (lv_deref e)
  | subexp_array_l : forall e e' (Hsub:subexp sub_e e),
    subexp_lv sub_e (lv_array e e')
  | subexp_array_r : forall e e' (Hsub:subexp sub_e e),
    subexp_lv sub_e (lv_array e' e)
  | subexp_struct : forall lv x (Hsub:subexp_lv sub_e lv),
    subexp_lv sub_e (lv_struct lv x).

  Lemma tacc_correct_acclocs_expr_strong : forall (fid : fidT) (mem : memT) 
    (e sub_e:exprT)
    (Hsub: subexp sub_e e)
    (Hexp: acc_correct_acclocs_expr acc fid e mem),
    SemAccSen.acc_correct_acclocs_expr acc fid sub_e mem
    with tacc_correct_acclocs_lv_strong : forall (fid : fidT) (mem : memT) 
      (l:lv) (sub_e:exprT)
      (Hsub: subexp_lv sub_e l)
      (Hexp: acc_correct_acclocs_lv acc fid l mem),
      SemAccSen.acc_correct_acclocs_expr acc fid sub_e mem.
  Proof.
    induction 1; intros.
    { apply tacc_correct_acclocs_expr. auto. }
    { simpl in Hexp. move/andP:Hexp=>[He He']. apply IHHsub. assumption. }
    { simpl in Hexp. move/andP:Hexp=>[He He']. apply IHHsub. assumption. }
    { simpl in Hexp. apply IHHsub. assumption. }
    { simpl in Hexp. move/andP:Hexp=>[Hl _]. 
      apply tacc_correct_acclocs_lv_strong with l; assumption. }
    { simpl in Hexp. apply tacc_correct_acclocs_lv_strong with l; assumption. }
    { simpl in Hexp. apply IHHsub. assumption. }
    { simpl in Hexp. apply IHHsub. assumption. }
    induction 1; intros.
    { simpl in Hexp. 
      apply tacc_correct_acclocs_expr_strong with e; assumption. }
    { simpl in Hexp. move/andP:Hexp=>[He He']. 
      apply tacc_correct_acclocs_expr_strong with e; assumption. }
    { simpl in Hexp. move/andP:Hexp=>[He He']. 
      apply tacc_correct_acclocs_expr_strong with e; assumption. }
    { simpl in Hexp.
      by apply IHHsub.
    }
  Qed.

  Lemma remove_uop_imply_subexp_n : forall m n bef o e1 e2
    (Hnm:n <= m) (Hsize:size_of_expr bef n)
    (Hrm: remove_uop bef = Some (expr_op2 o e1 e2)),
    subexp e1 bef /\ subexp e2 bef.
  Proof.
    induction m; intros.
    { assert (n = 0); first omega. subst. inversion Hsize; subst. 
      { simpl in Hrm. inversion Hrm. }
      { simpl in Hrm. inversion Hrm. }
      { simpl in Hrm. inversion Hrm. }
    }
    { destruct bef.
      { simpl in Hrm. inversion Hrm. }
      { simpl in Hrm. destruct op; try (inversion Hrm; fail).
        { inversion Hrm; subst. 
          split; [apply subexp_op2_l|apply subexp_op2_r]; apply subexp_eq. }
        { inversion Hrm; subst. 
          split; [apply subexp_op2_l|apply subexp_op2_r]; apply subexp_eq. }
        { inversion Hrm; subst. 
          split; [apply subexp_op2_l|apply subexp_op2_r]; apply subexp_eq. }
        { inversion Hrm; subst. 
          split; [apply subexp_op2_l|apply subexp_op2_r]; apply subexp_eq. }
        { inversion Hrm; subst. 
          split; [apply subexp_op2_l|apply subexp_op2_r]; apply subexp_eq. }
        { inversion Hrm; subst. 
          split; [apply subexp_op2_l|apply subexp_op2_r]; apply subexp_eq. }
      }      
      { simpl in Hrm. destruct op; try (inversion Hrm; fail).
        destruct bef; try (inversion Hrm; fail).
        { destruct op; try (inversion Hrm; fail).
          { inversion Hrm; subst. split; apply subexp_op1; [apply subexp_op2_l|apply subexp_op2_r]; apply subexp_eq. }
          { inversion Hrm; subst. split; apply subexp_op1; [apply subexp_op2_l|apply subexp_op2_r]; apply subexp_eq. }
          { inversion Hrm; subst. split; apply subexp_op1; [apply subexp_op2_l|apply subexp_op2_r]; apply subexp_eq. }
          { inversion Hrm; subst. split; apply subexp_op1; [apply subexp_op2_l|apply subexp_op2_r]; apply subexp_eq. }
          { inversion Hrm; subst. split; apply subexp_op1; [apply subexp_op2_l|apply subexp_op2_r]; apply subexp_eq. }
          { inversion Hrm; subst. split; apply subexp_op1; [apply subexp_op2_l|apply subexp_op2_r]; apply subexp_eq. } 
        }
        { destruct op; try (inversion Hrm; fail).
          inversion Hsize; subst. inversion H2; subst.
          assert (subexp e1 bef ∧ subexp e2 bef).
          apply IHm with n o; auto. omega. destruct H.
          split; apply subexp_op1; apply subexp_op1; assumption. }
      }  
      { simpl in Hrm. inversion Hrm. }
      { simpl in Hrm. inversion Hrm. }
      { simpl in Hrm. inversion Hrm. }
      { simpl in Hrm. inversion Hrm. }
    }
  Qed.

  Lemma remove_uop_imply_subexp : forall bef o e1 e2
    (Hrm: remove_uop bef = Some (expr_op2 o e1 e2)),
    subexp e1 bef /\ subexp e2 bef.
  Proof.
    intros. 
    apply remove_uop_imply_subexp_n with 
      (size_of_expr_f bef) (size_of_expr_f bef) o; auto.
    apply size_of_expr_f_correct.
  Qed.

  Lemma subexp_preserve_acc_cor : forall fid sub_e e mem
    (Hsub: subexp sub_e e)
    (He: acc_correct_acclocs_expr acc fid e mem),
    acc_correct_acclocs_expr acc fid sub_e mem
    with subexp_preserve_acc_cor_lv : forall fid sub_e l mem
      (Hsub: subexp_lv sub_e l)
      (He: acc_correct_acclocs_lv acc fid l mem),
      acc_correct_acclocs_expr acc fid sub_e mem.
  Proof.
    induction 1; subst; intros.
    { apply He. }
    { simpl in He. move/andP:He=>[He He']. apply IHHsub. assumption. }
    { simpl in He. move/andP:He=>[He' He]. apply IHHsub. assumption. }
    { simpl in He. apply IHHsub. assumption. }
    { simpl in He. move/andP:He=>[He _]. 
      apply subexp_preserve_acc_cor_lv with l; assumption. }
    { simpl in He. apply subexp_preserve_acc_cor_lv with l; assumption. }
    { simpl in He. apply IHHsub. assumption. }
    { simpl in He. apply IHHsub. assumption. }
    induction 1; subst; intros.
    { simpl in He. apply subexp_preserve_acc_cor with e; assumption. }
    { simpl in He. move/andP:He=>[He He']. 
      apply subexp_preserve_acc_cor with e; assumption. }
    { simpl in He. move/andP:He=>[He He']. 
      apply subexp_preserve_acc_cor with e; assumption. }
    { simpl in He.
      apply IHHsub.
      by apply He.
    }
  Qed.  

  Lemma tacc_correct_acclocs_prune_bop : forall 
    (fid : fidT) (mem : memT) (test : exprT) (mem' : Map! [locT, valueT])
    (Hinstr : acc_correct_acclocs_expr acc fid test mem)
    (Htest : 
      sem_expr fid test 
      (mrestrict_join (E:=valueT) 
        (set_to_powerset (coerce (acclocs acc) [fid]))
        mem (Elt mem')) 
      ≡ sem_expr fid test mem)
    (e : exprT)
    (Heqo : remove_uop test = Some e),
    prune_bop acc fid e
    (mrestrict_join (E:=valueT) 
      (set_to_powerset (coerce (acclocs acc) [fid]))
      mem (Elt mem')) 
    ⊑ mrestrict_join (E:=valueT) 
    (set_to_powerset (coerce (acclocs acc) [fid]))
    (prune_bop acc fid e mem) (Elt mem').
  Proof.
    intros. unfold prune_bop.
    destruct e; try (apply DomType.eq_le; reflexivity). 
    destruct e1; try (apply DomType.eq_le; reflexivity). 
    destruct l; try (apply DomType.eq_le; reflexivity). 
    destruct (is_strong acc (var_loc fid x nil)); 
      try (apply DomType.eq_le; reflexivity). 
    apply remove_uop_imply_subexp in Heqo. destruct Heqo.
    assert (subexp (expr_value_of (lv_id x)) test); auto.
    apply (tacc_correct_acclocs_expr_strong fid mem) in H; auto.
    apply (tacc_correct_acclocs_expr_strong fid mem) in H0; auto.
    unfold SemAccSen.acc_correct_acclocs_expr in H.
    unfold SemAccSen.acc_correct_acclocs_expr in H0.
    unfold mlookup,map_maplike in H. 
    unfold sem_expr in H. unfold sem_expr_lv.
    rewrite -> (H0 mem').

    unfold adjust. rewrite -> (H mem').
    unfold adjust_val.
    destruct (memory_lookup
                (Elt (SetInterface.singleton (var_loc fid x nil))) mem).
    destruct p.
    destruct (sem_expr fid e2 mem).
    destruct p.

    rewrite memory_update_mrestrict_join_commutative.
    apply DomType.eq_le. reflexivity.

    apply (subexp_preserve_acc_cor fid mem) in H1; auto.
    simpl in H1.
    apply SetFacts.subset_iff in H1.
    apply SetFacts.mem_iff.
    apply H1.
    by apply SetFacts.singleton_iff.
  Qed.

  Lemma tacc_correct_acclocs_prune
    (fid : fidT)
    (stack : list (fidT * bidT))
    (mem : memT)
    (Hc : pidx_consistent prog fid stack)
    (test : exprT)
    (Hinstr : acc_correct_acclocs_instr acc fid (instr_assume test) mem) :
    SemAccSen.acc_correct_acclocs_instr acc fid (instr_assume test) mem.
  Proof.
    rewrite/SemAccSen.acc_correct_acclocs_instr/sem_instr.
    move=> [mem'|]; last first.
    { rewrite Top_top ? mem_mrestrict_join_top_rhs.
      exact: top_prop.
    }
    rewrite/acc_correct_acclocs_instr in Hinstr.
    exploit tacc_correct_acclocs_expr; eauto.
    instantiate (1 := mem').
    move=> Htest.
    
    unfold memory_prune at 1.
    simpl in *.
    remember (leb
         (itv_of
            (sem_expr fid test
               (mrestrict_join (set_to_powerset (coerce (acclocs acc) [fid]))
                  mem (Elt mem'))))
         (Some
            (DomItv.itv'_z_z 0 0
               (memory_prune_obligation_1 fid test
                  (mrestrict_join
                     (set_to_powerset (coerce (acclocs acc) [fid])) mem
                     (Elt mem')))))). destruct b.
    { apply bot_prop. }

    unfold memory_prune.
    remember (leb (itv_of (sem_expr fid test mem))
             (Some
                (DomItv.itv'_z_z 0 0 (memory_prune_obligation_1 fid test mem)))).
    destruct b.
    { elimtype False.
      symmetry in Heqb. symmetry in Heqb0.
      move/lebP:Heqb=> Heqb. move/lebP:Heqb0=> Heqb0.
      elim Heqb. clear Heqb.
      rewrite (itv_of_mor Htest).
      rewrite <- Heqb0.
      apply DomType.eq_le.
      reflexivity. }

    remember (remove_uop test). destruct o; last first.
    { apply DomType.eq_le. reflexivity. }
    { clear Heqb Heqb0.
      apply tacc_correct_acclocs_prune_bop with test; auto. }
  Qed.
End Accs.
