Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.

Require Import LibTactics.
Require Import DomBridge.
Require Import DomType.
Require Import DomBasic.
Require Import DomMap.

Require Import Syntax.

Require Import DomZ.
Require Import DomSet.
Require Import DomConValue.
Require Import DomConMem.

Require Import SemCon.
Require Import DomConValue.
Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomPidx.
Require Import DomPidxSen.

Require Import SemAbsBlock.
Require Import SemAccSen.
Require Import SemSen.

Require Import CorBasic.
Require Import CorAbs.
Require Import CorItv.

Require Import Containers.SetProperties.
Require Import Containers.MapInterface.

Set Implicit Arguments.

(* Abstract Relations *)
(* For intervals, use Z_itvT_abs *)

Obligation Tactic := Simpless.tactic.

Section Correctness.
Variable (prog:progT) (acc:accT).
  
  Global Program Instance loc_abs : abs DomConValue.locT locsT := {
    abs_relation l ls := if ls is Elt ls' then (t_loc l) !\in ls' else true
  }.
  Next Obligation.
    move=> x y Hxy [z|] [w|] Hzw // Hin.
    usual_tac Hxy.
    rewrite/le/= in Hzw.
    apply SetFacts.mem_iff.
    apply Hzw.
    by apply SetFacts.mem_iff.
  Qed.

  Global Program Instance value_abs : abs DomConValue.valueT valueT := {
    abs_relation cv av := 
    match cv with
      | inl (inl z) => z \abs (itv_of av)
      | inl (inr l) => l \abs (locs_of av)
      | inr (a,u,oss) => oss \abs (mlookup (Elt:=DomAbsValue.ossT) a (arrv_of av))
    end
  }.
  Next Obligation.
    move=> a b Hab [[cz cl] ca] [[dz dl] da] [[Hz Hl] Ha].
    rewrite/fst/snd in Hz Hl Ha.
    move: a b Hab=> [[a|a]|a] [[b|b]|b] Hab;
      try (inversion Hab; fail);
      (do ? rewrite/eq/= in Hab); subst.
    { move=> Habs.
      by eapply abs_relation_proper; eauto; reflexivity.
    }
    { move=> Habs.
      by eapply abs_relation_proper; eauto; reflexivity.
    }
    { move: a b Hab=> [[aa au] ao] [[ba bu] bo] [[Hal Hu] Ho].
      rewrite/fst/snd in Hal Hu Ho.
      move=> Habs.
      eapply abs_relation_proper; eauto.
      rewrite Hal.
      apply mlookup_le_mor; first done.
      apply arrv_of_le_mor.
      by do ? split.
    }
  Qed.

  Global Program Instance memory_abs : abs DomConMem.memT memT := {
    abs_relation conMem absMem :=
      if absMem is Elt absMem'
      then forall loc,
             (mlookup (Key:=DomConValue.locT) (Elt:=option DomConValue.valueT) loc conMem) \abs absMem[t_loc loc]!
      else True
  }.
  Next Obligation.
    move=> x y Hxy [z|] [w|] Hzw //.
    rewrite/le/= in Hzw.
    apply: forall_implies=> loc.
    rewrite <- Hxy.
    rewrite/abs_relation/option_abs.
    case: (x[loc]!)=> [a|]; last done.
    move=> Habs.
    eapply abs_relation_proper; eauto; try reflexivity.
    by apply (Hzw (t_loc loc)).
  Qed.

  Lemma abs_top (m:DomConMem.memT) : m \abs (top:memT).
  Proof. by constructor. Qed.
  
  Lemma abs_value_top (v:DomConValue.valueT) : v \abs (top:valueT).
  Proof.
    move: v=> [[v|v]|v] //=.
    { exact: CorItv.abs_top. }
    { move: v=> [[a u] o].
      by do ? split; apply: CorItv.abs_top.
    }
  Qed.

  Lemma collect_key_in (l:list (idT*ossT)) (a:idT) (dset:Set![locT])
    (Hin:((lheap a nil):locT) !\in dset) :
    ((lheap a nil):locT)
    !\in ((fold_left
    (λ (a0 : Set![locT]) (p : idT * ossT),
           ((if ~~ eqb p.2 bot
           then SetInterface.add (lheap p.1 nil) a0
           else a0):Set![locT])) l dset):Set![locT]).
  Proof.
    move:l prog acc a dset Hin. induction l; intros.
    simpl. rewrite DomSet.set_smem_spec. assumption.
    simpl. apply IHl; auto. 
    rewrite/smem/DomSet.set_setlike.
    remember (~~ eqb a.2 bot). destruct b.
    rewrite SetFacts.add_b. apply/orP. right. apply Hin.
    apply Hin.
  Qed.

  Lemma collect_key_abs : forall (ofs:offsetT) (sz:sizeT) (st:strideT)
    (a:arrvT') (al:allocT)
    (H : (ofs, sz, st) \abs ((Elt a) [al ]! : ossT)) (uq:uniqueT)
    (dummy_set:Set![locT]),
    ((DomConValue.lheap al uq ofs):DomConValue.locT)
    \abs
    ((Elt (fold
        (λ (k : idT) (v : ossT) (s : SetInterface.set locT),
         if ~~ eqb v bot then SetInterface.add (lheap k nil) s else s) a
        dummy_set)):locsT).
  Proof.
    move => ofs sz st a al H uq dset.
    rewrite fold_1.
    rewrite/mlookup/withtop_maplike/mlookup/map_maplike in H.
    rewrite MF.elements_o in H.
    move:H. remember (elements _). 
    rewrite{2}/abs_relation/loc_abs/t_loc.
    move:prog acc ofs sz st al uq dset. clear a Heql.
    induction l; subst; intros.
    { simpl in H. inversion H. inversion H1. }
    { simpl in H. destruct a.
      rewrite/fold_left.
      remember (MapFacts.eqb _ _). destruct b.
      { symmetry in Heqb. 
        move/is_true_iff in Heqb.
        rewrite/MapFacts.eqb in Heqb.
        rewrite is_compare_eqb in Heqb.
        move/eqbP in Heqb. rewrite Heqb.
        rewrite/fst. apply collect_key_in.
        rewrite/smem/DomSet.set_setlike.
        remember (~~ eqb _ _); destruct b.
        { rewrite SetFacts.add_b. apply/orP. left. 
          rewrite is_compare_eqb. apply/eqbP. reflexivity. }
        { symmetry in Heqb0. apply Bool.negb_false_iff in Heqb0.
          move/eqbP:Heqb0 => Heqb0. destruct H. simpl in Heqb0.
          destruct o. destruct Heqb0.
          rewrite/coerce/snd in H0. 
          rewrite/bot/bot_/ossT_hasBot/pair_hasBot_pointwise/snd in H2.
          rewrite/bot/bot_/option_hasBot in H2.
          rewrite/eq/eq_/option_setoid/DomBasic.option_eq in H2.
          rewrite/abs_itvT in H0. destruct i.
          inversion H2. inversion H0.
        }
      }
      { apply IHl with sz st; try assumption. }
    }
  Qed.    

  Lemma cast_sound : forall t z (Hcon_cast : cast_type_in t z)
    abs_val (Habs : (vint z) \abs abs_val),
    (vint z) \abs cast t (abs_val).
  Proof.
    move=> t z Hcast.
    inv Hcast; move=> [[i a] s] Habs;
      rewrite/cast;
      rewrite/=/itv_of/itv_cast/range_of/leb/= in Habs *;
      try exact: abs_itvT'_top;
      by (case Hle: (DomType.leb _ _); first done; constructor; omega).
  Qed.

  Lemma cast_sound_not : forall t z z'
    (Hcon_cast : ¬cast_type_in t z)
    (Hcon_cast_new : cast_type_in t z')
    abs_val (Habs : (vint z) \abs abs_val),
    (vint z') \abs cast t (abs_val).
  Proof.
    move=> t z z' Hcast Hcast'.
    rewrite/vint.
    inv Hcast'; move=> [[i a] s] Habs;
      rewrite/cast/itv_cast/range_of;
      rewrite/abs_relation/value_abs/itv_of/fst in Habs *;
      try exact: abs_itvT'_top.
    { case Hle: (DomType.leb _ _); move/lebP in Hle; last by constructor; omega.
      rewrite -> Hle in Habs; inv Habs.
      by contradict Hcast; constructor; omega.
    }
    { case Hle: (DomType.leb _ _); move/lebP in Hle; last by constructor; omega.
      rewrite -> Hle in Habs; inv Habs.
      by contradict Hcast; constructor; omega.
    }
    { case Hle: (DomType.leb _ _); move/lebP in Hle; last by constructor; omega.
      rewrite -> Hle in Habs; inv Habs.
      by contradict Hcast; constructor; omega.
    }
    { case Hle: (DomType.leb _ _); move/lebP in Hle; last by constructor; omega.
      rewrite -> Hle in Habs; inv Habs.
      by contradict Hcast; constructor; omega.
    }
    { case Hle: (DomType.leb _ _); move/lebP in Hle; last by constructor; omega.
      rewrite -> Hle in Habs; inv Habs.
      by contradict Hcast; constructor; omega.
    }
    { case Hle: (DomType.leb _ _); move/lebP in Hle; last by constructor; omega.
      rewrite -> Hle in Habs; inv Habs.
      by contradict Hcast; constructor; omega.
    }
    { case Hle: (DomType.leb _ _); move/lebP in Hle; last by constructor; omega.
      rewrite -> Hle in Habs; inv Habs.
      by contradict Hcast; constructor; omega.
    }
    { case Hle: (DomType.leb _ _); move/lebP in Hle; last by constructor; omega.
      rewrite -> Hle in Habs; inv Habs.
      by contradict Hcast; constructor; omega.
    }
  Qed.

  Lemma cast_sound_array : forall t st' al uq ofs sz st abs_val
    (Hst : cast_stride_of t st')
    (Habs: (vheap al uq (ofs, sz, st)) \abs abs_val),
    (vheap al uq (Cdiv (ofs * st) st', Cdiv (sz * st) st', st'))
      \abs (cast (typ_pointer t) (abs_val)).
  Proof.
    move => t st' al uq ofs sz st [[i a] s] Hst Habs.
    rewrite/vheap/abs_relation/value_abs/itv_of/fst in Habs *.
    rewrite/cast/arrv_of/fst/snd in Habs *.
    clear -Hst Habs.
    move: s Habs=> [s|] Habs; last first.
    { rewrite/arrv_cast/mlookup.
      by do ? split; rewrite/fst/snd;
        apply: abs_itvT'_top.
    }
    rewrite/mlookup/withtop_maplike/mlookup/map_maplike in Habs.
    inv Hst.
    { rewrite/arrv_cast/ctype_stride_of.
      rewrite/mlookup/withtop_maplike/mlookup/map_maplike.
      rewrite MF.map_o /MF.option_map.
      case: (s[al]) Habs=> [oss|] Habs; last done.
      move: oss Habs=> [[ofs' sz'] st'] [[Hofs Hsz] Hst].
      rewrite/coerce/fst/snd in Hofs Hsz Hst.
      rewrite/coerce.
      case Heq: (eqb _ _); move/eqbP in Heq.
      { move: Heq=> [[Hofs' Hsz' Hst']].
        rewrite/fst/snd in Hofs' Hsz' Hst'.
        by rewrite -> Hst' in Hst.
      }
      do ? split; rewrite/fst/snd.
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by constructor; omega.
    }
    { rewrite/arrv_cast/ctype_stride_of.
      rewrite/mlookup/withtop_maplike/mlookup/map_maplike.
      rewrite MF.map_o /MF.option_map.
      case: (s[al]) Habs=> [oss|] Habs; last done.
      move: oss Habs=> [[ofs' sz'] st'] [[Hofs Hsz] Hst].
      rewrite/coerce/fst/snd in Hofs Hsz Hst.
      rewrite/coerce.
      case Heq: (eqb _ _); move/eqbP in Heq.
      { move: Heq=> [[Hofs' Hsz' Hst']].
        rewrite/fst/snd in Hofs' Hsz' Hst'.
        by rewrite -> Hst' in Hst.
      }
      do ? split; rewrite/fst/snd.
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by constructor; omega.
    }
    { rewrite/arrv_cast/ctype_stride_of.
      rewrite/mlookup/withtop_maplike/mlookup/map_maplike.
      rewrite MF.map_o /MF.option_map.
      case: (s[al]) Habs=> [oss|] Habs; last done.
      move: oss Habs=> [[ofs' sz'] st'] [[Hofs Hsz] Hst].
      rewrite/coerce/fst/snd in Hofs Hsz Hst.
      rewrite/coerce.
      case Heq: (eqb _ _); move/eqbP in Heq.
      { move: Heq=> [[Hofs' Hsz' Hst']].
        rewrite/fst/snd in Hofs' Hsz' Hst'.
        by rewrite -> Hst' in Hst.
      }
      do ? split; rewrite/fst/snd.
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by constructor; omega.
    }
    { rewrite/arrv_cast/ctype_stride_of.
      rewrite/mlookup/withtop_maplike/mlookup/map_maplike.
      rewrite MF.map_o /MF.option_map.
      case: (s[al]) Habs=> [oss|] Habs; last done.
      move: oss Habs=> [[ofs' sz'] st'] [[Hofs Hsz] Hst].
      rewrite/coerce/fst/snd in Hofs Hsz Hst.
      rewrite/coerce.
      case Heq: (eqb _ _); move/eqbP in Heq.
      { move: Heq=> [[Hofs' Hsz' Hst']].
        rewrite/fst/snd in Hofs' Hsz' Hst'.
        by rewrite -> Hst' in Hst.
      }
      do ? split; rewrite/fst/snd.
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by constructor; omega.
    }
    { rewrite/arrv_cast/ctype_stride_of.
      rewrite/mlookup/withtop_maplike/mlookup/map_maplike.
      rewrite MF.map_o /MF.option_map.
      case: (s[al]) Habs=> [oss|] Habs; last done.
      move: oss Habs=> [[ofs' sz'] st'] [[Hofs Hsz] Hst].
      rewrite/coerce/fst/snd in Hofs Hsz Hst.
      rewrite/coerce.
      case Heq: (eqb _ _); move/eqbP in Heq.
      { move: Heq=> [[Hofs' Hsz' Hst']].
        rewrite/fst/snd in Hofs' Hsz' Hst'.
        by rewrite -> Hst' in Hst.
      }
      do ? split; rewrite/fst/snd.
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by constructor; omega.
    }
    { rewrite/arrv_cast/ctype_stride_of.
      rewrite/mlookup/withtop_maplike/mlookup/map_maplike.
      rewrite MF.map_o /MF.option_map.
      case: (s[al]) Habs=> [oss|] Habs; last done.
      move: oss Habs=> [[ofs' sz'] st'] [[Hofs Hsz] Hst].
      rewrite/coerce/fst/snd in Hofs Hsz Hst.
      rewrite/coerce.
      case Heq: (eqb _ _); move/eqbP in Heq.
      { move: Heq=> [[Hofs' Hsz' Hst']].
        rewrite/fst/snd in Hofs' Hsz' Hst'.
        by rewrite -> Hst' in Hst.
      }
      do ? split; rewrite/fst/snd.
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by constructor; omega.
    }
    { rewrite/arrv_cast/ctype_stride_of.
      rewrite/mlookup/withtop_maplike/mlookup/map_maplike.
      rewrite MF.map_o /MF.option_map.
      case: (s[al]) Habs=> [oss|] Habs; last done.
      move: oss Habs=> [[ofs' sz'] st'] [[Hofs Hsz] Hst].
      rewrite/coerce/fst/snd in Hofs Hsz Hst.
      rewrite/coerce.
      case Heq: (eqb _ _); move/eqbP in Heq.
      { move: Heq=> [[Hofs' Hsz' Hst']].
        rewrite/fst/snd in Hofs' Hsz' Hst'.
        by rewrite -> Hst' in Hst.
      }
      do ? split; rewrite/fst/snd.
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by constructor; omega.
    }
    { rewrite/arrv_cast/ctype_stride_of.
      rewrite/mlookup/withtop_maplike/mlookup/map_maplike.
      rewrite MF.map_o /MF.option_map.
      case: (s[al]) Habs=> [oss|] Habs; last done.
      move: oss Habs=> [[ofs' sz'] st'] [[Hofs Hsz] Hst].
      rewrite/coerce/fst/snd in Hofs Hsz Hst.
      rewrite/coerce.
      case Heq: (eqb _ _); move/eqbP in Heq.
      { move: Heq=> [[Hofs' Hsz' Hst']].
        rewrite/fst/snd in Hofs' Hsz' Hst'.
        by rewrite -> Hst' in Hst.
      }
      do ? split; rewrite/fst/snd.
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by constructor; omega.
    }
    { rewrite/arrv_cast/ctype_stride_of.
      rewrite/mlookup/withtop_maplike/mlookup/map_maplike.
      rewrite MF.map_o /MF.option_map.
      case: (s[al]) Habs=> [oss|] Habs; last done.
      move: oss Habs=> [[ofs' sz'] st'] [[Hofs Hsz] Hst].
      rewrite/coerce/fst/snd in Hofs Hsz Hst.
      rewrite/coerce.
      case Heq: (eqb _ _); move/eqbP in Heq.
      { move: Heq=> [[Hofs' Hsz' Hst']].
        rewrite/fst/snd in Hofs' Hsz' Hst'.
        by rewrite -> Hst' in Hst.
      }
      do ? split; rewrite/fst/snd.
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by constructor; omega.
    }
    { rewrite/arrv_cast/ctype_stride_of.
      rewrite/mlookup/withtop_maplike/mlookup/map_maplike.
      rewrite MF.map_o /MF.option_map.
      case: (s[al]) Habs=> [oss|] Habs; last done.
      move: oss Habs=> [[ofs' sz'] st'] [[Hofs Hsz] Hst].
      rewrite/coerce/fst/snd in Hofs Hsz Hst.
      rewrite/coerce.
      case Heq: (eqb _ _); move/eqbP in Heq.
      { move: Heq=> [[Hofs' Hsz' Hst']].
        rewrite/fst/snd in Hofs' Hsz' Hst'.
        by rewrite -> Hst' in Hst.
      }
      do ? split; rewrite/fst/snd.
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by apply op_div_proper2; auto;
        [apply op_mul_proper2; auto|constructor; omega].
      by constructor; omega.
    }
  Qed.

  Lemma sound_memory_lookup2_ : forall (l:list locT) (ab ac:valueT) (am:memT)
    (Horder:ab ⊑ ac),
    ab ⊑
    fold_left (λ (a0 : valueT) (e : locT), ((am) [e ]!:valueT) ⊔ a0) l ac.
  Proof.
    induction l; intros. simpl. assumption.
    replace (fold_left (λ (a0 : valueT) (e : locT), (am) [e ]! ⊔ a0) (a :: l) ac)
      with (fold_left (λ (a0 : valueT) (e : locT), 
        (am) [e ]! ⊔ a0) l ((am) [a ]! ⊔ ac)); last reflexivity.
    apply IHl. apply DomType.le_trans with ac. assumption. apply join_rhs_.
  Qed.

  Lemma sound_memory_lookup_ : forall (pidx:pidxT) lst (m:DomConMem.memT) 
    (am:memT) (Hm : m \abs am)
    lval v (Hlookup : (m) [lval ]! = Some v)
    (H0 : existsb (λ y : locT, eqb (t_loc lval) y) lst) 
    acc_val,
    v \abs fold_left (λ (a : valueT) (e : locT), (am) [e ]! ⊔ a) lst acc_val.
  Proof.
    (* sound_memory_lookup_ *)
    induction lst; intros.
    { inversion H0. }
    { replace (fold_left (λ (a0 : valueT) (e : locT), (am) [e ]! ⊔ a0) 
      (a :: lst) acc_val) with
      (fold_left (λ (a0 : valueT) (e : locT), (am) [e ]! ⊔ a0) 
      lst ((am) [a ]! ⊔ acc_val)); last reflexivity.
      inversion H0. move/orP:H1 => H1. destruct H1.
      move/eqbP:H => H.
      assert (v \abs ((am) [a ]! ⊔ acc_val)).
      {
        assert (v \abs (am) [a ]!).
        { 
          rewrite/abs_relation/memory_abs in Hm.
          destruct am. 

          rewrite/abs_relation/option_abs in Hm.
          assert (match (m) [lval ]! with
                    | Some a => a \abs (Elt d) [t_loc lval ]!
                    | None => True
                  end). apply Hm.
          rewrite Hlookup in H1.
          move:H1. apply abs_relation_eq_mor_Proper.
          reflexivity. rewrite H. reflexivity.

          move: v Hlookup=> [[v|v]|v] Hlookup //=.
          exact: abs_itvT'_top.
          move: v Hlookup=> [[al u] o] Hlookup.
          by do ? split; apply: abs_itvT'_top.
        }
        move:H1. apply abs_relation_proper; try reflexivity.
        apply join_lhs_.
      }
      move:H1. apply abs_relation_proper; try reflexivity.
      apply sound_memory_lookup2_. apply DomType.eq_le. reflexivity.
      apply IHlst with m lval; assumption.
    }
  Qed.

  Lemma arrv_op_add_spec
        (z:Z) (i:DomItv.itvT) (Hz:z \abs i)
        ofs sz st
        (a:arrvT) (al:allocT)
        (Hoss: ((ofs, sz, st):DomConValue.ossT) \abs (a[al]!:ossT)) :
    (((ofs+z)%Z, sz, st):DomConValue.ossT)
      \abs
      ((arrv_op_add a i)[al]!:ossT).
  Proof.
    destruct i; last by inversion Hz.
    move: a Hoss=> [a|] Hoss; last by do ? split; apply: abs_itvT'_top.
    rewrite/arrv_op_add.
    rewrite/mlookup/withtop_maplike/mlookup/map_maplike in Hoss *.
    rewrite MF.map_o.
    move: a[al] Hoss=> [[[ofs' sz'] st']|] Hoss; last done.
    rewrite/MF.option_map/coerce in Hoss *.
    move: Hoss=> [[Hofs Hsz] Hst].
    do ? split; rewrite/fst/snd; try done.
    exact: op_add_proper2.
  Qed.

  Lemma arrv_op_sub_spec
        (z:Z) (i:DomItv.itvT) (Hz:z \abs i)
        ofs sz st
        (a:arrvT) (al:allocT)
        (Hoss: ((ofs, sz, st):DomConValue.ossT) \abs (a[al]!:ossT)) :
    (((ofs-z)%Z, sz, st):DomConValue.ossT)
      \abs
      ((arrv_op_sub a i)[al]!:ossT).
  Proof.
    destruct i; last by inversion Hz.
    move: a Hoss=> [a|] Hoss; last by do ? split; apply: abs_itvT'_top.
    rewrite/arrv_op_sub.
    rewrite/mlookup/withtop_maplike/mlookup/map_maplike in Hoss *.
    rewrite MF.map_o.
    move: a[al] Hoss=> [[[ofs' sz'] st']|] Hoss; last done.
    rewrite/MF.option_map/coerce in Hoss *.
    move: Hoss=> [[Hofs Hsz] Hst].
    do ? split; rewrite/fst/snd; try done.
    exact: op_sub_proper2.
  Qed.

  Lemma sem_expr_proper2_ : forall (pidx : pidxT) (e : exprT)
    (cm : DomConMem.memT) 
    (v : DomConValue.valueT) (Hv : SemCon.sem_expr pidx e cm v)
    (am : memT) (Hm : cm \abs am),
    v \abs sem_expr (pifid pidx) e am
  with sem_expr_lv_proper2_ : forall (pidx : pidxT) (l : lv)
    (m : DomConMem.memT) (am : memT) (Hmem : m \abs am)
    (con_loc : DomConValue.locT) (Hsem_lv : sem_lv pidx l m con_loc),
    con_loc \abs sem_expr_lv (pifid pidx) l am.
  Proof.
    (* sem_expr *)
    intros pidx e cm v Hv.
    induction Hv; intros; rewrite/vint/vheap/vloc.

    { inv Hlit; rewrite/sem_expr/=.
      { by constructor; omega. }
      { by constructor. }
      move: v=> [[v|v]|[[a u] o]]; [exact: abs_itvT'_top|done|].
      by do ? split; apply: abs_itvT'_top.
    }      
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_add_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      rewrite/sem_expr-/sem_expr.
      move: (sem_expr _ _ _) (sem_expr _ _ _) IHHv1 IHHv2.
      move=> [[i1 a1] s1] [[i2 a2] s2].
      rewrite/vop2.
      rewrite/abs_relation/value_abs/itv_of/arrv_of/fst/snd.
      move=> Habs Habs'.
      rewrite/vint/vheap in Habs Habs'.
      eapply abs_relation_proper; last by eapply arrv_op_add_spec; eauto.
      reflexivity.
      move: (arrv_op_add _ _) (arrv_op_add _ _)=> [l|] [r|];
      try exact: top_prop.
      rewrite/= ? map_mlookup_spec map_join_spec.
      apply join_rhs.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      rewrite/sem_expr-/sem_expr.
      move: (sem_expr _ _ _) (sem_expr _ _ _) IHHv1 IHHv2.
      move=> [[i1 a1] s1] [[i2 a2] s2].
      rewrite/vop2.
      rewrite/abs_relation/value_abs/itv_of/arrv_of/fst/snd.
      move=> Habs Habs'.
      rewrite/vint/vheap in Habs Habs'.
      eapply abs_relation_proper; last by eapply arrv_op_add_spec; eauto.
      reflexivity.
      move: (arrv_op_add _ _) (arrv_op_add _ _)=> [l|] [r|]; try exact: top_prop.
      rewrite/= ? map_mlookup_spec map_join_spec.
      apply join_lhs.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_sub_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      rewrite/sem_expr-/sem_expr.
      move: (sem_expr _ _ _) (sem_expr _ _ _) IHHv1 IHHv2.
      move=> [[i1 a1] s1] [[i2 a2] s2].
      rewrite/vop2.
      rewrite/abs_relation/value_abs/itv_of/arrv_of/fst/snd.
      move=> Habs Habs'.
      by eapply arrv_op_sub_spec; eauto.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_mul_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_div_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_mod_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_lt_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_le_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_gt_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_ge_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_eq_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_neq_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_and_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_or_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_band_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_bor_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_bxor_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_lshift_proper2.
    }
    { exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_rshift_proper2.
    }
    { exploit IHHv; eauto=> {IHHv} IHHv.
      by apply op_not_proper2.
    }
    { exploit IHHv; eauto=> {IHHv} IHHv.
      by apply op_bnot_proper2.
    }
    { replace (sem_expr (pifid pidx) (expr_value_of l) am) with
        (memory_lookup (sem_expr_lv (pifid pidx) l am) am);
        last reflexivity.
      rewrite/memory_lookup.
      assert (lval \abs sem_expr_lv (pifid pidx) l am).
      apply sem_expr_lv_proper2_ with m; assumption.
      remember (sem_expr_lv (pifid pidx) l am). destruct l0;
      last apply abs_value_top.
      clear Hlv Heql0.
      assert (SetInterface.mem (t_loc lval) s); first apply H.
        
      rewrite SetInterface.fold_1.
      rewrite SetFacts.elements_b in H0.
      remember (SetInterface.elements s) as lst.

      apply sound_memory_lookup_ with m lval; try assumption.
      clear Heqlst. move:lst H0.
      { induction lst; intros.
        inversion H0.
        inversion H0. move/orP:H2 => H2. destruct H2.
        simpl. apply/orP. left. 
        rewrite is_compare_eqb in H1. assumption.
        simpl. apply/orP. right.
        apply IHlst. assumption. }
    }
    { replace (sem_expr (pifid pidx) (expr_addr_of l) am) with
        ((bot, sem_expr_lv (pifid pidx) l am, bot):valueT);
        last reflexivity.
      apply sem_expr_lv_proper2_ with m; auto. }
    { replace (sem_expr (pifid pidx) (expr_cast t e) am) with
      (cast t (sem_expr (pifid pidx) e am));
      last reflexivity.
      apply cast_sound. assumption.
      apply IHHv. assumption. }
    { replace (sem_expr (pifid pidx) (expr_cast t e) am) with
      (cast t (sem_expr (pifid pidx) e am));
      last reflexivity.
      apply cast_sound_not with z. assumption.
      assumption.
      apply IHHv. assumption. }
    { replace (sem_expr (pifid pidx) (expr_cast (typ_pointer t) e) am) with 
      (cast (typ_pointer t) (sem_expr (pifid pidx) e am));
      last reflexivity.
      apply cast_sound_array. assumption.
      apply IHHv. assumption. }
    { exact: abs_itvT'_top. }

  (* lv *)
    intros pidx l m am Hv con_loc Hsem_lv.
    induction Hsem_lv; intros.
    { simpl. unfold lglobal.
      rewrite SetFacts.singleton_b. 
      rewrite DomBridge.is_compare_eqb.
      apply/eqbP. reflexivity. }
    { simpl. unfold llocal.
      rewrite SetFacts.singleton_b. 
      rewrite DomBridge.is_compare_eqb.
      apply/eqbP. reflexivity. }
    { replace (sem_expr_lv (pifid pidx) (lv_deref e) am) with 
      (let v := sem_expr (pifid pidx) e am in 
        locs_of v ⊔ locs_from_arrvT (arrv_of v)); 
      last reflexivity.
      assert ((vloc l) \abs sem_expr (pifid pidx) e am).
      apply sem_expr_proper2_ with m; assumption.
      rewrite/abs_relation/value_abs in H.
      move:H. apply abs_relation_proper. reflexivity.
      apply join_lhs_.
    }
    { replace (sem_expr_lv (pifid pidx) (lv_deref e) am) with
      (let v := sem_expr (pifid pidx) e am in 
        locs_of v ⊔ locs_from_arrvT (arrv_of v)); 
      last reflexivity.
      assert ((vheap al uq (ofs, sz, st)) \abs sem_expr (pifid pidx) e am).
      apply sem_expr_proper2_ with m; assumption.
      rewrite/abs_relation/value_abs in H.
      rewrite/ locs_from_arrvT.
      destruct (arrv_of (sem_expr (pifid pidx) e am)).
      unfold lheap.

      assert((DomConValue.lheap al uq ofs) \abs
        (Elt (fold
          (λ (k : idT) (v : ossT) (s : SetInterface.set locT),
            if ~~ eqb v bot then
              SetInterface.add (lheap k nil) s else s) a
          SetInterface.empty):locsT)).
      { apply collect_key_abs with sz st. assumption. }

      move:H0. apply abs_relation_proper. reflexivity.
      apply join_rhs_.
      
      rewrite join_top_rhs. constructor.
    }
    { replace (sem_expr_lv (pifid pidx) (Syntax.lv_array arr idx) am) with
      (let v := sem_expr (pifid pidx) arr am in locs_from_arrvT (arrv_of v));
      last reflexivity.
      rewrite/locs_from_arrvT.
      assert ((DomConValue.vheap al uq (ofs, sz, st))
                \abs sem_expr (pifid pidx) arr am).
      apply sem_expr_proper2_ with m; assumption.
      remember (sem_expr (pifid pidx) arr am).
      move: v Heqv H=> [[i a] s] Heqv H.
      rewrite/arrv_of/fst/snd. destruct s.

      assert ((DomConValue.lheap al uq ofs) \abs
        (Elt (fold
          (λ (k : idT) (v : ossT) (s : Set! [locT]),
            if ~~ eqb v bot then SetInterface.add (lheap k nil) s else s) a0
          SetInterface.empty) : locsT)).
      apply collect_key_abs with sz st. apply H.

      move:H0. rewrite/abs_relation/loc_abs/t_loc. apply.

      constructor.
    }
    { unfold sem_expr_lv; fold sem_expr_lv.
      move: (sem_expr_lv _ _ _) (IHHsem_lv Hv)=> [alocs|] Hlocs; last done.
      clear Hsem_lv IHHsem_lv.
      move: loc Hlocs=> [loc poses] Hlocs.
      rewrite/abs_relation/loc_abs/t_loc/fst/snd in Hlocs *.
      replace
        (match loc with
           | inl (inl (a, _, _)) => lheap a (pos :: poses)
           | inl (inr gvar) => lglobal gvar (pos :: poses)
           | inr (fid, bid, _) => llocal fid bid (pos :: poses)
         end)
        with
        (proj1_sig (append_field pos)
                   match loc with
                     | inl (inl (a, _, _)) => lheap a (poses)
                     | inl (inr gvar) => lglobal gvar (poses)
                     | inr (fid, bid, _) => llocal fid bid (poses)
                   end); last first.
      { rewrite/append_field/=.
        by move: loc Hlocs=> [[[[a u] o]|loc]|[[fvar vvar] dvar]] Hlocs.
      }
      by apply DomSet.mapP.
    }
  Qed.

  Lemma sem_expr_proper2 :
    Proper2
      (eq ===> eq ===> abs_relation ===> abs_relation ===|')
      SemCon.sem_expr
      (fun pidx => sem_expr (pifid pidx)).
  Proof.
    move=> cpidx apidx Hpidx cexpr aexpr Hexpr cmem amem Hmem.
    usual_tac Hpidx; usual_tac Hexpr.
    move=> val Hval.
    by eapply sem_expr_proper2_; eauto.
  Qed.

  Lemma sem_expr_lv_proper2 :
    Proper2
      (eq ===> eq ===> abs_relation ===> abs_relation ===|')
      SemCon.sem_lv
      (fun pidx => sem_expr_lv (pifid pidx)).
  Proof.
    move => pidx pidy Hpidx; usual_tac Hpidx.
    move => lex ley Hlv. usual_tac Hlv.
    move => x y Hxy.
    move => a b. 
    by eapply sem_expr_lv_proper2_; eauto.
  Qed.

  Lemma sem_exprs_proper2 :
    Proper2
      (eq ===> eq ===> abs_relation ===> (let (f,_) := list_abs in f) ===|')
      SemCon.sem_exprs
      (fun pidx => sem_exprs (pifid pidx)).
  Proof.
    move=> cpidx apidx Hpidx cexpr aexpr Hexpr cmem amem Hmem convals Hsem.
    usual_tac Hpidx; usual_tac Hexpr.
    elim: aexpr convals Hsem => [|aexpr aexprs IHexpr] [|conval convals] Hsem;
    inversion Hsem; clear Hsem; subst; rewrite/list_abs//.
    split.
    { apply (sem_expr_proper2 (eq_refl apidx) (eq_refl aexpr) amem Hmem Hexpr). }
    { by rewrite/list_abs in IHexpr; apply IHexpr. }
  Qed.

  Lemma abs_option_value_top : forall (v:option DomConValue.valueT), v \abs top.
  Proof. 
    intros. rewrite/abs_relation/option_abs. destruct v; last auto.
    apply abs_value_top.
  Qed.

  Lemma cardinal_1_iff : forall (x y:locT) (s : Set! [locT])
    (Hx:x !\in s) (Hy:y !\in s) (Hcar:SetInterface.cardinal s ≡ 1),
    x ≡ y.
  Proof.
    intros.
    rewrite SetInterface.cardinal_1 in Hcar.
    rewrite/smem/set_setlike in Hx,Hy.
    rewrite SetFacts.elements_b in Hx.
    rewrite SetFacts.elements_b in Hy.
    move: (SetInterface.elements s) Hcar Hx Hy=> {s} [|a s] Hcar Hx Hy.
    done.
    inv Hcar.
    destruct s; last done.
    rewrite/existsb in Hx Hy.
    case Hx': (_ == _) in Hx; last done.
    case Hy': (_ == _) in Hy; last done.
    move/is_true_iff in Hx'.
    move/is_true_iff in Hy'.
    rewrite is_compare_eqb in Hx'.
    rewrite is_compare_eqb in Hy'.
    move/eqbP in Hx'; rewrite Hx'.
    move/eqbP in Hy'; rewrite Hy'.
    reflexivity.
  Qed.

  Lemma pidx_consistent_imply_callfroms_star : forall l (f g:fidT) gb
    (Hprog:acc_correct_callfroms_prog prog acc)
    (Hstar:acc_correct_callfroms_star acc)
    (Hin:List.In (g,gb) l)
    (Hpidx:pidx_consistent prog f l),
    g !\in (((callfroms_star acc) [f ]!):Set! [fidT]).
  Proof.
    induction l; intros; first inversion Hin.
    apply in_inv in Hin. destruct Hin.
    {
      subst. 
      rewrite simpl_pidx_consistent in Hpidx.
      move/andP:Hpidx => [Hpidx _].
      remember ((pfuncs prog) [g]) as funcname.
      destruct funcname as [func|]; last inversion Hpidx.
      remember ((fcmds func) [gb]) as cmds.
      destruct cmds as [cmd|]; last inversion Hpidx.
      remember (ctmn cmd) as terminator.
      destruct terminator; try (inversion Hpidx; fail).
      
      rewrite/acc_correct_callfroms_prog in Hprog.
      assert (Hacc_func:MapsTo g func (pfuncs prog)).
      apply MF.find_mapsto_iff. symmetry. assumption.
      apply Hprog in Hacc_func. clear Hprog.
      
      rewrite/acc_correct_callfroms_func in Hacc_func.
      assert (Hacc_blk:MapsTo gb cmd (fcmds func)).
      apply MF.find_mapsto_iff. symmetry. assumption.
      apply Hacc_func in Hacc_blk. clear Hacc_func.

      rewrite/acc_correct_callfroms_blk in Hacc_blk.
      rewrite <- Heqterminator in Hacc_blk.
      move/set_forallP in Hacc_blk.
      exploit (Hacc_blk _ f); auto.
      move=> {Hacc_blk} Hacc_blk.

      rewrite/acc_correct_callfroms_star in Hstar.
      apply Hstar in Hacc_blk. clear Hstar.
      destruct Hacc_blk as [Hin Horder].
      apply Hin.
   }
   {
      rewrite simpl_pidx_consistent in Hpidx.
      destruct a as [h hb].
      move/andP:Hpidx => [Hpidx Hl].
      remember ((pfuncs prog) [h]) as funcname.
      destruct funcname as [func|]; last inversion Hpidx.
      remember ((fcmds func) [hb]) as cmds.
      destruct cmds as [cmd|]; last inversion Hpidx.
      remember (ctmn cmd) as terminator.
      destruct terminator; try (inversion Hpidx; fail).
      
      rewrite/acc_correct_callfroms_prog in Hprog.
      assert (Hacc_func:MapsTo h func (pfuncs prog)).
      apply MF.find_mapsto_iff. symmetry. assumption.
      apply Hprog in Hacc_func. 
      
      rewrite/acc_correct_callfroms_func in Hacc_func.
      assert (Hacc_blk:MapsTo hb cmd (fcmds func)).
      apply MF.find_mapsto_iff. symmetry. assumption.
      apply Hacc_func in Hacc_blk. 

      rewrite/acc_correct_callfroms_blk in Hacc_blk.
      rewrite <- Heqterminator in Hacc_blk.
      move/set_forallP in Hacc_blk.
      exploit (Hacc_blk _ f); auto.
      move=> {Hacc_blk} Hacc_blk.

      rewrite/acc_correct_callfroms_star in Hstar.
      apply Hstar in Hacc_blk. 
      destruct Hacc_blk as [Hin Horder].
      
      rewrite/fst in Hin,Horder.
      rewrite/le/le_/set_poset in Horder.
      rewrite/SetInterface.Subset in Horder.
      rewrite/smem/set_setlike.
      apply SetFacts.mem_iff. apply Horder. apply SetFacts.mem_iff.
      apply IHl with gb; auto.
    }
  Qed.

  Lemma nth_fid_imply_some: forall l n fid bid g
    (Hn : nth_fid l n fid = Some g),
    exists gb, 
      List.find (λ fb : fidT * bidT, eqb g fb.1) ((fid, bid) :: l) 
      = Some (g,gb).
  Proof.
    induction l; intros.
    {
      move:Hn. rewrite/nth_fid. destruct (eqb (length nil) n).
      { move => Hg; inv Hg. exists bid. simpl. 
        rewrite (eqb_refl g). reflexivity.
      }
      rewrite/rev/nth'. move => Hf; inv Hf.
    }
    {
      move:Hn. destruct a. rewrite nth_fid_subsume.
      case Heqn: (eqb (S (length l)) n).
      {
        move/eqbP:Heqn=>Heqn. usual_tac Heqn.
        move => Heq. inv Heq.
        exists bid. 
        rewrite/List.find/fst.
        assert (Heqbg:eqb g g). apply/eqbP. reflexivity.
        rewrite Heqbg. reflexivity.
      }
      {
        move => Hnth_fid. exploit IHl; eauto.
        instantiate (1 := b).
        move => Hex. destruct Hex.
        rewrite/List.find/fst.
        case Hg: (eqb g fid). 
        {
          move/eqbP:Hg=>Hg. usual_tac Hg.
          exists bid. reflexivity.
        }          
        exploit IHl; eauto.
      }
    }
  Qed.

  Lemma not_recursive_imply_same_nth_fid : forall l (f g:fidT) 
    (n m:nat)
    (Hprog:acc_correct_callfroms_prog prog acc)
    (Hstar:acc_correct_callfroms_star acc)
    (Hpidx:pidx_consistent prog f l)
    (Hrec:not_recursive acc g)
    (Hn:nth_fid l n f = Some g)
    (Hm:nth_fid l m f = Some g),
    n = m.
  Proof.
    elim=> [| [fid bid] l IH] f g n m Hprog Hstar Hpidx Hrec Hn Hm.
    { 
      move:Hn Hm. rewrite/nth_fid/length/rev.
      case Hn: (eqb 0 n).
      {
        move/eqbP:Hn => Hn; inv Hn.
        move => Hfg; inv Hfg.
        case Hm: (eqb 0 m).
        { move/eqbP:Hm => Hm; inv Hm. auto. }
        { rewrite/nth'. move=>Hf; inv Hf. }
      }
      rewrite/nth'. move=>Hf; inv Hf.
    }
    rewrite nth_fid_subsume in Hn.
    rewrite nth_fid_subsume in Hm.
    case Heq: (eqb _ _) in Hn; last first.
    { case Heq': (eqb _ _) in Hm; last first.
      { eapply IH; eauto.
        by move/andP: Hpidx=> [_ Hpidx].
      }
      { move/eqbP in Heq'; usual_tac Heq'.
        inv Hm.
        elimtype False.
        exploit acc_correct_consistent; eauto.
        move=> Hfind. clear -Hn Hfind.
        eapply nth_fid_imply_some with (bid:=bid) in Hn.
        destruct Hn as [b Hn]. rewrite Hn in Hfind. inv Hfind.
      }
    }
    { move/eqbP in Heq; usual_tac Heq.
      inv Hn.
      case Heq': (eqb _ _) in Hm; last first.
      { elimtype False.
        exploit acc_correct_consistent; eauto.
        move=> Hfind. clear -Hm Hfind.
        eapply nth_fid_imply_some with (bid:=bid) in Hm.
        destruct Hm as [b Hm]. rewrite Hm in Hfind. inv Hfind.
      }
      { by move/eqbP in Heq'. }
    }
  Qed.

  Lemma memory_update_sound (pidx:pidxT) : forall 
    (Hpidx_cons: pidx_consistent prog (pifid pidx) pidx.2)
    (Hacc_prog: acc_correct_callfroms_prog prog acc)
    (Hacc_star: acc_correct_callfroms_star acc)
    (m:DomConMem.memT) (m':memT) 
    (l:DomConValue.locT) (l':locsT) (v:DomConValue.valueT) (v':valueT)
    (Hmem: m \abs m') (Hloc: l \abs l') (Hval: Some v \abs v')
    (Hmem_wf: mem_wf pidx (madd (Key:=DomConValue.locT) (Elt:=option DomConValue.valueT) (M:=DomConMem.memT) l (Some v) m)),
    (madd (Key:=DomConValue.locT) (Elt:=option DomConValue.valueT) (M:=DomConMem.memT) l (Some v) m)
    \abs (memory_update acc l' v' m').
  Proof.
    move=> Hpidx Hacc_prog Hacc_star m m' l l' v v' Hmem Hloc Hval Hwf.
    move: m' Hmem=> [m'|] Hmem; last done.
    move: l' Hloc=> [l'|] Hloc; last done.
    move=> loc.
    rewrite morphism_madd_spec memory_update_spec.
    case Heq: (eqb loc l).
    { move/eqbP in Heq; usual_tac Heq.
      rewrite Hloc.
      case: (_ && _); first done.
      eapply abs_relation_proper; [reflexivity|idtac|by eauto].
      exact: join_rhs.
    }
    case Hin: (_ !\in _); last done.
    case Heq': (eqb _ _); last first.
    { eapply abs_relation_proper; [reflexivity|idtac|by eauto].
      exact: join_lhs.
    }
    case Hstrong: (is_strong _ _); last first.
    { eapply abs_relation_proper; [reflexivity|idtac|by eauto].
      exact: join_lhs.
    }
    rewrite/andb.
    move/eqbP in Heq'.
    exploit cardinal_1_iff; [exact Hloc|exact Hin|done|move=> Hloc'].
    clear l' Hloc Hin Heq'.
    exploit (Hwf l); first by rewrite/= eqb_refl.
    move=> Hwf'.
    move: (Hwf loc)=> {Hwf} Hwf.
    rewrite morphism_madd_spec Heq in Hwf.
    move: (_[_]!) Hwf=> [cv|]; last done.
    move=> Hwf; exploit Hwf; [done|move=> {Hwf} Hwf].

    move: loc Heq Hstrong Hloc' Hwf=>
    [[[[[a u] o]|i]|[[fid id] depth]] fl] Heq Hstrong Hloc' Hwf;
    rewrite/= in Hstrong; (try by inv Hstrong);
    move: l Hwf Hwf' Heq Hloc'=>
    [[[[[a' u'] o']|i']|[[fid' id'] depth']] fl'] Hwf Hwf' Heq Hloc';
    usual_tac Hloc'; inv Hloc'.
    { move/eqbP in Heq; contradict Heq; reflexivity. }
    { contradict Heq.
      do ? split; try reflexivity.
      rewrite/snd.
      exploit not_recursive_imply_same_nth_fid; eauto.
      { by usual_tac Hwf; eauto. }
      { by usual_tac Hwf'; eauto. }
      by move=> ->; rewrite eqb_refl.
    }
  Qed.
End Correctness.
