Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.

Require Import LibTactics.
Require Import DomBridge.
Require Import DomType.
Require Import DomBasic.
Require Import DomMap.

Require Import Syntax.

Require Import DomConValue.

Require Import SemCon.
Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomPidxSen.

Require Import SemAbsBlock.
Require Import SemAccSen.
Require Import SemSen.

Require Import DomZ.
Require Import DomSet.
Require Import CorBasic.
Require Import CorAbs.
Require Import CorItv.
Require Import CorConExpr.

Set Implicit Arguments.

Obligation Tactic := Simpless.tactic.

Section Correctness.
Variable (prog:progT) (acc:accT).
  
  Lemma ge_not_lt : forall l r, con_op_ge l r = con_op_not (con_op_lt l r).
  Proof.
    intros. rewrite/con_op_ge/con_op_not/con_op_lt. rewrite/ltb.
    remember (DomType.leb r l). destruct b.
    remember (DomType.leb l r && ~~ eqb l r). destruct b.
    { symmetry in Heqb. symmetry in Heqb0.
      move/lebP:Heqb => Hrl.
      move/andP:Heqb0 => [Hlr Hnot].
      move/lebP:Hlr => Hlr.
      move/negP:Hnot => Hnot.
      elim Hnot. apply/eqbP. 
      apply DomType.le_antisym; assumption. }
    { reflexivity. }
    remember (DomType.leb l r && ~~ eqb l r). destruct b.
    { reflexivity. }
    { symmetry in Heqb. symmetry in Heqb0.
      move/lebP:Heqb => Hrl.
      move/andP:Heqb0 => Hlrnot.
      elim Hlrnot. split.
      apply/lebP. 
      rewrite/le/le_/Z_poset.
      rewrite/le/le_/Z_poset in Hrl.
      omega.
      apply/eqbP.
      rewrite/eq/eq_/Z_setoid.
      rewrite/le/le_/Z_poset in Hrl.
      omega. }
  Qed.

  Lemma gt_not_le : forall l r, con_op_gt l r = con_op_not (con_op_le l r).
  Proof.
    intros. rewrite/con_op_gt/con_op_not/con_op_le. rewrite/ltb.
    remember (DomType.leb l r). destruct b.
    remember (DomType.leb r l && ~~ eqb r l). destruct b.
    { symmetry in Heqb. symmetry in Heqb0.
      move/lebP:Heqb => Hrl.
      move/andP:Heqb0 => [Hlr Hnot].
      move/lebP:Hlr => Hlr.
      move/negP:Hnot => Hnot.
      elim Hnot. apply/eqbP. 
      apply DomType.le_antisym; assumption. }
    { reflexivity. }
    remember (DomType.leb r l && ~~ eqb r l). destruct b.
    { reflexivity. }
    { symmetry in Heqb. symmetry in Heqb0.
      move/lebP:Heqb => Hrl.
      move/andP:Heqb0 => Hlrnot.
      elim Hlrnot. split.
      apply/lebP. 
      rewrite/le/le_/Z_poset.
      rewrite/le/le_/Z_poset in Hrl.
      omega.
      apply/eqbP.
      rewrite/eq/eq_/Z_setoid.
      rewrite/le/le_/Z_poset in Hrl.
      omega. }
  Qed.

  Lemma le_not_gt : forall l r, con_op_le l r = con_op_not (con_op_gt l r).
  Proof.
    intros. rewrite/con_op_gt/con_op_not/con_op_le. rewrite/ltb.
    remember (DomType.leb l r). destruct b.
    remember (DomType.leb r l && ~~ eqb r l). destruct b.
    { symmetry in Heqb. symmetry in Heqb0.
      move/lebP:Heqb => Hrl.
      move/andP:Heqb0 => [Hlr Hnot].
      move/lebP:Hlr => Hlr.
      move/negP:Hnot => Hnot.
      elim Hnot. apply/eqbP. 
      apply DomType.le_antisym; assumption. }
    { reflexivity. }
    remember (DomType.leb r l && ~~ eqb r l). destruct b.
    { reflexivity. }
    { symmetry in Heqb. symmetry in Heqb0.
      move/lebP:Heqb => Hrl.
      move/andP:Heqb0 => Hlrnot.
      elim Hlrnot. split.
      apply/lebP. 
      rewrite/le/le_/Z_poset.
      rewrite/le/le_/Z_poset in Hrl.
      omega.
      apply/eqbP.
      rewrite/eq/eq_/Z_setoid.
      rewrite/le/le_/Z_poset in Hrl.
      omega. }
  Qed.

  Lemma lt_not_ge : forall l r, con_op_lt l r = con_op_not (con_op_ge l r).
  Proof.
    intros. rewrite/con_op_ge/con_op_not/con_op_lt. rewrite/ltb.
    remember (DomType.leb r l). destruct b.
    remember (DomType.leb l r && ~~ eqb l r). destruct b.
    { symmetry in Heqb. symmetry in Heqb0.
      move/lebP:Heqb => Hrl.
      move/andP:Heqb0 => [Hlr Hnot].
      move/lebP:Hlr => Hlr.
      move/negP:Hnot => Hnot.
      elim Hnot. apply/eqbP. 
      apply DomType.le_antisym; assumption. }
    { reflexivity. }
    remember (DomType.leb l r && ~~ eqb l r). destruct b.
    { reflexivity. }
    { symmetry in Heqb. symmetry in Heqb0.
      move/lebP:Heqb => Hrl.
      move/andP:Heqb0 => Hlrnot.
      elim Hlrnot. split.
      apply/lebP. 
      rewrite/le/le_/Z_poset.
      rewrite/le/le_/Z_poset in Hrl.
      omega.
      apply/eqbP.
      rewrite/eq/eq_/Z_setoid.
      rewrite/le/le_/Z_poset in Hrl.
      omega. }
  Qed.

  Lemma neq_not_eq : forall l r, con_op_neq l r = con_op_not (con_op_eq l r).
  Proof.
    intros. rewrite/con_op_neq/con_op_not/con_op_eq. 
    remember (eqb l r). destruct b.
    { reflexivity. }
    { reflexivity. }
  Qed.

  Lemma eq_not_neq : forall l r, con_op_eq l r = con_op_not (con_op_neq l r).
  Proof.
    intros. rewrite/con_op_neq/con_op_not/con_op_eq. 
    remember (eqb l r). destruct b.
    { reflexivity. }
    { reflexivity. }
  Qed.

  Lemma con_op_not_not : forall z, con_op_not (con_op_not z) ≠ 0%Z ->
    z ≠ 0%Z.
  Proof.
    intros. rewrite/con_op_not in H.
    remember (eqb z 0%Z). destruct b.
    simpl in H. elim H. reflexivity.
    symmetry in Heqb. move/eqbP:Heqb => Heqb. apply Heqb.
  Qed.

  Inductive size_of_expr : exprT -> nat -> Prop :=
    | size_const : forall l, size_of_expr (expr_const l) O
    | size_op2 : forall o lex rex n m, 
      size_of_expr lex n -> size_of_expr rex m ->
      size_of_expr (expr_op2 o lex rex) (S (n+m))
    | size_op1 : forall o e n, 
      size_of_expr e n -> size_of_expr (expr_op1 o e) (S n)
    | size_value_of : forall l, size_of_expr (expr_value_of l) O
    | size_addr_of : forall l, size_of_expr (expr_addr_of l) O
    | size_cast : forall c e n, 
      size_of_expr e n -> size_of_expr (expr_cast c e) (S n)
    | size_size_of : forall e n, 
      size_of_expr e n -> size_of_expr (expr_size_of e) (S n).

  Lemma remove_uop_preserve_con_sem_ (pidx:pidxT) : forall m n e e' mem v
    (Hnm:n <= m)
    (Hsize:size_of_expr e n)
    (Huop:remove_uop e = Some e')
    (Hcon:SemCon.sem_expr pidx e mem (vint v))
    (Hv:v≠0%Z),
    exists v', v'≠0%Z ∧ SemCon.sem_expr pidx e' mem (vint v').
  Proof.
    induction m.
    { destruct e; simpl; intros; try (inversion Huop; fail).
      inversion Hsize; subst. inversion Hnm.
      inversion Hsize; subst. inversion Hnm. }

    destruct e; intros; inversion Huop.
    { destruct op; try inversion H0.
      { subst. clear H0. exists v. split; assumption. }
      { subst. clear H0. exists v. split; assumption. }
      { subst. clear H0. exists v. split; assumption. }
      { subst. clear H0. exists v. split; assumption. }
      { subst. clear H0. exists v. split; assumption. }
      { subst. clear H0. exists v. split; assumption. }
    }
    { destruct op; last inversion H0.
      destruct e; try (inversion H0; fail).
      { destruct op; try (inversion H0; fail).
        { inversion H0; subst; clear H0.
          inversion Hcon; subst. inversion He; subst.
          exists (con_op_not (con_op_lt lz rz)). split; first assumption.
          replace (con_op_not (con_op_lt lz rz)) with (con_op_ge lz rz);
          last apply ge_not_lt.
          constructor; assumption. }
        { inversion H0; subst; clear H0.
          inversion Hcon; subst. inversion He; subst.
          exists (con_op_not (con_op_le lz rz)). split; first assumption.
          replace (con_op_not (con_op_le lz rz)) with (con_op_gt lz rz);
          last apply gt_not_le.
          constructor; assumption. }
        { inversion H0; subst; clear H0.
          inversion Hcon; subst. inversion He; subst.
          exists (con_op_not (con_op_gt lz rz)). split; first assumption.
          replace (con_op_not (con_op_gt lz rz)) with (con_op_le lz rz);
          last apply le_not_gt.
          constructor; assumption. }
        { inversion H0; subst; clear H0.
          inversion Hcon; subst. inversion He; subst.
          exists (con_op_not (con_op_ge lz rz)). split; first assumption.
          replace (con_op_not (con_op_ge lz rz)) with (con_op_lt lz rz);
          last apply lt_not_ge.
          constructor; assumption. }
        { inversion H0; subst; clear H0.
          inversion Hcon; subst. inversion He; subst.
          exists (con_op_not (con_op_eq lz rz)). split; first assumption.
          replace (con_op_not (con_op_eq lz rz)) with (con_op_neq lz rz);
          last apply neq_not_eq.
          constructor; assumption. }
        { inversion H0; subst; clear H0.
          inversion Hcon; subst. inversion He; subst.
          exists (con_op_not (con_op_neq lz rz)). split; first assumption.
          replace (con_op_not (con_op_neq lz rz)) with (con_op_eq lz rz);
          last apply eq_not_neq.
          constructor; assumption. }
      }          
      { destruct op; last inversion H0.
        inversion Hsize; subst; clear Hsize.
        inversion H3; subst; clear H3.
        inversion Hcon; subst; clear Hcon.
        inversion He; subst; clear He.
        apply IHm with n e z0%Z; try assumption; try omega.
        apply con_op_not_not; assumption. }
    }        
  Qed.

  Fixpoint size_of_expr_f (e:exprT): nat :=
    match e with
      | expr_const _ => O
      | expr_op2 _ lex rex => S (size_of_expr_f lex + size_of_expr_f rex)
      | expr_op1 _ e' => S (size_of_expr_f e')
      | expr_value_of _ => O
      | expr_addr_of _ => O
      | expr_cast _ e' => S (size_of_expr_f e')
      | expr_size_of e' => S (size_of_expr_f e')
    end.

  Lemma size_of_expr_f_correct : forall e, size_of_expr e (size_of_expr_f e).
  Proof. induction e; simpl; constructor; try assumption. Qed.

  Lemma remove_uop_preserve_con_sem (pidx:pidxT) : forall e e' m v
    (Huop:remove_uop e = Some e')
    (Hcon:SemCon.sem_expr pidx e m (vint v))
    (Hv:v≠0%Z),
    exists v', v' ≠ 0%Z ∧ SemCon.sem_expr pidx e' m (vint v').
  Proof.
    intros. 
    apply remove_uop_preserve_con_sem_ 
      with (size_of_expr_f e) (size_of_expr_f e) e v; try omega;
        try assumption.
    apply size_of_expr_f_correct.
  Qed.

  Lemma itv_restrict_ub: forall lz i ub (Habs:lz \abs i) (Hle:(lz <= ub)%Z),
    lz \abs itv_restrict_ub i ub.
  Proof.
    intros. rewrite/itv_restrict_ub.
    destruct i; last inversion Habs.
    destruct i.
    { constructor. auto. }
    { constructor. apply Z.min_glb; auto. inversion Habs. auto. }
    { destruct (Z_le_dec l ub). 
      constructor; auto. inversion Habs. omega.
      elim n. inversion Habs. omega. }
    { destruct (Z_le_dec l ub). 
      constructor; auto. inversion Habs. omega.
      apply Z.min_glb; auto. inversion Habs. omega.
      elim n. inversion Habs. omega. }
  Qed.    

  Lemma itv_restrict_lb: forall lz i lb (Habs:lz \abs i) (Hle:(lb <= lz)%Z),
    lz \abs itv_restrict_lb i lb.
  Proof.
    intros. rewrite/itv_restrict_lb.
    destruct i; last inversion Habs.
    destruct i.
    { constructor. auto. }
    { destruct (Z_le_dec lb u). 
      constructor; auto. inversion Habs. omega.
      elim n. inversion Habs. omega. }
    { constructor. apply Z.max_lub; auto. inversion Habs. auto. }
    { destruct (Z_le_dec lb u). 
      constructor; auto.
      apply Z.max_lub; auto. inversion Habs. omega.
      inversion Habs. omega.
      elim n. inversion Habs. omega. }
  Qed.    

  Lemma adjust_inter_lt_sound : forall lz rz i j
    (Hv : con_op_lt lz rz ≠ 0%Z) (Hlz : lz \abs i) (Hrz : rz \abs j),
    lz \abs adjust_inter_lt i j.
  Proof.
    intros.
    rewrite/con_op_lt in Hv. remember (ltb lz rz). destruct b.
    symmetry in Heqb. apply ltb_true_imply_lt in Heqb.

    rewrite/adjust_inter_lt. destruct j.
    destruct i0; auto.
    apply itv_restrict_ub; auto. inversion Hrz; omega.
    apply itv_restrict_ub; auto. inversion Hrz; omega.
    inversion Hrz.
    elim Hv. reflexivity.
  Qed.

  Lemma adjust_inter_le_sound : forall lz rz i j
    (Hv : con_op_le lz rz ≠ 0%Z) (Hlz : lz \abs i) (Hrz : rz \abs j),
    lz \abs adjust_inter_le i j.
  Proof.
    intros.
    rewrite/con_op_le in Hv. remember (DomType.leb lz rz). destruct b.
    symmetry in Heqb. apply leb_true_imply_le in Heqb.

    rewrite/adjust_inter_le. destruct j.
    destruct i0; auto.
    apply itv_restrict_ub; auto. inversion Hrz; omega.
    apply itv_restrict_ub; auto. inversion Hrz; omega.
    inversion Hrz.
    elim Hv. reflexivity.
  Qed.

  Lemma adjust_inter_gt_sound : forall lz rz i j
    (Hv : con_op_gt lz rz ≠ 0%Z) (Hlz : lz \abs i) (Hrz : rz \abs j),
    lz \abs adjust_inter_gt i j.
  Proof.
    intros.
    rewrite/con_op_gt in Hv. remember (ltb rz lz). destruct b.
    symmetry in Heqb. apply ltb_true_imply_lt in Heqb.

    rewrite/adjust_inter_gt. destruct j.
    destruct i0; auto.
    apply itv_restrict_lb; auto. inversion Hrz; omega.
    apply itv_restrict_lb; auto. inversion Hrz; omega.
    inversion Hrz.
    elim Hv. reflexivity.
  Qed.

  Lemma adjust_inter_ge_sound : forall lz rz i j
    (Hv : con_op_ge lz rz ≠ 0%Z) (Hlz : lz \abs i) (Hrz : rz \abs j),
    lz \abs adjust_inter_ge i j.
  Proof.
    intros.
    rewrite/con_op_ge in Hv. remember (DomType.leb rz lz). destruct b.
    symmetry in Heqb. apply leb_true_imply_le in Heqb.

    rewrite/adjust_inter_ge. destruct j.
    destruct i0; auto.
    apply itv_restrict_lb; auto. inversion Hrz; omega.
    apply itv_restrict_lb; auto. inversion Hrz; omega.
    inversion Hrz.
    elim Hv. reflexivity.
  Qed.

  Lemma con_op_ge_id : forall lz, con_op_ge lz lz ≠ 0%Z.
  Proof.
    intros. rewrite/con_op_ge.
    remember (DomType.leb lz lz). destruct b.
    intro. inversion H.
    symmetry in Heqb. move/lebP:Heqb=>Heqb. elim Heqb. apply DomType.eq_le.
    reflexivity.
  Qed.

  Lemma adjust_inter_eq_sound : forall lz rz i j
    (Hv : con_op_eq lz rz ≠ 0%Z) (Hlz : lz \abs i) (Hrz : rz \abs j),
    lz \abs adjust_inter_eq i j.
  Proof.
    intros.
    rewrite/con_op_eq in Hv. remember (eqb rz lz). destruct b.
    symmetry in Heqb. move/eqbP:Heqb=>Heqb.

    usual_tac Heqb.
    rewrite/adjust_inter_eq/adjust_inter_le. destruct j.
    destruct i0; auto.
    apply itv_restrict_ub; auto. inversion Hrz; omega.
    apply adjust_inter_ge_sound with lz; auto.
    apply con_op_ge_id.
    apply itv_restrict_ub; auto. 
    apply adjust_inter_ge_sound with lz; auto.
    apply con_op_ge_id.
    inversion Hrz; omega.
    inversion Hrz.
    remember (eqb lz rz). destruct b.
    symmetry in Heqb0. symmetry in Heqb.
    move/eqbP:Heqb0 => Heqb0. move/eqbP:Heqb => Heqb.
    elim Heqb. by symmetry.
    elim Hv. reflexivity.
  Qed.

  Lemma adjust_inter_neq_sound : forall lz rz i j
    (Hv : con_op_neq lz rz ≠ 0%Z) (Hlz : lz \abs i) (Hrz : rz \abs j),
    lz \abs adjust_inter_neq i j.
  Proof.
    intros.
    rewrite/con_op_neq in Hv. remember (eqb lz rz). destruct b.
    symmetry in Heqb. move/eqbP:Heqb=>Heqb.

    usual_tac Heqb.
    rewrite/adjust_inter_neq/adjust_inter_lt/adjust_inter_gt. destruct j.
    destruct i0; auto.
    move:Hlz. apply abs_relation_proper. reflexivity. apply join_lhs_.
    move:Hlz. apply abs_relation_proper. reflexivity. apply join_rhs_.
    move:Hlz. apply abs_relation_proper. reflexivity. apply join_lhs_.
    elim Hv. reflexivity.
    inversion Hrz.

    rewrite/adjust_inter_neq/adjust_inter_lt/adjust_inter_gt. destruct j.
    destruct i0; auto.
    move:Hlz. apply abs_relation_proper. reflexivity. apply join_lhs_.
    move:Hlz. apply abs_relation_proper. reflexivity. apply join_rhs_.
    move:Hlz. apply abs_relation_proper. reflexivity. apply join_lhs_.
    
    symmetry in Heqb. move/eqbP:Heqb=>Heqb.
    destruct (Z_lt_dec lz u).
    assert (lz \abs SemAbsBlock.itv_restrict_ub i (u - 1)).
    apply itv_restrict_ub; auto. omega.
    move: H. apply abs_relation_proper. reflexivity. apply join_lhs_.
    assert (u <= lz)%Z. omega.
    apply Z_le_lt_eq_dec in H. destruct H.
    assert (lz \abs SemAbsBlock.itv_restrict_lb i (l + 1)).
    apply itv_restrict_lb; auto. assert (l <= u)%Z. apply l0. omega.
    move: H. apply abs_relation_proper. reflexivity. apply join_rhs_.
    subst. 
    assert (l <= lz)%Z. apply l0. apply Z_le_lt_eq_dec in H. destruct H.
    assert (lz \abs SemAbsBlock.itv_restrict_lb i (l + 1)).
    apply itv_restrict_lb; auto. omega.
    move: H. apply abs_relation_proper. reflexivity. apply join_rhs_.
    subst.
    inversion Hrz; subst. elim Heqb. 
    apply DomType.le_antisym; rewrite/le/le_/Z_poset; omega.
    inversion Hrz.
  Qed.  

  Lemma adjust_val_op2_global_var (pidx:pidxT) :
    forall v x op e
           (m:DomConMem.memT) (am:memT)
           (Hmem: m \abs am)
           (Hv : v ≠ 0%Z)
           (Hcon : SemCon.sem_expr
                     pidx (expr_op2 op (expr_value_of (lv_id (global x))) e)
                     m (vint v)),
      (m[(DomConValue.lglobal x) : DomConValue.locT]!
       : option DomConValue.valueT)
      \abs
      (adjust_val
         (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                     (lv_id (global x)) am) am) op
      (sem_expr (DomPidx.pifid pidx) e am) : valueT).
  Proof.
    intros. inversion Hcon; subst; clear Hcon;
    rewrite/DomConValue.lglobal.
    { rewrite/adjust_val/adjust_inter/adjust_loc.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (global x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup.
      clear Hlookup Hcon_e i l a.

      inversion Hl; subst.
      inversion Hlv; subst. rewrite Hlookup.
      clear Hl Hlv.
      rewrite/DomConValue.lglobal in Hlookup.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (global x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((inl (inr x), nil):DomConValue.locT); try assumption.
      inversion Heql. rewrite/t_loc/lglobal.
      apply existsb_exists.
      exists ((inl (inr x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      clear Hlookup Hcon_e i l a.
      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup/global.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      set fl := (fold_left _ _ _).
      case Hfl: fl => [[flix fllx] flrx].
      rewrite <- Hfl.
      apply sound_memory_lookup_ with 
        m ((DomConValue.lglobal x):DomConValue.locT); try assumption.
      inversion Heql. rewrite/t_loc/DomConValue.lglobal/lglobal.
      apply existsb_exists.
      exists ((inl (inr x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      clear Hlookup Hcon_e i l a.
      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup/global.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      set fl := (fold_left _ _ _).
      case Hfl: fl => [[flix fllx] flrx].
      rewrite <- Hfl.
      apply sound_memory_lookup_ with 
        m ((DomConValue.lglobal x):DomConValue.locT); try assumption.
      inversion Heql. rewrite/t_loc/DomConValue.lglobal/lglobal.
      apply existsb_exists.
      exists ((inl (inr x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      clear Hlookup Hcon_e i l a.
      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup/global.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      set fl := (fold_left _ _ _).
      case Hfl: fl => [[flix fllx] flrx].
      rewrite <- Hfl.
      apply sound_memory_lookup_ with 
        m ((DomConValue.lglobal x):DomConValue.locT); try assumption.
      inversion Heql. rewrite/t_loc/DomConValue.lglobal/lglobal.
      apply existsb_exists.
      exists ((inl (inr x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      clear Hlookup Hcon_e i l a.
      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup/global.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      set fl := (fold_left _ _ _).
      case Hfl: fl => [[flix fllx] flrx].
      rewrite <- Hfl.
      apply sound_memory_lookup_ with 
        m ((DomConValue.lglobal x):DomConValue.locT); try assumption.
      inversion Heql. rewrite/t_loc/DomConValue.lglobal/lglobal.
      apply existsb_exists.
      exists ((inl (inr x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am).
      destruct v. destruct p.
      remember (sem_expr (DomPidx.pifid pidx) e am).
      destruct v. destruct p.
      inversion Hl; subst.
      inversion Hlv; subst.

      assert ((vint lz) \abs (i, l, a)) as Hlz. rewrite Heqv.
      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l1; last apply abs_value_top.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with m (DomConValue.lglobal x); auto.
      
      assert (DomConValue.lglobal x \abs Elt s) as Hx. rewrite Heql1.
      apply sem_expr_lv_proper2_ with m; auto.
      inversion Hx as [Hxs].
      rewrite SetFacts.elements_b in Hxs. apply existsb_exists in Hxs.
      destruct Hxs as [x0 [Hin Hxeq]]. rewrite is_compare_eqb in Hxeq.
      apply existsb_exists. exists x0; auto.

      assert ((vint rz) \abs (i0, l0, a0)) as Hrz.
      rewrite Heqv0.
      apply sem_expr_proper2_ with m; auto.

      rewrite/global.
      rewrite <- Heqv.

      rewrite/DomConValue.lglobal in Hlv Hlookup.
      rewrite/vint in Hlz Hrz Hlookup.
      rewrite Hlookup.
      rewrite/abs_relation/option_abs.
      rewrite/abs_relation/value_abs/itv_of/fst.
      rewrite/abs_relation/value_abs/itv_of/fst in Hlz,Hrz.
      rewrite/adjust_inter/global.
      
      apply adjust_inter_lt_sound with rz; auto.
    }

    {
      rewrite/adjust_val.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am).
      destruct v. destruct p.
      remember (sem_expr (DomPidx.pifid pidx) e am).
      destruct v. destruct p.
      inversion Hl; subst.
      inversion Hlv; subst.
      
      assert ((vint lz) \abs (i, l, a)) as Hlz. rewrite Heqv.
      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l1; last apply abs_value_top.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with m (DomConValue.lglobal x); auto.
      
      assert (DomConValue.lglobal x \abs Elt s) as Hx. rewrite Heql1.
      apply sem_expr_lv_proper2_ with m; auto.
      inversion Hx as [Hxs].
      rewrite SetFacts.elements_b in Hxs. apply existsb_exists in Hxs.
      destruct Hxs as [x0 [Hin Hxeq]]. rewrite is_compare_eqb in Hxeq.
      apply existsb_exists. exists x0; auto.

      assert ((vint rz) \abs (i0, l0, a0)) as Hrz.
      rewrite Heqv0.
      apply sem_expr_proper2_ with m; auto.

      rewrite/global.
      rewrite <- Heqv.

      rewrite Hlookup.
      rewrite/abs_relation/option_abs.
      rewrite/abs_relation/value_abs/itv_of/fst.
      rewrite/abs_relation/value_abs/itv_of/fst in Hlz,Hrz.
      rewrite/adjust_inter.

      apply adjust_inter_le_sound with rz; auto.
    }
    {
      rewrite/adjust_val.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am).
      destruct v. destruct p.
      remember (sem_expr (DomPidx.pifid pidx) e am).
      destruct v. destruct p.
      inversion Hl; subst.
      inversion Hlv; subst.
      
      assert ((vint lz) \abs (i, l, a)) as Hlz. rewrite Heqv.
      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l1; last apply abs_value_top.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with m (DomConValue.lglobal x); auto.
      
      assert (DomConValue.lglobal x \abs Elt s) as Hx. rewrite Heql1.
      apply sem_expr_lv_proper2_ with m; auto.
      inversion Hx as [Hxs].
      rewrite SetFacts.elements_b in Hxs. apply existsb_exists in Hxs.
      destruct Hxs as [x0 [Hin Hxeq]]. rewrite is_compare_eqb in Hxeq.
      apply existsb_exists. exists x0; auto.

      assert ((vint rz) \abs (i0, l0, a0)) as Hrz.
      rewrite Heqv0.
      apply sem_expr_proper2_ with m; auto.

      rewrite/global.
      rewrite <- Heqv.

      rewrite Hlookup.
      rewrite/abs_relation/option_abs.
      rewrite/abs_relation/value_abs/itv_of/fst.
      rewrite/abs_relation/value_abs/itv_of/fst in Hlz,Hrz.
      rewrite/adjust_inter.

      apply adjust_inter_gt_sound with rz; auto.
    }
    {
      rewrite/adjust_val.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am).
      destruct v. destruct p.
      remember (sem_expr (DomPidx.pifid pidx) e am).
      destruct v. destruct p.
      inversion Hl; subst.
      inversion Hlv; subst.
      
      assert ((vint lz) \abs (i, l, a)) as Hlz. rewrite Heqv.
      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l1; last apply abs_value_top.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with m (DomConValue.lglobal x); auto.
      
      assert (DomConValue.lglobal x \abs Elt s) as Hx. rewrite Heql1.
      apply sem_expr_lv_proper2_ with m; auto.
      inversion Hx as [Hxs].
      rewrite SetFacts.elements_b in Hxs. apply existsb_exists in Hxs.
      destruct Hxs as [x0 [Hin Hxeq]]. rewrite is_compare_eqb in Hxeq.
      apply existsb_exists. exists x0; auto.

      assert ((vint rz) \abs (i0, l0, a0)) as Hrz.
      rewrite Heqv0.
      apply sem_expr_proper2_ with m; auto.

      rewrite/global.
      rewrite <- Heqv.

      rewrite Hlookup.
      rewrite/abs_relation/option_abs.
      rewrite/abs_relation/value_abs/itv_of/fst.
      rewrite/abs_relation/value_abs/itv_of/fst in Hlz,Hrz.
      rewrite/adjust_inter.

      apply adjust_inter_ge_sound with rz; auto.
    }
    {
      rewrite/adjust_val.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am).
      destruct v. destruct p.
      remember (sem_expr (DomPidx.pifid pidx) e am).
      destruct v. destruct p.
      inversion Hl; subst.
      inversion Hlv; subst.
      
      assert ((vint lz) \abs (i, l, a)) as Hlz. rewrite Heqv.
      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l1; last apply abs_value_top.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with m (DomConValue.lglobal x); auto.
      
      assert (DomConValue.lglobal x \abs Elt s) as Hx. rewrite Heql1.
      apply sem_expr_lv_proper2_ with m; auto.
      inversion Hx as [Hxs].
      rewrite SetFacts.elements_b in Hxs. apply existsb_exists in Hxs.
      destruct Hxs as [x0 [Hin Hxeq]]. rewrite is_compare_eqb in Hxeq.
      apply existsb_exists. exists x0; auto.

      assert ((vint rz) \abs (i0, l0, a0)) as Hrz.
      rewrite Heqv0.
      apply sem_expr_proper2_ with m; auto.

      rewrite/global.
      rewrite <- Heqv.

      rewrite Hlookup.
      rewrite/abs_relation/option_abs.
      rewrite/abs_relation/value_abs/itv_of/fst.
      rewrite/abs_relation/value_abs/itv_of/fst in Hlz,Hrz.
      rewrite/adjust_inter.

      apply adjust_inter_eq_sound with rz; auto.
    }
    {
      rewrite/adjust_val.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am).
      destruct v. destruct p.
      remember (sem_expr (DomPidx.pifid pidx) e am).
      destruct v. destruct p.
      inversion Hl; subst.
      inversion Hlv; subst.
      
      assert ((vint lz) \abs (i, l, a)) as Hlz. rewrite Heqv.
      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l1; last apply abs_value_top.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with m (DomConValue.lglobal x); auto.
      
      assert (DomConValue.lglobal x \abs Elt s) as Hx. rewrite Heql1.
      apply sem_expr_lv_proper2_ with m; auto.
      inversion Hx as [Hxs].
      rewrite SetFacts.elements_b in Hxs. apply existsb_exists in Hxs.
      destruct Hxs as [x0 [Hin Hxeq]]. rewrite is_compare_eqb in Hxeq.
      apply existsb_exists. exists x0; auto.

      assert ((vint rz) \abs (i0, l0, a0)) as Hrz.
      rewrite Heqv0.
      apply sem_expr_proper2_ with m; auto.

      rewrite/global.
      rewrite <- Heqv.

      rewrite Hlookup.
      rewrite/abs_relation/option_abs.
      rewrite/abs_relation/value_abs/itv_of/fst.
      rewrite/abs_relation/value_abs/itv_of/fst in Hlz,Hrz.
      rewrite/adjust_inter.

      apply adjust_inter_neq_sound with rz; auto.
    }

    { rewrite/adjust_val/adjust_inter/adjust_loc/global.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      rewrite Hlookup.
      clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((DomConValue.lglobal x):DomConValue.locT); try assumption.
      inversion Heql. rewrite/t_loc/lglobal.
      apply existsb_exists.
      exists (lglobal x nil). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc/global.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((DomConValue.lglobal x):DomConValue.locT); try assumption.
      inversion Heql. rewrite/t_loc/lglobal.
      apply existsb_exists.
      exists (lglobal x nil). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc/global.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((DomConValue.lglobal x):DomConValue.locT); try assumption.
      inversion Heql. rewrite/t_loc/lglobal.
      apply existsb_exists.
      exists (lglobal x nil). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc/global.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((DomConValue.lglobal x):DomConValue.locT); try assumption.
      inversion Heql. rewrite/t_loc/lglobal.
      apply existsb_exists.
      exists (lglobal x nil). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc/global.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((DomConValue.lglobal x):DomConValue.locT); try assumption.
      inversion Heql. rewrite/t_loc/lglobal.
      apply existsb_exists.
      exists (lglobal x nil). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc/global.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((DomConValue.lglobal x):DomConValue.locT); try assumption.
      inversion Heql. rewrite/t_loc/lglobal.
      apply existsb_exists.
      exists (lglobal x nil). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc/global.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inl x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inl x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((DomConValue.lglobal x):DomConValue.locT); try assumption.
      inversion Heql. rewrite/t_loc/lglobal.
      apply existsb_exists.
      exists (lglobal x nil). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }
  Qed.

  Lemma adjust_val_op2_local_var (pidx:pidxT) :
    forall v x op e (fid:fidT)
           (m:DomConMem.memT) (am:memT)
           (Hmem: m \abs am)
           (Hv : v ≠ 0%Z)
           (Hcon : SemCon.sem_expr
                     pidx (expr_op2 op (expr_value_of (lv_id (local x))) e)
                     m (vint v)),
      (m[(DomConValue.llocal pidx x) : DomConValue.locT]!
       : option DomConValue.valueT)
      \abs
      (adjust_val (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                              (lv_id (local x)) am) am) op
                  (sem_expr (DomPidx.pifid pidx) e am) : valueT).
  Proof.
    intros. inversion Hcon; subst; clear Hcon.
    { rewrite/adjust_val/adjust_inter/adjust_loc.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (local x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (local x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((inr (DomPidx.pifid pidx, x, length pidx.2), nil):DomConValue.locT);
        try assumption.
      inversion Heql. rewrite/t_loc/llocal.
      apply existsb_exists.
      exists ((inr (DomPidx.pifid pidx, x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (local x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (local x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((inr (DomPidx.pifid pidx, x, length pidx.2), nil):DomConValue.locT);
        try assumption.
      inversion Heql. rewrite/t_loc/llocal.
      apply existsb_exists.
      exists ((inr (DomPidx.pifid pidx, x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (local x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (local x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((inr (DomPidx.pifid pidx, x, length pidx.2), nil):DomConValue.locT);
        try assumption.
      inversion Heql. rewrite/t_loc/llocal.
      apply existsb_exists.
      exists ((inr (DomPidx.pifid pidx, x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (local x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (local x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((inr (DomPidx.pifid pidx, x, length pidx.2), nil):DomConValue.locT);
        try assumption.
      inversion Heql. rewrite/t_loc/llocal.
      apply existsb_exists.
      exists ((inr (DomPidx.pifid pidx, x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (local x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (local x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((inr (DomPidx.pifid pidx, x, length pidx.2), nil):DomConValue.locT);
        try assumption.
      inversion Heql. rewrite/t_loc/llocal.
      apply existsb_exists.
      exists ((inr (DomPidx.pifid pidx, x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { 
      rewrite/adjust_val.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inr x)) am) am).
      destruct v. destruct p.
      remember (sem_expr (DomPidx.pifid pidx) e am).
      destruct v. destruct p.
      inversion Hl; subst.
      inversion Hlv; subst.
      
      assert ((vint lz) \abs (i, l, a)) as Hlz. rewrite Heqv.
      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inr x)) am).
      destruct l1; last apply abs_value_top.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with m (DomConValue.llocal pidx x); auto.
      
      assert (DomConValue.llocal pidx x \abs Elt s) as Hx. rewrite Heql1.
      apply sem_expr_lv_proper2_ with m; auto.
      inversion Hx as [Hxs].
      rewrite SetFacts.elements_b in Hxs. apply existsb_exists in Hxs.
      destruct Hxs as [x0 [Hin Hxeq]]. rewrite is_compare_eqb in Hxeq.
      apply existsb_exists. exists x0; auto.

      assert ((vint rz) \abs (i0, l0, a0)) as Hrz.
      rewrite Heqv0.
      apply sem_expr_proper2_ with m; auto.

      rewrite/local.
      rewrite <- Heqv.

      rewrite Hlookup.
      rewrite/abs_relation/option_abs.
      rewrite/abs_relation/value_abs/itv_of/fst.
      rewrite/abs_relation/value_abs/itv_of/fst in Hlz,Hrz.
      rewrite/adjust_inter.
      
      apply adjust_inter_lt_sound with rz; auto.
    }

    { 
      rewrite/adjust_val.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inr x)) am) am).
      destruct v. destruct p.
      remember (sem_expr (DomPidx.pifid pidx) e am).
      destruct v. destruct p.
      inversion Hl; subst.
      inversion Hlv; subst.
      
      assert ((vint lz) \abs (i, l, a)) as Hlz. rewrite Heqv.
      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inr x)) am).
      destruct l1; last apply abs_value_top.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with m (DomConValue.llocal pidx x); auto.
      
      assert (DomConValue.llocal pidx x \abs Elt s) as Hx. rewrite Heql1.
      apply sem_expr_lv_proper2_ with m; auto.
      inversion Hx as [Hxs].
      rewrite SetFacts.elements_b in Hxs. apply existsb_exists in Hxs.
      destruct Hxs as [x0 [Hin Hxeq]]. rewrite is_compare_eqb in Hxeq.
      apply existsb_exists. exists x0; auto.

      assert ((vint rz) \abs (i0, l0, a0)) as Hrz.
      rewrite Heqv0.
      apply sem_expr_proper2_ with m; auto.

      rewrite/local.
      rewrite <- Heqv.

      rewrite Hlookup.
      rewrite/abs_relation/option_abs.
      rewrite/abs_relation/value_abs/itv_of/fst.
      rewrite/abs_relation/value_abs/itv_of/fst in Hlz,Hrz.
      rewrite/adjust_inter.
      
      apply adjust_inter_le_sound with rz; auto.
    }

    { 
      rewrite/adjust_val.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inr x)) am) am).
      destruct v. destruct p.
      remember (sem_expr (DomPidx.pifid pidx) e am).
      destruct v. destruct p.
      inversion Hl; subst.
      inversion Hlv; subst.
      
      assert ((vint lz) \abs (i, l, a)) as Hlz. rewrite Heqv.
      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inr x)) am).
      destruct l1; last apply abs_value_top.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with m (DomConValue.llocal pidx x); auto.
      
      assert (DomConValue.llocal pidx x \abs Elt s) as Hx. rewrite Heql1.
      apply sem_expr_lv_proper2_ with m; auto.
      inversion Hx as [Hxs].
      rewrite SetFacts.elements_b in Hxs. apply existsb_exists in Hxs.
      destruct Hxs as [x0 [Hin Hxeq]]. rewrite is_compare_eqb in Hxeq.
      apply existsb_exists. exists x0; auto.

      assert ((vint rz) \abs (i0, l0, a0)) as Hrz.
      rewrite Heqv0.
      apply sem_expr_proper2_ with m; auto.

      rewrite/local.
      rewrite <- Heqv.

      rewrite Hlookup.
      rewrite/abs_relation/option_abs.
      rewrite/abs_relation/value_abs/itv_of/fst.
      rewrite/abs_relation/value_abs/itv_of/fst in Hlz,Hrz.
      rewrite/adjust_inter.
      
      apply adjust_inter_gt_sound with rz; auto.
    }

    { 
      rewrite/adjust_val.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inr x)) am) am).
      destruct v. destruct p.
      remember (sem_expr (DomPidx.pifid pidx) e am).
      destruct v. destruct p.
      inversion Hl; subst.
      inversion Hlv; subst.
      
      assert ((vint lz) \abs (i, l, a)) as Hlz. rewrite Heqv.
      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inr x)) am).
      destruct l1; last apply abs_value_top.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with m (DomConValue.llocal pidx x); auto.
      
      assert (DomConValue.llocal pidx x \abs Elt s) as Hx. rewrite Heql1.
      apply sem_expr_lv_proper2_ with m; auto.
      inversion Hx as [Hxs].
      rewrite SetFacts.elements_b in Hxs. apply existsb_exists in Hxs.
      destruct Hxs as [x0 [Hin Hxeq]]. rewrite is_compare_eqb in Hxeq.
      apply existsb_exists. exists x0; auto.

      assert ((vint rz) \abs (i0, l0, a0)) as Hrz.
      rewrite Heqv0.
      apply sem_expr_proper2_ with m; auto.

      rewrite/local.
      rewrite <- Heqv.

      rewrite Hlookup.
      rewrite/abs_relation/option_abs.
      rewrite/abs_relation/value_abs/itv_of/fst.
      rewrite/abs_relation/value_abs/itv_of/fst in Hlz,Hrz.
      rewrite/adjust_inter.
      
      apply adjust_inter_ge_sound with rz; auto.
    }

    { 
      rewrite/adjust_val.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inr x)) am) am).
      destruct v. destruct p.
      remember (sem_expr (DomPidx.pifid pidx) e am).
      destruct v. destruct p.
      inversion Hl; subst.
      inversion Hlv; subst.
      
      assert ((vint lz) \abs (i, l, a)) as Hlz. rewrite Heqv.
      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inr x)) am).
      destruct l1; last apply abs_value_top.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with m (DomConValue.llocal pidx x); auto.
      
      assert (DomConValue.llocal pidx x \abs Elt s) as Hx. rewrite Heql1.
      apply sem_expr_lv_proper2_ with m; auto.
      inversion Hx as [Hxs].
      rewrite SetFacts.elements_b in Hxs. apply existsb_exists in Hxs.
      destruct Hxs as [x0 [Hin Hxeq]]. rewrite is_compare_eqb in Hxeq.
      apply existsb_exists. exists x0; auto.

      assert ((vint rz) \abs (i0, l0, a0)) as Hrz.
      rewrite Heqv0.
      apply sem_expr_proper2_ with m; auto.

      rewrite/local.
      rewrite <- Heqv.

      rewrite Hlookup.
      rewrite/abs_relation/option_abs.
      rewrite/abs_relation/value_abs/itv_of/fst.
      rewrite/abs_relation/value_abs/itv_of/fst in Hlz,Hrz.
      rewrite/adjust_inter.
      
      apply adjust_inter_eq_sound with rz; auto.
    }

    { 
      rewrite/adjust_val.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inr x)) am) am).
      destruct v. destruct p.
      remember (sem_expr (DomPidx.pifid pidx) e am).
      destruct v. destruct p.
      inversion Hl; subst.
      inversion Hlv; subst.
      
      assert ((vint lz) \abs (i, l, a)) as Hlz. rewrite Heqv.
      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inr x)) am).
      destruct l1; last apply abs_value_top.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with m (DomConValue.llocal pidx x); auto.
      
      assert (DomConValue.llocal pidx x \abs Elt s) as Hx. rewrite Heql1.
      apply sem_expr_lv_proper2_ with m; auto.
      inversion Hx as [Hxs].
      rewrite SetFacts.elements_b in Hxs. apply existsb_exists in Hxs.
      destruct Hxs as [x0 [Hin Hxeq]]. rewrite is_compare_eqb in Hxeq.
      apply existsb_exists. exists x0; auto.

      assert ((vint rz) \abs (i0, l0, a0)) as Hrz.
      rewrite Heqv0.
      apply sem_expr_proper2_ with m; auto.

      rewrite/local.
      rewrite <- Heqv.

      rewrite Hlookup.
      rewrite/abs_relation/option_abs.
      rewrite/abs_relation/value_abs/itv_of/fst.
      rewrite/abs_relation/value_abs/itv_of/fst in Hlz,Hrz.
      rewrite/adjust_inter.
      
      apply adjust_inter_neq_sound with rz; auto.
    }

    { rewrite/adjust_val/adjust_inter/adjust_loc/local.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inr x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inr x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((inr (DomPidx.pifid pidx, x, length pidx.2), nil):DomConValue.locT);
        try assumption.
      inversion Heql. rewrite/t_loc/llocal.
      apply existsb_exists.
      exists ((inr (DomPidx.pifid pidx, x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc/local.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inr x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inr x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((inr (DomPidx.pifid pidx, x, length pidx.2), nil):DomConValue.locT);
        try assumption.
      inversion Heql. rewrite/t_loc/llocal.
      apply existsb_exists.
      exists ((inr (DomPidx.pifid pidx, x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc/local.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inr x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inr x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((inr (DomPidx.pifid pidx, x, length pidx.2), nil):DomConValue.locT);
        try assumption.
      inversion Heql. rewrite/t_loc/llocal.
      apply existsb_exists.
      exists ((inr (DomPidx.pifid pidx, x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc/local.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inr x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inr x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((inr (DomPidx.pifid pidx, x, length pidx.2), nil):DomConValue.locT);
        try assumption.
      inversion Heql. rewrite/t_loc/llocal.
      apply existsb_exists.
      exists ((inr (DomPidx.pifid pidx, x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc/local.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inr x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inr x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((inr (DomPidx.pifid pidx, x, length pidx.2), nil):DomConValue.locT);
        try assumption.
      inversion Heql. rewrite/t_loc/llocal.
      apply existsb_exists.
      exists ((inr (DomPidx.pifid pidx, x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc/local.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inr x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inr x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((inr (DomPidx.pifid pidx, x, length pidx.2), nil):DomConValue.locT);
        try assumption.
      inversion Heql. rewrite/t_loc/llocal.
      apply existsb_exists.
      exists ((inr (DomPidx.pifid pidx, x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }

    { rewrite/adjust_val/adjust_inter/adjust_loc/local.
      remember (memory_lookup (sem_expr_lv (DomPidx.pifid pidx)
                                           (lv_id (inr x)) am) am)
      as v eqn:Hlookup.
      destruct v. destruct p. 
      remember (sem_expr (DomPidx.pifid pidx) e am) as v eqn:Hcon_e.
      destruct v. destruct p.
      symmetry in Hlookup. symmetry in Hcon_e.
      rewrite <- Hlookup. clear Hlookup Hcon_e i l a.

      inversion Hl; subst. inversion Hlv; subst. rewrite Hlookup. 
      clear Hl Hlv.

      rewrite/memory_lookup.
      remember (sem_expr_lv (DomPidx.pifid pidx) (lv_id (inr x)) am).
      destruct l; last apply abs_option_value_top.

      rewrite/abs_relation/option_abs.
      rewrite SetInterface.fold_1.
      apply sound_memory_lookup_ with 
        m ((inr (DomPidx.pifid pidx, x, length pidx.2), nil):DomConValue.locT);
        try assumption.
      inversion Heql. rewrite/t_loc/llocal.
      apply existsb_exists.
      exists ((inr (DomPidx.pifid pidx, x), nil)). split.
      rewrite elements_singleton. 
      left. reflexivity.
      apply/eqbP. reflexivity. }
  Qed.

  Lemma contradict_by_none_sem_expr_global :
    forall (pidx:pidxT) con_l (i:idT) (m:DomConMem.memT)
           (e:exprT) (v:Z) (op:op2)
           (Heqcon_l : con_l = ((DomConValue.lglobal i) : DomConValue.locT))
           (Heqvconl : None = (m) [con_l]!)
           (Hv : v ≠ 0%Z)
           (Hcon : SemCon.sem_expr
                     pidx (expr_op2 op (expr_value_of (lv_id (global i))) e) m
                     (vint v)),
      False.
  Proof.
    intros. inversion Hcon; subst;
    inversion Hl; inversion Hlv; subst;
    rewrite/DomConValue.lglobal in Hlookup;
    rewrite Hlookup in Heqvconl; inversion Heqvconl.
  Qed.

  Lemma contradict_by_none_sem_expr_local : forall pidx con_l i m e v op 
    (Heqcon_l : con_l = ((DomConValue.llocal pidx i) : DomConValue.locT))
    (Heqvconl : None = (m) [con_l ]!)
    (Hv : v ≠ 0%Z)
    (Hcon : SemCon.sem_expr pidx
           (expr_op2 op (expr_value_of (lv_id (inr i))) e) m
           (vint v)),
    False.
  Proof.
    intros. inversion Hcon; subst;
    inversion Hl; inversion Hlv; subst;
    rewrite/DomConValue.llocal in Hlookup;
    rewrite Hlookup in Heqvconl; inversion Heqvconl.
  Qed.

  Lemma pruning_sound (pidx:pidxT) : forall m am e 
    (Hpidx: pidx_consistent prog (DomPidx.pifid pidx) pidx.2)
    (Hprog: acc_correct_callfroms_prog prog acc)
    (Hstar: acc_correct_callfroms_star acc)
    (Hmem_wf:DomConMem.mem_wf pidx m)
    (Hmem: m \abs am)
    (Hcon: assume pidx e m),
    m \abs memory_prune acc (DomPidx.pifid pidx) e am.
  Proof.
    intros.
    rewrite/memory_prune.
    remember (DomType.leb (itv_of (sem_expr (DomPidx.pifid pidx) e am))
      (Some
        (DomItv.itv'_z_z 0 0
          (memory_prune_obligation_1 (DomPidx.pifid pidx) e am)))).
    symmetry in Heqb. destruct b.
    {
      move/lebP:Heqb=>Heqb.
      inversion Hcon; subst; clear Hcon.
      remember (sem_expr_proper2_ prog acc Hsem am Hmem) as Hprop.
      clear HeqHprop.
      unfold abs_relation,value_abs in Hprop.
      rewrite/vint in Hprop.
      rewrite -> Heqb in Hprop.
      rewrite/abs_relation/Z_itvT_abs/abs_itvT in Hprop.
      inversion Hprop.
      elim Hnonzero.
      rewrite/eq/eq_/DomConValue.valueT_setoid/sum_setoid/DomBasic.sum_eq.
      rewrite/eq/eq_/Z_setoid; omega.
    }
    remember (remove_uop e). destruct o; auto.
    symmetry in Heqo.

    rewrite/prune_bop. destruct e0; auto.
    destruct e0_1; auto.
    destruct l; auto.
    remember (is_strong acc (var_loc (DomPidx.pifid pidx) x nil)).
    destruct b; auto.
    symmetry in Heqb0. 
    inversion Hcon; subst; clear Hcon.
    rewrite/adjust.

    remember (var_loc (DomPidx.pifid pidx) x) as l.
    rewrite/var_loc in Heql. destruct x.
    { (* l is global *)
      rewrite/lglobal in Heql.
      remember ((inl (inr i), nil) : DomConValue.locT) as con_l.

      assert (Hcon_l:
        m ≡ m[con_l <- mlookup (Elt:=option DomConValue.valueT) con_l m]!).
      { intro. simpl. remember (eqb z con_l). destruct b; last reflexivity.
        symmetry in Heqb1. move/eqbP:Heqb1 => Hz. 
        unfold proj1_sig. destruct m. rewrite Hz. reflexivity. }
      
      rewrite Hcon_l. 
      remember ((m) [con_l ]!) as vconl. destruct vconl.
      {
        eapply memory_update_sound with prog pidx; auto.
        {
          rewrite Heqcon_l. 
          rewrite/sem_expr_lv/var_loc/lglobal/abs_relation/loc_abs/t_loc.
          pose (ssingleton_spec ((inl (inr i), nil)) ((inl (inr i), nil))).
          destruct o. destruct H. apply H0.
          destruct H. elim H. reflexivity.
        }
        { assert (exists v', v' ≠ 0%Z
          ∧ SemCon.sem_expr pidx
          (expr_op2 op (expr_value_of (lv_id (inl i))) e0_2) m
          (vint v'))
            as Hnew_con.
          apply remove_uop_preserve_con_sem with e v; assumption.
          clear Heqo Hsem Hnonzero.
          destruct Hnew_con as [v' [Hv' Hcon]].

          subst. rewrite Heqvconl.
          apply adjust_val_op2_global_var with v'; auto.
        }
        { rewrite <- Hcon_l. auto. }
      }
      { assert (exists v', v' ≠ 0%Z
          ∧ SemCon.sem_expr pidx
          (expr_op2 op (expr_value_of (lv_id (inl i))) e0_2) m
          (vint v'))
          as Hnew_con.
        apply remove_uop_preserve_con_sem with e v; assumption.
        clear Heqo Hsem Hnonzero.
        destruct Hnew_con as [v' [Hv' Hcon]].
        elimtype False.
        eapply contradict_by_none_sem_expr_global; eauto.
      }
    }

    { (* l is local *)
      rewrite/llocal in Heql.
      remember ((inr (DomPidx.pifid pidx, i, length pidx.2), nil)
                : DomConValue.locT)
        as con_l.
      assert (Hcon_l:
        m ≡ m[con_l <- mlookup (Elt:=option DomConValue.valueT) con_l m]!).
      { intro. simpl. remember (eqb z con_l). destruct b; last reflexivity.
        symmetry in Heqb1. move/eqbP:Heqb1 => Hz. 
        unfold proj1_sig. destruct m. rewrite Hz. reflexivity. }
      
      rewrite Hcon_l. 
      remember ((m) [con_l ]!) as vconl. destruct vconl.
      {
        apply memory_update_sound with prog pidx; auto.
        {
          rewrite Heqcon_l. 
          rewrite/sem_expr_lv/var_loc/lglobal/abs_relation/loc_abs/t_loc.
          pose (ssingleton_spec ((inr (DomPidx.pifid pidx, i), nil))
                                ((inr (DomPidx.pifid pidx, i), nil))).
          destruct o. destruct H. apply H0.
          destruct H. elim H. reflexivity.
        }
        { assert (exists v', v' ≠ 0%Z
          ∧ SemCon.sem_expr pidx
          (expr_op2 op (expr_value_of (lv_id (inr i))) e0_2) m
          (vint v'))
        as Hnew_con.
          apply remove_uop_preserve_con_sem with e v; assumption.
          clear Heqo Hsem Hnonzero.
          destruct Hnew_con as [v' [Hv' Hcon]].

          subst. rewrite Heqvconl.
          apply adjust_val_op2_local_var with v'; auto.
        }
        { rewrite <- Hcon_l. auto. }
      }
      { assert (exists v', v' ≠ 0%Z
          ∧ SemCon.sem_expr pidx
          (expr_op2 op (expr_value_of (lv_id (inr i))) e0_2) m
          (vint v'))
      as Hnew_con.
        apply remove_uop_preserve_con_sem with e v; assumption.
        clear Heqo Hsem Hnonzero.
        destruct Hnew_con as [v' [Hv' Hcon]].
        elimtype False.
        eapply contradict_by_none_sem_expr_local; eauto.
      }
    }
  Qed.

End Correctness.
