Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.
Require Import ZArith.

Require Import DomItv.
Require Import DomType.
Require Import DomUtil.
Require Import DomBasic.
Require Import DomZ.
Require Import CorBasic.
Require Import CorAbs.

Open Scope Z.

Inductive abs_itvT' (n:Z) : itvT' -> Prop :=
| abs_itvT'_top : abs_itvT' n itv'_top
| abs_itvT'_inf_z : ∀u, n <= u -> abs_itvT' n (itv'_inf_z u)
| abs_itvT'_z_inf : ∀l, l <= n -> abs_itvT' n (itv'_z_inf l)
| abs_itvT'_z_z : ∀l u (H:l <= u), 
                    l <= n -> n <= u -> abs_itvT' n (itv'_z_z l u H).

Definition abs_itvT (n:Z) (i:itvT) :=
  if i is Some i'
  then abs_itvT' n i'
  else False.

Program Instance Z_itvT_abs : abs Z itvT := {
  abs_relation := abs_itvT
}.
Next Obligation.
  move=> x y ->.
  move=> z w; case: w => [w |]; case: z => [z |] Hzw //= Hyz.
  move: w Hzw; inversion Hyz; subst; clear Hyz; move=> [|a|a|a b] Hle; constructor; try done.
  { by rewrite -> H. }
  { by rewrite <- H. }
  { by rewrite -> H1. }
  { by rewrite <- H0. }
  { case: Hzw => Hal Hub.
    by rewrite <- H0.
  }
  { case: Hzw => Hal Hub.
    by rewrite -> H1.
  }
Qed.

Lemma abs_top (z:Z) : z \abs (top:itvT).
Proof. by constructor. Qed.
  
Lemma op_not'_prop : ∀n gn, 
  abs_itvT' n gn -> abs_itvT' (con_op_not n) (op_not' gn).
Proof.
  intros n gn Hn; destruct Hn; unfold op_not'.
  { destif; constructor. inversion C. inversion C.
    unfold con_op_not. destif; omega.
    unfold con_op_not. destif; omega. }
  { repeat destif; constructor. inversion C. inversion C. 
    unfold con_op_not; destif; omega.
    unfold con_op_not; destif; omega.
    unfold con_op_not; destif; omega.
    unfold con_op_not; destif; try omega.
    move/lebP:C0=>C0. elim C0. move/eqbP:C1=>C1.
    unfold le; simpl.
    apply le_trans with n; auto. rewrite C1. apply eq_le; reflexivity. }
  { repeat destif; constructor. inversion C. 
    unfold con_op_not; destif; omega.
    unfold con_op_not; destif; omega.
    unfold con_op_not; destif; omega.
    unfold con_op_not; destif; omega.
    unfold con_op_not; destif; try omega.
    move/lebP:C0=>C0. elim C0. move/eqbP:C1=>C1.
    unfold le; simpl.
    apply le_trans with n; auto. rewrite C1. apply eq_le; reflexivity. }
  { repeat destif; constructor. 
    unfold con_op_not; destif; try omega.
    move/eqbP:C=>C. inversion C. rewrite <- H3 in H1. rewrite <- H2 in H0.
    move/eqbP:C0=>C0. elim C0. apply le_antisym; auto. 
    unfold con_op_not; destif; try omega.
    unfold con_op_not; destif; try omega.
    unfold con_op_not; destif; try omega.
    unfold con_op_not; destif; try omega.
    unfold con_op_not; destif; try omega.
    move/lebP:C0=>C0. 
    move/eqbP:C1=>C1. rewrite -> C1 in H0. rewrite -> C1 in H1.
    elim C0. constructor; auto. }
Qed.

Lemma op_bnot'_prop : ∀n gn, 
  abs_itvT' n gn -> abs_itvT' (con_op_bnot n) (op_bnot' gn).
Proof. intros. unfold op_bnot'. constructor. Qed.

Lemma op_add'_prop : ∀n m gn gm, 
                   abs_itvT' n gn -> abs_itvT' m gm -> abs_itvT' (n+m) (op_add' gn gm).
Proof. induction 1; intros Hm; inversion Hm; constructor; omega. Qed.

Lemma op_sub'_prop : ∀n m gn gm, 
                   abs_itvT' n gn -> abs_itvT' m gm -> abs_itvT' (n-m) (op_sub' gn gm).
Proof. induction 1; intros Hm; inversion Hm; constructor; omega. Qed.

Lemma op_mul'_prop : ∀n m gn gm, 
                   abs_itvT' n gn -> abs_itvT' m gm -> abs_itvT' (n*m) (op_mul' gn gm).
Proof with leauto. 
  induction 1; intros Hm; inversion Hm; try constructor.

  unfold op_mul'; destif; last constructor.
  move/andP:C => [Cl Cu]; move/eqbP:Cu; move/eqbP:Cl; move => Cl Cu; repeat inv_eq'; subst.
  assert (m=0); first omega; subst.
  replace (n * 0) with 0; last omega.
  constructor; omega.

  unfold op_mul'; destif; constructor.
  move/andP:C => [Cl Cu]; move/lebP:Cu; move/lebP:Cl; move => Cl Cu; rewrite/le in Cl Cu; simpl in Cl,Cu.
  apply Zle_trans with (n * u0)...
  assert (n <= 0)...

  unfold op_mul'; destif; constructor.
  move/andP:C => [Cl Cu]; move/lebP:Cu; move/lebP:Cl; move => Cl Cu; rewrite/le in Cl Cu; simpl in Cl,Cu.
  apply Zle_trans with (n * l)...
  assert (n <= 0)...

  unfold op_mul'; destif.
  {
    move/andP:C => [Cl Cu]; move/eqbP:Cu; move/eqbP:Cl; move => Cl Cu; repeat inv_eq'; subst.
    assert (m=0); first omega; subst.
    replace (n * 0) with 0; last omega.
    constructor; omega.
  }
  {
    destif.
    move/lebP:C0 => C0; rewrite/le in C0; simpl in C0; constructor; destruct (Z_le_dec 0 n).
    {
      apply Zle_trans with (n * u0)...
      apply Zle_trans with (u * u0)...
      assert (0 <= u0)...
    }
    {
      apply Zle_trans with (n * l)...
      apply Zle_trans with (u * l)...
    }
    destif.
    move/lebP:C0 => C0; rewrite/le in C0; simpl in C0.
    move/lebP:C1 => C1; rewrite/le in C1; simpl in C1; constructor; 
      destruct (Z_le_dec 0 n).
    {
      apply Zle_trans with (n * l)...
      apply Zle_trans with (u * l)...
    }
    {
      apply Zle_trans with (n * u0)...
      apply Zle_trans with (u * u0)...
    }    
    constructor.
  }
  
  unfold op_mul'; destif; constructor. 
  move/andP:C => [Cl Cu]; move/lebP:Cu; move/lebP:Cl; move => Cl Cu; rewrite/le in Cl Cu; simpl in Cl,Cu.
  apply Zle_trans with (n * u)...
  assert (0 <= n)...

  unfold op_mul'; destif; constructor.
  move/andP:C => [Cl Cu]; move/lebP:Cu; move/lebP:Cl; move => Cl Cu; rewrite/le in Cl Cu; simpl in Cl,Cu.
  apply Zle_trans with (l * m)...
  assert (0 <= m)...

  unfold op_mul'; repeat destif.
  {
    move/andP:C => [Cl Cu]; move/eqbP:Cu; move/eqbP:Cl; move => Cl Cu; repeat inv_eq'; subst.
    assert (m=0); first omega; subst.
    replace (n * 0) with 0; last omega.
    constructor; omega.
  }
  { 
    move/lebP:C0 => C0; rewrite/le in C0; simpl in C0. constructor.
    destruct (Z_le_dec 0 n).
    {
      apply Zle_trans with (l * l0)...
      apply Zle_trans with (n * l0)...
    }
    {
      apply Zle_trans with (l * u)...
      apply Zle_trans with (n * u)...
      assert (0 <= u)...
    }
  }
  {
    move/lebP:C0 => Cl; move/lebP:C1 => Cu; rewrite/le in Cl Cu; simpl in Cl,Cu.
    constructor.
    destruct (Z_le_dec 0 n).
    {
      apply Zle_trans with (n * u)...
      apply Zle_trans with (l * u)...
    }
    {
      apply Zle_trans with (n * l0)...
      apply Zle_trans with (l * l0)...
    }
  }

  constructor.

  unfold op_mul'; destif; last constructor;
  move/andP:C => [Cl Cu]; move/eqbP:Cu; move/eqbP:Cl; move => Cl Cu;
    repeat inv_eq'; subst.
  assert (n=0); first omega; subst.
  replace (0 * m) with 0; constructor; omega.

  unfold op_mul'; repeat destif.
  {
    move/andP:C => [Cl Cu]; move/eqbP:Cu; move/eqbP:Cl; move => Cl Cu; 
      repeat inv_eq'; subst.
    assert (n=0); first omega; subst.
    replace (0 * m) with 0; last omega.
    constructor; omega.
  }
  {
    move/lebP:C0 => C0; rewrite/le in C0; simpl in C0. constructor.
    destruct (Z_le_dec 0 m).
    {
      apply Zle_trans with (u * m)...
      apply Zle_trans with (u * u0)...
      assert (0 <= u)...
    }
    {
      apply Zle_trans with (l * m)...
      apply Zle_trans with (l * u0)...
    }
  }
  {
    move/lebP:C0 => Cl; move/lebP:C1 => Cu; rewrite/le in Cl Cu; simpl in Cl,Cu.
    constructor.
    destruct (Z_le_dec 0 m).
    {
      apply Zle_trans with (l * m)...
      apply Zle_trans with (l * u0)...
    }
    {
      apply Zle_trans with (u * m)...
      apply Zle_trans with (u * u0)...
    }
  }

  constructor.

  unfold op_mul'; repeat destif.
  {
    move/andP:C => [Cl Cu]; move/eqbP:Cu; move/eqbP:Cl; move => Cl Cu; 
      repeat inv_eq'; subst.
    assert (n=0); first omega; subst.
    replace (0 * m) with 0; last omega.
    constructor; omega.
  }
  {
    move/lebP:C0 => C0; rewrite/le in C0; simpl in C0.
    constructor.
    destruct (Z_le_dec 0 m).
    {
      apply Zle_trans with (l * l0)...
      apply Zle_trans with (l * m)...
    }
    {
      apply Zle_trans with (u * l0)...
      apply Zle_trans with (u * m)...
      assert (0 <= u)...
    }
  }
  {
    move/lebP:C0 => Cl; move/lebP:C1 => Cu; rewrite/le in Cl Cu; simpl in Cl,Cu.
    constructor.
    destruct (Z_le_dec 0 m).
    {
      apply Zle_trans with (u * m)...
      apply Zle_trans with (u * l0)...
    }
    {
      apply Zle_trans with (l * m)...
      apply Zle_trans with (l * l0)...
    }
  }

  constructor.
  
  destruct (Z_le_dec 0 n).
  {
    apply Zle_trans with (n * l0)...
    destruct (Z_le_dec 0 l0).
    {
      apply Zle_trans with (Z.min (l * l0) (l * u0))...
      apply Zle_trans with (l * l0)...
    }
    {
      apply Zle_trans with (Z.min (u * l0) (u * u0))...
      apply Zle_trans with (u * l0)...
    }
  }
  {
    apply Zle_trans with (n * u0)...
    destruct (Z_le_dec 0 u0).
    {
      apply Zle_trans with (Z.min (l * l0) (l * u0))...
      apply Zle_trans with (l * u0)...
    }
    {
      apply Zle_trans with (Z.min (u * l0) (u * u0))...
      apply Zle_trans with (u * u0)...
    }
  }
  
  destruct (Z_le_dec 0 n).
  {
    apply Zle_trans with (n * u0)...
    destruct (Z_le_dec 0 u0).
    {
      apply Zle_trans with (Z.max (u * l0) (u * u0))...
      apply Zle_trans with (u * u0)...
    }
    {
      apply Zle_trans with (Z.max (l * l0) (l * u0))...
      apply Zle_trans with (l * u0)...
    }
  }
  {
    apply Zle_trans with (n * l0)...
    destruct (Z_le_dec 0 l0).
    {
      apply Zle_trans with (Z.max (u * l0) (u * u0))...
      apply Zle_trans with (u * l0)...
    }
    {
      apply Zle_trans with (Z.max (l * l0) (l * u0))...
      apply Zle_trans with (l * l0)...
    }
  }
Qed.

Lemma op_div'_prop : ∀n m gn gm, 
                   abs_itvT' n gn -> abs_itvT' m gm -> abs_itvT' (Cdiv n m) (op_div' gn gm).
Proof with leauto. 
  induction 1; intros Hm; inversion Hm; try constructor.
  
  unfold op_div'; destif; move/lebP:C => C; rewrite/le in C; simpl in C; constructor.
  destruct (Z_le_dec 0 n).
  {
    apply Zle_trans with (Cdiv u u0)...
    apply Zle_trans with (Cdiv n u0)...
  }
  {
    apply Zle_trans with 0...
    rewrite <- Z.opp_involutive, <- Cdiv_opp, <- Cdiv_opp_l. 
    apply Cdiv_pos_pos; omega.
  }
  
  unfold op_div'; destif; move/lebP:C => C; rewrite/le in C; simpl in C; constructor.
  destruct (Z_le_dec 0 n).
  {
    apply Zle_trans with (Cdiv u l)...
    apply Zle_trans with (Cdiv n l)...
  }
  {
    apply Zle_trans with 0...
    apply Z.opp_le_mono; rewrite <- Cdiv_opp_l. 
    apply Cdiv_pos_pos; omega.
  }
  
  unfold op_div'; repeat destif; try constructor.
  unfold ltb in C0; move/andP:C0 => C0; destruct C0;
                                    move/lebP:H4 => H4; move/negP:H5 => H5.
  rewrite/le/= in H4.
  destruct (Z_le_dec 0 n).
  {
    assert (0 ≠ l); first (intro; elim H5; apply/eqbP; auto).
    assert (0 < l); first omega.
    apply Zle_trans with (Cdiv n l)...
    apply Zle_trans with (Cdiv u l)...
  }
  {
    assert (0 ≠ l); first (intro; elim H5; apply/eqbP; auto).
    assert (0 < m); first omega.
    assert (0 <= u0); first omega.
    apply Zle_trans with (Cdiv n u0)...
    apply Zle_trans with (Cdiv u u0)...
  }
  unfold ltb in C0.
  destruct (Z_le_dec u0 0).
  {
    assert (u0 ≠ 0); first intro.
    assert (Hleb_zu:leb 0 u0); first (subst; auto).
    rewrite Hleb_zu in C; rewrite Bool.andb_true_r in C.
    move/lebP:C => /= C; rewrite/le/= in C.
    assert (Hleb_zl:leb 0 l); simpl; first omega.
    rewrite Hleb_zl in C0; rewrite Bool.andb_true_l in C0.
    move/negP:C0 => C0; elim C0; intro.
    move/eqbP:H5 => /= H5; subst; omega.
    assert (u0 < 0); first omega.

    destruct (Z_le_dec 0 n).
    {
      apply Zle_trans with (Cdiv u u0)...
      apply Zle_trans with (Cdiv n u0)...
    }
    {
      assert (n <= 0); first omega.
      assert (m < 0); first omega.
      assert (l <= 0); first omega.
      apply Zle_trans with (Cdiv u l)...
      apply Zle_trans with (Cdiv n l)...
    }
  }
  {
    assert (Hzu: leb 0 u0); first (apply/lebP; rewrite/le/=; omega).
    rewrite Hzu in C; rewrite Bool.andb_true_r in C;
    move/lebP:C => C; rewrite/le/= in C.
    assert (Hzl: leb 0 l); first (apply/lebP; rewrite/le/=; omega).
    rewrite Hzl in C0; rewrite Bool.andb_true_l in C0.
    move/negP:C0 => C0; elim C0; intro; move/eqbP:H4 => H4; rewrite/eq/= in H4; omega.
  }

  unfold op_div'; destif; constructor.
  apply leb_false_imply_lt in C.
  destruct (Z_le_dec 0 n).
  {
    apply Zle_trans with 0...
    apply Z.opp_le_mono; rewrite <- Cdiv_opp; apply Cdiv_pos_pos; omega.
  }
  {
    assert (u <= 0); first omega.
    apply Zle_trans with (Cdiv n u)...
    apply Zle_trans with (Cdiv l u)...
  }

  unfold op_div'; destif; constructor.
  apply leb_false_imply_lt in C.
  destruct (Z_le_dec 0 n).
  {
    apply Zle_trans with 0...
    apply Cdiv_pos_pos; omega.
  }
  {
    assert (0 <= l0); first omega.
    apply Zle_trans with (Cdiv l l0)...
    apply Zle_trans with (Cdiv n l0)...
  }

  unfold op_div'; destif; first constructor; destif; constructor.
  {
    apply ltb_true_imply_lt in C0.
    destruct (Z_le_dec 0 n).
    {
      assert (0 <= u); first omega.
      assert (0 < m); first omega.
      apply Zle_trans with (Cdiv l u)...
      apply Zle_trans with (Cdiv n u)...
    }      
    {
      assert (0 <= l0); first omega.
      apply Zle_trans with (Cdiv l l0)...
      apply Zle_trans with (Cdiv n l0)...
    }      
  }
  {
    apply ltb_false_imply_le in C0.
    apply Bool.andb_false_iff in C; elim C; intros Hleb_false.
    assert (Hleb_lz:leb l0 0); first (apply/lebP; rewrite/le/=; omega).
    rewrite Hleb_lz in Hleb_false; inversion Hleb_false.
    apply leb_false_imply_lt in Hleb_false.
    destruct (Z_le_dec 0 n).
    {
      assert (m < 0); first omega.
      apply Zle_trans with (Cdiv n l0)...
      apply Zle_trans with (Cdiv l l0)...
    }
    {
      assert (n <= 0); first omega.
      assert (u <= 0); first omega.
      apply Zle_trans with (Cdiv n u)...
      apply Zle_trans with (Cdiv l u)...
    }        
  }

  unfold op_div'; destif; constructor.
  {
    apply leb_false_imply_lt in C.
    destruct (Z_le_dec 0 n).
    {
      assert (u0 <= 0); first omega.
      apply Zle_trans with (Cdiv n u0)...
      apply Zle_trans with (Cdiv u u0)...
    }
    {
      assert (n <= 0); first omega.
      apply Zle_trans with 0...
      rewrite <- Z.opp_involutive, <- Cdiv_opp, <- Cdiv_opp_l.
      apply Cdiv_pos_pos; omega.
    }
  }
  {
    apply leb_false_imply_lt in C.
    destruct (Z_le_dec 0 n).
    {
      apply Zle_trans with 0...
      rewrite Z.opp_le_mono; rewrite <- Cdiv_opp.
      apply Cdiv_pos_pos; omega.
    }
    {
      assert (u0 <= 0); first omega.
      apply Zle_trans with (Cdiv n u0)...
      apply Zle_trans with (Cdiv l u0)...
    }
  }
  
  unfold op_div'; destif; constructor.
  {
    apply leb_false_imply_lt in C.
    destruct (Z_le_dec 0 n).
    {
      apply Zle_trans with 0...
      apply Cdiv_pos_pos; omega.
    }
    {
      assert (n <= 0); first omega.
      assert (0 <= l0); first omega.
      apply Zle_trans with (Cdiv l l0)...
      apply Zle_trans with (Cdiv n l0)...
    }
  }
  {
    apply leb_false_imply_lt in C.
    destruct (Z_le_dec 0 n).
    {
      assert (0 <= l0); first omega.
      apply Zle_trans with (Cdiv n l0)...
      apply Zle_trans with (Cdiv u l0)...
    }
    {
      assert (n <= 0); first omega.
      apply Zle_trans with 0...
      apply Z.opp_le_mono; rewrite <- Cdiv_opp_l.
      apply Cdiv_pos_pos; omega.
    }
  }      

  unfold op_div'; destif; first constructor.
  apply Bool.andb_false_iff in C; elim C; clear C; intro C;
  apply leb_false_imply_lt in C.
  {
    destruct (Z_le_dec 0 n).
    {
      constructor.
      {
        apply Zle_trans with (Z.min (Cdiv l l0) (Cdiv l u0))...
        apply Zle_trans with (Cdiv l u0)...
        assert (0 <= u0); first omega.
        assert (0 < m); first omega.
        apply Zle_trans with (Cdiv n u0)...
      }
      {
        apply Zle_trans with (Z.max (Cdiv u l0) (Cdiv u u0))...
        apply Zle_trans with (Cdiv u l0)...
        assert (0 <= l0); first omega.
        apply Zle_trans with (Cdiv n l0)...
      }          
    }
    {
      assert (n <= 0); first omega.
      constructor.
      {
        apply Zle_trans with (Z.min (Cdiv l l0) (Cdiv l u0))...
        apply Zle_trans with (Cdiv l l0)...
        assert (0 <= l0); first omega.
        apply Zle_trans with (Cdiv n l0)...
      }
      {
        apply Zle_trans with (Z.max (Cdiv u l0) (Cdiv u u0))...
        apply Zle_trans with (Cdiv u u0)...
        assert (0 <= u0); first omega.
        assert (0 < m); first omega.
        apply Zle_trans with (Cdiv n u0)...
      }          
    }
  }          
  {
    destruct (Z_le_dec 0 n).
    {
      constructor.
      {
        apply Zle_trans with (Z.min (Cdiv u l0) (Cdiv u u0))...
        apply Zle_trans with (Cdiv u u0)...
        assert (u0 <= 0); first omega.
        assert (m < 0); first omega.
        apply Zle_trans with (Cdiv n u0)...
      }
      {
        apply Zle_trans with (Z.max (Cdiv l l0) (Cdiv l u0))...
        apply Zle_trans with (Cdiv l l0)...
        assert (m < 0); first omega.
        assert (l0 <= 0); first omega.
        apply Zle_trans with (Cdiv n l0)...
      }
    }
    {
      assert (n <= 0); first omega.
      constructor.
      {
        apply Zle_trans with (Z.min (Cdiv u l0) (Cdiv u u0))...
        apply Zle_trans with (Cdiv u l0)...
        assert (l0 <= 0); first omega.
        assert (m < 0); first omega.
        apply Zle_trans with (Cdiv n l0)...
      }
      {
        apply Zle_trans with (Z.max (Cdiv l l0) (Cdiv l u0))...
        apply Zle_trans with (Cdiv l u0)...
        assert (m < 0); first omega.
        assert (u0 <= 0); first omega.
        apply Zle_trans with (Cdiv n u0)...
      }
    }
  }
Qed.

Ltac cmod_sound :=
  repeat 
  match goal with
    | [Hlu: 0 < ?l ∨ ?u < 0, Hlm: ?l <= ?m, Hmu: ?m <= ?u |- _] =>
      let Hm := fresh "Hm" in
      assert (Hm:m ≠ 0); first (destruct Hlu; omega); clear Hlu
    | [Hm: ?m ≠ 0, Hlm: ?l <= ?m, Hmu: ?m <= ?u |-
      - (Z.max (Z.abs ?l) (Z.abs ?u) - 1) <= Cmod ?n ?m] =>
      pose (cmod_le_m l u n m Hm Hlm Hmu); omega
    | [Hm: ?m ≠ 0, Hlm: ?l <= ?m, Hmu: ?m <= ?u |-
      - (Z.max (Z.abs ?l) (Z.abs ?u) - 1) <= - Cmod ?n ?m] =>
      rewrite cmod_opp
    | [Hm: ?m ≠ 0, Hlm: ?l <= ?m, Hmu: ?m <= ?u |- 
      Cmod ?n ?m <= Z.max (Z.abs ?l) (Z.abs ?u) - 1] =>
      pose (cmod_le_m l u n m Hm Hlm Hmu); omega
    | [Hm: ?m ≠ 0, Hlm: ?l <= ?m, Hmu: ?m <= ?u |- 
      - Cmod ?n ?m <= Z.max (Z.abs ?l) (Z.abs ?u) - 1] =>
      rewrite cmod_opp
    | [ |- Cmod ?n ?m <= 0] => apply cmod_neg; omega
    | [ |- 0 <= Cmod ?n ?m] => apply cmod_pos; omega
    | [Hnu: ?n <= ?u |- Cmod ?n ?m <= ?u] => apply cmod_n_u; omega
    | [Hnu: ?l <= ?n |- ?l <= Cmod ?n ?m] => apply cmod_l_n; omega
    | [ |- _ <= Z.min _ _] => apply Z.min_glb
    | [ |- - Z.min _ _ <= _] => apply Z.opp_le_mono; rewrite Z.opp_involutive
    | [ |- context [- - _]] => rewrite Z.opp_involutive
    | [ |- - Cmod ?n ?m <= - ?l] => apply Z.opp_le_mono
    | [ |- -?u <= - Cmod ?n ?m] => apply Z.opp_le_mono
  end.

Lemma op_mod'_prop : ∀n m gn gm, 
  abs_itvT' n gn -> abs_itvT' m gm -> abs_itvT' (Cmod n m) (op_mod' gn gm).
Proof with cmod_sound.
  intros n m gn gm Hn Hm; destruct Hn; destruct Hm; 
    unfold op_mod'; try constructor.

  { destruct (l_z_u_dec l u); constructor... }

  {
    destruct (Z_le_dec 0 u0); first constructor.
    destruct (Z_le_dec u 0); constructor...
  }

  {
    destruct (Z_le_dec l 0); first constructor.
    destruct (Z_le_dec u 0); constructor...
  }

  {
    destruct (l_z_u_dec l u0); first constructor.
    destruct (Z_le_dec u 0); constructor...
  }

  {
    destruct (Z_le_dec 0 u); first constructor.
    destruct (Z_le_dec 0 l); constructor...
  }

  {
    destruct (Z_le_dec l0 0); first constructor.
    destruct (Z_le_dec 0 l); constructor...
  }

  {
    destruct (l_z_u_dec l0 u); first constructor.
    destruct (Z_le_dec 0 l); constructor...
  }
  
  {
    destruct (Z_le_dec 0 u0); first constructor.
    destruct (Z_le_dec 0 l); first constructor...
    destruct (Z_le_dec u 0); constructor...
  }

  {
    destruct (Z_le_dec l0 0); first constructor.
    destruct (Z_le_dec 0 l); first constructor...
    destruct (Z_le_dec u 0); constructor...
  }

  {
    destruct (l_z_u_dec l0 u0); first constructor.
    destruct (Z_le_dec 0 l); first constructor...
    destruct (Z_le_dec u 0); constructor...
  }
Qed.

Lemma op_lt'_prop : ∀n m gn gm, 
  abs_itvT' n gn -> abs_itvT' m gm -> abs_itvT' (con_op_lt n m) (op_lt' gn gm).
Proof.
  intros n m gn gm Hn Hm; destruct Hn; destruct Hm;
    unfold op_lt'; try constructor.
  { unfold con_op_lt. destif; omega. }
  { unfold con_op_lt. destif; omega. }
  { unfold con_op_lt. destif; omega. }
  { unfold con_op_lt. destif; omega. }
  { unfold con_op_lt. destif; omega. }
  { unfold con_op_lt. destif; omega. }
  { unfold con_op_lt. destif; omega. }
  { unfold con_op_lt. destif; omega. }
  { unfold con_op_lt. destif; omega. }
  { unfold con_op_lt. destif; omega. }
  { unfold con_op_lt. destif; omega. }
  { unfold con_op_lt. destif; omega. }
  { destif; try constructor.
    unfold con_op_lt; destif; try omega.
    move/ltP:C => [C C']. move/ltP:C0 => C0.
    elim C0; split. 
    apply le_trans with l; auto. apply le_trans with u; auto.
    intro. elim C'. apply le_antisym; auto. apply le_trans with m; auto.
    rewrite <- H1; auto.
    unfold con_op_lt; destif; try omega. 
    { unfold con_op_lt. destif; omega. }
    { unfold con_op_lt. destif; omega. }
  }
  { destif; try constructor.
    unfold con_op_lt; destif; try omega.
    move/ltP:C => [C C']. move/ltP:C0 => C0.
    elim C0; split. 
    apply le_trans with l; auto. apply le_trans with u; auto.
    intro. elim C'. apply le_antisym; auto. apply le_trans with m; auto.
    rewrite <- H3; auto.
    unfold con_op_lt; destif; try omega.
    { unfold con_op_lt. destif; omega. }
    { unfold con_op_lt. destif; omega. }
  }
  { unfold con_op_lt. destif; omega. }
  { unfold con_op_lt. destif; omega. }
  { destif; try constructor.
    unfold con_op_lt; destif; try omega.
    unfold con_op_lt; destif; try omega.
    move/lebP:C => C. move/ltP:C0 => C0.
    elim C0; intros.
    elim H2. apply le_antisym; auto.
    apply le_trans with u; auto. apply le_trans with l; auto.
    { unfold con_op_lt. destif; omega. }
    { unfold con_op_lt. destif; omega. }
  }
  { unfold con_op_lt. destif; omega. }
  { unfold con_op_lt. destif; omega. }
  { destif; try constructor.
    unfold con_op_lt; destif; try omega.
    unfold con_op_lt; destif; try omega.
    move/lebP:C => C. move/ltP:C0 => C0.
    elim C0; intros.
    elim H4. apply le_antisym; auto.
    apply le_trans with u; auto. apply le_trans with l; auto.
    { unfold con_op_lt. destif; omega. }
    { unfold con_op_lt. destif; omega. }
  }
  { unfold con_op_lt. destif; omega. }
  { unfold con_op_lt. destif; omega. }
  { destif; try constructor.
    unfold con_op_lt; destif; try omega.
    unfold con_op_lt; destif; try omega.
    move/lebP:C => C. move/ltP:C0 => C0.
    elim C0; intros.
    elim H4. apply le_antisym; auto.
    apply le_trans with u0; auto. apply le_trans with l; auto.
    { unfold con_op_lt. destif; omega. }
    { unfold con_op_lt. destif; omega. }
  }
  { destif; try constructor.
    unfold con_op_lt; destif; try omega.
    move/ltP:C => [C C']. move/ltP:C0 => C0.
    elim C0; split. 
    apply le_trans with l0; auto. apply le_trans with u; auto.
    intro. elim C'. apply le_antisym; auto. apply le_trans with m; auto.
    rewrite <- H3; auto.
    unfold con_op_lt; destif; try omega.
    { unfold con_op_lt. destif; omega. }
    { unfold con_op_lt. destif; omega. }
  }
  { destif; try constructor.
    unfold con_op_lt; destif; try omega.
    move/ltP:C => [C C']. move/ltP:C0 => C0.
    elim C0; split. 
    apply le_trans with l0; auto. apply le_trans with u; auto.
    intro. elim C'. apply le_antisym; auto. apply le_trans with m; auto.
    rewrite <- H5; auto.
    unfold con_op_lt; destif; try omega. 
    destif; try constructor.
    unfold con_op_lt; destif; try omega.
    unfold con_op_lt; destif; try omega.
    move/lebP:C0 => C0. move/ltP:C1 => C1.
    elim C1; intros.
    elim H6. apply le_antisym; auto.
    apply le_trans with u0; auto. apply le_trans with l; auto.
    { unfold con_op_lt. destif; omega. }
    { unfold con_op_lt. destif; omega. }
  }
Qed.

Lemma op_le'_prop : ∀n m gn gm, 
  abs_itvT' n gn -> abs_itvT' m gm -> abs_itvT' (con_op_le n m) (op_le' gn gm).
Proof.
  intros n m gn gm Hn Hm; destruct Hn; destruct Hm;
    unfold op_le'; try constructor.
  { unfold con_op_le. destif; omega. }
  { unfold con_op_le. destif; omega. }
  { unfold con_op_le. destif; omega. }
  { unfold con_op_le. destif; omega. }
  { unfold con_op_le. destif; omega. }
  { unfold con_op_le. destif; omega. }
  { unfold con_op_le. destif; omega. }
  { unfold con_op_le. destif; omega. }
  { unfold con_op_le. destif; omega. }
  { unfold con_op_le. destif; omega. }
  { unfold con_op_le. destif; omega. }
  { unfold con_op_le. destif; omega. }
  { destif; try constructor.
    unfold con_op_le; destif; try omega.
    move/lebP:C => C. move/lebP:C0 => C0.
    elim C0.
    apply le_trans with l; auto. apply le_trans with u; auto.
    unfold con_op_le; destif; try omega.
    { unfold con_op_le. destif; omega. }
    { unfold con_op_le. destif; omega. }
  }
  { destif; try constructor.
    unfold con_op_le; destif; try omega.
    move/lebP:C => C. move/lebP:C0 => C0.
    elim C0.
    apply le_trans with l; auto. apply le_trans with u; auto.
    unfold con_op_le; destif; try omega.
    { unfold con_op_le. destif; omega. }
    { unfold con_op_le. destif; omega. }
  }
  { unfold con_op_le. destif; omega. }
  { unfold con_op_le. destif; omega. }
  { destif; try constructor.
    unfold con_op_le; destif; try omega.
    unfold con_op_le; destif; try omega.
    move/ltP:C => [C C']. move/lebP:C0 => C0.
    elim C'. apply le_antisym; auto.
    apply le_trans with n; auto. apply le_trans with m; auto.
    { unfold con_op_le. destif; omega. }
    { unfold con_op_le. destif; omega. }
  }
  { unfold con_op_le. destif; omega. }
  { unfold con_op_le. destif; omega. }
  { destif; try constructor.
    unfold con_op_le; destif; try omega.
    unfold con_op_le; destif; try omega.
    move/ltP:C => [C C']. move/lebP:C0 => C0.
    elim C'. apply le_antisym; auto.
    apply le_trans with n; auto. apply le_trans with m; auto.
    { unfold con_op_le. destif; omega. }
    { unfold con_op_le. destif; omega. }
  }
  { unfold con_op_le. destif; omega. }
  { unfold con_op_le. destif; omega. }
  { destif; try constructor.
    unfold con_op_le; destif; try omega.
    unfold con_op_le; destif; try omega.
    move/ltP:C => [C C']. move/lebP:C0 => C0.
    elim C'. apply le_antisym; auto.
    apply le_trans with n; auto. apply le_trans with m; auto.
    { unfold con_op_le. destif; omega. }
    { unfold con_op_le. destif; omega. }
  }
  { destif; try constructor.
    unfold con_op_le; destif; try omega.
    move/lebP:C => C. move/lebP:C0 => C0.
    elim C0.
    apply le_trans with l0; auto. apply le_trans with u; auto.
    unfold con_op_le; destif; try omega.
    { unfold con_op_le. destif; omega. }
    { unfold con_op_le. destif; omega. }
  }
  { destif; try constructor.
    unfold con_op_le; destif; try omega.
    move/lebP:C => C. move/lebP:C0 => C0.
    elim C0.
    apply le_trans with l0; auto. apply le_trans with u; auto.
    unfold con_op_le; destif; try omega.
    destif; try constructor.
    unfold con_op_le; destif; try omega.
    unfold con_op_le; destif; try omega.
    move/ltP:C0 => [C0 C0']. move/lebP:C1 => C1.
    elim C0'. apply le_antisym; auto.
    apply le_trans with n; auto. apply le_trans with m; auto.
    { unfold con_op_le. destif; omega. }
    { unfold con_op_le. destif; omega. }
  }
Qed.

Lemma op_gt'_prop : ∀n m gn gm, 
  abs_itvT' n gn -> abs_itvT' m gm -> abs_itvT' (con_op_gt n m) (op_gt' gn gm).
Proof.
  unfold op_gt'.
  intros n m gn gm Hn Hm; destruct Hn; destruct Hm;
    unfold op_lt'; try constructor.
  { unfold con_op_gt. destif; omega. }
  { unfold con_op_gt. destif; omega. }
  { unfold con_op_gt. destif; omega. }
  { unfold con_op_gt. destif; omega. }
  { unfold con_op_gt. destif; omega. }
  { unfold con_op_gt. destif; omega. }
  { unfold con_op_gt. destif; omega. }
  { unfold con_op_gt. destif; omega. }
  { unfold con_op_gt. destif; omega. }
  { unfold con_op_gt. destif; omega. }
  { unfold con_op_gt. destif; omega. }
  { unfold con_op_gt. destif; omega. }
  { destif; try constructor.
    unfold con_op_gt. destif; omega.
    unfold con_op_gt. destif; try omega.
    move/lebP:C => C. move/ltP:C0 => [C0 C0'].
    elim C0'. apply le_antisym; auto.
    apply le_trans with l; auto. apply le_trans with u; auto. 
    { unfold con_op_gt. destif; omega. }
    { unfold con_op_gt. destif; omega. }
  }
  { destif; try constructor.
    unfold con_op_gt; destif; try omega.
    unfold con_op_gt; destif; try omega.
    move/lebP:C => C. move/ltP:C0 => [C0 C0'].
    elim C0'. apply le_antisym; auto.
    apply le_trans with l; auto. apply le_trans with u; auto.
    { unfold con_op_gt. destif; omega. }
    { unfold con_op_gt. destif; omega. }
  }
  { unfold con_op_gt. destif; omega. }
  { unfold con_op_gt. destif; omega. }
  { destif; try constructor.
    unfold con_op_gt; destif; try omega.
    move/ltP:C => [C C']. move/ltP:C0 => C0.
    elim C0; split. 
    apply le_trans with l; auto. apply le_trans with u; auto. 
    intro. elim C'.
    apply le_antisym; auto. 
    apply le_trans with n; auto. rewrite <- H1; auto. 
    unfold con_op_gt; destif; try omega.
    { unfold con_op_gt. destif; omega. }
    { unfold con_op_gt. destif; omega. }
  }
  { unfold con_op_gt. destif; omega. }
  { unfold con_op_gt. destif; omega. }
  { destif; try constructor.
    unfold con_op_gt; destif; try omega.
    move/ltP:C => [C C']. move/ltP:C0 => C0.
    elim C0; split. 
    apply le_trans with l; auto. apply le_trans with u; auto. 
    intro. elim C'.
    apply le_antisym; auto. 
    apply le_trans with n; auto. rewrite <- H3; auto. 
    unfold con_op_gt; destif; try omega. 
    { unfold con_op_gt. destif; omega. }
    { unfold con_op_gt. destif; omega. }
  }
  { unfold con_op_gt. destif; omega. }
  { unfold con_op_gt. destif; omega. }
  { destif; try constructor.
    unfold con_op_gt; destif; try omega.
    move/ltP:C => [C C']. move/ltP:C0 => C0.
    elim C0; split. 
    apply le_trans with l; auto. apply le_trans with u0; auto. 
    intro. elim C'.
    apply le_antisym; auto. 
    apply le_trans with n; auto. rewrite <- H3; auto. 
    unfold con_op_gt; destif; try omega. 
    { unfold con_op_gt. destif; omega. }
    { unfold con_op_gt. destif; omega. }
  }
  { destif; try constructor.
    unfold con_op_gt; destif; try omega.
    unfold con_op_gt; destif; try omega.
    move/lebP:C => C. move/ltP:C0 => [C0 C0'].
    elim C0'. apply le_antisym; auto.
    apply le_trans with l0; auto. apply le_trans with u; auto. 
    { unfold con_op_gt. destif; omega. }
    { unfold con_op_gt. destif; omega. }
  }
  { destif; try constructor.
    unfold con_op_gt; destif; try omega.
    move/ltP:C => [C C']. move/ltP:C0 => C0.
    elim C0; split. 
    apply le_trans with l; auto. apply le_trans with u0; auto. 
    intro. elim C'.
    apply le_antisym; auto. 
    apply le_trans with n; auto. rewrite <- H5; auto. 
    unfold con_op_gt; destif; try omega. 
    destif; try constructor.
    unfold con_op_gt; destif; try omega.
    unfold con_op_gt; destif; try omega.
    move/lebP:C0 => C0. move/ltP:C1 => [C1 C1'].
    elim C1'. apply le_antisym; auto.
    apply le_trans with l0; auto. apply le_trans with u; auto. 
    { unfold con_op_gt. destif; omega. }
    { unfold con_op_gt. destif; omega. }
  }
Qed.

Lemma op_ge'_prop : ∀n m gn gm, 
  abs_itvT' n gn -> abs_itvT' m gm -> abs_itvT' (con_op_ge n m) (op_ge' gn gm).
Proof.
  unfold op_ge'.
  intros n m gn gm Hn Hm; destruct Hn; destruct Hm;
    unfold op_le'; try constructor.
  { unfold con_op_ge. destif; omega. }
  { unfold con_op_ge. destif; omega. }
  { unfold con_op_ge. destif; omega. }
  { unfold con_op_ge. destif; omega. }
  { unfold con_op_ge. destif; omega. }
  { unfold con_op_ge. destif; omega. }
  { unfold con_op_ge. destif; omega. }
  { unfold con_op_ge. destif; omega. }
  { unfold con_op_ge. destif; omega. }
  { unfold con_op_ge. destif; omega. }
  { unfold con_op_ge. destif; omega. }
  { unfold con_op_ge. destif; omega. }
  { destif; try constructor.
    unfold con_op_ge; destif; try omega.
    unfold con_op_ge; destif; try omega.
    move/ltP:C => [C C']. move/lebP:C0 => C0.
    elim C'. apply le_antisym; auto.
    apply le_trans with m; auto. apply le_trans with n; auto. 
    { unfold con_op_ge. destif; omega. }
    { unfold con_op_ge. destif; omega. }
  }
  { destif; try constructor.
    unfold con_op_ge; destif; try omega.
    unfold con_op_ge; destif; try omega.
    move/ltP:C => [C C']. move/lebP:C0 => C0.
    elim C'. apply le_antisym; auto.
    apply le_trans with m; auto. apply le_trans with n; auto. 
    { unfold con_op_ge. destif; omega. }
    { unfold con_op_ge. destif; omega. }
  }
  { unfold con_op_ge. destif; omega. }
  { unfold con_op_ge. destif; omega. }
  { destif; try constructor.
    unfold con_op_ge; destif; try omega.
    move/lebP:C => C. move/lebP:C0 => C0.
    elim C0. apply le_trans with u; auto. apply le_trans with l; auto. 
    unfold con_op_ge; destif; try omega. 
    { unfold con_op_ge. destif; omega. }
    { unfold con_op_ge. destif; omega. }
  }
  { unfold con_op_ge. destif; omega. }
  { unfold con_op_ge. destif; omega. }
  { destif; try constructor.
    unfold con_op_ge; destif; try omega.
    move/lebP:C => C. move/lebP:C0 => C0.
    elim C0. apply le_trans with u; auto. apply le_trans with l; auto. 
    unfold con_op_ge; destif; try omega. 
    { unfold con_op_ge. destif; omega. }
    { unfold con_op_ge. destif; omega. }
  }
  { unfold con_op_ge. destif; omega. }
  { unfold con_op_ge. destif; omega. }
  { destif; try constructor.
    unfold con_op_ge; destif; try omega.
    move/lebP:C => C. move/lebP:C0 => C0.
    elim C0. apply le_trans with u0; auto. apply le_trans with l; auto. 
    unfold con_op_ge; destif; try omega. 
    { unfold con_op_ge. destif; omega. }
    { unfold con_op_ge. destif; omega. }
  }
  { destif; try constructor.
    unfold con_op_ge; destif; try omega.
    unfold con_op_ge; destif; try omega.
    move/ltP:C => [C C']. move/lebP:C0 => C0.
    elim C'. apply le_antisym; auto.
    apply le_trans with m; auto. apply le_trans with n; auto. 
    { unfold con_op_ge. destif; omega. }
    { unfold con_op_ge. destif; omega. }
  }
  { destif; try constructor.
    unfold con_op_ge; destif; try omega.
    move/lebP:C => C. move/lebP:C0 => C0.
    elim C0. apply le_trans with u0; auto. apply le_trans with l; auto. 
    unfold con_op_ge; destif; try omega. 
    destif; try constructor.
    unfold con_op_ge; destif; try omega.
    unfold con_op_ge; destif; try omega.
    move/ltP:C0 => [C0 C0']. move/lebP:C1 => C1.
    elim C0'. apply le_antisym; auto.
    apply le_trans with m; auto. apply le_trans with n; auto. 
    { unfold con_op_ge. destif; omega. }
    { unfold con_op_ge. destif; omega. }
  }
Qed.

Lemma op_eq'_prop : ∀n m gn gm, 
  abs_itvT' n gn -> abs_itvT' m gm -> abs_itvT' (con_op_eq n m) (op_eq' gn gm).
Proof.
  intros n m gn gm Hn Hm; destruct Hn; destruct Hm;
    unfold op_eq'; try constructor.
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif; omega. } 
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif.
    { destif; try (constructor; omega).
      move/eqbP:C => C. move/ltP:C0 => [C0 C0'].
      elim C0'. apply le_antisym; auto.
      apply le_trans with m; auto. rewrite <- C; auto. }
    { destif; try (constructor; omega). }
  }
  { unfold con_op_eq; destif.
    { destif; try (constructor; omega).
      move/eqbP:C => C. move/ltP:C0 => [C0 C0'].
      elim C0'. apply le_antisym; auto.
      apply le_trans with m; auto. rewrite <- C; auto. }
    { destif; try (constructor; omega). }
  }
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif.
    { destif; try (constructor; omega).
      move/eqbP:C => C. move/ltP:C0 => [C0 C0'].
      elim C0'. apply le_antisym; auto.
      apply le_trans with m; auto. rewrite <- C; auto. }
    { destif; try (constructor; omega). }
  }
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif.
    { destif; try (constructor; omega).
      move/eqbP:C => C. move/ltP:C0 => [C0 C0'].
      elim C0'. apply le_antisym; auto.
      apply le_trans with m; auto. rewrite <- C; auto. }
    { destif; try (constructor; omega). }
  }
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif; omega. }
  { unfold con_op_eq; destif.
    { destif; try (constructor; omega).
      move/eqbP:C => C. move/ltP:C0 => [C0 C0'].
      elim C0'. apply le_antisym; auto.
      apply le_trans with m; auto. rewrite <- C; auto. }
    { destif; try (constructor; omega). }
  }
  { unfold con_op_eq; destif.
    { destif; try (constructor; omega).
      move/eqbP:C => C. move/ltP:C0 => [C0 C0'].
      elim C0'. apply le_antisym; auto.
      apply le_trans with m; auto. rewrite <- C; auto. }
    { destif; try (constructor; omega). }
  }
  { unfold con_op_eq; destif.
    { destif; try (constructor; omega).
      move/eqbP:C => C. move/ltP:C0 => [C0 C0'].
      elim C0'. apply le_antisym; auto.
      apply le_trans with m; auto. rewrite <- C; auto. 
      destif; try (constructor; omega).
      move/eqbP:C => C. move/ltP:C1 => [C1 C1'].
      elim C1'. apply le_antisym; auto.
      apply le_trans with m; auto. rewrite <- C; auto. 
      destif; try (constructor; omega). }
    { destif; try (constructor; omega).
      destif; try (constructor; omega).
      destif; try (constructor; omega).
      move/andP:C2 => [C2 C2']. move/andP:C2 => [C2 C2''].
      move/eqbP:C2 => C2. move/eqbP:C2' => C2'. move/eqbP:C2'' => C2''. 
      move/eqbP:C => C. elim C. apply le_antisym; auto.
      apply le_trans with u; auto. rewrite C2''; auto. 
      apply le_trans with u0; auto. rewrite <- C2'; auto. 
      rewrite <- C2''; auto. rewrite <- C2; auto. }
  }
Qed.

Lemma op_neq'_prop : ∀n m gn gm, 
  abs_itvT' n gn -> abs_itvT' m gm -> 
  abs_itvT' (con_op_neq n m) (op_neq' gn gm).
Proof.
  intros n m gn gm Hn Hm; destruct Hn; destruct Hm;
    unfold op_neq'; try constructor.
  { unfold con_op_neq. destif; omega. }
  { unfold con_op_neq. destif; omega. }
  { unfold con_op_neq. destif; omega. }
  { unfold con_op_neq. destif; omega. }
  { unfold con_op_neq. destif; omega. }
  { unfold con_op_neq. destif; omega. }
  { unfold con_op_neq. destif; omega. }
  { unfold con_op_neq. destif; omega. }
  { unfold con_op_neq. destif; omega. }
  { unfold con_op_neq. destif; omega. }
  { unfold con_op_neq. destif; omega. }
  { unfold con_op_neq. destif; omega. }
  { unfold op_eq'. destif.
    { unfold op_not'; simpl.
      unfold con_op_neq. destif; try (constructor; omega).
      { move/ltP:C => [C C']. move/eqbP:C0 => C0.
        elim C'. apply le_antisym; auto.
        apply le_trans with m; auto.
        rewrite <- C0; auto. }
    }
    { unfold op_not'; simpl; constructor. 
      { unfold con_op_neq. destif; omega. }
      { unfold con_op_neq. destif; omega. }
    }
  }
  { unfold op_eq'. destif.
    { unfold op_not'; simpl.
      unfold con_op_neq. destif; try (constructor; omega).
      { move/ltP:C => [C C']. move/eqbP:C0 => C0.
        elim C'. apply le_antisym; auto.
        apply le_trans with m; auto.
        rewrite <- C0; auto. }
    }
    { unfold op_not'; simpl; constructor. 
      { unfold con_op_neq. destif; omega. }
      { unfold con_op_neq. destif; omega. }
    }
  }
  { unfold con_op_neq. destif; omega. }
  { unfold con_op_neq. destif; omega. }
  { unfold op_eq'. destif.
    { unfold op_not'; simpl.
      unfold con_op_neq. destif; try (constructor; omega).
      { move/ltP:C => [C C']. move/eqbP:C0 => C0.
        elim C'. apply le_antisym; auto.
        apply le_trans with m; auto.
        rewrite <- C0; auto. }
    }
    { unfold op_not'; simpl; constructor. 
      { unfold con_op_neq. destif; omega. }
      { unfold con_op_neq. destif; omega. }
    }
  }
  { unfold con_op_neq. destif; omega. }
  { unfold con_op_neq. destif; omega. }
  { unfold op_eq'. destif.
    { unfold op_not'; simpl.
      unfold con_op_neq. destif; try (constructor; omega).
      { move/ltP:C => [C C']. move/eqbP:C0 => C0.
        elim C'. apply le_antisym; auto.
        apply le_trans with m; auto.
        rewrite <- C0; auto. }
    }
    { unfold op_not'; simpl; constructor. 
      { unfold con_op_neq. destif; omega. }
      { unfold con_op_neq. destif; omega. }
    }
  }
  { unfold con_op_neq. destif; omega. }
  { unfold con_op_neq. destif; omega. }
  { unfold op_eq'. destif.
    { unfold op_not'; simpl.
      unfold con_op_neq. destif; try (constructor; omega).
      { move/ltP:C => [C C']. move/eqbP:C0 => C0.
        elim C'. apply le_antisym; auto.
        apply le_trans with m; auto.
        rewrite <- C0; auto. }
    }
    { unfold op_not'; simpl; constructor. 
      { unfold con_op_neq. destif; omega. }
      { unfold con_op_neq. destif; omega. }
    }
  }
  { unfold op_eq'. destif.
    { unfold op_not'; simpl.
      unfold con_op_neq. destif; try (constructor; omega).
      { move/ltP:C => [C C']. move/eqbP:C0 => C0.
        elim C'. apply le_antisym; auto.
        apply le_trans with m; auto.
        rewrite <- C0; auto. }
    }
    { unfold op_not'; simpl; constructor. 
      { unfold con_op_neq. destif; omega. }
      { unfold con_op_neq. destif; omega. }
    }
  }
  { unfold op_eq'. destif.
    { unfold op_not'; simpl.
      unfold con_op_neq. destif; try (constructor; omega).
      { move/ltP:C => [C C']. move/eqbP:C0 => C0.
        elim C'. apply le_antisym; auto.
        apply le_trans with m; auto.
        rewrite <- C0; auto. }
    }
    destif.
    { unfold op_not'; simpl.
      unfold con_op_neq. destif; try (constructor; omega).
      { move/ltP:C0 => [C0 C0']. move/eqbP:C1 => C1.
        elim C0'. apply le_antisym; auto.
        apply le_trans with m; auto.
        rewrite <- C1; auto. }
    }
    destif.
    { unfold op_not'; simpl.
      unfold con_op_neq. destif; try (constructor; omega).
      { move/andP:C1 => [C1 C1']. move/andP:C1 => [C1 C1''].
        move/eqbP:C1 => C1. move/eqbP:C1' => C1'. move/eqbP:C1'' => C1''.
        move/eqbP:C2 => C2. elim C2.
        apply le_antisym.
        apply le_trans with u; auto. rewrite C1''; auto.
        apply le_trans with u0; auto. 
        rewrite <- C1'. rewrite <- C1''. rewrite <- C1; auto. }
    }
    { unfold op_not'; simpl; constructor. 
      { unfold con_op_neq. destif; omega. }
      { unfold con_op_neq. destif; omega. }
    }
  }
Qed. 

Lemma op_and'_prop : ∀n m gn gm, 
  abs_itvT' n gn -> abs_itvT' m gm -> 
  abs_itvT' (con_op_and n m) (op_and' gn gm).
Proof.
  intros n m gn gm Hn Hm; destruct Hn; destruct Hm;
    unfold op_and'; try constructor.
  { unfold con_op_and. destif; try omega. destif; try omega. }
  { unfold con_op_and. destif; try omega. destif; try omega. }
  { unfold con_op_and. destif; try omega. destif; try omega. }
  { unfold con_op_and. destif; try omega. destif; try omega. }
  { unfold con_op_and. destif; try omega. destif; try omega. }
  { unfold con_op_and. destif; try omega. destif; try omega. }
  { destif. inversion C.
    destif. 
    unfold con_op_and. destif; try (constructor; omega).
    destif; try (constructor; omega).
    move/eqbP:C0 => C0. move/eqbP:C1 => C1. move/eqbP:C2 => C2.
    inversion C0. elim C2.
    apply le_antisym. rewrite H3; auto. rewrite H2; auto.
    destif.
    move/andP:C1 => [C1 C1']. move/negP:C1 => C1. elim C1. constructor.
    constructor. 
    { unfold con_op_and. destif; try omega. destif; try omega. }
    { unfold con_op_and. destif; try omega. destif; try omega. }
  }
  { destif. inversion C.
    destif. inversion C0.
    destif.
    move/andP:C1 => [C1 C1']. move/negP:C1' => C1'. elim C1'. constructor.
    constructor.
    { unfold con_op_and. destif; try omega. destif; try omega. }
    { unfold con_op_and. destif; try omega. destif; try omega. }
  }
  { destif. inversion C.
    destif. inversion C0.
    destif; last constructor.
    move/andP:C1 => [C1 C1']. move/negP:C1 => C1. move/negP:C1' => C1'.
    unfold con_op_and. destif.
    move/eqbP:C2 => C2. elim C1. apply/lebP. unfold le; simpl.
    rewrite <- C2; auto.
    destif; try (constructor; omega).
    move/eqbP:C3 => C3. elim C1'. apply/lebP. unfold le; simpl.
    rewrite <- C3; auto. 
    { unfold con_op_and. destif; try omega. destif; try omega. }
    { unfold con_op_and. destif; try omega. destif; try omega. }
  }
  { destif. inversion C.
    destif. inversion C0.
    destif; last constructor.
    move/andP:C1 => [C1 C1']. move/negP:C1 => C1. move/negP:C1' => C1'.
    unfold con_op_and. destif.
    move/eqbP:C2 => C2. elim C1. apply/lebP. unfold le; simpl.
    rewrite <- C2; auto.
    destif; try (constructor; omega).
    move/eqbP:C3 => C3. elim C1'. apply/lebP. unfold le; simpl.
    rewrite <- C3; auto. 
    { unfold con_op_and. destif; try omega. destif; try omega. }
    { unfold con_op_and. destif; try omega. destif; try omega. }
  }
  { destif. inversion C.
    destif. 
    unfold con_op_and. destif; try (constructor; omega).
    destif; try (constructor; omega).
    move/eqbP:C2 => C2. elim C2. move/eqbP:C0 => C0. inversion C0.
    apply le_antisym. rewrite H4; auto. rewrite H3; auto.
    destif; last constructor.
    unfold con_op_and.
    destif.
    move/eqbP:C2 => C2. 
    move/andP:C1 => [C1 _].
    move/negP:C1 => C1. elim C1. apply/lebP. unfold le; simpl.
    rewrite <- C2; auto.
    destif; last (constructor; omega).
    move/eqbP:C3 => C3. 
    move/andP:C1 => [_ C1].
    move/negP:C1 => C1. elim C1. apply/lebP. unfold le; simpl; split; auto.
    rewrite <- C3; auto. rewrite <- C3; auto. 
    { unfold con_op_and. destif; try omega. destif; try omega. }
    { unfold con_op_and. destif; try omega. destif; try omega. }
  }
  { destif. inversion C.
    destif. inversion C0.
    destif; last constructor.
    move/andP:C1 => [C1 C1']. move/negP:C1 => C1. move/negP:C1' => C1'.
    elim C1'. constructor. 
    { unfold con_op_and. destif; try omega. destif; try omega. }
    { unfold con_op_and. destif; try omega. destif; try omega. }
  }
  { destif. inversion C.
    destif. inversion C0.
    destif; last constructor.
    unfold con_op_and.
    destif.
    move/eqbP:C2 => C2. 
    move/andP:C1 => [C1 _].
    move/negP:C1 => C1. elim C1. apply/lebP. unfold le; simpl.
    rewrite <- C2; auto.
    destif; last (constructor; omega).
    move/eqbP:C3 => C3. 
    move/andP:C1 => [_ C1].
    move/negP:C1 => C1. elim C1. apply/lebP. unfold le; simpl.
    rewrite <- C3; auto.
    { unfold con_op_and. destif; try omega. destif; try omega. }
    { unfold con_op_and. destif; try omega. destif; try omega. }
  }
  { destif. inversion C.
    destif. inversion C0.
    destif; last constructor.
    unfold con_op_and.
    destif.
    move/eqbP:C2 => C2. 
    move/andP:C1 => [C1 _].
    move/negP:C1 => C1. elim C1. apply/lebP. unfold le; simpl.
    rewrite <- C2; auto.
    destif; last (constructor; omega).
    move/eqbP:C3 => C3. 
    move/andP:C1 => [_ C1].
    move/negP:C1 => C1. elim C1. apply/lebP. unfold le; simpl.
    rewrite <- C3; auto. 
    { unfold con_op_and. destif; try omega. destif; try omega. }
    { unfold con_op_and. destif; try omega. destif; try omega. }
  }
  { destif. inversion C.
    destif. 
    unfold con_op_and.
    destif; try (constructor; omega).
    destif; try (constructor; omega).
    move/eqbP:C2 => C2. elim C2.
    move/eqbP:C0 => C0. inversion C0.
    apply le_antisym.
    rewrite H4; auto. rewrite H3; auto.
    destif; last (constructor; try omega).
    unfold con_op_and.
    destif.
    move/eqbP:C2 => C2. 
    move/andP:C1 => [C1 _].
    move/negP:C1 => C1. elim C1. apply/lebP. unfold le; simpl.
    rewrite <- C2; auto.
    destif; last (constructor; omega).
    move/eqbP:C3 => C3. 
    move/andP:C1 => [_ C1].
    move/negP:C1 => C1. elim C1. apply/lebP. unfold le; simpl; split.
    rewrite <- C3; auto. rewrite <- C3; auto. 
    { unfold con_op_and. destif; try omega. destif; try omega. }
    { unfold con_op_and. destif; try omega. destif; try omega. }
  }
  { destif. 
    unfold con_op_and.
    destif; try (constructor; omega).
    destif; try (constructor; omega).
    move/eqbP:C0 => C0. elim C0.
    move/eqbP:C => C. inversion C.
    apply le_antisym.
    rewrite H3; auto. rewrite H2; auto.
    destif. inversion C0.
    destif; last constructor.
    unfold con_op_and.
    move/andP:C1 => [_ C1]. inversion C1. 
    { unfold con_op_and. destif; try omega. destif; try omega. }
    { unfold con_op_and. destif; try omega. destif; try omega. }
  }
  { destif. 
    unfold con_op_and.
    destif; try (constructor; omega).
    destif; try (constructor; omega).
    move/eqbP:C0 => C0. elim C0.
    move/eqbP:C => C. inversion C.
    apply le_antisym.
    rewrite H4; auto. rewrite H3; auto.
    destif. inversion C0.
    destif; last constructor.
    unfold con_op_and.
    destif.
    move/eqbP:C2 => C2. 
    move/andP:C1 => [C1 _].
    move/negP:C1 => C1. elim C1. apply/lebP. unfold le; simpl; split.
    rewrite <- C2; auto. rewrite <- C2; auto.
    destif; last (constructor; omega).
    move/eqbP:C3 => C3. 
    move/andP:C1 => [_ C1].
    move/negP:C1 => C1. elim C1. apply/lebP. unfold le; simpl.
    rewrite <- C3; auto. 
    { unfold con_op_and. destif; try omega. destif; try omega. }
    { unfold con_op_and. destif; try omega. destif; try omega. }
  }
  { destif. 
    unfold con_op_and.
    destif; try (constructor; omega).
    destif; try (constructor; omega).
    move/eqbP:C0 => C0. elim C0.
    move/eqbP:C => C. inversion C.
    apply le_antisym.
    rewrite H4; auto. rewrite H3; auto.
    destif. inversion C0.
    destif; last constructor.
    unfold con_op_and.
    destif.
    move/eqbP:C2 => C2. 
    move/andP:C1 => [C1 _].
    move/negP:C1 => C1. elim C1. apply/lebP. unfold le; simpl; split.
    rewrite <- C2; auto. rewrite <- C2; auto.
    destif; last (constructor; omega).
    move/eqbP:C3 => C3. 
    move/andP:C1 => [_ C1].
    move/negP:C1 => C1. elim C1. apply/lebP. unfold le; simpl.
    rewrite <- C3; auto. 
    { unfold con_op_and. destif; try omega. destif; try omega. }
    { unfold con_op_and. destif; try omega. destif; try omega. }
  }
  { destif. 
    unfold con_op_and.
    destif; try (constructor; omega).
    destif; try (constructor; omega).
    move/eqbP:C0 => C0. elim C0.
    move/eqbP:C => C. inversion C.
    apply le_antisym.
    rewrite H6; auto. rewrite H5; auto.
    destif. 
    unfold con_op_and.
    destif; try (constructor; omega).
    destif; try (constructor; omega).
    move/eqbP:C2 => C2. elim C2.
    move/eqbP:C0 => C0. inversion C0.
    apply le_antisym.
    rewrite H6; auto. rewrite H5; auto.
    destif; last constructor.
    unfold con_op_and.
    destif.
    move/eqbP:C2 => C2. 
    move/andP:C1 => [C1 _].
    move/negP:C1 => C1. elim C1. apply/lebP. unfold le; simpl; split.
    rewrite <- C2; auto. rewrite <- C2; auto.
    destif; last (constructor; omega).
    move/eqbP:C3 => C3. 
    move/andP:C1 => [_ C1].
    move/negP:C1 => C1. elim C1. apply/lebP. unfold le; simpl; split.
    rewrite <- C3; auto. rewrite <- C3; auto. 
    { unfold con_op_and. destif; try omega. destif; try omega. }
    { unfold con_op_and. destif; try omega. destif; try omega. }
  }
Qed.

Lemma op_or'_prop : ∀n m gn gm, 
  abs_itvT' n gn -> abs_itvT' m gm -> 
  abs_itvT' (con_op_or n m) (op_or' gn gm).
Proof.
  intros n m gn gm Hn Hm; destruct Hn; destruct Hm;
    unfold op_or'; try constructor.
  { unfold con_op_or. destif; try omega. }
  { unfold con_op_or. destif; try omega. }
  { destif. 
    move/andP:C => [C C']. inversion C.
    destif; last constructor.
    move/orP:C0 => C0. destruct C0.
    inversion H0.
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0' => C0'. 
    move/negP:H0 => H0. elim H0. apply/lebP. unfold le; simpl.
    rewrite <- C0'; auto.
    { unfold con_op_or. destif; try omega. }
    { unfold con_op_or. destif; try omega. }
  }
  { destif. 
    move/andP:C => [C C']. inversion C.
    destif; last constructor.
    move/orP:C0 => C0. destruct C0.
    inversion H0.
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0' => C0'. 
    move/negP:H0 => H0. elim H0. apply/lebP. unfold le; simpl.
    rewrite <- C0'; auto. 
    { unfold con_op_or. destif; try omega. }
    { unfold con_op_or. destif; try omega. }
  }
  { destif. 
    move/andP:C => [C C']. inversion C.
    destif; last constructor.
    move/orP:C0 => C0. destruct C0.
    inversion H2.
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0' => C0'. 
    move/negP:H2 => H2. elim H2. apply/lebP. unfold le; simpl; split.
    rewrite <- C0'; auto. rewrite <- C0'; auto. 
    { unfold con_op_or. destif; try omega. }
    { unfold con_op_or. destif; try omega. }
  }
  { destif. 
    move/andP:C => [C C']. inversion C.
    destif; last constructor.
    move/orP:C0 => C0. destruct C0.
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0 => C0. 
    move/negP:H0 => H0. elim H0. apply/lebP. unfold le; simpl.
    rewrite <- C0; auto. 
    inversion H0. 
    { unfold con_op_or. destif; try omega. }
    { unfold con_op_or. destif; try omega. }
  }
  { destif. 
    move/andP:C => [C C']. inversion C.
    destif; last constructor.
    move/orP:C0 => C0. destruct C0.
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0 => C0. 
    move/negP:H1 => H1. elim H1. apply/lebP. unfold le; simpl.
    rewrite <- C0; auto. 
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0' => C0'. 
    move/negP:H1 => H1. elim H1. apply/lebP. unfold le; simpl.
    rewrite <- C0'; auto. 
    { unfold con_op_or. destif; try omega. }
    { unfold con_op_or. destif; try omega. }
  }
  { destif. 
    move/andP:C => [C C']. inversion C.
    destif; last constructor.
    move/orP:C0 => C0. destruct C0.
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0 => C0. 
    move/negP:H1 => H1. elim H1. apply/lebP. unfold le; simpl.
    rewrite <- C0; auto. 
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0' => C0'. 
    move/negP:H1 => H1. elim H1. apply/lebP. unfold le; simpl.
    rewrite <- C0'; auto. 
    { unfold con_op_or. destif; try omega. }
    { unfold con_op_or. destif; try omega. }
  }
  { destif. 
    move/andP:C => [C C']. inversion C.
    destif; last constructor.
    move/orP:C0 => C0. destruct C0.
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0 => C0. 
    move/negP:H3 => H3. elim H3. apply/lebP. unfold le; simpl.
    rewrite <- C0; auto. 
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0' => C0'. 
    move/negP:H3 => H3. elim H3. apply/lebP. unfold le; simpl.
    rewrite <- C0'; auto. 
    { unfold con_op_or. destif; try omega. }
    { unfold con_op_or. destif; try omega. }
  }
  { destif. 
    move/andP:C => [C C']. inversion C.
    destif; last constructor.
    move/orP:C0 => C0. destruct C0.
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0 => C0. 
    move/negP:H0 => H0. elim H0. apply/lebP. unfold le; simpl.
    rewrite <- C0; auto. 
    move/negP:H0 => H0. elim H0. constructor. 
    { unfold con_op_or. destif; try omega. }
    { unfold con_op_or. destif; try omega. }
  }
  { destif. 
    move/andP:C => [C C']. inversion C.
    destif; last constructor.
    move/orP:C0 => C0. destruct C0.
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0 => C0. 
    move/negP:H1 => H1. elim H1. apply/lebP. unfold le; simpl.
    rewrite <- C0; auto. 
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0' => C0'. 
    move/negP:H1 => H1. elim H1. apply/lebP. unfold le; simpl.
    rewrite <- C0'; auto. 
    { unfold con_op_or. destif; try omega. }
    { unfold con_op_or. destif; try omega. }
  }
  { destif. 
    move/andP:C => [C C']. inversion C.
    destif; last constructor.
    move/orP:C0 => C0. destruct C0.
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0 => C0. 
    move/negP:H1 => H1. elim H1. apply/lebP. unfold le; simpl.
    rewrite <- C0; auto. 
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0' => C0'. 
    move/negP:H1 => H1. elim H1. apply/lebP. unfold le; simpl.
    rewrite <- C0'; auto. 
    { unfold con_op_or. destif; try omega. }
    { unfold con_op_or. destif; try omega. }
  }
  { destif. 
    move/andP:C => [C C']. inversion C.
    destif; last constructor.
    move/orP:C0 => C0. destruct C0.
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0 => C0. 
    move/negP:H3 => H3. elim H3. apply/lebP. unfold le; simpl.
    rewrite <- C0; auto. 
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0' => C0'. 
    move/negP:H3 => H3. elim H3. apply/lebP. unfold le; simpl.
    rewrite <- C0'; auto. 
    { unfold con_op_or. destif; try omega. }
    { unfold con_op_or. destif; try omega. }
  }
  { destif. 
    move/andP:C => [C C']. inversion C'.
    destif; last constructor.
    move/orP:C0 => C0. destruct C0.
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0 => C0. 
    move/negP:H2 => H2. elim H2. apply/lebP. unfold le; simpl.
    rewrite <- C0; auto. 
    move/negP:H2 => H2. elim H2. constructor. 
    { unfold con_op_or. destif; try omega. }
    { unfold con_op_or. destif; try omega. }
  }
  { destif. 
    move/andP:C => [C C']. inversion C'.
    destif; last constructor.
    move/orP:C0 => C0. destruct C0.
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0 => C0. 
    move/negP:H3 => H3. elim H3. apply/lebP. unfold le; simpl.
    rewrite <- C0; auto. 
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0' => C0'. 
    move/negP:H3 => H3. elim H3. apply/lebP. unfold le; simpl.
    rewrite <- C0'; auto. 
    { unfold con_op_or. destif; try omega. }
    { unfold con_op_or. destif; try omega. }
  }
  { destif. 
    move/andP:C => [C C']. inversion C'.
    destif; last constructor.
    move/orP:C0 => C0. destruct C0.
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0 => C0. 
    move/negP:H3 => H3. elim H3. apply/lebP. unfold le; simpl.
    rewrite <- C0; auto. 
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0' => C0'. 
    move/negP:H3 => H3. elim H3. apply/lebP. unfold le; simpl.
    rewrite <- C0'; auto. 
    { unfold con_op_or. destif; try omega. }
    { unfold con_op_or. destif; try omega. }
  }
  { destif. 
    move/andP:C => [C C']. 
    unfold con_op_or.
    destif; try (constructor; omega).
    move/andP:C0 => C0. elim C0; split.
    move/eqbP:C => C. inversion C.
    apply/eqbP. apply le_antisym. 
    rewrite H6; auto. rewrite H5; auto.
    move/eqbP:C' => C'. inversion C'.
    apply/eqbP. apply le_antisym. 
    rewrite H6; auto. rewrite H5; auto.
    destif; last constructor.
    move/orP:C0 => C0. destruct C0.
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0 => C0. 
    move/negP:H5 => H5. elim H5. apply/lebP. unfold le; simpl.
    rewrite <- C0; auto. 
    unfold con_op_or.
    destif; last (constructor; omega).
    move/andP:C0 => [C0 C0']. move/eqbP:C0' => C0'. 
    move/negP:H5 => H5. elim H5. apply/lebP. unfold le; simpl.
    rewrite <- C0'; auto. 
    { unfold con_op_or. destif; try omega. }
    { unfold con_op_or. destif; try omega. }
  }
Qed.

Lemma op_band'_prop : ∀n m gn gm, 
  abs_itvT' n gn -> abs_itvT' m gm -> 
  abs_itvT' (con_op_band n m) (op_band' gn gm).
Proof. intros. unfold op_band'. constructor. Qed.

Lemma op_bor'_prop : ∀n m gn gm, 
  abs_itvT' n gn -> abs_itvT' m gm -> 
  abs_itvT' (con_op_bor n m) (op_bor' gn gm).
Proof. intros. constructor. Qed.

Lemma op_bxor'_prop : ∀n m gn gm, 
  abs_itvT' n gn -> abs_itvT' m gm -> 
  abs_itvT' (con_op_bxor n m) (op_bxor' gn gm).
Proof. intros. constructor. Qed.

Lemma op_lshift'_prop : ∀n m gn gm, 
  abs_itvT' n gn -> abs_itvT' m gm -> 
  abs_itvT' (con_op_lshift n m) (op_lshift' gn gm).
Proof. intros. constructor. Qed.

Lemma op_rshift'_prop : ∀n m gn gm, 
  abs_itvT' n gn -> abs_itvT' m gm -> 
  abs_itvT' (con_op_rshift n m) (op_rshift' gn gm).
Proof. intros. constructor. Qed.

Close Scope Z.

Lemma op_add_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) Z.add DomItv.op_add.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_add'_prop.
Qed.

Lemma op_sub_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) Z.sub DomItv.op_sub.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_sub'_prop.
Qed.

Lemma op_mul_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) Z.mul DomItv.op_mul.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_mul'_prop.
Qed.

Lemma op_div_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) Cdiv DomItv.op_div.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_div'_prop.
Qed.

Lemma op_mod_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) Cmod DomItv.op_mod.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_mod'_prop.
Qed.

Lemma op_lt_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) con_op_lt DomItv.op_lt.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_lt'_prop.
Qed.

Lemma op_le_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) con_op_le DomItv.op_le.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_le'_prop.
Qed.

Lemma op_gt_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) con_op_gt DomItv.op_gt.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_gt'_prop.
Qed.

Lemma op_ge_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) con_op_ge DomItv.op_ge.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_ge'_prop.
Qed.

Lemma op_eq_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) con_op_eq DomItv.op_eq.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_eq'_prop.
Qed.

Lemma op_neq_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) con_op_neq DomItv.op_neq.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_neq'_prop.
Qed.

Lemma op_and_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) con_op_and DomItv.op_and.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_and'_prop.
Qed.

Lemma op_or_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) con_op_or DomItv.op_or.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_or'_prop.
Qed.

Lemma op_band_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) con_op_band DomItv.op_band.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_band'_prop.
Qed.

Lemma op_bor_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) con_op_bor DomItv.op_bor.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_bor'_prop.
Qed.

Lemma op_bxor_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) con_op_bxor DomItv.op_bxor.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_bxor'_prop.
Qed.

Lemma op_lshift_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) con_op_lshift DomItv.op_lshift.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_lshift'_prop.
Qed.

Lemma op_rshift_proper2 : Proper2 (abs_relation ===> abs_relation ===> abs_relation) con_op_rshift DomItv.op_rshift.
Proof.
  move=> x [y|] Hxy z [w|] Hzw //=.
  exact/op_rshift'_prop.
Qed.

Lemma op_not_proper2 : Proper2 (abs_relation ===> abs_relation) con_op_not DomItv.op_not.
Proof.
  move=> x [y|] Hxy //=.
  exact/op_not'_prop.
Qed.

Lemma op_bnot_proper2 : Proper2 (abs_relation ===> abs_relation) con_op_bnot DomItv.op_bnot.
Proof.
  move=> x [y|] Hxy //=.
  exact/op_bnot'_prop.
Qed.
