Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import List.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomSet.
Require Import DomMap.
Require Import DomUtil.

Require Import Syntax.

Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomPidxSen.
Require Import DomPidxInsen.

Require Import SemAbsBlock.
Require Import SemAccSen.
Require Import SemAccInsen.

Require Import SemSenLocal.
Require Import SemInsenLocal.

Require Import CorBasic.
Require Import TrsSen.
Require Import CorAcc.

Set Implicit Arguments.

Section Postfixs.
  Variable (prog:progT) (acc:accT) (state:stateT) (Hacc: acc_correct prog acc state).
  Definition Hacc' := tacc_correct prog acc state Hacc.

  Lemma tstate_instrs fid bid stack instrs
        (Hc : pidx_consistent prog fid stack)
        (Hacc_instrs : acc_correct_acclocs_instrs acc fid instrs (state[(fid, bid, false)]!))
        (Hinstrs : sem_instrs acc fid instrs
                               (state) [(fid, bid, false)]!
                   ⊑ (state) [(fid, bid, true)]!) :
    postfix_instrs
      acc fid instrs
      (tstate prog state) [(fid, bid, false, stack) ]!
      (tstate prog state) [(fid, bid, true, stack) ]!.
  Proof.
    rewrite ? tstate_lookup /fst/snd.
    rewrite/DomPidx.pifid/DomPidxSen.pidxT_pidx/fst/snd Hc.
    elim: instrs (state[_]!) (state[_]!) Hacc_instrs Hinstrs=>
    [|istr istrs IH] mnew mold Hinstrs Hsem.
    { by constructor. }
    { rewrite/= in Hinstrs; move/andP: Hinstrs=> [Hinstr Hinstrs].
      econstructor; last first.
      apply IH; eauto.
      exact/DomType.eq_le/eq_refl.
      done.
    }
  Qed.

  Lemma tstate_jump
        fid func (Hfunc : MapsTo fid func (pfuncs prog))
        bid instrs nextbids (Hcmd: MapsTo bid (instrs, tmn_jump nextbids) (fcmds func))
        stack (Hc : pidx_consistent prog fid stack)
        (Hjump : postfix_jump state fid func bid nextbids) :
    SemSenLocal.postfix_jump (tstate prog state) fid func bid stack nextbids.
  Proof.
    move: Hjump.
    move/set_forallP.
    refine (implies_right_turn (Proper _ _) _ _ _ _).
    refine (forall_implies _ _ _ _) => nbid.
    refine (forall_implies _ _ _ _) => Hnbid.
    rewrite ? tstate_lookup /fst.
    rewrite/DomPidx.pifid/DomPidxSen.pidxT_pidx/fst/snd Hc.
    case Hmem2: (state[(fid, nbid, false)]!)=> [mem2|]; last by move=> _; apply: top_prop.
    case Hmem1: (state[(fid, bid, true)]!)=> [mem1|]; last done.
    by dom_tac; move/lebP=> Hmem.
  Qed.

  Lemma tstate_bind_params
    fid func (Hfunc : MapsTo fid func (pfuncs prog))
    bid instrs nbid (nfid:fidT) (nfids:Set![fidT]) args retto
    (Hcmd : MapsTo bid (instrs, tmn_call nbid nfids args retto) (fcmds func))
    (Hnfid: nfid !\in nfids)
    stack (Hc : pidx_consistent prog fid stack)
    nfunc (Hnfunc : MapsTo nfid nfunc (pfuncs prog))
    (Hcf : (fid, bid) !\in ((callfroms acc) [nfid]! : Set![fidT*bidT]))
    (argvals:list valueT) (params: list idT)
    (Hparams : forall i : idT,
                 List.In i params →
                 (llocal nfid i nil) !\in (acclocs acc) [nfid]!)
    (mem1 mem2 : Map! [locT, valueT])
    (Hsem : le (sem_bind_params
                  acc nfid argvals params
                  (mrestrict (E:=valueT)
                             (set_to_powerset
                                ((acclocs acc) [nfid]! : Set![locT]))
                             (Elt mem1)))
               (Elt mem2)) :
    sem_bind_params
      acc nfid argvals params
      (mrestrict (E:=valueT) (set_to_powerset
                                ((acclocs acc) [nfid ]! : Set![locT]))
                 (Elt mem1))
      ⊑ (Elt mem2).
  Proof.
    elim: argvals params Hparams mem1 mem2 Hsem=> [|argval argvals IH] params Hparams mem1 mem2.
    { rewrite/List.map/sem_exprs/sem_bind_params.
      by move=> Hle; rewrite/le/le_/memT_poset/withtop_poset/withtop_le in Hle.
    }
    move: params Hparams=> [|param params] Hparams.
    { rewrite/sem_exprs/sem_bind_params.
      by move=> Hle; rewrite/le/le_/memT_poset/withtop_poset/withtop_le in Hle.
    }
    move=> Hle.
    rewrite/List.map.
    rewrite sem_bind_params_simpl in Hle.
    exploit (IH params); auto.
    { move=> i Hi.
      apply Hparams.
        by right.
    }
    { instantiate (1 := mem2).
      eapply le_trans; last by apply Hle.
      apply DomType.eq_le.
      eapply sem_bind_params_mor; try reflexivity.
      rewrite <- memory_update_mrestrict_commutative.
      apply mrestrict_eq_mor; try reflexivity.
      rewrite/memory_update.
      rewrite elements_singleton.
      reflexivity.
      by apply Hparams; left.
    }
  Qed.

  Lemma tstate_call
        fid func (Hfunc : MapsTo fid func (pfuncs prog))
        bid instrs nextbid (nfid:fidT) (nfids:Set![fidT]) args retto (Hcmd : MapsTo bid (instrs, tmn_call nextbid nfids args retto) (fcmds func))
        (Hnfid: nfid !\in nfids)
        stack (Hc : pidx_consistent prog fid stack)
        (Hcall : postfix_call prog acc state fid func bid nextbid nfid args retto) :
    SemSenLocal.postfix_call prog acc (tstate prog state) fid func bid stack nextbid nfid args.
  Proof.
    rewrite/postfix_call in Hcall.
    move=> nfunc Hnfunc.
    apply MF.find_mapsto_iff in Hnfunc; rewrite Hnfunc in Hcall; apply MF.find_mapsto_iff in Hnfunc.
    move/and3P: Hcall=> [Hcf Hin Hout].
    do ? rewrite ? tstate_lookup
       /pifid/pibid/DomPidxSen.pidxT_pidx/fst/snd ? Hc.
    split; last first.
    { move: Hout=> {Hin}.
      rewrite ? tstate_lookup /DomPidx.pifid.
      rewrite/DomPidxSen.pidxT_pidx/fst/snd Hc.
      by move/lebP.
    }
    { move: Hin=> {Hout}.
      rewrite ? tstate_lookup /DomPidx.pifid.
      rewrite/DomPidxSen.pidxT_pidx/fst/snd Hc.
      rewrite {1}simpl_pidx_consistent.
      apply MF.find_mapsto_iff in Hfunc; rewrite Hfunc.
      apply MF.find_mapsto_iff in Hcmd; rewrite Hcmd.
      rewrite/ctmn/snd Hnfid /andb Hc.

      case Hmem2: ((state) [(nfid, fentry nfunc, false) ]!)=> [mem2|]; last by move=> _; apply: top_prop.

      case Hmem1: ((state) [(fid, bid, true) ]!)=> [mem1|]; dom_tac; last first.
      { move=> Hcon {Hcmd}; elimtype False.
        elim: args (fparams nfunc) Hcon=> [|arg args IH] params Hcon; first done.
        move: params Hcon=> [|param params]; first done.
        rewrite sem_exprs_simpl.
        rewrite sem_bind_params_simpl.
        rewrite/memory_update/=.
        exact: IH.
      }

      rewrite/memT_restricting_maplike/withtop_restricting_maplike.
      move: (sem_exprs _ _ _)=> argvals.
      by move/lebP.
    }
  Qed.
  
  Lemma tstate_exit
        fid func (Hfunc : MapsTo fid func (pfuncs prog))
        bid instrs ret (Hcmd : MapsTo bid (instrs, tmn_exit ret) (fcmds func))
        stack (Hc : pidx_consistent prog fid stack)
        (Hexit : postfix_exit prog acc state fid func bid ret) :
    SemSenLocal.postfix_exit prog acc (tstate prog state)
                             fid func bid stack ret.
  Proof.
    move: Hexit.
    move: stack Hc=> [|[pfid pbid] pstack] Hc; first done.
    move/set_forallP.
    refine (implies_right_turn (Proper _ _) _ _ _ _).
    { move=> x y Hxy; usual_tac Hxy.
      exact: eq_refl.
    }
    move=> Hpostfix pfunc Hpfunc pcmd Hpcmd.
    move: Hacc'=> [Hcf _ _ _].
    move: {Hcf} (Hcf _ _ Hpfunc _ _ Hpcmd).
    move: pcmd Hpcmd=> [pinstrs [|nbid nfid args retto|]]; try done; move=> Hpcmd.
    rewrite/SemAccSen.acc_correct_callfroms_blk/ctmn/snd=> Hcf.
    move: (Hc); rewrite simpl_pidx_consistent; move/andP=> [Hco Hcostack].
    apply MF.find_mapsto_iff in Hpfunc; rewrite Hpfunc in Hco.
    apply MF.find_mapsto_iff in Hpcmd; rewrite Hpcmd in Hco.
    rewrite/ctmn/snd/= in Hco.
    move/set_forallP in Hcf.
    exploit (Hcf _ fid); auto.
    move=> {Hcf} Hcf.
    move: {Hpostfix} (Hpostfix _ Hcf)=> Hpostfix.
    rewrite Hpfunc Hpcmd in Hpostfix.
    move/lebP: Hpostfix.
    rewrite/sem_bind_ret.
    rewrite ? tstate_lookup /DomPidx.pifid.
    rewrite/DomPidxSen.pidxT_pidx/fst/snd Hc Hcostack.
    move: (sem_expr _ _ _)=> retval.
    rewrite {1}/memory_update.
    rewrite/is_strong/var_loc/lglobal/llocal.
    rewrite/DomPidx.pifid/DomPidxSen.pidxT_pidx/pidxT_pidx/fst/snd.
    rewrite/memory_update.
    rewrite/var_loc/lglobal/llocal.

    case Hmem2: ((state) [(pfid, nbid, false) ]!)=>[mem2|]; dom_tac; last by move=> _; apply: top_prop.

    case Hmem1: ((state) [(fid, bid, true) ]!)=>[mem1|]; dom_tac; last by case: (not_recursive acc fid).                

    rewrite ? if_Elt; dom_tac.
    rewrite ? DomMap.if_madd; dom_tac.
    rewrite/memory_lookup; dom_tac.
    rewrite/mlookup/memT_maplike/withtop_maplike.
    rewrite/map_maplike/memT_poset/withtop_poset/withtop_le.
    by move=> Hle.
  Qed.

  Lemma tstate_correct (Hpostfix: SemInsenLocal.postfix prog acc state) :
    SemSenLocal.postfix prog acc (tstate prog state).
  Proof.
    move/map_forallP: Hpostfix.
    refine (implies_right_turn (Proper _ _) _ _ _ _).
    refine (forall_implies _ _ _ _) => fid.
    refine (forall_implies _ _ _ _) => func.
    refine (forall_implies _ _ _ _) => Hfunc.
    move/map_forallP.
    refine (implies_right_turn (Proper _ _) _ _ _ _).
    refine (forall_implies _ _ _ _) => bid Hpostfix stack [instrs tmn] Hcmd.
    move: {Hpostfix} (Hpostfix _ Hcmd).

    rewrite ? tstate_lookup /DomPidx.pifid.
    rewrite/DomPidxSen.pidxT_pidx/fst/snd.
    
    case Hc: (pidx_consistent _ _ _); last by move=> _; split; reflexivity.
    move/andP=> [Hinstrs Htmn].
    split.
    { move/lebP in Hinstrs.
      apply: tstate_instrs; auto.
      move/and4P: Hacc=> [_ _ Hal _].
      move/map_forallP: Hal.
      refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hal.
      move: {Hal} (Hal _ _ Hfunc); move/andP=> [Hal _].
      move/map_forallP: Hal.
      refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hal.
      by move: {Hal} (Hal _ _ Hcmd); move/andP=> [Hal _].
    }
    { move: Htmn=> {Hinstrs}.
      move: tmn Hcmd=> [nextbids|nextbids nextfid args retto|nextbids] Hcmd.
      { by eapply tstate_jump; eauto. }
      { move/set_forallP.
        refine (implies_right_turn (Proper _ _) _ _ _ _).
        refine (forall_implies _ _ _ _) => nbid.
        refine (forall_implies _ _ _ _) => Hnbid.
        by eapply tstate_call; eauto.
      }
      { by eapply tstate_exit; eauto. }
    }
  Qed.
End Postfixs.
