Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomBridge.

Require Import DomAbsValue.
Require Import DomSet.
Require Import DomMap.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomAbsState.

Require Import Syntax.

Require Import SemAbsBlock.
Require Import SemAccInsen.
Require Import SemInsenLocal.

Require Import Main.

Require Import SemCon.
Require Import CorAbs.
Require Import CorCon.
Require Import CorConExpr.
Require Import CorAcc.
Require Import CorAccExpr.
Require Import CorBasic.
Require Import DomPidxSen.
Require Import DomPidx.

Set Implicit Arguments.


Section Imports.
  Require Import DomZ.
  Require Import DomSet.
  Require Import DomConValue.
  Require Import DomConMem.

  Require Import SemCon.
  Require Import DomConValue.
  Require Import DomAbsValue.
  Require Import DomAcc.
  Require Import DomAbsMem.
  Require Import DomPidx.
  Require Import DomPidxSen.

  Require Import SemAbsBlock.
  Require Import SemAccSen.
  Require Import SemSen.

  Require Import CorBasic.
  Require Import CorAbs.
  Require Import CorItv.
End Imports.

Require Import SemUse.


Section AccGammaUses.
  Variables (prog:progT) (acc:accT).

  Definition gamma_local (pidx:pidxT) (lhs:DomConMem.memT) (rhs:memT): Prop :=
    forall (cloc: DomConValue.locT) (Hcloc: (t_loc cloc) !\in (acclocs acc)[pifid pidx]!),
      let cv: option DomConValue.valueT := lhs[cloc]! in
      let av: DomAbsValue.valueT := rhs[t_loc cloc]! in
      cv \abs av.

  Global Add Parametric Morphism: gamma_local with
      signature eq ==> eq ==> le ==> le as gamma_local_mor.
  Proof.
    move=> x y Hxy z w Hzw a b Hab.
    usual_tac Hxy.
    rewrite/gamma_local.
    refine (forall_implies _ _ _ _) => cloc.
    refine (forall_implies _ _ _ _) => Hcloc.
    move=> Habs.
    by rewrite <- Hzw; rewrite <- Hab.
  Qed.

  Definition finite_con_mem (mem:DomConMem.memT): Prop :=
    exists locs: list DomConValue.locT,
      forall loc (Hloc: (mem[loc]!: option DomConValue.valueT) <> None), List.In loc locs.

  Global Add Parametric Morphism: finite_con_mem with
      signature eq ==> eq as finite_con_mem_mor.
  Proof.
    move=> x y Hxy.
    constructor=> Hfin.
    { move: Hfin=> [locs Hlocs].
      exists locs.
      move=> loc Hloc.
      apply Hlocs.
      by rewrite Hxy.
    }
    { move: Hfin=> [locs Hlocs].
      exists locs.
      move=> loc Hloc.
      apply Hlocs.
      by rewrite <- Hxy.
    }
  Qed.

  Definition finite_con_mem_madd
             (mem:DomConMem.memT)
             (loc: DomConValue.locT) (val: option DomConValue.valueT)
             (Hfin: finite_con_mem mem):
    finite_con_mem mem[loc <- val]!.
  Proof.
    move: Hfin=> [locs Hlocs].
    exists (loc::locs)=> l.
    rewrite morphism_madd_spec.
    case Heq: (eqb _ _).
    { move/eqbP in Heq; usual_tac Heq.
      by move=> _; left.
    }
    { move=> Hmem.
      by right; apply Hlocs.
    }    
  Qed.

  Definition finite_con_mem_mrestrict
             (mem:DomConMem.memT)
             (ls: \powerset DomConValue.locT)
             (Hfin: finite_con_mem mem):
    finite_con_mem (mrestrict (E:=option DomConValue.valueT) ls mem).
  Proof.
    move: Hfin=> [locs Hlocs].
    exists locs=> loc Hloc.
    apply Hlocs.
    rewrite morphism_mrestrict_spec in Hloc.
    by case: (_ !\in _) in Hloc.
  Qed.
  
  Lemma finite_con_mem_instr
        (pidx:pidxT) (istr:instrT) (mem1 mem2:DomConMem.memT)
        (Hsem: SemCon.sem_instr pidx istr mem1 mem2)
        (Hfin: finite_con_mem mem1):
    finite_con_mem mem2.
  Proof.
    inv Hsem; try rewrite <- Hsubst.
    { by apply finite_con_mem_madd. }
    { by apply finite_con_mem_madd. }
    { by rewrite <- Heq. }
    { inv Hlcall; (try rewrite <- Heq); (try by apply finite_con_mem_madd); try done.
      { apply finite_con_mem_madd.
        rewrite <- Hcopy.
        by apply finite_con_mem_madd.
      }
      { rewrite <- Hcopy.
        by apply finite_con_mem_madd.
      }
      { apply finite_con_mem_madd.
        elim: args0 {Hwf_mem' Heq} mem' IH=> [|arg args IH]  mem' Hargs; inv Hargs.
        { by rewrite <- Heq. }
        { rewrite <- Heq.
          apply finite_con_mem_madd.
          by apply IH.
        }
      }
    }
    { rewrite <- Hforget.
      by apply finite_con_mem_mrestrict.
    }
    { by rewrite <- Heq. }
  Qed.

  Lemma finite_con_mem_instrs
        (pidx:pidxT) (istrs:list instrT) (mem1 mem2:DomConMem.memT)
        (Hsem: SemCon.sem_instrs pidx istrs mem1 mem2)
        (Hfin: finite_con_mem mem1):
    finite_con_mem mem2.
  Proof.
    induction Hsem.
    { by rewrite <- Heq. }
    { apply IHHsem.
      by eapply finite_con_mem_instr; eauto.
    }
  Qed.

  Lemma finite_con_mem_step
        (pidx1 pidx2:pidxT) (mem1 mem2:DomConMem.memT)
        (Hsem: sem_step prog (pidx1,mem1) (pidx2,mem2))
        (Hfin: finite_con_mem mem1):
    finite_con_mem mem2.
  Proof.
    inv Hsem.
    { by eapply finite_con_mem_instrs; eauto. }
    { by rewrite <- Heq. }
    { elim: argvals mem1 Hfin {Hwf_mnew Hargs} mem2 (fparams nfunc) Hbind=> [|argval argvals IH] mem1 Hfin mem2 params Hbind; inv Hbind.
      { by rewrite <- Heq. }
      { exploit finite_con_mem_madd; eauto.
        erewrite Hsubst.
        move=> Hmmid.
        by eapply IH; eauto.
      }
    }
    { inv Hbind.
      move:retto Hcall H=> [retto|] Hcall H; last first.
      { rewrite <- H.
        by apply finite_con_mem_mrestrict.
      }
      { rewrite <- H.
        apply finite_con_mem_madd.
        by apply finite_con_mem_mrestrict.
      }
    }
  Qed.

  Lemma finite_con_mem_sem
        (pidx:pidxT) (mem:DomConMem.memT)
        step (Hsem: sem prog step (pidx,mem)):
    finite_con_mem mem.
  Proof.
    elim: step pidx mem Hsem=> [|step IH] pidx mem Hsem; inv Hsem.
    { rewrite Hmemeq.
      exists nil.
      move=> loc Hloc.
      by rewrite/= in Hloc.
    }
    { move: midstate {Hwf_mid} Hprog Hstep=> [ppp mmm] Hprog Hstep.
      eapply finite_con_mem_step; eauto.
    }
  Qed.

  Lemma gamma_local_extension
        (pidx:pidxT) (lhs:DomConMem.memT) (Hlhs: finite_con_mem lhs) (rhs:memT)
        (Hlocal: gamma_local pidx lhs rhs):
    exists mem,
      lhs \abs (mrestrict_join (K:=locT) (E:=valueT) (set_to_powerset (acclocs acc)[pifid pidx]!) rhs (Elt mem)).
  Proof.
    move: Hlhs=> [locs Hlocs].
    move: rhs Hlocal=> [rhs|] Hlocal; last by exists bot.
    exists (List.fold_right
              (fun loc mem => madd (t_loc loc) (top:valueT) mem)
              bot
              locs).
    move=> loc.
    case Hv: _[_]!=> [v|]; last done.
    rewrite/mrestrict_join mem_join_spec.
    case Hin: (t_loc loc !\in (acclocs acc)[pifid pidx]!).
    { rewrite <- join_lhs.
      rewrite mem_mrestrict_spec set_to_powerset_spec Hin.
      rewrite <- Hv.
      by apply Hlocal.
    }
    { rewrite <- join_rhs.
      exploit Hlocs.
      { instantiate (1 := loc).
          by rewrite Hv.
      }
      move=> {Hlocs} Hlocs.
      rewrite mem_mrestrict_spec powerset_neg_spec set_to_powerset_spec Hin /negb.
      rewrite/mlookup/DomAbsMem.memT_maplike/withtop_maplike.
      elim: locs Hlocs=> [|l locs IH] Hlocs; first done.
      unfold fold_right; fold fold_right.
      rewrite map_madd_spec.
      case Heq: (eqb _ _).
      { by apply abs_value_top. }
      { apply IH.
        inv Hlocs; last done.
        move/eqbP in Heq.
        contradict Heq.
        reflexivity.
      }
    }
  Qed.

  Variables
    (Hcf: SemAccSen.acc_correct_callfroms_prog prog acc)
    (Hcs: SemAccSen.acc_correct_callfroms_star acc).

  Lemma gamma_local_instr
        (pidx:pidxT) (Hpidx: pidx_consistent prog (pifid pidx) pidx.2)
        (istr:instrT)
        (cm1 cm2:DomConMem.memT)
        (Hcm1fin: finite_con_mem cm1)
        (Hsem: SemCon.sem_instr pidx istr cm1 cm2)
        (am1:memT) (Hacc: SemAccInsen.acc_correct_acclocs_instr acc (pifid pidx) istr am1)
        (Hgamma: gamma_local pidx cm1 am1):
    gamma_local pidx cm2 (SemAbsBlock.sem_instr acc (pifid pidx) istr am1).
  Proof.
    exploit gamma_local_extension; [by eauto|by eauto|].
    move=> [mem Hmem].
    move: (tacc_correct_acclocs_instr prog acc (pifid pidx) pidx.2 istr am1 Hpidx Hacc)=> {Hacc} Hacc.
    rewrite/SemAccSen.acc_correct_acclocs_instr in Hacc.
    exploit sem_instr_proper2; eauto.
    move=> Hcm2.
    move: am1 Hcm2 Hgamma Hmem Hacc=> [am1|] Hcm2 Hgamma Hmem Hacc; last first.
    { move=> cloc Hcloc.
      exploit sem_instr_proper2; try by eauto.
      move=> Hcm2'.
      move: (SemAbsBlock.sem_instr _ _ _ _) Hcm2'=> [mem'|] Hcm2'; last first.
      { case: _[_]!; last done.
          by apply abs_value_top.
      }
      by apply Hcm2'.
    }
    eapply abs_relation_proper in Hcm2; first last.
    { by apply Hacc. }
    { reflexivity. }
    rewrite/gamma_local.
    move: (SemAbsBlock.sem_instr _ _ _ _) ((acclocs acc)[_]!) Hcm2=> [am2|] locs Hcm2 cloc Hcloc; last first.
    { case: (_[_]!); last done.
      by apply abs_value_top.
    }
    eapply abs_relation_proper; [reflexivity|idtac|apply Hcm2].
    rewrite/mrestrict_join mem_join_spec.
    rewrite ? mem_mrestrict_spec.
    rewrite powerset_neg_spec set_to_powerset_spec Hcloc /negb.
    apply join_lub; last by apply bot_prop.
    by apply DomType.eq_le; reflexivity.
  Qed.

  Lemma gamma_local_instrs
        (pidx:pidxT) (Hpidx: pidx_consistent prog (pifid pidx) pidx.2)
        (istrs:list instrT)
        (cm1 cm2:DomConMem.memT)
        (Hcm1fin: finite_con_mem cm1)
        (Hsem: SemCon.sem_instrs pidx istrs cm1 cm2)
        (am1:memT) (Hacc: acc_correct_acclocs_instrs acc (pifid pidx) istrs am1)

        (Hgamma: gamma_local pidx cm1 am1):
    gamma_local pidx cm2 (SemAbsBlock.sem_instrs acc (pifid pidx) istrs am1).
  Proof.
    elim: istrs cm1 cm2 Hcm1fin Hsem am1 Hacc Hgamma=> [|istr istrs IH] cm1 cm2 Hcm1fin Hsem am1 Hacc Hgamma.
    { inv Hsem.
      by rewrite <- Heq.
    }
    { inv Hsem.
      move/andP: Hacc=> [Histr Histrs].
      eapply IH; first last; eauto.
      { by eapply gamma_local_instr; eauto. }
      { by eapply finite_con_mem_instr; eauto. }
    }
  Qed.

  Lemma acc_gamma_use_expr (pidx:pidxT) (e: exprT)
        (lhs:DomConMem.memT) (Hlhsfin: finite_con_mem lhs)
        (rhs:memT)
        (Hlocal: gamma_local pidx lhs rhs)
        (Hacc: SemAccInsen.acc_correct_acclocs_expr acc (pifid pidx) e rhs):
    gamma_use_expr pidx lhs rhs e

  with acc_gamma_use_lv (pidx:pidxT) (l:lv)
                        (lhs:DomConMem.memT) (Hlhsfin: finite_con_mem lhs)
                        (rhs:memT)
                        (Hlocal: gamma_local pidx lhs rhs)
                        (Hacc: SemAccInsen.acc_correct_acclocs_lv acc (pifid pidx) l rhs):
         gamma_use_lv pidx lhs rhs l.
  Proof.
    elim: e Hacc=> [l|op e1 IHe1 e2 IHe2|op e1 IHe1|l|l|ct e1 IHe1|e1 IHe1]; rewrite/acc_correct_acclocs_expr=> Hacc.
    { by constructor. }
    { move/andP: Hacc=> [He1' He2'].
      constructor.
      { by apply IHe1. }
      { by apply IHe2. }
    }
    { constructor.
      by apply IHe1.
    }
    { move/andP: Hacc=> [Hacc Hacc'].
      econstructor; eauto.
      move=> cloc Hcloc.
      exploit gamma_local_extension; first eauto; first eauto.
      move=> [mem Hmem].
      move: Hacc (tacc_correct_acclocs_lv acc l (pifid pidx) rhs Hacc)=> Hacc0 Hacc.
      move: (Hacc mem)=> {Hacc} Hacc.
      move: rhs Hlocal Hacc Hacc' Hmem Hacc0=> [rhs|] Hlocal Hacc Hacc' Hmem Hacc0; last first.
      { rewrite{2}/mlookup/DomAbsMem.memT_maplike/withtop_maplike.
        rewrite/abs_relation/option_abs.
        case: _[_]!; last done.
        by apply abs_value_top.
      }
      move: (_[_]!) Hmem Hacc Hacc'=> locs Hmem Hacc Hacc'.
      exploit sem_expr_lv_proper2_; [by auto|by auto|by eauto|by eauto|].
      move=> Hcloc'.
      rewrite -> Hacc in Hcloc'.
      move: (sem_expr_lv _ _ _) Hacc' Hcloc'=> [locs'|] Hacc' Hcloc'; last done.
      apply SetFacts.subset_iff in Hacc'.
      rewrite/abs_relation/loc_abs/smem/= in Hcloc'.
      apply SetFacts.mem_iff in Hcloc'.
      apply Hacc' in Hcloc'.
      apply SetFacts.mem_iff in Hcloc'.
      move:Hcloc'; dom_tac=> Hcloc'.
      eapply abs_relation_proper; [reflexivity|idtac|by apply Hmem].
      rewrite/mrestrict_join mem_join_spec.
      rewrite/mrestrict/memT_restricting_maplike/withtop_restricting_maplike.
      rewrite/mlookup/DomAbsMem.memT_maplike/withtop_maplike.
      apply join_lub.
      { rewrite map_mrestrict_spec set_to_powerset_spec.
        rewrite/mrestrict/map_restricting_maplike Hcloc'.
        apply DomType.eq_le.
        reflexivity.
      }
      { rewrite map_mrestrict_spec powerset_neg_spec set_to_powerset_spec Hcloc' /negb.
        by apply bot_prop.
      }
    }
    { by constructor; auto. }
    { constructor.
      by apply IHe1.
    }
    { constructor.
      by apply IHe1.
    }

    elim: l Hacc=> [x|e|arr idx|strct IH pos] Hacc.
    { by constructor. }
    { by constructor; auto. }
    { move/andP: Hacc=> [He1 He2].
      by constructor; auto.
    }
    { constructor.
      by apply IH.
    }
  Qed.

  Lemma acc_gamma_use_exprs (pidx:pidxT) (es: list exprT)
        (lhs:DomConMem.memT) (Hlhsfin: finite_con_mem lhs)
        (rhs:memT)
        (Hlocal: gamma_local pidx lhs rhs)
        (Hacc: SemAccInsen.acc_correct_acclocs_exprs acc (pifid pidx) es rhs):
    gamma_use_exprs pidx lhs rhs es.
  Proof.
    elim: es Hacc=> [|e es IH] Hacc; first by constructor.
    move/andP: Hacc=> [He Hes].
    constructor.
    by apply acc_gamma_use_expr.
    by apply IH.
  Qed.

  Lemma gamma_local_top (pidx:pidxT) (mem:DomConMem.memT):
    gamma_local pidx mem top.
  Proof.
    move=> cloc Hcloc.
    case: _[_]!; last done.
    move=> v cv av.
    by apply abs_value_top.
  Qed.

  Lemma gamma_local_memory_update (pidx:pidxT) : forall 
    (Hpidx_cons: pidx_consistent prog (pifid pidx) pidx.2)
    (Hacc_prog: acc_correct_callfroms_prog prog acc)
    (Hacc_star: acc_correct_callfroms_star acc)
    (m:DomConMem.memT) (m':memT) 
    (l:DomConValue.locT) (l':locsT) (v:DomConValue.valueT) (v':valueT)
    (Hmem: gamma_local pidx m m') (Hloc: l \abs l') (Hval: Some v \abs v')
    (Hmem_wf: mem_wf pidx (madd (Key:=DomConValue.locT) (Elt:=option DomConValue.valueT) (M:=DomConMem.memT) l (Some v) m)),
    gamma_local pidx
                (madd (Key:=DomConValue.locT) (Elt:=option DomConValue.valueT) (M:=DomConMem.memT) l (Some v) m)
                (memory_update acc l' v' m').
  Proof.
    move=> Hpidx Hacc_prog Hacc_star m m' l l' v v' Hmem Hloc Hval Hwf.
    move: m' Hmem=> [m'|] Hmem; last by apply gamma_local_top.
    move: l' Hloc=> [l'|] Hloc; last by apply gamma_local_top.
    move=> loc Hloc''.
    intros; unfold cv, av; clear cv av.
    rewrite morphism_madd_spec memory_update_spec.
    case Heq: (eqb loc l).
    { move/eqbP in Heq; usual_tac Heq.
      rewrite Hloc.
      case: (_ && _); first done.
      eapply abs_relation_proper; [reflexivity|idtac|by eauto].
      exact: join_rhs.
    }
    case Hin: (_ !\in _); last by apply Hmem.
    case Heq': (eqb _ _); last first.
    { eapply abs_relation_proper; [reflexivity|idtac|by eauto].
      exact: join_lhs.
    }
    case Hstrong: (is_strong _ _); last first.
    { eapply abs_relation_proper; [reflexivity|idtac|by eauto].
      exact: join_lhs.
    }
    rewrite/andb.
    move/eqbP in Heq'.
    exploit cardinal_1_iff; [exact Hloc|exact Hin|done|move=> Hloc'].
    clear l' Hloc Hin Heq'.
    exploit (Hwf l); first by rewrite/= eqb_refl.
    move=> Hwf'.
    move: (Hwf loc)=> {Hwf} Hwf.
    rewrite morphism_madd_spec Heq in Hwf.
    move: (_[_]!) Hwf=> [cv|]; last done.
    move=> Hwf; exploit Hwf; [done|move=> {Hwf} Hwf].

    move: loc Heq Hstrong Hloc' Hloc'' Hwf=>
    [[[[[a u] o]|i]|[[fid id] depth]] fl] Heq Hstrong Hloc' Hloc'' Hwf;
    rewrite/= in Hstrong; (try by inv Hstrong);
    move: l Hwf Hwf' Heq Hloc'=>
    [[[[[a' u'] o']|i']|[[fid' id'] depth']] fl'] Hwf Hwf' Heq Hloc';
    usual_tac Hloc'; inv Hloc'.
    { move/eqbP in Heq; contradict Heq; reflexivity. }
    { contradict Heq.
      do ? split; try reflexivity.
      rewrite/snd.
      exploit not_recursive_imply_same_nth_fid; eauto.
      { by usual_tac Hwf; eauto. }
      { by usual_tac Hwf'; eauto. }
      by move=> ->; rewrite eqb_refl.
    }
  Qed.  

  Lemma acc_gamma_use_instr
        (pidx:pidxT) (Hpidx: pidx_consistent prog (pifid pidx) pidx.2)
        (instr: instrT)
        (lhs:DomConMem.memT) (Hlhsfin: finite_con_mem lhs)
        (rhs:memT)
        (Hacc_instr: SemAccInsen.acc_correct_acclocs_instr acc (pifid pidx) instr rhs)
        (Hlocal: gamma_local pidx lhs rhs):
    gamma_use_instr acc pidx lhs rhs instr.
  Proof.
    case: instr Hacc_instr Hlocal=> [l e|l a site|test|ol lc le site|l|] Hacc Hlocal; try econstructor.
    { move/andP: Hacc=> [Hacc _].
      move/andP: Hacc=> [Hacc _].
      by apply acc_gamma_use_lv.
    }
    { move/andP: Hacc=> [Hacc _].
      move/andP: Hacc=> [_ Hacc].
      by apply acc_gamma_use_expr.
    }
    { case: a Hacc=> [e s] Hacc.
      move/andP: Hacc=> [Hacc _].
      move/andP: Hacc=> [Hacc2 Hacc1].
      move/andP: Hacc2=> [Hacc3 Hacc2].
      move/andP: Hacc3=> [Hacc4 Hacc3].
      constructor.
      { by apply acc_gamma_use_lv. }
      { by apply acc_gamma_use_expr. }
      { by apply acc_gamma_use_expr. }
    }
    { by apply acc_gamma_use_expr. }
    { case: lc Hacc=> Hacc; (try done); rewrite/= in Hacc.
      { move/andP: Hacc=> [Hacc _].
        move/andP: Hacc=> [Hacc2 Hacc1].
        move/andP: Hacc2=> [Hacc3 Hacc2].
        move: ol Hacc1 Hacc3=> [ol|] Hacc1 Hacc3; last done.
        by apply acc_gamma_use_lv.
      }
      { move/andP: Hacc=> [Hacc _].
        move/andP: Hacc=> [Hacc2 Hacc1].
        move/andP: Hacc2=> [Hacc3 Hacc2].
        move: ol Hacc1 Hacc3=> [ol|] Hacc1 Hacc3; last done.
        by apply acc_gamma_use_lv.
      }
      { move/andP: Hacc=> [Hacc _].
        move/andP: Hacc=> [Hacc2 Hacc1].
        move/andP: Hacc2=> [Hacc3 Hacc2].
        move: ol Hacc1 Hacc3=> [ol|] Hacc1 Hacc3; last done.
        by apply acc_gamma_use_lv.
      }
      { move/andP: Hacc=> [Hacc _].
        move/andP: Hacc=> [Hacc2 Hacc1].
        move/andP: Hacc2=> [Hacc3 Hacc2].
        move: ol Hacc1 Hacc3=> [ol|] Hacc1 Hacc3; last done.
        by apply acc_gamma_use_lv.
      }
      { move/andP: Hacc=> [Hacc2 Hacc1].
        move/andP: Hacc2=> [Hacc3 Hacc2].
        move/andP: Hacc3=> [Hacc4 Hacc3].
        move=> ldst Hldst lsrc Hlsrc lsrcValue HlsrcValue lnewMem HlnewMem Hwf.
        move: ol Hacc2 Hacc4=> [ol|] Hacc2 Hacc4; last done.
        apply acc_gamma_use_lv; auto.
        { rewrite HlnewMem.
          by apply finite_con_mem_madd.
        }
        rewrite HlnewMem.
        move: rhs Hlocal Hacc1 Hacc3 Hacc2 Hacc4=> [rhs|] Hlocal Hacc1 Hacc3 Hacc2 Hacc4; last first.
        { move=> cloc Hcloc.
          intro cv; unfold cv; clear cv.
          rewrite memory_update_spec.
          case: _[_]!; last done.
          by apply abs_value_top.
        }
        move: le Hacc1 Hacc3 Hldst Hlsrc Hacc2 Hacc4=> [|lehd [|leid hetl]] Hacc1 Hacc3 Hldst Hlsrc Hacc2 Hacc4; unfold hd, tl in *.
        { inv Hldst.
          by inv Hlit.
        }
        { inv Hlsrc.
          by inv Hlit.
        }
        move/andP: Hacc1=> [Hacc1 Hacc1'].
        move/andP: Hacc1'=> [Hacc1' Hacc1''].
        move/andP: Hacc3=> [Hacc3 Hacc3'].
        move/andP: Hacc3'=> [Hacc3' Hacc3''].
        exploit acc_gamma_use_expr; [by eauto|by eauto|by apply Hacc3|idtac].
        exploit acc_gamma_use_expr; [by eauto|by eauto|by apply Hacc3'|idtac].
        move=> Hleid Hlehd.
        exploit gamma_use_expr_proper2_; [by auto|by auto|by apply Hldst|by eauto|idtac].
        exploit gamma_use_expr_proper2_; [by auto|by auto|by apply Hlsrc|by eauto|idtac].
        move=> Habsid Habshd.
        move=> cloc Hcloc.
        intro cv; unfold cv; clear cv.
        rewrite/abs_relation/value_abs in Habsid.
        move: (locs_of _) Habsid Hacc1'=> [locs|] Habsid Hacc1'; last done.
        exploit (Hlocal lsrc).
        { rewrite/smem/abs_relation/= in Habsid Hacc1' *.
          apply SetFacts.mem_iff.
          apply SetFacts.subset_iff in Hacc1'.
          apply Hacc1'.
          by apply SetFacts.mem_iff.
        }
        move=> Habs.
        eapply gamma_local_memory_update; eauto.
        { rewrite HlsrcValue.
          eapply abs_relation_proper; eauto.
          { apply memory_lookup_ub.
            by apply Habsid.
          }
          { by rewrite HlsrcValue. }
        }
        { by rewrite <- HlnewMem. }
      }
      { move/andP: Hacc=> [Hacc _].
        move/andP: Hacc=> [Hacc2 Hacc1].
        move/andP: Hacc2=> [Hacc3 Hacc2].
        move: ol Hacc1 Hacc3=> [ol|] Hacc1 Hacc3; last done.
        by apply acc_gamma_use_lv.
      }
      { move/andP: Hacc=> [Hacc _].
        move/andP: Hacc=> [Hacc2 Hacc1].
        move/andP: Hacc2=> [Hacc3 Hacc2].
        move: ol Hacc1 Hacc3=> [ol|] Hacc1 Hacc3; last done.
        by apply acc_gamma_use_lv.
      }
      { move/andP: Hacc=> [Hacc _].
        move/andP: Hacc=> [Hacc2 Hacc1].
        move/andP: Hacc2=> [Hacc3 Hacc2].
        move: ol Hacc1 Hacc3=> [ol|] Hacc1 Hacc3; last done.
        by apply acc_gamma_use_lv.
      }
      { move/andP: Hacc=> [Hacc _].
        move/andP: Hacc=> [Hacc2 Hacc1].
        move/andP: Hacc2=> [Hacc3 Hacc2].
        move: ol Hacc1 Hacc3=> [ol|] Hacc1 Hacc3; last done.
        by apply acc_gamma_use_lv.
      }
      { move/andP: Hacc=> [Hacc _].
        move/andP: Hacc=> [Hacc2 Hacc1].
        move/andP: Hacc2=> [Hacc3 Hacc2].
        move: ol Hacc1 Hacc3=> [ol|] Hacc1 Hacc3; last done.
        by apply acc_gamma_use_lv.
      }
    }
    { case: lc Hacc=> Hacc; (try done); rewrite/= in Hacc.
      { apply acc_gamma_use_exprs; auto.
        move/andP: Hacc=> [Hacc _].
        move/andP: Hacc=> [Hacc _].
        by move/andP: Hacc=> [_ Hacc].
      }
      { apply acc_gamma_use_exprs; auto.
        move/andP: Hacc=> [Hacc _].
        by move/andP: Hacc=> [Hacc _].
      }
      { apply acc_gamma_use_exprs; auto.
        move/andP: Hacc=> [Hacc _].
        move/andP: Hacc=> [Hacc _].
        by move/andP: Hacc=> [_ Hacc].
      }
      { apply acc_gamma_use_exprs; auto.
        move/andP: Hacc=> [Hacc _].
        move/andP: Hacc=> [Hacc _].
        by move/andP: Hacc=> [_ Hacc].
      }
    }
    { case: lc Hacc=> Hacc; (try done); rewrite/= in Hacc.
      move/andP: Hacc=> [Hacc' Hacc].
      move/andP: Hacc'=> [Hacc'' _].
      move/andP: Hacc''=> [_ Hacc'].
      elim: le Hacc Hacc'=> [|arg args IH] Hacc Hacc'; first by constructor.
      move/andP: Hacc=> [Hacc1 Hacc2].
      move/andP: Hacc'=> [Hacc1' Hacc2'].
      constructor.
      { move=> v Hv.
        move: v Hv=> [[v|v]|v] Hv; try done.
        clear IH Hacc2.
        exploit gamma_use_expr_proper2_; eauto.
        { 
          apply acc_gamma_use_expr; eauto.
        }
        done.
      }
      { by apply IH. }
    }
  Qed.

  Lemma acc_gamma_use_instrs
        (pidx:pidxT) (Hpidx: pidx_consistent prog (pifid pidx) pidx.2)
        (instrs: list instrT)
        (lhs:DomConMem.memT) (Hlhsfin: finite_con_mem lhs)
        (rhs:memT)
        (Hacc_instrs: SemAccInsen.acc_correct_acclocs_instrs acc (pifid pidx) instrs rhs)
        (Hlocal: gamma_local pidx lhs rhs):
    gamma_use_instrs acc pidx lhs rhs instrs.
  Proof.
    elim: instrs lhs rhs Hacc_instrs Hlocal Hlhsfin=> [|istr istrs IH] lhs rhs Hacc_instrs Hlocal Hlhsfin; first by constructor.
    move/andP: Hacc_instrs=> [Hacc_instr Hacc_instrs].
    econstructor; last first.
    { move=> cmem Hcmem.
      eapply IH.
      by apply Hacc_instrs.
      by eapply gamma_local_instr; eauto.
      by eapply finite_con_mem_instr; eauto.
    }
    by eapply acc_gamma_use_instr.
  Qed.

  Lemma acc_gamma_use_tmn (pidx:pidxT) (tmn:tmnT)
        (lhs:DomConMem.memT) (Hlhsfin: finite_con_mem lhs)
        (rhs:memT)
        (Hacc_tmn: SemAccInsen.acc_correct_acclocs_tmn acc (pifid pidx) tmn rhs)
        (Hlocal: gamma_local pidx lhs rhs):
    gamma_use_tmn pidx lhs rhs tmn.
  Proof.
    case: tmn Hacc_tmn=> [nbs|nb nfs args retto|ret] Hacc.
    { by constructor. }
    { move/andP: Hacc=> [Hargs Hretto].
      constructor.
        by apply acc_gamma_use_exprs.
        move: retto Hretto=> [retto|] Hretto; last done.
        rewrite/acc_correct_acclocs_var in Hretto.
        rewrite/var_loc/DomAbsValue.var_loc
               /DomConValue.lglobal/DomConValue.llocal
               /lglobal/llocal.
        move: retto Hretto=> [retto|retto] Hretto.
        { by apply Hlocal. }
        { by apply Hlocal. }
    }
    { constructor.
      by apply acc_gamma_use_expr.
    }      
  Qed.

  Lemma acc_gamma_use_blk
        (pidx:pidxT) (Hpidx: pidx_consistent prog (pifid pidx) pidx.2)
        (cmd:cmdT)
        (lhs:DomConMem.memT) (Hlhsfin: finite_con_mem lhs)
        (rhs:memT)
        (Hacc_instrs: SemAccInsen.acc_correct_acclocs_instrs acc (pifid pidx) (cinstrs cmd) rhs)
        (Hacc_tmn: SemAccInsen.acc_correct_acclocs_tmn acc (pifid pidx) (ctmn cmd) (SemAbsBlock.sem_instrs acc (pifid pidx) (cinstrs cmd) rhs))
        (Hlocal: gamma_local pidx lhs rhs):
    gamma_use_blk acc pidx cmd lhs rhs.
  Proof.
    move: cmd pidx Hpidx Hacc_instrs Hacc_tmn Hlocal=> [instrs tmn] [[[fid bid] epilogue] stack].
    rewrite/gamma_use_blk/cinstrs/ctmn/pifid/pibid/pidxT_pidx/fst/snd.
    move=> Hacc_instrs Hacc_tmn Hacc Hlocal.
    split; first by apply acc_gamma_use_instrs.
    move=> lout Hlout.
    apply acc_gamma_use_tmn; auto.
    { by eapply finite_con_mem_instrs; eauto. }
    by eapply gamma_local_instrs; eauto.
  Qed.
End AccGammaUses.


Section GammaUseLes.
  Variables (prog:progT) (acc:accT).

  Lemma gamma_use_le_expr (pidx:pidxT) (e:exprT) (lhs:DomConMem.memT) (mhs rhs:memT)
        (Hgamma_use: gamma_use_expr pidx lhs mhs e)
        (Huse_le: use_le_expr (pifid pidx) e mhs rhs):
    gamma_use_expr pidx lhs rhs e

  with gamma_use_le_lv (pidx:pidxT) (l:lv) (lhs:DomConMem.memT) (mhs rhs:memT)
                       (Hgamma_use: gamma_use_lv pidx lhs mhs l)
                       (Huse_le: use_le_lv (pifid pidx) l mhs rhs):
         gamma_use_lv pidx lhs rhs l.
  Proof.
    elim: Hgamma_use Huse_le; rewrite/use_le_expr.
    { by constructor. }
    { move=> o e1 e2 He1 IHe1 He2 IHe2.
      move/andP=> [He1' He2'].
      constructor.
      by apply IHe1.
      by apply IHe2.
    }
    { move=> o e1 He1 IHe1 He1'.
      constructor.
      by apply IHe1.
    }
    { move=> lv Hlv Habs.
      move/andP=> [Hlv' Heq].
      econstructor; eauto.
      move=> cloc Hcloc.
      eapply abs_relation_proper; eauto; first reflexivity.
      exploit gamma_use_lv_proper2_; eauto.
      move=> Hcloc'.
      move: (sem_expr_lv _ _ _) Heq Hcloc'=> [locs|] Heq Hcloc'; last done.
      rewrite SetInterface.fold_1 in Heq.
      rewrite <- fold_left_rev_right in Heq.
      rewrite/abs_relation/= in Hcloc'.
      rewrite SetFacts.elements_b in Hcloc'.
      rewrite existsb_rev in Hcloc'.
      elim: (rev (SetInterface.elements locs)) Heq Hcloc'=> [|loc' locs' IH] Heq Hcloc'.
      { by inv Hcloc'. }
      { rewrite/= in Hcloc'.
        move/andP: Heq=> [Heq' Heq].
        case Hloc: (t_loc cloc == loc') in Hcloc'; last by apply IH.
        clear -Hloc Heq.
        rewrite (is_compare_eqb (H2:=_)) in Hloc.
        move/eqbP in Hloc.
        usual_tac Hloc.
        by move/lebP in Heq.
      }
    }
    { move=> lv Hlv Hlv'.
      constructor; eauto.
    }
    { move=> ct e1 He1 IHe1 He1'.
      constructor.
      by apply IHe1.
    }
    { move=> e1 He1 IHe1 He1'.
      constructor.
      by apply IHe1.
    }

    elim: Hgamma_use Huse_le; rewrite/use_le_lv.
    { constructor. }
    { move=> e1 He1 He1'.
      constructor; eauto.
    }
    { move=> e1 e2 He1 He2.
      move/andP=> [He1' He2'].
      constructor; eauto.
    }
    { move=> l1 pos Hl1 IHl1 Hl1'.
      constructor.
      by apply IHl1.
    }
  Qed.

  Lemma gamma_use_le_exprs (pidx:pidxT) (es: list exprT) (lhs:DomConMem.memT) (mhs rhs:memT)
        (Hgamma_use: gamma_use_exprs pidx lhs mhs es)
        (Huse_le: use_le_exprs (pifid pidx) es mhs rhs):
    gamma_use_exprs pidx lhs rhs es.
  Proof.
    elim: es Hgamma_use Huse_le=> [|e es IH] Hgamma_use Huse_le; first by constructor.
    inv Hgamma_use.
    move/andP: Huse_le=> [He' Hes'].
    constructor.
    { by eapply gamma_use_le_expr; first by eauto. }
    { by apply IH. }
  Qed.

  Lemma gamma_use_le_instr (pidx:pidxT) (istr: instrT) (lhs:DomConMem.memT) (mhs rhs:memT)
        (Hgamma_use: gamma_use_instr acc pidx lhs mhs istr)
        (Huse_le: use_le_instr acc (pifid pidx) istr mhs rhs):
    gamma_use_instr acc pidx lhs rhs istr.
  Proof.
    case: istr Hgamma_use Huse_le=> [l e|l a site|test|ol lc le site|l|] Hgamma Hequiv; inv Hgamma.
    { move/andP: Hequiv=> [Hequiv2 Hequiv1].
      constructor.
      by eapply gamma_use_le_lv; eauto.
      by eapply gamma_use_le_expr; eauto.
    }
    { move/andP: Hequiv=> [Hequiv2 Hequiv1].
      move/andP: Hequiv2=> [Hequiv3 Hequiv2].
      constructor.
      { by eapply gamma_use_le_lv; eauto. }
      { by eapply gamma_use_le_expr; eauto. }
      { by eapply gamma_use_le_expr; eauto. }
    }
    { constructor.
      by eapply gamma_use_le_expr; eauto.
    }
    { constructor.
      { case: lc ol Hlvalue Hargs Hargs' Hequiv;
        move=> [ol|] Hlvalue Hargs Hargs' Hequiv; (try done);
               try (eapply gamma_use_le_lv; eauto);
               try by do 2! move/andP: Hequiv=> [Hequiv _].
        move=> ldst Hldst lsrc Hlsrc lsrcValue HlsrcValue lnewMem HnewMem Hwf.
        eapply gamma_use_le_lv; eauto.
        move: le Hlvalue Hargs Hargs' Hequiv Hldst Hlsrc=> [|le1 [|le2 le3]] Hlvalue Hargs Hargs' Hequiv Hldst Hlsrc; unfold hd,tl in *.
        { by inv Hldst; inv Hlit. }
        { by inv Hlsrc; inv Hlit. }
        inv Hargs; inv Hes.
        inv Hargs'; inv Hes.
        move/andP: Hequiv=> [Hequiv2 Hequiv1].
        move/andP: Hequiv2=> [Hequiv3 Hequiv2].
        move/andP: Hequiv2=> [Hequivle1 Hequivle2].
        move/andP: Hequivle2=> [Hequivle2 Hequivle3].
        exploit gamma_use_le_expr; eauto.
      }
      { case: lc ol Hlvalue Hargs Hargs' Hequiv;
        move=> [ol|] Hlvalue Hargs Hargs' Hequiv; (try done);
               try by (move/andP: Hequiv=> [Hequiv _];
                       move/andP: Hequiv=> [_ Hequiv];
                       eapply gamma_use_le_exprs; eauto).
      }
      { case: lc ol Hlvalue Hargs Hargs' Hequiv;
        move=> ol Hlvalue Hargs Hargs' Hequiv; (try done).
        move/andP: Hequiv=> [Hequiv2 Hequiv1].
        move/andP: Hequiv2=> [Hequiv3 Hequiv2].
        elim: {Hlvalue Hequiv3} le Hargs Hargs' Hequiv1 Hequiv2=> [|le1 le2 IH] Hargs Hargs' Hequiv1 Hequiv2; unfold hd,tl in *; first constructor.
        move/andP: Hequiv1=> [Hequiv11 Hequiv12].
        move/andP: Hequiv2=> [Hequiv21 Hequiv22].
        inv Hargs; inv Hargs'.
        constructor.
        { move=> [[v|v]|v] Hv; try done.
          exploit gamma_use_le_expr; eauto.
          move=> Hgamma.
          by exploit gamma_use_expr_proper2_; eauto.
        }
        { by apply IH. }
      }
    }
    { by constructor. }
    { by constructor. }
  Qed.

  Lemma gamma_use_le_instrs (pidx:pidxT) (istrs: list instrT) (lhs:DomConMem.memT) (mhs rhs:memT)
        (Hgamma_use: gamma_use_instrs acc pidx lhs mhs istrs)
        (Huse_le: use_le_instrs acc (pifid pidx) istrs mhs rhs):
    gamma_use_instrs acc pidx lhs rhs istrs.
  Proof.
    elim: istrs lhs mhs rhs Hgamma_use Huse_le=> [|istr istrs IH] lhs mhs rhs Hgamma Huse; first constructor.
    inv Hgamma.
    move/andP: Huse=> [Husei Huseis].
    constructor.
    { by eapply gamma_use_le_instr; eauto. }
    { move=> cmem Hcmem.
      by eapply IH; eauto.
    }      
  Qed.

  Lemma gamma_use_le_tmn (pidx:pidxT) (tmn:tmnT) (lhs:DomConMem.memT) (mhs rhs:memT)
        (Hgamma_use: gamma_use_tmn pidx lhs mhs tmn)
        (Huse_le: use_le_tmn (pifid pidx) tmn mhs rhs):
    gamma_use_tmn pidx lhs rhs tmn.
  Proof.
    inv Hgamma_use; constructor.
    { move/andP: Huse_le=> [Huse _].
      by eapply gamma_use_le_exprs; eauto.
    }
    { move: ret Hret Huse_le=> [ret|] Hret Huse_le; last done.
      move/andP: Huse_le=> [_ Hvar].
      eapply abs_relation_proper; last by eauto.
      reflexivity.
      by move/lebP in Hvar.
    }
    { by eapply gamma_use_le_expr; eauto. }
  Qed.

  Lemma gamma_use_le_blk (pidx:pidxT) (cmd:cmdT) (lhs:DomConMem.memT) (mhs rhs:memT)
        (Hgamma_use: gamma_use_blk acc pidx cmd lhs mhs)
        (Huse_le: use_le_blk acc (pifid pidx) (pibid pidx) cmd mhs rhs):
    gamma_use_blk acc pidx cmd lhs rhs.
  Proof.
    move: cmd pidx Hgamma_use Huse_le=> [instrs tmn] [[[fid bid] epilogue] stack].
    rewrite/gamma_use_blk/use_le_blk/cinstrs/ctmn/pifid/pibid/pidxT_pidx/fst/snd.
    move=> [Hgamma_use_instrs Hgamma_use_tmn].
    move/andP=> [Huse_le_instrs Huse_le_tmn].
    split.
    { by eapply gamma_use_le_instrs; eauto. }
    { move=> lout Hlout.
      by eapply gamma_use_le_tmn; eauto.
    }
  Qed.
End GammaUseLes.
