Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomBridge.

Require Import DomAbsValue.
Require Import DomSet.
Require Import DomMap.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomAbsState.

Require Import Syntax.

Require Import SemAbsBlock.
Require Import SemAccInsen.
Require Import SemInsenLocal.

Require Import Main.
Require Import SemUse.
Require Import CorUse.

Set Implicit Arguments.


Section Translations.
  Variables (prog:progT) (acc:accT).

  Definition t_state (state':stateT') : stateT :=
    MapInterface.fold
      (fun (fid:fidT) (func:funcT) (state:stateT) =>
         MapInterface.fold
           (fun (be:bidT * bool) (mem:memT) (state:stateT) =>
              state[(fid, be.1, be.2) <- mem]!)
           state'[fid]!
           state)
      (pfuncs prog)
      bot.

  Lemma simpl_t_state (state':stateT') (fid:fidT) (bid:bidT) (epilogue:bool) :
    eq (t_state state')[(fid,bid,epilogue)]!
       match (pfuncs prog)[fid] with
         | Some func => 
           state'[fid]![(bid,epilogue)]!
         | None => (bot:memT)
       end.
  Proof.
    rewrite/t_state.
    rewrite fold_1; rewrite <- fold_left_rev_right.
    rewrite MF.elements_o MF.findA_rev; last by apply elements_3w.
    assert (Hnodup: NoDupA (eq_key (H:=order_OrderedType)) (rev (elements (pfuncs prog)))).
    { apply NoDupA_rev; last by apply elements_3w.
      constructor=> * //=.
                   by move=> [x ?] [y ?] [z ?] /= -> ->.
    }
    elim: (rev _) Hnodup => [|[hdfid hdfunc] tl IH] Hnodup; first reflexivity.
    rewrite/= fold_1; rewrite <- fold_left_rev_right.
    rewrite/MF.eqb is_compare_eqb.
    case Heq: (eqb _ _).
    { move/eqbP in Heq; usual_tac Heq.
      rewrite [X in eq _ (coerce X)]MF.elements_o.
      rewrite [X in eq _ (coerce X)]MF.findA_rev; last by apply elements_3w.
      elim: (rev _)=> [|[[hdbid' hde'] hdmem'] tl' IH'].
      { rewrite/= in IH *.
        inv Hnodup.
        rewrite IH; last done.
        case Hfind: (findA _ _); last reflexivity.
        contradict H1.
        elim: tl {IH H2} Hfind=> [|[hda hdb] tl IH]; first done.
        rewrite/=/MF.eqb is_compare_eqb.
        case Heq: (eqb _ _); last by move=> Hfind; right; apply IH.
        move/eqbP in Heq; usual_tac Heq.
          by move=> _; left.
      }
      rewrite MF.add_o /=/MF.eqb ? is_compare_eqb.
      case Heq: (eqb _ _).
      { move/eqbP in Heq; usual_tac Heq; inv Heq.
        rewrite eqb_refl.
        reflexivity.
      }
      case Heq': (eqb _ _); last done.
      move/eqbP in Heq; contradict Heq.
      move/eqbP in Heq'; usual_tac Heq'; inv Heq'.
      reflexivity.
    }
    inv Hnodup.
    elim: (rev _)=> [|[[hdbid' hde'] hdmem'] tl' IH']; first by apply IH.
    rewrite MF.add_o is_compare_eqb.
    rewrite/fst/snd.
    case Heq': (eqb _ _); last done.
    move/eqbP in Heq; contradict Heq.
      by move/eqbP in Heq'; usual_tac Heq'; inv Heq'.
  Qed.

  Require Import SemAccSen.
  Require Import TrsLocal.
  Require Import TrsSen.

  Lemma TrsLocal_tmem_notop
        (state: SemAccSen.stateT)
        (fid: fidT) (bid: bidT) (epilogue: bool) (stack: list (fidT * bidT))
        (Hnotop: forall pidx, neq state[pidx]! (top: memT)):
    neq (TrsLocal.tmem acc state fid bid epilogue stack) (top:memT).
  Proof.
    elim: stack fid bid epilogue=> [|[pfid pbid] pstack IH] fid bid epilogue.
    { rewrite TrsLocal.simpl_tmem.
      move Hmem: (state[_]!)=> [mem|]; first done.
      elimtype False.
      apply (Hnotop (fid,bid,epilogue,nil)).
      by rewrite Hmem.
    }
    rewrite TrsLocal.simpl_tmem.
    move Hmem: (state[_]!)=> [mem|]; last first.
    { elimtype False.
      apply (Hnotop (fid,bid,epilogue,(pfid,pbid)::pstack)).
      by rewrite Hmem.
    }
    by move: (TrsLocal.tmem _ _ _ _ _ _) (IH pfid pbid true)=> {IH} [mem'|] Hmem'.
  Qed.

  Lemma TrsLocal_tstate_preserve
        (state: SemAccSen.stateT)
        (fid: fidT) (bid: bidT) (epilogue: bool) (stack: list (fidT * bidT))
        (Hpidx: DomPidxSen.pidx_consistent prog fid stack)
        (Hnotop: forall pidx, neq state[pidx]! (top: memT))
        (loc: DomAbsValue.locT) (Hloc: loc !\in (acclocs acc)[fid]!):
    eq (A:=valueT)
       (TrsLocal.tstate prog acc state)[(fid,bid,epilogue,stack)]![loc]!
       state[(fid,bid,epilogue,stack)]![loc]!.
  Proof.
    rewrite TrsLocal.tstate_lookup Hpidx.
    rewrite TrsLocal.simpl_tmem.
    move Hmem: (state[_]!)=> [mem|]; last first.
    { elimtype False.
      apply (Hnotop (fid,bid,epilogue,stack)).
      by rewrite Hmem.
    }
    case Hmem': (match stack with
                  | nil => bot
                  | (fid0, bid0) :: stack0 =>
                    TrsLocal.tmem acc state fid0 bid0 true stack0
                end)=> [mem'|]; last first.
    { case: stack Hpidx Hmem Hmem'=>[|[pfid pbid] pstack] Hpidx Hmem Hmem'; first done.
      exploit (TrsLocal_tmem_notop state pfid pbid true pstack); try done.
      by rewrite Hmem'.
    }
    rewrite/mrestrict_join.
    rewrite mem_join_spec.
    rewrite ? mem_mrestrict_spec.
    rewrite powerset_neg_spec set_to_powerset_spec Hloc /negb.
    by apply join_bot.
  Qed.

  Lemma TrsSen_tstate_preserve
        (state: SemAccInsen.stateT)
        (fid: fidT) (bid: bidT) (epilogue: bool) (stack: list (fidT * bidT))
        (Hpidx: DomPidxSen.pidx_consistent prog fid stack)
        (loc: DomAbsValue.locT):
    eq (A:=valueT)
       (TrsSen.tstate prog state)[(fid,bid,epilogue,stack)]![loc]!
       state[(fid,bid,epilogue)]![loc]!.
  Proof.
    rewrite TrsSen.tstate_lookup Hpidx.
    reflexivity.
  Qed.
End Translations.


Section Postfixs.
  Variables (prog:progT) (acc:accT).

  Lemma validate_postfix' (state propagated_state propagated_inter_state: stateT')
        (Hvalidate: validate prog acc state propagated_state propagated_inter_state) :
    validate_postfix prog acc (t_state prog propagated_state).
  Proof.
    move/and3P: Hvalidate=> [Hvalidate Hcfprog Hnotop].
    move/map_forallP: Hvalidate.
    refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hvalidate.
    apply/andP; split.
    { apply/and4P; do ? split.
      { apply/map_forallP.
        move: Hvalidate.
        refine (forall_implies _ _ _ _) => fid.
        refine (forall_implies _ _ _ _) => func.
        refine (forall_implies _ _ _ _) => Hfunc.
        move=> Hvalidate.
        move/andP: Hvalidate=> [Hvalidate _].
        by move/andP: Hvalidate=> [Hcf _].
      }
      { apply/map_forallP.
        move=> fid cf Hcf.
        move/map_forallP: Hcfprog.
        refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hcfprog.
        apply/equivP.
        { apply set_forallP.
            by move=> x y Hxy; usual_tac Hxy.
        }
        { reflexivity. }
        move=> [pfid pbid] Hp.
        rewrite/fst.
        move: (Hcfprog _ _ Hcf).
        move Hfunc: (pfuncs prog)[fid]=> [func|]; last done.
        move=> _.
        apply MF.find_mapsto_iff in Hfunc.
        move: (Hvalidate _ _ Hfunc)=> {Hvalidate} Hvalidate.
        move/andP: Hvalidate=> [Hvalidate _].
        move/and5P: Hvalidate=> [_ Hcs _ _ _].
        rewrite/acc_correct_acclocs_star_func in Hcs.
        move/set_forallP: Hcs.
        refine (implies_right_turn (Proper _ _) _ _ _ _).
        { by move=> x y Hxy; usual_tac Hxy. }
        move=> Hvalidate.
        rewrite/= in Hvalidate.
        apply MF.find_mapsto_iff in Hcf.
        rewrite Hcf /coerce in Hvalidate.
          by move: (Hvalidate _ Hp)=> {Hvalidate} Hvalidate.
      }
      { apply/map_forallP.
        move: Hvalidate.
        refine (forall_implies _ _ _ _) => fid.
        refine (forall_implies _ _ _ _) => func.
        refine (forall_implies _ _ _ _) => Hfunc.
        move=> Hvalidate.
        move/andP: Hvalidate=> [Hvalidate _].
        move/and4P: Hvalidate=> [_ _ Hal _].
        move/andP: Hal=> [Hblk Hparams].
        apply/andP; split; last done.
        apply/map_forallP.
        move/map_forallP: Hblk.
        refine (implies_right_turn (Proper _ _) _ _ _ _).
        refine (forall_implies _ _ _ _) => bid.
        refine (forall_implies _ _ _ _) => cmd.
        refine (forall_implies _ _ _ _) => Hcmd.
        apply MF.find_mapsto_iff in Hfunc.
        move/andP=> [Hinstrs Htmn].
        move/andP: Hinstrs=> [Hinstrs Htmn'].
        apply/andP; split.
        { by rewrite simpl_t_state Hfunc. }
        { by rewrite simpl_t_state Hfunc. }
      }
      { apply/map_forallP.
        move=> fid cf Hcf.
        move/map_forallP: Hcfprog.
        refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hcfprog.
        apply/equivP.
        { apply set_forallP.
            by move=> x y Hxy; usual_tac Hxy.
        }
        { reflexivity. }
        move=> [pfid pbid] Hp.
        rewrite/fst.
        move: (Hcfprog _ _ Hcf).
        move Hfunc: (pfuncs prog)[fid]=> [func|]; last done.
        move=> _.
        apply MF.find_mapsto_iff in Hfunc.
        move: (Hvalidate _ _ Hfunc)=> {Hvalidate} Hvalidate.
        move/andP: Hvalidate=> [Hvalidate _].
        move/and5P: Hvalidate=> [_ _ _ Hac _].
        rewrite/acc_correct_acclocs_callfroms_func in Hac.
        move/set_forallP: Hac.
        refine (implies_right_turn (Proper _ _) _ _ _ _).
        { by move=> x y Hxy; usual_tac Hxy. }
        move=> Hvalidate.
        rewrite/= in Hvalidate.
        apply MF.find_mapsto_iff in Hcf.
        rewrite Hcf /coerce in Hvalidate.
          by move: (Hvalidate _ Hp)=> {Hvalidate} Hvalidate.
      }
    }
    { apply/map_forallP.
      move: Hvalidate (Hvalidate)=> Hvalidate.
      refine (forall_implies _ _ _ _) => fid.
      refine (forall_implies _ _ _ _) => func.
      refine (forall_implies _ _ _ _) => Hfunc.
      move=> Hv.
      move/andP: Hv=> [Hv Hpost'].
      move/and5P: Hv=> [_ _ _ _ Hv].
      move/and3P: Hv=> [Hpost Hinter Hpro].
      move/map_forallP: Hpost.
      refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hpost.
      move/map_forallP: Hpost'.
      refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hpost'.
      apply/map_forallP.
      move=> bid [instrs tmn] Hcmd.
      move: (Hpost _ _ Hcmd)=> {Hpost} Hpost.
      move: (Hpost' _ _ Hcmd)=> {Hpost'} Hpost'.
      apply/andP; split.
      { move/andP: Hpost=> [Hle _].
        rewrite ? simpl_t_state.
        apply MF.find_mapsto_iff in Hfunc.
          by rewrite Hfunc.
      }
      { move/andP: Hpost=> [_ Htmn].
        rewrite/postfix_inter_block/snd in Hpost'.
        move: tmn Hcmd Htmn Hpost'=> [nbids|nbid nfids args retto|ret] Hcmd Htmn Hpost'.
        { rewrite/postfix_jump in Htmn.
          move/set_forallP: Htmn.
          refine (implies_right_turn (Proper _ _) _ _ _ _)=> Htmn.
          rewrite/SemInsenLocal.postfix_jump.
          apply/set_forallP=> nbid Hnbid.
          rewrite ? simpl_t_state.
          apply MF.find_mapsto_iff in Hfunc.
          rewrite Hfunc.
            by apply Htmn.
        }
        { move/set_forallP: Htmn.
          refine (implies_right_turn (Proper _ _) _ _ _ _)=> Htmn.
          move/set_forallP: Hpost'.
          refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hpost'.
          apply/set_forallP.
          move=> nfid Hnfid.
          move: (Htmn nfid Hnfid)=> {Htmn} Htmn.
          move: (Hpost' nfid Hnfid)=> {Hpost'} Hpost'.
          rewrite/postfix_call in Htmn.
          rewrite/postfix_inter_call in Hpost'.
          rewrite/SemInsenLocal.postfix_call.
          move Hnfunc: (pfuncs prog)[nfid] Htmn Hpost'=> [nfunc|] Htmn Hpost'; last done.
          apply/and3P; do ? split.
          { by move/andP: Htmn=> [? _]. }
          { rewrite ? simpl_t_state.
            rewrite Hnfunc.
            apply MF.find_mapsto_iff in Hfunc.
            rewrite Hfunc.
            move/lebP in Hpost'.
            apply/lebP.
            move/andP: Hinter=> [_ Hinter].
            move/map_forallP: Hinter.
            refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hinter.
            move: (Hinter _ _ Hcmd)=> {Hinter} Hinter.
            move/andP: Hinter=> [_ Hinter].
            move/eqbP in Hinter.
            rewrite Hinter.
            move: Hvalidate (Hvalidate)=> Hvalidate Hv.
            apply MF.find_mapsto_iff in Hnfunc.
            move: (Hv _ _ Hnfunc)=> {Hv} Hv.
            move/andP: Hv=> [Hv _].
            move/and5P: Hv=> [_ _ _ _ Hv].
            move/and3P: Hv=> [_ Hinter' _].
            move/andP: Hinter'=> [Hinter' _].
            move/eqbP in Hinter'; rewrite Hinter'.
            done.
          }
          { rewrite ? simpl_t_state.
            apply MF.find_mapsto_iff in Hfunc.
            rewrite Hfunc.
              by move/andP: Htmn=> [_ ?].
          }
        }
        { rewrite/postfix_inter_exit in Hpost'.
          move/set_forallP: Hpost'.
          refine (implies_right_turn (Proper _ _) _ _ _ _).
          { by move=> x y Hxy; usual_tac Hxy. }
          move=> Hpost'.
          rewrite/postfix_exit.
          apply/equivP.
          { apply set_forallP.
              by move=> x y Hxy; usual_tac Hxy.
          }
          { reflexivity. }
          move=> [pfid pbid] Hp.
          move: (Hpost' _ Hp)=> {Hpost'} Hpost'.
          rewrite/fst/snd in Hpost' *.
          move Hpfunc: (pfuncs prog)[pfid] Hpost'=> [pfunc|] Hpost'; last done.
          move Hpcmd: (fcmds pfunc)[pbid] Hpost'=> [[pinstrs pcmd]|] Hpost'; last done.
          rewrite/ctmn/snd in Hpost' *.
          move: pcmd Hpcmd Hpost'=> [nbids|nbid nfids args retto|pret] Hpcmd Hpost'; [done|idtac|done].
          move/lebP in Hpost'.
          apply/lebP.
          eapply le_trans; last first.
          { rewrite simpl_t_state.
            rewrite Hpfunc.
            move: Hvalidate (Hvalidate)=> Hvalidate Hv.
            apply MF.find_mapsto_iff in Hpfunc.
            move: (Hv _ _ Hpfunc)=> {Hv} Hv.
            move/andP: Hv=> [Hv _].
            move/and5P: Hv=> [_ _ _ _ Hv].
            move/and3P: Hv=> [_ Hinter' _].
            move/andP: Hinter'=> [_ Hinter'].
            move/map_forallP: Hinter'.
            refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hinter'.
            apply MF.find_mapsto_iff in Hpcmd.
            move: (Hinter' _ _ Hpcmd)=> {Hinter'} Hinter'.
            move/andP: Hinter'=> [Hinter' _].
            move/eqbP in Hinter'; rewrite Hinter'.
            by apply Hpost'.
          }

          move/andP: Hinter=> [_ Hinter].
          move/map_forallP: Hinter.
          refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hinter.
          move: (Hinter _ _ Hcmd)=> {Hinter} Hinter.
          move/eqbP in Hinter.

          apply DomType.eq_le.
          apply MF.find_mapsto_iff in Hfunc.
          apply sem_bind_ret_mor; try reflexivity.
          { apply sem_expr_mor; try reflexivity.
            by rewrite simpl_t_state Hfunc.
          }
          { case: (not_recursive _ _).
            { rewrite simpl_t_state Hfunc.
              rewrite <- Hinter.
              reflexivity.
            }
            { rewrite simpl_t_state Hfunc.
              rewrite <- Hinter.
              reflexivity.
            }
          }
        }
      }
    }
  Qed.
End Postfixs.


Section Correctness.
  Variables (prog:progT) (acc:accT).

  Require Import SemCon.
  Require Import CorAbs.
  Require Import CorConExpr.
  Require Import CorBasic.
  Require Import CorPostfix.

  Definition t_pidx (pidx: DomPidxSen.pidxT): DomPidxInsen.pidxT :=
    (pidx.1.1.1, pidx.1.1.2, pidx.1.2).

  Lemma notop_prop
        (state: func_stateT) (Hnotop: notop state) (be: bidT * bool):
    neq state[be]! (top: memT).
  Proof.
    move/map_forallP: Hnotop.
    refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hnotop.
    { move=> [bid epilogue] Hbe.
      by move=> [a|] [b|].
    }
    case Hbe: state[be]=> [mem|]; last by rewrite/= Hbe.
    rewrite/= Hbe /=.
    apply MF.find_mapsto_iff in Hbe.
    move: (Hnotop _ _ Hbe)=> {Hnotop} Hnotop.
    move: {Hbe} mem Hnotop=> [mem|]; last done.
    by move=> _ /=.
  Qed.

  Theorem validate_correct (state propagated_state propagated_inter_state: stateT')
          (Hvalidate: validate prog acc state propagated_state propagated_inter_state) :
    forall step fid bid stack cmem
           (Hsem : SemCon.sem prog step ((fid,bid,false,stack), cmem))
           func (Hfunc: (pfuncs prog)[fid] = Some func)
           cmd (Hcmd: (fcmds func)[bid] = Some cmd),
      gamma_use_blk
        acc (fid,bid,false,stack) cmd
        cmem
        state[fid]![(bid,false)]!.
  Proof.
    move=> step fid bid stack cmem Hsem func Hfunc cmd Hcmd.
    exploit sem_consistent; eauto.
    rewrite/DomPidx.pifid/DomPidxSen.pidxT_pidx/fst/snd in Hfunc *=> Hpidx.
    move/and3P: Hvalidate (Hvalidate) (Hvalidate) (Hvalidate)=> [Hvalidate Hcfprog Hnotop] Hv Hv' Hv''.
    move/map_forallP: Hvalidate.
    refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hvalidate.
    apply MF.find_mapsto_iff in Hfunc.
    move: (Hvalidate _ _ Hfunc)=> {Hvalidate} Hvalidate.
    apply MF.find_mapsto_iff in Hfunc.
    move/andP: Hvalidate=> [Hvalidate Hinter].
    move/and5P: Hvalidate=> [_ _ _ _ Hvalidate].
    move/and3P: Hvalidate=> [Hfix Hinter' Huse].
    eapply gamma_use_le_blk; auto; last first.
    { move/map_forallP: Huse.
      refine (implies_right_turn (Proper _ _) _ _ _ _)=> Huse.
      apply MF.find_mapsto_iff in Hcmd.
      by apply Huse.
    }
    move/and3P: Hv=> [Hv _ _].
    move/map_forallP: Hv.
    refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hv.
    apply MF.find_mapsto_iff in Hfunc.
    move: (Hv _ _ Hfunc)=> {Hv} Hv.
    move/andP: Hv=> [Hv Hinter''].
    move/and5P: Hv=> [_ _ Hv _ _].
    move/andP: Hv=> [Hv _].
    move/map_forallP: Hv.
    refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hv.
    apply MF.find_mapsto_iff in Hcmd.
    move: (Hv _ _ Hcmd)=> {Hv} Hv.
    move/andP: Hv=> [Hinstrs Htmn].
    move/andP: Hinstrs=> [Hinstrs Htmn'].
    exploit validate_postfix'; eauto.
    move/andP=> [Hacc _].
    exploit CorAcc.tacc_correct; eauto.
    move=> [Hcf Hcs _ _].
    eapply acc_gamma_use_blk; (try done); [by eauto|by auto|idtac|idtac].
    { by eapply finite_con_mem_sem; eauto. }
    move=> cloc Hcloc.
    exploit correct_postfix; [eapply validate_postfix'; eauto|by eauto|idtac].
    move=> Habs.
    eapply abs_relation_proper; [reflexivity|idtac|idtac]; eauto; last first.
    { rewrite/abs_relation/CorCon.state_abs/fst/snd in Habs.
      instantiate (1 :=
                     (TrsLocal.tstate
                        prog acc
                        (tstate prog (t_state prog propagated_state)))
                       [(fid, bid, false, stack) ]!
                       [t_loc cloc]!).
      move: ((TrsLocal.tstate
                prog acc
                (TrsSen.tstate prog (t_state prog propagated_state)))
               [(fid, bid, false, stack) ]!) Habs=> [amem|] Habs; last first.
      { move: ((cmem) [cloc ]!)=> [avalue|]; last done.
        by apply abs_value_top.
      }
      by apply Habs.
    }    
    { rewrite TrsLocal_tstate_preserve; auto; first last.
      { move=> [[[pfid pbid] pepilogue] pstack].
        rewrite TrsSen.tstate_lookup /fst.
        case (_: bool); last done.
        rewrite simpl_t_state.
        case: ((pfuncs prog) [pfid])=> [_|]; last done.
        apply notop_prop.
        move/map_forallP: Hnotop.
        refine (implies_right_turn (Proper _ _) _ _ _ _)=> Hnotop.
        case Hfs: (propagated_state[pfid])=> [Hfunc_state|]; rewrite/= Hfs /=; last done.
        apply MF.find_mapsto_iff in Hfs.
        by apply (Hnotop _ _ Hfs).
      }
      rewrite TrsSen_tstate_preserve; last done.
      rewrite simpl_t_state.
      apply MF.find_mapsto_iff in Hfunc.
      rewrite Hfunc.
      apply DomType.eq_le.
      reflexivity.
    }
  Qed.
End Correctness.
