Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomUtil.
Require Import DomZ.
Require Import DomMap.
Require Import DomPidx.
Require Import DomPidxSen.

Set Implicit Arguments.

Section Locs.
  Definition offsetT := Z.
  Definition sizeT := Z.
  Definition strideT := Z.
  Definition ossT := (offsetT * sizeT * strideT)%type.

  Definition to_ossT (o:offsetT) (s:sizeT) (st:strideT) : ossT :=
    (o, s, st).

  Global Add Parametric Morphism : to_ossT with
      signature eq ==> eq ==> eq ==> eq as to_ossT_mor.
  Proof. done. Qed.

  Definition uniqueT := positive.
  Definition fieldsT := list idT.
  Definition locT :=
    ((allocT * uniqueT * offsetT + idT + fidT * idT * nat) * fieldsT)%type.

  Global Instance locT_setoid : setoid locT := _.
  Global Instance locT_usual_setoid : @usual_setoid locT _ := _.
  Global Instance locT_setoidDec : @setoidDec locT _ := _.
  Global Instance locT_poset : @poset locT _ := _.
  Global Instance locT_posetDec : @posetDec locT _ _ _ := _.
  Global Instance locT_order : @DomType.order locT _ _ := _.

  Definition lheap (a:allocT) (u:uniqueT) (o:offsetT) : locT :=
    (inl (inl (a, u, o)), nil).
  Definition lglobal (id:idT) : locT := (inl (inr id), nil).
  Definition llocal (pidx:pidxT) (id:idT) : locT :=
    (inr ((pifid pidx), id, length pidx.2), nil).

  Global Add Parametric Morphism : lheap with
      signature eq ==> eq ==> eq ==> eq as lheap_mor.
  Proof. done. Qed.

  Global Add Parametric Morphism : lglobal with
      signature eq ==> eq as lglobal_mor.
  Proof. done. Qed.

  Global Add Parametric Morphism : llocal with
      signature eq ==> eq ==> eq as llocal_mor.
  Proof.
    move=> lp rp Hp; usual_tac Hp.
    move=> li ri Hi; usual_tac Hi.
    reflexivity.
  Qed.
End Locs.


Section Values.
  Definition valueT := (Z + locT + allocT * uniqueT * ossT)%type.

  Definition vint (z:Z) : valueT := inl (inl z).
  Definition vloc (l:locT) : valueT := inl (inr l).
  Definition vheap (a:allocT) (u:uniqueT) (oss:ossT) : valueT :=
    inr (a, u, oss).

  Global Add Parametric Morphism : vint with
      signature eq ==> eq as vint_mor.
  Proof. done. Qed.

  Global Add Parametric Morphism : vloc with
      signature eq ==> eq as vloc_mor.
  Proof. done. Qed.
  
  Global Add Parametric Morphism : vheap with
      signature eq ==> eq ==> eq ==> eq as vheap_mor.
  Proof. done. Qed.

  Global Instance valueT_setoid : setoid valueT := _.
  Global Instance valueT_usual_setoid : @usual_setoid valueT _ := _.
  Global Instance valueT_setoidDec : @setoidDec valueT _ := _.
End Values.
