Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import List.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomUtil.
Require Import DomBridge.
Require Import DomSet.
Require Import DomMap.
Require Import DomZ.
Require Import DomItv.

Require Import Syntax.

Require Import Containers.SetProperties.
Require Import Containers.MapInterface.
Require Import DomBasic.

Set Implicit Arguments.

Section Pidxs.
  Context (pidxT : Type) `{setoid pidxT}.

  Class pidx : Type := Pidx {
    piinitial : forall prog:progT, option pidxT;
    pifid : pidxT -> fidT;
    pibid : pidxT -> bidT;
    piepilogue : pidxT -> bool;

    pifid_proper : Proper (eq ==> eq) pifid;
    pibid_proper : Proper (eq ==> eq) pibid;
    piepilogue_proper : Proper (eq ==> eq) piepilogue
  }.

  Global Add Parametric Morphism `{pidx} : pifid with
    signature eq ==> eq as pifid_mor.
  Proof. exact: pifid_proper. Qed.
  
  Global Add Parametric Morphism `{pidx} : pibid with
    signature eq ==> eq as pibid_mor.
  Proof. exact: pibid_proper. Qed.
  
  Global Add Parametric Morphism `{pidx} : piepilogue with
    signature eq ==> eq as piepilogue_mor.
  Proof. exact: piepilogue_proper. Qed.
  
  Context (prog:progT) `{pidx}.
  Definition pifunc pidx := (pfuncs prog)[pifid pidx].

  Definition picmd pidx :=
    if pifunc pidx is Some func
      then (fcmds func)[pibid pidx]
      else None.

  Add Parametric Morphism : picmd with
      signature eq ==> eq as picmd_mor.
  Proof.
    rewrite/picmd/pifunc.
    move=> x y Hxy.
    rewrite -> Hxy.
    case: ((pfuncs prog) [pifid y]) => [func|]; last done.
    by rewrite -> Hxy; apply/eq_refl.
  Qed.

  Definition piinstrs pidx :=
    if picmd pidx is Some cmd
      then Some (cinstrs cmd)
      else None.

  Add Parametric Morphism : piinstrs with
      signature eq ==> eq as piinstrs_mor.
  Proof.
    rewrite/piinstrs/picmd/pifunc.
    move=> x y Hxy.
    rewrite -> Hxy.
    case: ((pfuncs prog) [pifid y]) => [func|]; last done.
    rewrite -> Hxy.
    case: ((fcmds func) [pibid y]) => [cmd|]; last done.
    exact/eq_refl.
  Qed.

  Definition pitmn pidx :=
    if picmd pidx is Some cmd
      then Some (ctmn cmd)
      else None.

  Add Parametric Morphism : pitmn with
      signature eq ==> eq as pitmn_mor.
  Proof.
    rewrite/pitmn/picmd/pifunc.
    move=> x y Hxy.
    rewrite -> Hxy.
    case: ((pfuncs prog) [pifid y]) => [func|]; last done.
    rewrite -> Hxy.
    case: ((fcmds func) [pibid y]) => [cmd|]; last done.
    exact/eq_refl.
  Qed.
End Pidxs.
