Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import List.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomUtil.
Require Import DomBridge.
Require Import DomSet.
Require Import DomMap.
Require Import DomZ.
Require Import DomItv.
Require Import DomPidx.

Require Import Syntax.

Require Import Containers.SetProperties.
Require Import Containers.MapInterface.

Set Implicit Arguments.

Definition pidxT := (fidT * bidT * bool * list (fidT * bidT))%type.

Global Instance pidxT_setoid : setoid pidxT := _.
Global Instance pidxT_usual_setoid : @usual_setoid pidxT _ := _.
Global Instance pidxT_setoidDec : @setoidDec pidxT _ := _.
Global Instance pidxT_poset : @poset pidxT _ := _.
Global Instance pidxT_posetDec : @posetDec pidxT _ _ _ := _. 

Global Program Instance pidxT_pidx : @pidx pidxT _ := {
  piinitial prog :=
  if (pfuncs prog)[pentry prog] is Some func
  then Some (pentry prog, fentry func, false, nil)
  else None;
  pifid := nosimpl (fun pidx => pidx.1.1.1);
  pibid := nosimpl (fun pidx => pidx.1.1.2);
  piepilogue := nosimpl (fun pidx => pidx.1.2)
}.
Next Obligation.
  by move=> lp rp [[[? ?] ?] ?].
Qed.
Next Obligation.
  by move=> lp rp [[[? ?] ?] ?].
Qed.
Next Obligation.
  by move=> lp rp [[[? ?] ?] ?].
Qed.


Section Consistents.
  Fixpoint pidx_consistent (prog:progT) (cfid:fidT) (stack:list (fidT*bidT)) : bool :=
    if stack is (fid,bid)::stack'
    then andb
           (if (pfuncs prog)[fid] is Some func
            then if (fcmds func)[bid] is Some cmd
                 then if ctmn cmd is tmn_call _ nfids _ _
                      then cfid !\in nfids
                      else false
                 else false
            else false)
           (pidx_consistent prog fid stack')
    else eqb cfid (pentry prog).

  Lemma simpl_pidx_consistent (prog:progT) (cfid:fidT) (stack:list (fidT*bidT)) :
    pidx_consistent prog cfid stack =
    if stack is (fid,bid)::stack'
    then andb
           (if (pfuncs prog)[fid] is Some func
            then if (fcmds func)[bid] is Some cmd
                 then if ctmn cmd is tmn_call _ nfids _ _
                      then cfid !\in nfids
                      else false
                 else false
            else false)
           (pidx_consistent prog fid stack')
    else eqb cfid (pentry prog).
  Proof. by case: stack. Qed.

  Global Add Parametric Morphism : pidx_consistent with
      signature Logic.eq ==> eq ==> eq ==> eq 
        as pidx_consistent_mor.
  Proof.
    move=> prog x' x -> y' y -> {x' y'}.
    exact: eq_refl.
  Qed.
End Consistents.
