Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomBridge.

Require Import Containers.Sets.

Set Implicit Arguments.

Section Sets.
  Context (A : Type) `{setoidA : setoid A, posetA : @poset A _, orderA : @DomType.order A _ _, setoidDecA : @setoidDec A _, posetDecA : @posetDec A _ _ _}.

  Notation "'Set!' '[' A ']'" := (@set A order_OrderedType _) (at level 0, no associativity).

  Global Program Instance set_setoid : setoid Set![A] := {
    eq_ x y := x [=] y
  }.
  
  Global Program Instance set_setlike : setlike (Elt:=A) (S:=Set![A]) := {
    sempty := empty;
    ssingleton := singleton;
    smem := mem;
    sadd := add;
    sremove := remove
  }.
  Next Obligation.
    case Heqb: (eqb e e').
    { left; split.
      exact/eqbP.
      apply mem_iff.
      apply singleton_iff.
      exact/eq_sym/eqbP.
    }
    { right; split.
      apply/eqbP.
      by rewrite Heqb.
      apply/negP=> Hc.
      apply mem_iff in Hc.
      apply singleton_iff in Hc.
      move/eqbP in Heqb.
      rewrite -> Hc in Heqb.
      contradict Heqb.
      exact: eq_refl.
    }
  Qed.
  Next Obligation.
  Proof.
    remember (eqb e' e) as Beqb; destruct Beqb; symmetry in HeqBeqb.
    {
      left; split.
      move/eqbP:HeqBeqb => HeqBeqb. 
      symmetry; tauto.
      apply mem_iff. apply add_iff. left. 
      move/eqbP:HeqBeqb; apply.
    }        
    {
      right; split.
      move/eqbP:HeqBeqb => HeqBeqb. 
      intro. apply HeqBeqb. symmetry; assumption.
      remember (e \in s) as res; destruct res.
      apply mem_iff. apply add_iff. right. 
      symmetry in Heqres. apply mem_iff in Heqres. apply Heqres.
      apply not_mem_iff. intro H. apply add_iff in H. 
      destruct H.
      {
        move/eqbP:HeqBeqb => HeqBeqb. elim HeqBeqb.
        unfold Equivalence.equiv,_eq,order_OrderedType in H.
        apply H.
      }
      {
        symmetry in Heqres.
        apply not_mem_iff in Heqres.
        elim Heqres; auto.
      }
    }
  Qed.
  Next Obligation.
  Proof.
    remember (eqb e' e) as Beqb; destruct Beqb; symmetry in HeqBeqb.
    {
      left; split.
      move/eqbP:HeqBeqb => HeqBeqb. 
      symmetry; tauto.
      unfold negb. remember (e \in {s ~ e'}) as b; destruct b; auto.
      symmetry in Heqb. apply mem_iff in Heqb. 
      apply remove_iff in Heqb. destruct Heqb as [Hin Hneq].
      move/eqbP:HeqBeqb => HeqBeqb. 
      unfold complement in Hneq. elim Hneq.
      unfold Equivalence.equiv,_eq,order_OrderedType; auto.
    }        
    {
      right; split.
      move/eqbP:HeqBeqb => HeqBeqb. 
      intro. apply HeqBeqb. symmetry; assumption.
      remember (e \in s) as res; destruct res.
      {
        apply mem_iff. apply remove_iff. split.
        { symmetry in Heqres. apply mem_iff in Heqres. assumption. }
        {
          unfold complement. intro H.
          move/eqbP:HeqBeqb => HeqBeqb. 
          unfold Equivalence.equiv,_eq,order_OrderedType in H; auto.
        }
      }
      {
        apply not_mem_iff. intro H.
        apply remove_iff in H. destruct H.
        symmetry in Heqres.
        apply not_mem_iff in Heqres. tauto.
      }
    }
  Qed.

  Lemma set_smem_spec (a:A) (m:Set![A]) :
    mem a m = smem a m.
  Proof. done. Qed.
  
  Definition set_forall := for_all (A:=A).

  Definition map (f: A ->> A) (m:Set![A]) : Set![A] :=
    fold
      (fun a m => add (proj1_sig f a) m)
      m
      empty.

  Lemma mapP (f : A ->> A) (m:Set![A])
        (a:A) (Ha:a !\in m) :
    (proj1_sig f a) !\in (map f m).
  Proof.
    move: f=> [f fp] /=.
    rewrite/map.
    rewrite/smem/= in Ha *.
    rewrite fold_1; rewrite <- fold_left_rev_right.
    rewrite SetFacts.elements_b in Ha.
    rewrite existsb_rev in Ha.
    elim: (rev _) Ha=> [|b l IH] Ha; first done.
    rewrite/= in Ha.
    move/orP: Ha=> [Ha|Ha].
    { rewrite is_compare_eqb in Ha; move/eqbP in Ha.
      rewrite add_b.
      rewrite is_compare_eqb fp; last by symmetry; eauto.
      by rewrite eqb_refl.
    }
    { rewrite add_b IH; last done.
      by case: (_ == _).
    }
  Qed.

  Lemma mapP' (f : A ->> A) (m:Set![A])
        (a:A) (Ha:a !\in map f m) :
    exists (b:A), b !\in m /\ eq (proj1_sig f b) a.
  Proof.
    Lemma mapP'' (f : A ->> A) (m:Set![A])
          (a:A) (Ha:a !\in map f m) :
      exists (b:A), (existsb (fun y => b == y) (rev (elements m))) /\ eq (proj1_sig f b) a.
    Proof.
      move: f Ha=> [f fp] /=.
      rewrite/map fold_1 /smem/=.
      rewrite <- fold_left_rev_right.
      elim: (rev _)=> [|hd tl IH] Ha; first done.
      simpl in Ha.
      rewrite SetFacts.add_b in Ha; move/orP in Ha; inv Ha.
      { exists hd; split.
        { by apply/orP; left; rewrite is_compare_eqb eqb_refl. }
        { rewrite is_compare_eqb in H.
            by move/eqbP in H.
        }
      }
      exploit IH; first done.
      move=> [x [Hx Hx']].
      exists x; split; last done.
      by apply/orP; right.
    Qed.

    exploit mapP''; first by eauto.
    move=> [x [Hx Hx']].
    exists x; split; last done.
    by rewrite /smem/= SetFacts.elements_b existsb_rev.
  Qed.

  Lemma set_forallP
        (test : A -> bool)
        (Hproper : Proper (eq ==> eq) test)
        (s : Set![A]) :
    reflect (forall elt (Hin : elt !\in s), test elt) (set_forall test s).
  Proof.
    rewrite/set_forall.
    apply/equivP; first apply idP.
    rewrite <- is_true_iff.
    rewrite <- for_all_iff; last exact/Hproper.
    rewrite/For_all.
    apply/forall_iff => elt.
    rewrite <- is_true_iff.
    by split=> Hin_t Hin; apply/Hin_t/mem_iff.
  Qed.

  Global Program Instance set_poset : @poset Set![A] _ := {
    le_ x y := x [<=] y
  }.
  Next Obligation.
    move=> x y Hxy elt Hx.
    by apply/Hxy.
  Qed.
  Next Obligation.
    move=> x y Hxy Hyx elt; constructor => Hin.
    by apply/Hxy.
    by apply/Hyx.
  Qed.

  Global Add Parametric Morphism : map with
      signature eq ==> le ==> le as map_le_mor.
  Proof.
    move=> x y Hxy; rewrite/eq/= in Hxy.
    move=> z w Hzw.
    move=> a Ha.
    apply SetFacts.mem_iff in Ha.
    apply SetFacts.mem_iff.
    exploit mapP'; first by eauto.
    move=> [b [Hb Hb']].
    rewrite <- Hb'.
    rewrite Hxy.
    apply mapP.
    apply SetFacts.mem_iff.
    rewrite <- Hzw.
    by apply SetFacts.mem_iff.
  Qed.

  Global Add Parametric Morphism : map with
      signature eq ==> eq ==> eq as map_mor.
  Proof.
    move=> x y Hxy z w Hzw.
    apply DomType.le_antisym; apply map_le_mor.
    done.
    by apply DomType.eq_le.
    by symmetry.
    exact/DomType.eq_le/eq_sym.
  Qed.

  Global Program Instance set_hasBot : @hasBot Set![A] _ _ := {
    bot_ := {}
  }.
  Next Obligation.
    move=> ?; by case/empty_iff.
  Qed.

  Global Program Instance set_hasJoin : @hasJoin Set![A] _ _ := {
    join_ x y := x ++ y
  }.
  Next Obligation.
    by move=> elt Hin; rewrite -> union_iff; auto.
  Qed.
  Next Obligation.
    by move=> elt Hin; rewrite -> union_iff; auto.
  Qed.
  Next Obligation.
    rewrite/le in Hx Hy.
    move=> z; rewrite -> union_iff.
    move=> [Hin|Hin].
    exact/Hx.
    exact/Hy.
  Qed.

  Global Program Instance set_setoidDec : @setoidDec Set![A] _ := {
    eqb_ x y := equal x y
  }.
  Next Obligation.
    case Heq: (equal x y); constructor.
    exact/equal_iff.
    rewrite/eq; simpl.
    move=> Hxy.
    apply equal_iff in Hxy.
    by rewrite Hxy in Heq.
  Qed.
  
  Global Program Instance set_posetDec : @posetDec Set![A] _ _ _ := {
    leb_ x y := subset x y
  }.
  Next Obligation.
    case Hle: (subset x y); constructor.
    exact/subset_iff.
    rewrite/le; simpl.
    move=> Hxy.
    apply subset_iff in Hxy.
    by rewrite Hxy in Hle.
  Qed.

  Global Add Parametric Morphism : mem with
    signature eq ==> eq ==> eq as mem_mor.
  Proof.
    move=> x y Hxy z w Hzw.
    exact/mem_m.
  Qed.
  Global Add Parametric Morphism : mem with
    signature eq ==> le ==> le as mem_le_mor.
  Proof.
    move=> x y Hxy z w Hzw.
    apply/le_trans; first by apply DomType.eq_le; rewrite -> Hxy.
    case Hz: (y \in z); case Hw: (y \in w); firstorder.
    rewrite <- mem_iff in Hz.
    rewrite <- not_mem_iff in Hw.
    by elim Hw; apply/Hzw.
  Qed.
End Sets.
Notation "'Set!' '[' A ']'" := (@set A order_OrderedType _) (at level 0, no associativity).


Program Definition set_to_powerset
        A `{setoid A, @poset A _, @DomType.order A _ _,
            @setoidDec A _, @posetDec A _ _ _}
        (s : set A) : \powerset A :=
  fun (x:A) => x !\in s.
Next Obligation.
  move=> x y Hxy.
  red; rewrite -/eq.
  rewrite -> Hxy.
  exact: eq_refl.
Qed.

Global Add Parametric Morphism A `{setoid A, @poset A _, @DomType.order A _ _, @setoidDec A _, @posetDec A _ _ _} : (@set_to_powerset A _ _ _ _ _) with
    signature le ==> le as set_to_powerset_le_mor.
Proof.
  move=> x y Hxy z /=.
  by rewrite <- Hxy.
Qed.

Global Add Parametric Morphism A `{setoid A, @poset A _, @DomType.order A _ _, @setoidDec A _, @posetDec A _ _ _} : (@set_to_powerset A _ _ _ _ _) with
    signature eq ==> eq as set_to_powerset_mor.
Proof.
  move=> x y Hxy.
  apply/le_antisym.
  by rewrite Hxy; apply/DomType.eq_le.
  by rewrite <- Hxy; apply/DomType.eq_le.
Qed.

Lemma set_to_powerset_spec
  A `{setoid A, @poset A _, @DomType.order A _ _,
    @setoidDec A _, @posetDec A _ _ _}
  (s : set A) (a : A) :
  (a !\in (set_to_powerset s)) = (a !\in s).
Proof. done. Qed.
