Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Classical_Prop.

Require Import Utf8.
Require Import Basics.
Require Import List.
Require Import Setoid.
Require Import Morphisms.

Require Import DomType.
Require Import DomBasic.

Set Implicit Arguments.

Section Pairs.
  Lemma pair_mor3 :
    forall la lb lc ra rb rc,
      eq la ra -> eq lb rb -> eq lc rc ->
      eq (la, lb, lc) (ra, rb, rc).
  Proof. done. Qed.
End Pairs.
            
Section HasBots.
  Lemma bot_trivial : forall elt, le elt bot -> eq elt bot.
  Proof. by move=> elt Hle; apply/le_antisym/bot_prop. Qed.
End HasBots.

Section HasTops.
  Lemma top_trivial : forall A `{hasTop A} elt, le top elt -> eq top elt.
  Proof. by move=> A H H0 H1 elt Hle; apply/le_antisym/top_prop. Qed.
End HasTops.

Section Lifts.
  Definition lift1 {A} (op : A -> A) (val : option A) : option A :=
    match val with
      | Some val' => Some (op val')
      | None => None
    end.

  Lemma lift1_le_mor A `{setoid A,poset A} (op : A -> A) (proper : Proper (le ==> le) op) : Proper (le ==> le) (lift1 op).
  Proof.
    move=> [x |] [y |] Hxy //=.
    exact/proper.
  Qed.  

  Lemma lift1_mor A `{setoid A} (op : A -> A) (proper : Proper (eq ==> eq) op) : Proper (eq ==> eq) (lift1 op).
  Proof.
    move=> [x |] [y |] Hxy //=.
    exact/proper.
  Qed.  

  Definition lift2 {A} (op : A -> A -> A) (lhs rhs : option A) : option A :=
    match lhs, rhs with
      | Some lhs', Some rhs' => Some (op lhs' rhs')
      | _, _ => None
    end.

  Lemma lift2_le_mor A `{setoid A,poset A} (op : A -> A -> A) (proper : Proper (le ==> le ==> le) op) : Proper (le ==> le ==> le) (lift2 op).
  Proof.
    move=> [x |] [y |] Hxy [z |] [w |] Hzw //=.
    exact/proper.
  Qed.  

  Lemma lift2_mor A `{setoid A} (op : A -> A -> A) (proper : Proper (eq ==> eq ==> eq) op) : Proper (eq ==> eq ==> eq) (lift2 op).
  Proof.
    move=> [x |] [y |] Hxy [z |] [w |] Hzw //=.
    exact/proper.
  Qed.  
End Lifts.


Section Lists.
  Lemma list_length_mor {A} `{setoid A} : Proper (eq ==> eq) (length (A:=A)).
  Proof.
    elim; first by case.
    move=> lh lt IH [] //= rh rt [_ ?].
      by congr S; apply/IH.
  Qed.
End Lists.


Section Coerces.
  Context (A : Type) `{setoid A, @poset A _, @hasBot A _ _}.
  
  Definition coerce (oe : option A) := if oe is Some e then e else bot.
  
  Global Add Parametric Morphism : coerce with
      signature le ==> le as coerce_mor.
  Proof.
    move=> [x|] [y|] Hle //=.
    exact/bot_prop.
    exact/DomType.eq_le/eq_refl.
  Qed.

  Global Add Parametric Morphism : coerce with
      signature eq ==> eq as coerce_eq_mor.
  Proof.
    move=> [x|] [y|] Heq //=.
    exact/eq_refl.
  Qed.
End Coerces.
