Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomUtil.
Require Import DomSet.
Require Import DomMap.

Require Import Syntax.

Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomPidx.
Require Import DomPidxInsen.

Require Import SemAbsBlock.

Definition stateT := Map![pidxT, memT].

Global Instance stateT_setoid : setoid stateT := _.
Global Instance stateT_setoidDec : @setoidDec stateT _ := _.
Global Instance stateT_poset : @poset stateT _ := _.
Global Instance stateT_posetDec : @posetDec stateT _ _ _ := _. 
Global Instance stateT_hasJoin : @hasJoin stateT _ _ := _.
Global Instance stateT_maplike : maplike (Key:=pidxT) (Elt:=memT) (M:=stateT) := _.

Section Accs.
  Variable (prog:progT) (acc:accT).

  Definition acc_correct_callfroms_blk (fid:fidT) (bid:bidT) (cmd:cmdT) : bool := 
    match ctmn cmd with
      | tmn_call _ nfids _ _ =>
        set_forall
          (fun nfid => smem (S:=Set![fidT*bidT]) (fid,bid) (callfroms acc)[nfid]!)
          nfids
      | _ => true
    end.

  Definition acc_correct_callfroms_func (fid: fidT) (fbody:funcT) : bool := 
    map_forall
      (fun bid cmd => acc_correct_callfroms_blk fid bid cmd)
      (fcmds fbody).

  Definition acc_correct_callfroms_prog : bool := 
    map_forall 
      (fun fid fbody => acc_correct_callfroms_func fid fbody)
      (pfuncs prog).

  Definition acc_correct_callfroms_star : bool := 
    map_forall
      (fun (fid:fidT) (callfroms:Set![fidT*bidT]) =>
         set_forall
           (fun (callinfo:fidT*bidT) =>
              andb (callinfo.1 !\in ((callfroms_star acc)[fid]!:Set![fidT]))
                   (leb (A:=Set![fidT])
                        (callfroms_star acc)[callinfo.1]!
                        (callfroms_star acc)[fid]!)
           )
           callfroms)
      (callfroms acc).

  Definition acc_correct_acclocs_var
    (fid:fidT) (x:varT) (mem:memT): bool :=
    let alocs := mlookup (Key:=fidT) (Elt:=Set! [locT]) fid (acclocs acc) in
      (var_loc fid x nil) !\in alocs.

  Global Add Parametric Morphism : acc_correct_acclocs_var with
      signature eq ==> eq ==> eq ==> eq as acc_correct_acclocs_var_mor.
  Proof.
    move=> x y Hxy; usual_tac Hxy.
    by move=> z w Hzw; usual_tac Hzw.
  Qed.

  Definition acc_correct_acclocs_site
    (fid:fidT) (site:allocT) (mem:memT): bool :=
    let alocs := mlookup (Key:=fidT) (Elt:=Set! [locT]) fid (acclocs acc) in
    (lheap site nil) !\in alocs.

  Global Add Parametric Morphism : acc_correct_acclocs_site with
      signature eq ==> eq ==> eq ==> eq as acc_correct_acclocs_site_mor.
  Proof.
    move=> x y Hxy; usual_tac Hxy.
    by move=> z w Hzw; usual_tac Hzw.
  Qed.

  Fixpoint acc_correct_acclocs_expr (fid:fidT) (e:exprT) (mem:memT): bool :=
    match e with
      | expr_const _ => true
      | expr_op2 o e1 e2 => 
        acc_correct_acclocs_expr fid e1 mem &&
        acc_correct_acclocs_expr fid e2 mem
      | expr_op1 o e1 => acc_correct_acclocs_expr fid e1 mem
      | expr_value_of lv =>
        let alocs := mlookup (Key:=fidT) (Elt:=Set! [locT]) fid (acclocs acc) in
        acc_correct_acclocs_lv fid lv mem &&
        if sem_expr_lv fid lv mem is Elt locs
        then SetInterface.subset locs alocs
        else false
      | expr_addr_of lv => acc_correct_acclocs_lv fid lv mem
      | expr_cast _ e1 => acc_correct_acclocs_expr fid e1 mem
      | expr_size_of e1 => acc_correct_acclocs_expr fid e1 mem
    end

  with acc_correct_acclocs_lv (fid:fidT) (l:lv) (mem:memT): bool :=
         match l with
           | lv_id x => true
           | lv_deref e1 => acc_correct_acclocs_expr fid e1 mem
           | lv_array e1 e2 =>
             acc_correct_acclocs_expr fid e1 mem &&
             acc_correct_acclocs_expr fid e2 mem
           | lv_struct l pos =>
             acc_correct_acclocs_lv fid l mem
         end.

  Lemma acc_correct_acclocs_expr_mor' :
    forall w x
           (a b: memT) (Hab: eq a b),
      eq (acc_correct_acclocs_expr x w a)
         (acc_correct_acclocs_expr x w b)
  with acc_correct_acclocs_lv_mor' :
    forall w x
           (a b: memT) (Hab: eq a b),
      eq (acc_correct_acclocs_lv x w a)
         (acc_correct_acclocs_lv x w b).
  Proof.
    { induction w; intros; simpl.
      { apply eq_refl. }
      { rewrite IHw1; last apply Hab.
        rewrite IHw2; last apply Hab.
        apply eq_refl.
      }
      { apply IHw; apply Hab. }
      { rewrite acc_correct_acclocs_lv_mor'; [|apply Hab].
        case: (acc_correct_acclocs_lv _ _ _); rewrite/andb;
        last apply eq_refl.
        pose (sem_expr_lv_mor (eq_refl x) (eq_refl l) a b Hab) as Heqsemlv.
        case: (sem_expr_lv _ _ _) (sem_expr_lv _ _ _) Heqsemlv =>
        [llocs|] [rlocs|] Heqsemlv.
        { rewrite/eq/eq_/withtop_setoid/withtop_eq in Heqsemlv.
          rewrite Heqsemlv.
          apply eq_refl.
        }
        { elim Heqsemlv. }
        { elim Heqsemlv. }
        { apply eq_refl. }
      }
      { apply acc_correct_acclocs_lv_mor'; try reflexivity.
        apply Hab.
      }
      { apply IHw; apply Hab. }
      { apply IHw; apply Hab. }
    }

    { induction w; intros; simpl.
      { done. }
      { apply acc_correct_acclocs_expr_mor'; try reflexivity.
        apply Hab.
      }
      { rewrite acc_correct_acclocs_expr_mor'; [|apply Hab].
        rewrite/andb.
        case: (acc_correct_acclocs_expr _ _ _); last apply eq_refl.
        apply acc_correct_acclocs_expr_mor'; try reflexivity.
        apply Hab.
      }
      { apply acc_correct_acclocs_lv_mor'; try reflexivity.
        apply Hab.
      }
    }
  Qed.

  Global Add Parametric Morphism : acc_correct_acclocs_expr with
      signature eq ==> eq ==> eq ==> eq as acc_correct_acclocs_expr_mor.
  Proof.
    move=> x y Hxy w z Hwz a b Hab.
    usual_tac Hxy; usual_tac Hwz.
    by apply acc_correct_acclocs_expr_mor'.
  Qed.

  Global Add Parametric Morphism : acc_correct_acclocs_lv with
      signature eq ==> eq ==> eq ==> eq as acc_correct_acclocs_lv_mor.
  Proof.
    move=> x y Hxy w z Hwz a b Hab.
    usual_tac Hxy; usual_tac Hwz.
    by apply acc_correct_acclocs_lv_mor'.
  Qed.

  Definition acc_correct_acclocs_lv_opt (fid:fidT) (opt_l:option lv) (mem:memT)
    : bool :=
    match opt_l with
      | None => true
      | Some l => acc_correct_acclocs_lv fid l mem
    end.

  Global Add Parametric Morphism : acc_correct_acclocs_lv_opt with
      signature eq ==> eq ==> eq ==> eq as acc_correct_acclocs_lv_opt_mor.
  Proof.
    move=> x y Hxy; usual_tac Hxy.
    move=> [z|] [w|] Hzw //=.
    rewrite/eq/= in Hzw.
    usual_tac Hzw.
    by apply acc_correct_acclocs_lv_mor'.
  Qed.

  Fixpoint acc_correct_acclocs_exprs
           (fid:fidT) (es:list exprT) (mem:memT): bool :=
    match es with
      | nil => true
      | cons e es1 => 
        acc_correct_acclocs_expr fid e mem &&
        acc_correct_acclocs_exprs fid es1 mem
    end.

  Global Add Parametric Morphism : acc_correct_acclocs_exprs with
      signature eq ==> eq ==> eq ==> eq as acc_correct_acclocs_exprs_mor.
  Proof.
    move=> x y Hxy; usual_tac Hxy.
    move=> z w Hzw; usual_tac Hzw.
    elim: w=> [|istr istrs IH] a b Hab; first done.
    rewrite/=.
    rewrite acc_correct_acclocs_expr_mor; eauto; try reflexivity.
    by rewrite IH.
  Qed.

  (* Arguments of library calls are encoded specially
    so we should check their validness directly. *)
  Fixpoint acc_correct_acclocs_libcall_args
           (fid:fidT) (args:list exprT) (mem:memT): bool :=
    match args with
      | nil => true
      | cons arg tl =>
        (if locs_of (sem_expr fid arg mem) is Elt locs
         then
           let alocs := mlookup (Key:=fidT) (Elt:=Set! [locT])
                                fid (acclocs acc) in
           SetInterface.subset locs alocs
         else
           false
        ) && (acc_correct_acclocs_libcall_args fid tl mem)
    end.

  Global Add Parametric Morphism : acc_correct_acclocs_libcall_args with
      signature eq ==> eq ==> eq ==> eq as acc_correct_acclocs_libcall_args_mor.
  Proof.
    move=> x y Hxy; usual_tac Hxy.
    move=> z w Hzw; usual_tac Hzw.
    elim: w=> [|istr istrs IH] a b Hab; first done.
    rewrite/=.
    rewrite IH; last by eauto.
    exploit locs_of_mor; first by eapply sem_expr_mor; eauto; reflexivity.
    instantiate (1 := istr).
    instantiate (1 := y).
    move: (locs_of _) (locs_of _)=> [llocs|] [rlocs|] Hlocs //=.
    rewrite/eq/= in Hlocs.
    by rewrite Hlocs.
  Qed.

  Definition acc_correct_acclocs_instr
             (fid:fidT) (instr:instrT) (mem:memT): bool :=
    match instr with
      | instr_assign lv e => 
        acc_correct_acclocs_lv fid lv mem &&
        acc_correct_acclocs_expr fid e mem &&
        let alocs := mlookup (Key:=fidT) (Elt:=Set! [locT]) fid (acclocs acc) in
        if sem_expr_lv fid lv mem is Elt locs
        then SetInterface.subset locs alocs
        else false
      | instr_alloc lv (alloc_arr sz str) site =>
        acc_correct_acclocs_lv fid lv mem &&
        acc_correct_acclocs_site fid site mem &&
        acc_correct_acclocs_expr fid sz mem &&
        acc_correct_acclocs_expr fid str mem &&
        let alocs := mlookup (Key:=fidT) (Elt:=Set! [locT]) fid (acclocs acc) in
        if sem_expr_lv fid lv mem is Elt locs
        then SetInterface.subset locs alocs
        else false
      | instr_assume e => acc_correct_acclocs_expr fid e mem
      | instr_libcall ol lc args site =>

        (* Hypothesis for return lvalue: except for scanf *)
        match lc with
          | scanf => true
          | _ =>
            (if lc is strcpy
             then
               let dstLocs :=
                   (sem_expr fid (hd (expr_const lit_bot) args) mem) in
               let dst := locs_of dstLocs in
               let srcLocs := (sem_expr fid (hd (expr_const lit_bot)
                                                (tl args)) mem) in
               let src := locs_of srcLocs in
               let srcValue := memory_lookup src mem in
               let newMem := memory_update acc dst srcValue mem in
               (if ol is Some lv
                then acc_correct_acclocs_lv fid lv newMem
                else true)
             else
               (if ol is Some lv
                then acc_correct_acclocs_lv fid lv mem
                else true)
            )
        end
        &&

        (* Hypothesis for argument expressions:
          strcpy, scanf, builtin_alloca, and builtin_choose_expr *)
        match lc with
          | strcpy => acc_correct_acclocs_exprs fid args mem
          | scanf => acc_correct_acclocs_exprs fid args mem
          | builtin_alloca => acc_correct_acclocs_exprs fid args mem
          | builtin_choose_expr => acc_correct_acclocs_exprs fid args mem
          | _ => true
        end
        &&

        (* Hypothesis for return lvalue with memory:
          except for scanf *)
        match lc with
          | scanf => true
          | _ =>
            (if lc is strcpy
             then
               let dstLocs := (sem_expr fid (hd (expr_const lit_bot)
                                                args) mem) in
               let dst := locs_of dstLocs in
               let srcLocs := (sem_expr fid (hd (expr_const lit_bot)
                                                (tl args)) mem) in
               let src := locs_of srcLocs in
               let srcValue := memory_lookup src mem in
               let newMem := memory_update acc dst srcValue mem in
               (if ol is Some lv
                then
                  let alocs := mlookup (Key:=fidT) (Elt:=Set! [locT])
                                       fid (acclocs acc) in
                  if sem_expr_lv fid lv newMem is Elt locs
                  then SetInterface.subset locs alocs
                  else false
                else true)
             else
               (if ol is Some lv
                then
                  let alocs := mlookup (Key:=fidT) (Elt:=Set! [locT])
                                       fid (acclocs acc) in
                  if sem_expr_lv fid lv mem is Elt locs
                  then SetInterface.subset locs alocs
                  else false
                else true)
            )
        end
        &&

        (* Hypothesis for argument lvalues:
          strcpy and scanf *)
        match lc with
          | strcpy => acc_correct_acclocs_libcall_args fid args mem
          | scanf => acc_correct_acclocs_libcall_args fid args mem
          | _ => true
        end
      | instr_forget aloc =>
        let alocs := mlookup (Key:=fidT) (Elt:=Set! [locT]) fid (acclocs acc) in
        aloc !\in alocs
      | instr_skip => true
    end.

  Ltac solver :=
    repeat
      (match goal with
         | [|- context[acc_correct_acclocs_site _ _ _]] =>
           (rewrite acc_correct_acclocs_site_mor; eauto; try reflexivity;
            case: (acc_correct_acclocs_site _ _ _); rewrite/andb; last done)
         | [|- context[acc_correct_acclocs_lv _ _ _]] =>
           (rewrite acc_correct_acclocs_lv_mor; eauto; try reflexivity;
            case: (acc_correct_acclocs_lv _ _ _); rewrite/andb; last done)
         | [|- context[acc_correct_acclocs_expr _ _ _]] =>
           (rewrite acc_correct_acclocs_expr_mor; eauto; try reflexivity;
            case: (acc_correct_acclocs_expr _ _ _); rewrite/andb; last done)
         | [|- context[acc_correct_acclocs_exprs _ _ _]] =>
           (rewrite acc_correct_acclocs_exprs_mor; eauto; try reflexivity;
            case: (acc_correct_acclocs_exprs _ _ _); rewrite/andb; last done)
         | [|- context[acc_correct_acclocs_libcall_args _ _ _]] =>
           (rewrite acc_correct_acclocs_libcall_args_mor; eauto; try reflexivity;
            case: (acc_correct_acclocs_libcall_args _ _ _); rewrite/andb; last done)
       end; auto).

  Lemma andb_mor (a b c d: bool)
        (Hac: eq a c) (Hbd: eq b d) :
    eq (a && b) (c && d).
  Proof. by move: a b c d Hac Hbd; do 4! case. Qed.

  Lemma sub_alocs_mor (llocs rlocs:locsT) (Hlocs: eq llocs rlocs) (alocs:Set![locT]) :
    eq (match llocs with
          | Elt locs => SetInterface.subset locs alocs
          | Top => false
        end)
       (match rlocs with
          | Elt locs => SetInterface.subset locs alocs
          | Top => false
        end).
  Proof.
    move: llocs rlocs Hlocs=> [llocs|] [rlocs|] Hlocs //=.
    rewrite/eq/= in Hlocs.
    by rewrite Hlocs.
  Qed.
  
  Lemma lv_sub_alocs_mor (ol: option lv) (alocs:Set![locT]) (fid:fidT)
        (a b: memT) (Hab: eq a b) :
    eq
      (match ol with
         | Some lv =>
           match sem_expr_lv fid lv a with
             | Elt locs => SetInterface.subset locs alocs
             | Top => false
           end
         | None => true
       end)
      (match ol with
         | Some lv =>
           match sem_expr_lv fid lv b with
             | Elt locs => SetInterface.subset locs alocs
             | Top => false
           end
         | None => true
       end).
  Proof.
    move: ol=> [l|]; last done.
    exploit sem_expr_lv_mor; eauto; try reflexivity.
    instantiate (1 := l).
    instantiate (1 := fid).
    move: (sem_expr_lv _ _ _) (sem_expr_lv _ _ _)=> llocs rlocs Hlocs.
    by apply sub_alocs_mor.
  Qed.

  Ltac mortac :=
    repeat
      (match goal with
         | [|- eq (acc_correct_acclocs_exprs _ _ _) (acc_correct_acclocs_exprs _ _ _)]=> apply acc_correct_acclocs_exprs_mor
         | [|- eq (acc_correct_acclocs_lv _ _ _) (acc_correct_acclocs_lv _ _ _)]=> apply acc_correct_acclocs_lv_mor
         | [|- eq (acc_correct_acclocs_libcall_args _ _ _) (acc_correct_acclocs_libcall_args _ _ _)]=> apply acc_correct_acclocs_libcall_args_mor
         | [|- eq (memory_update _ _ _ _) (memory_update _ _ _ _)]=> apply memory_update_mor
         | [|- eq (memory_lookup _ _) (memory_lookup _ _)]=> apply memory_lookup_mor
         | [|- eq (locs_of _) (locs_of _)] => apply locs_of_mor
         | [|- eq (sem_expr _ _ _) (sem_expr _ _ _)] => apply sem_expr_mor
         | [|- eq (sem_expr_lv _ _ _) (sem_expr_lv _ _ _)] => apply sem_expr_lv_mor
       end; eauto; try reflexivity).
  
  Global Add Parametric Morphism : acc_correct_acclocs_instr with
      signature eq ==> eq ==> eq ==> eq as acc_correct_acclocs_instr_mor.
  Proof.
    move=> x y Hxy; usual_tac Hxy.
    move=> z w Hzw; usual_tac Hzw.
    move=> a b Hab.
    destruct w; intros.
    { simpl; solver.
      apply sub_alocs_mor.
      by apply sem_expr_lv_mor.
    }
    { simpl; solver.
      destruct a0; intros; solver.
      apply sub_alocs_mor.
      by apply sem_expr_lv_mor.
    }
    { simpl; solver.
      by apply acc_correct_acclocs_expr_mor.
    }
    { case: lc; intros; (do ? apply andb_mor);
        (try done);
        (try by move: ol=> [l|]; first rewrite Hab);
        (try by apply lv_sub_alocs_mor);
        (try by simpl; mortac).
      move: ol=> [l|]; last done.
      { by simpl; mortac. }
      move: ol=> [l|]; last done.
      apply sub_alocs_mor.
      by mortac.
    }
    { done. }
    { done. }
  Qed.
  
  Fixpoint acc_correct_acclocs_instrs
           (fid:fidT) (instrs:list instrT) (mem:memT): bool :=
    match instrs with
      | nil => true
      | cons instr next_instrs =>
        acc_correct_acclocs_instr fid instr mem &&
        (let next_mem := sem_instr acc fid instr mem in
         acc_correct_acclocs_instrs fid next_instrs next_mem)
    end.

  Global Add Parametric Morphism : acc_correct_acclocs_instrs with
      signature eq ==> eq ==> eq ==> eq as acc_correct_acclocs_instrs_mor.
  Proof.
    move=> x y Hxy; usual_tac Hxy.
    move=> z w Hzw; usual_tac Hzw.
    elim: w=> [|istr istrs IH] a b Hab; first done.
    rewrite/=.
    rewrite acc_correct_acclocs_instr_mor; eauto; try reflexivity.
    rewrite IH; first reflexivity.
    by apply sem_instr_mor.
  Qed.

  Definition acc_correct_acclocs_tmn (fid:fidT) (tmn:tmnT) (mem:memT): bool :=
    match tmn with
      | tmn_jump _ => true
      | tmn_call _ _ es ret => 
        acc_correct_acclocs_exprs fid es mem &&
        (if ret is Some ret'
         then acc_correct_acclocs_var fid ret' mem
         else true)
      | tmn_exit e => acc_correct_acclocs_expr fid e mem
    end.

  Global Add Parametric Morphism : acc_correct_acclocs_tmn with
      signature eq ==> eq ==> eq ==> eq as acc_correct_acclocs_tmn_mor.
  Proof.
    move=> x y Hxy; usual_tac Hxy.
    move=> z w Hzw.
    move: z w Hzw=> [nbids|nbid nfids args retto|ret] [nbids'|nbid' nfids' args' retto'|ret'] Hzw a b Hab //=.
    move/andP: Hzw=> [Hzw Hzw'].
    move/andP: Hzw=> [Hzw Hzw''].
    move/andP: Hzw=> [Hzw Hzw'''].
    move/eqbP in Hzw; usual_tac Hzw.
    move/eqbP in Hzw'; usual_tac Hzw'.
    move/eqbP in Hzw''; usual_tac Hzw''.
    apply andb_mor; first by rewrite Hab.
    move: retto'=> [retto'|]; last done.
    by rewrite Hab.
    move/eqbP in Hzw.
    by rewrite Hab; rewrite Hzw.
  Qed.    

  Definition acc_correct_acclocs_blk (fid:fidT) (bid:bidT) (cmd:cmdT) (state:stateT): bool :=
    let instrs := cinstrs cmd in
    let tmn := ctmn cmd in
    let instrs_mem := state[(fid,bid,false)]! in
    let tmn_mem := state[(fid,bid,true)]! in
    acc_correct_acclocs_instrs fid instrs instrs_mem &&
    acc_correct_acclocs_tmn fid tmn tmn_mem.

  Global Add Parametric Morphism : acc_correct_acclocs_blk with
      signature eq ==> eq ==> eq ==> eq ==> eq as acc_correct_acclocs_blk_mor.
  Proof.
    move=> x y Hxy; usual_tac Hxy.
    move=> z w Hzw; usual_tac Hzw.
    move=> [a b] [c d] [Hac Hbd] e f Hef.
    rewrite/fst/snd in Hac Hbd.
    rewrite/acc_correct_acclocs_blk/cinstrs/=.
    rewrite acc_correct_acclocs_instrs_mor; eauto; try reflexivity.
    by rewrite acc_correct_acclocs_tmn_mor; eauto; reflexivity.
  Qed.

  Definition acc_correct_acclocs_func (fid: fidT) (fbody:funcT) (state:stateT): bool :=
    let alocs := mlookup (Key:=fidT) (Elt:=Set! [locT]) fid (acclocs acc) in
    andb
    (map_forall
      (fun bid cmd => acc_correct_acclocs_blk fid bid cmd state)
      (fcmds fbody))
    (list_forall
      (fun i => (llocal fid i nil) !\in alocs)
      (fparams fbody)).

  Definition acc_correct_acclocs_prog (state:stateT): bool :=
    map_forall
      (fun fid fbody => acc_correct_acclocs_func fid fbody state)
      (pfuncs prog).
  
  Definition acc_correct_acclocs_callfroms : bool :=
    map_forall
      (fun fid callfrom =>
         set_forall
           (fun callinfo =>
              leb (A:=Set![locT]) (acclocs acc)[fid]! (acclocs acc)[callinfo.1]!)
           callfrom)
      (callfroms acc).

  Definition acc_correct (state:stateT) : bool :=
    [&& acc_correct_callfroms_prog,
        acc_correct_callfroms_star,
        acc_correct_acclocs_prog state &
        acc_correct_acclocs_callfroms].
End Accs.
