Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomBridge.
Require Import DomSet.
Require Import DomMap.

Require Import Syntax.

Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomPidxSen.

Require Import SemAbsBlock.
Require Import SemAccSen.

Require Import SemSenLocal.
Require Import SemSen.

Set Implicit Arguments.

Obligation Tactic := Simpless.tactic.

Section States.
  Variable (prog:progT) (acc:accT).

  Fixpoint tmem (state:stateT) (cfid:fidT) (cbid:bidT) (cepilogue:bool) (cstack:list (fidT*bidT)) : memT :=
    mrestrict_join (E:=valueT)
    (set_to_powerset (acclocs acc)[cfid]!)
    state[(cfid,cbid,cepilogue,cstack)]!
    (if cstack is (fid,bid)::stack
     then tmem state fid bid true stack
     else bot).

  Definition tmem_mem (state:stateT) 
    (cfid:fidT) (cstack:list (fidT*bidT)) (m:memT) : memT :=
    mrestrict_join (E:=valueT)
    (set_to_powerset (acclocs acc)[cfid]!)
    m
    (if cstack is (fid,bid)::stack
     then tmem state fid bid true stack
     else bot).

  Lemma simpl_tmem (state:stateT) (cfid:fidT) (cbid:bidT) (cepilogue:bool) (cstack:list (fidT*bidT)) :
    (tmem state cfid cbid cepilogue cstack) =
    mrestrict_join (E:=valueT)
    (set_to_powerset (acclocs acc)[cfid]!)
    state[(cfid,cbid,cepilogue,cstack)]!
    (if cstack is (fid,bid)::stack
     then tmem state fid bid true stack
     else bot).
  Proof. by case: cstack. Qed.
  
  Add Parametric Morphism : tmem with
      signature eq ==> eq ==> eq ==> eq ==> eq ==> eq as tmem_mor.
  Proof.
    move=> x y Hxy
           fid' fid -> {fid'}
           bid' bid -> {bid'}
           epilogue' epilogue -> {epilogue'}
           stack' stack Hstack.
    usual_tac Hstack.
    elim: stack fid bid epilogue=> [|[pfid pbid] pstack IH] fid bid epilogue.
    by rewrite 2!simpl_tmem; rewrite -> Hxy; apply: eq_refl.
    rewrite [X in eq X _]simpl_tmem.
    rewrite [X in eq _ X]simpl_tmem.
    rewrite -> IH.
    rewrite -> Hxy.
    exact: eq_refl.
  Qed.

  Lemma update_tmem_commutative
        stack x (v:valueT) state (fid:fidT) bid nfid
        (Hxin: (var_loc nfid x nil) !\in (acclocs acc) [fid ]!) :
    memory_update acc (Elt (ssingleton (var_loc nfid x nil))) v
                  (tmem state fid bid true stack)
    ≡ tmem_mem state fid stack
    (memory_update acc (Elt (ssingleton (var_loc nfid x nil)))
                   v (state[(fid,bid,true,stack)]!)).
  Proof.
    rewrite/tmem_mem simpl_tmem.
    symmetry.
    by apply memory_update_mrestrict_join_commutative.
  Qed.

  Lemma update_tmem_mem_commutative stack x (v:valueT) state (fid:fidT) nfid (mem:memT)
        (Hxin: var_loc nfid x nil
         !\in set_to_powerset (acclocs acc) [fid ]!) :
    memory_update acc (Elt (ssingleton (var_loc nfid x nil))) v
                  (tmem_mem state fid stack mem)
    ≡ tmem_mem state fid stack
    (memory_update acc (Elt (ssingleton (var_loc nfid x nil))) v (mem)).
  Proof.
    rewrite/tmem_mem.
    symmetry.
    by apply memory_update_mrestrict_join_commutative.
  Qed.

  Lemma bind_tmem_mem_commutative (argvals:list valueT) params 
    (stack:list (fidT * bidT)) state (fid:fidT) nfid mem
    (Hpin: forall i, List.In i params ->
      var_loc nfid (local i) nil
      !\in set_to_powerset (acclocs acc) [fid ]!) :
    sem_bind_params acc nfid argvals params 
    (tmem_mem state fid stack mem)
    ≡ tmem_mem state fid stack 
    (sem_bind_params acc nfid argvals params mem).
  Proof.
    rewrite/tmem_mem.
    symmetry.
    by apply bind_mrestrict_join_commutative.
  Qed.    

  Lemma bind_tmem_commutative (argvals:list valueT) params state (fid:fidT) nfid bid stack
    (Hpin: forall i, List.In i params ->
      var_loc nfid (local i) nil
      !\in set_to_powerset (acclocs acc) [fid ]!) :
      sem_bind_params acc nfid argvals params 
      (tmem state fid bid true stack)
      ≡ tmem_mem state fid stack 
      (sem_bind_params acc nfid argvals params (state[(fid,bid,true,stack)]!)).
  Proof.
    rewrite/tmem_mem simpl_tmem.
    symmetry.
    by apply bind_mrestrict_join_commutative.
  Qed.

  Lemma tmem_mem_monotone : 
    forall stack state fid m m' (Hmorder:m ⊑ m'),
      tmem_mem state fid stack m ⊑ 
      tmem_mem state fid stack m'.
  Proof.
    move=> stack state fid m m' Hm.
    apply mrestrict_join_mor; auto.
    reflexivity.
    exact/DomType.eq_le/eq_refl.
  Qed.

  Lemma tmem_mrestrict_join_commutative : forall cstack state cfid l m m'
    (Hsubset:subset l (set_to_powerset (acclocs acc) [cfid ]!)),
    tmem_mem state cfid cstack (mrestrict_join (E:=valueT) l m m') ≡
    mrestrict_join (E:=valueT) l m (tmem_mem state cfid cstack m').
  Proof.
    move=> stack state fid locs m m' Hlocs.
    rewrite/tmem_mem.
    rewrite/mrestrict_join.
    do ? rewrite -> (mrestrict_distributive (E:=valueT) (M:=memT)).
    rewrite ? join_associative.
    apply join_mor; first by apply mrestrict_subset.
    apply join_mor; first by rewrite (mrestrict_commutative (E:=valueT) (M:=memT)); reflexivity.
    symmetry; apply mrestrict_subset.
    apply subset_neg; auto.
  Qed.

  Program Definition tstate (state:stateT) : SemAccSen.stateT :=
    fun pidx =>
      let: (fid, bid, epilogue, stack) := pidx in      
      if pidx_consistent prog fid stack
      then 
           tmem state fid bid epilogue stack
      else bot.
  Next Obligation.
    move=> x y Hxy.
    move: x y Hxy=>
    [[[fid' bid'] epilogue'] stack']
    [[[fid bid] epilogue] stack]
    [[[Hfid Hbid] Hepilogue] Hstack].
    rewrite/= in Hfid Hbid Hepilogue Hstack.
    apply: congr_if.
    by rewrite -> Hfid; rewrite -> Hstack.
    by rewrite -> Hfid; rewrite -> Hbid; rewrite -> Hepilogue; rewrite -> Hstack; apply: eq_refl.
    exact: eq_refl.
  Qed.

  Add Parametric Morphism : tstate with
      signature eq ==> eq as tstate_mor.
  Proof.
    move=> x y Hxy [[[fid bid] epilogue] stack].
    rewrite/proj1_sig/tstate.
    apply: congr_if; first done.
    by rewrite -> Hxy; apply: eq_refl.
    exact: eq_refl.
  Qed.

  Lemma tstate_lookup (state:stateT) (pidx:pidxT) :
    (tstate state)[pidx]! = 
    let: (fid, bid, epilogue, stack) := pidx in      
      if pidx_consistent prog fid stack
        then 
          tmem state fid bid epilogue stack
        else bot.
  Proof. by move: pidx=> [[[fid bid] epilogue] stack]. Qed.
End States.
