(*******************************)
(* types for using map and set *)
(*******************************)

(* int properties *)
let int_setoid = DomType.Setoid
let int_poset = DomType.Poset
let int_setoidDec =
{
  DomType.eqb_ = (=);
  DomType.eqbP_ = (fun x y -> if x = y then Ssrbool.ReflectT else Ssrbool.ReflectF)
}
let int_posetDec =
{
  DomType.leb_ = (<=);
  DomType.lebP_ = (fun x y -> if x <= y then Ssrbool.ReflectT else Ssrbool.ReflectF)
}

(* int utils *)

let int_ordered_type : int OrderedType.coq_OrderedType =
  DomBridge.order_OrderedType 
    int_setoid int_poset
    int_setoidDec int_posetDec

let int_fset : int SetInterface.coq_FSet =
  SetAVLInstance.coq_SetAVL_FSet int_ordered_type
let intSetEmpty : int SetInterface.set =
  SetInterface.empty int_ordered_type int_fset
let intSetAdd : int -> int SetInterface.set -> int SetInterface.set =
  SetInterface.add int_ordered_type int_fset
let intSetFold (f: int -> 'a -> 'a) (s: int SetInterface.set) (a: 'a) : 'a =
  SetInterface.fold int_ordered_type int_fset f s a
let intSetUnion (s: int SetInterface.set) (t: int SetInterface.set) : int SetInterface.set =
  SetInterface.union int_ordered_type int_fset s t

let int_fmap : int MapInterface.coq_FMap =
  MapAVLInstance.coq_MapAVL_FMap int_ordered_type
let intMapEmpty : (int, 'elt) MapInterface.dict = 
  MapInterface.empty int_ordered_type int_fmap
let intMapAdd (p:int) (e:'elt) (m:(int, 'elt) MapInterface.dict)
: (int, 'elt) MapInterface.dict = 
  MapInterface.add int_ordered_type int_fmap p e m
let intMapFind (p:int) (m:(int, 'elt) MapInterface.dict) : 'elt option = 
  MapInterface.find int_ordered_type int_fmap p m 
let intMapFold (f: int -> 'v -> 'a -> 'a) (m: (int, 'v) MapInterface.dict) (a: 'a) : 'a = 
  MapInterface.fold int_ordered_type int_fmap f m a

(* int * int *)

type intint = int * int

let intint_setoid : intint DomType.setoid =
  DomBasic.pair_setoid int_setoid int_setoid
let intint_poset : intint DomType.poset =
  DomBasic.pair_poset 
    int_setoid int_setoid
    int_poset int_poset
let intint_setoidDec : intint DomType.setoidDec =
  DomBasic.pair_setoidDec
    int_setoid int_setoid
    int_setoidDec int_setoidDec
let intint_posetDec : intint DomType.posetDec =
  DomBasic.pair_posetDec
    int_setoid int_setoid
    int_poset int_poset
    int_setoidDec int_setoidDec
    int_posetDec int_posetDec

let intint_ordered_type : intint OrderedType.coq_OrderedType = 
  DomBridge.order_OrderedType
    intint_setoid intint_poset
    intint_setoidDec intint_posetDec

let intint_fset : intint SetInterface.coq_FSet = 
  SetAVLInstance.coq_SetAVL_FSet intint_ordered_type
let intintSetEmpty : intint SetInterface.set =
  SetInterface.empty intint_ordered_type intint_fset
let intintSetAdd : intint -> intint SetInterface.set -> intint SetInterface.set =
  SetInterface.add intint_ordered_type intint_fset
let intintSetFold (f: intint -> 'a -> 'a) (s: intint SetInterface.set) (a: 'a) : 'a =
  SetInterface.fold intint_ordered_type intint_fset f s a

(* locT *)

let locT_ordered_type : DomAbsValue.locT OrderedType.coq_OrderedType = 
  DomBridge.order_OrderedType 
    DomAbsValue.locT_setoid DomAbsValue.locT_poset
    DomAbsValue.locT_setoidDec DomAbsValue.locT_posetDec

let locT_fset : DomAbsValue.locT SetInterface.coq_FSet =
  SetAVLInstance.coq_SetAVL_FSet locT_ordered_type
let locTSetEmpty : DomAbsValue.locT SetInterface.set = 
  SetInterface.empty locT_ordered_type locT_fset
let locTSetAdd : DomAbsValue.locT -> DomAbsValue.locT SetInterface.set -> DomAbsValue.locT SetInterface.set =
  SetInterface.add locT_ordered_type locT_fset
let locTSetRemove : DomAbsValue.locT -> DomAbsValue.locT SetInterface.set -> DomAbsValue.locT SetInterface.set =
  SetInterface.remove locT_ordered_type locT_fset
let locTSetFold (f: DomAbsValue.locT -> 'a -> 'a) (s: DomAbsValue.locT SetInterface.set) (a: 'a) : 'a =
  SetInterface.fold locT_ordered_type locT_fset f s a

let locT_fmap : DomAbsValue.locT MapInterface.coq_FMap =
  MapAVLInstance.coq_MapAVL_FMap locT_ordered_type
let locTMapEmpty : (DomAbsValue.locT,'elt) MapInterface.dict = 
  MapInterface.empty locT_ordered_type locT_fmap
let locTMapAdd (l:DomAbsValue.locT) (e:'elt) (m:(DomAbsValue.locT,'elt) MapInterface.dict) : (DomAbsValue.locT,'elt) MapInterface.dict = 
  MapInterface.add locT_ordered_type locT_fmap l e m
let locTMapFind (l:DomAbsValue.locT) (m:(DomAbsValue.locT,'elt) MapInterface.dict) : 'elt option = 
  MapInterface.find locT_ordered_type locT_fmap l m
let locTMapFold (f: DomAbsValue.locT -> 'v -> 'a -> 'a) (m: (DomAbsValue.locT, 'v) MapInterface.dict) (a: 'a) : 'a =
  MapInterface.fold locT_ordered_type locT_fmap f m a
let locTMapCardinal (m:(DomAbsValue.locT,'elt) MapInterface.dict) : int = 
  MapInterface.cardinal locT_ordered_type locT_fmap m
let locTMapRemove (l:DomAbsValue.locT) (m:(DomAbsValue.locT,'elt) MapInterface.dict) : (DomAbsValue.locT,'elt) MapInterface.dict =
  MapInterface.remove locT_ordered_type locT_fmap l m
let locTMapMem (l:DomAbsValue.locT) (m:(DomAbsValue.locT,'elt) MapInterface.dict) : bool =
  MapInterface.mem locT_ordered_type locT_fmap l m

(* bidT * bool *)

type bidBool = int * bool

let bidBool_setoid : bidBool DomType.setoid = 
  DomBasic.pair_setoid int_setoid DomType.bool_setoid
let bidBool_poset : bidBool DomType.poset = 
  DomBasic.pair_poset 
    int_setoid DomType.bool_setoid
    int_poset DomType.bool_poset
let bidBool_setoidDec : bidBool DomType.setoidDec =
  DomBasic.pair_setoidDec
    int_setoid DomType.bool_setoid
    int_setoidDec DomType.bool_setoidDec
let bidBool_posetDec : bidBool DomType.posetDec = 
  DomBasic.pair_posetDec
    int_setoid DomType.bool_setoid
    int_poset DomType.bool_poset
    int_setoidDec DomType.bool_setoidDec
    int_posetDec DomType.bool_posetDec

let bidBool_ordered_type : bidBool OrderedType.coq_OrderedType = 
  DomBridge.order_OrderedType
    bidBool_setoid bidBool_poset
    bidBool_setoidDec bidBool_posetDec

let bidBool_fmap : bidBool MapInterface.coq_FMap = 
  MapAVLInstance.coq_MapAVL_FMap bidBool_ordered_type
let bidBoolMapEmpty : (bidBool,'elt) MapInterface.dict = 
  MapInterface.empty bidBool_ordered_type bidBool_fmap
let bidBoolMapAdd (bb:bidBool) (e:'elt) (m:(bidBool,'elt) MapInterface.dict) : (bidBool,'elt) MapInterface.dict = 
  MapInterface.add bidBool_ordered_type bidBool_fmap bb e m
let bidBoolMapFind (bb:bidBool) (m:(bidBool,'elt) MapInterface.dict) : 'elt option = 
  MapInterface.find bidBool_ordered_type bidBool_fmap bb m
let bidBoolMapFold (f: bidBool -> 'v -> 'a -> 'a) (m: (bidBool, 'v) MapInterface.dict) (a: 'a) : 'a =
  MapInterface.fold bidBool_ordered_type bidBool_fmap f m a
