open BinInt
open Datatypes
open OrderedType
open Peano
open ZArith_dec

type __ = Obj.t
let __ = let rec f _ = Obj.repr f in Obj.repr f

module MapAVL = 
 struct 
  type ('key, 'elt) tree =
  | Leaf
  | Node of ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree * int
  
  (** val tree_rect :
      'a3 -> (('a1, 'a2) tree -> 'a3 -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> 'a3
      -> int -> 'a3) -> ('a1, 'a2) tree -> 'a3 **)
  
  let rec tree_rect f f0 = function
  | Leaf -> f
  | Node (t1, k, e, t2, z) ->
    f0 t1 (tree_rect f f0 t1) k e t2 (tree_rect f f0 t2) z
  
  (** val tree_rec :
      'a3 -> (('a1, 'a2) tree -> 'a3 -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> 'a3
      -> int -> 'a3) -> ('a1, 'a2) tree -> 'a3 **)
  
  let rec tree_rec f f0 = function
  | Leaf -> f
  | Node (t1, k, e, t2, z) ->
    f0 t1 (tree_rec f f0 t1) k e t2 (tree_rec f f0 t2) z
  
  (** val height : ('a1, 'a2) tree -> int **)
  
  let height = function
  | Leaf -> 0
  | Node (t0, k, e, t1, h) -> h
  
  (** val cardinal : ('a1, 'a2) tree -> int **)
  
  let rec cardinal = function
  | Leaf -> 0
  | Node (l, k, e, r, z) -> Pervasives.succ (plus (cardinal l) (cardinal r))
  
  (** val empty : ('a1, 'a2) tree **)
  
  let empty =
    Leaf
  
  (** val is_empty : ('a1, 'a2) tree -> bool **)
  
  let is_empty = function
  | Leaf -> true
  | Node (t0, k, e, t1, z) -> false
  
  (** val mem : 'a1 coq_OrderedType -> 'a1 -> ('a1, 'a2) tree -> bool **)
  
  let rec mem key_OT x = function
  | Leaf -> false
  | Node (l, y, e, r, z) ->
    (match compare key_OT x y with
     | Eq -> true
     | Lt -> mem key_OT x l
     | Gt -> mem key_OT x r)
  
  type ('key, 'elt) coq_R_mem =
  | R_mem_0 of ('key, 'elt) tree
  | R_mem_1 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int
  | R_mem_2 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int * bool * ('key, 'elt) coq_R_mem
  | R_mem_3 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int * bool * ('key, 'elt) coq_R_mem
  
  (** val coq_R_mem_rect :
      'a1 coq_OrderedType -> 'a1 -> (('a1, 'a2) tree -> __ -> 'a3) -> (('a1,
      'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int ->
      __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2
      -> ('a1, 'a2) tree -> int -> __ -> __ -> bool -> ('a1, 'a2) coq_R_mem
      -> 'a3 -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> int -> __ -> __ -> bool -> ('a1, 'a2) coq_R_mem ->
      'a3 -> 'a3) -> ('a1, 'a2) tree -> bool -> ('a1, 'a2) coq_R_mem -> 'a3 **)
  
  let rec coq_R_mem_rect key_OT x f f0 f1 f2 m b = function
  | R_mem_0 m0 -> f m0 __
  | R_mem_1 (m0, l, y, _x, r0, _x0) -> f0 m0 l y _x r0 _x0 __ __
  | R_mem_2 (m0, l, y, _x, r0, _x0, res, r1) ->
    f1 m0 l y _x r0 _x0 __ __ res r1
      (coq_R_mem_rect key_OT x f f0 f1 f2 l res r1)
  | R_mem_3 (m0, l, y, _x, r0, _x0, res, r1) ->
    f2 m0 l y _x r0 _x0 __ __ res r1
      (coq_R_mem_rect key_OT x f f0 f1 f2 r0 res r1)
  
  (** val coq_R_mem_rec :
      'a1 coq_OrderedType -> 'a1 -> (('a1, 'a2) tree -> __ -> 'a3) -> (('a1,
      'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int ->
      __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2
      -> ('a1, 'a2) tree -> int -> __ -> __ -> bool -> ('a1, 'a2) coq_R_mem
      -> 'a3 -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> int -> __ -> __ -> bool -> ('a1, 'a2) coq_R_mem ->
      'a3 -> 'a3) -> ('a1, 'a2) tree -> bool -> ('a1, 'a2) coq_R_mem -> 'a3 **)
  
  let rec coq_R_mem_rec key_OT x f f0 f1 f2 m b = function
  | R_mem_0 m0 -> f m0 __
  | R_mem_1 (m0, l, y, _x, r0, _x0) -> f0 m0 l y _x r0 _x0 __ __
  | R_mem_2 (m0, l, y, _x, r0, _x0, res, r1) ->
    f1 m0 l y _x r0 _x0 __ __ res r1
      (coq_R_mem_rec key_OT x f f0 f1 f2 l res r1)
  | R_mem_3 (m0, l, y, _x, r0, _x0, res, r1) ->
    f2 m0 l y _x r0 _x0 __ __ res r1
      (coq_R_mem_rec key_OT x f f0 f1 f2 r0 res r1)
  
  (** val find :
      'a1 coq_OrderedType -> 'a1 -> ('a1, 'a2) tree -> 'a2 option **)
  
  let rec find key_OT x = function
  | Leaf -> None
  | Node (l, y, d, r, z) ->
    (match compare key_OT x y with
     | Eq -> Some d
     | Lt -> find key_OT x l
     | Gt -> find key_OT x r)
  
  type ('key, 'elt) coq_R_find =
  | R_find_0 of ('key, 'elt) tree
  | R_find_1 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int
  | R_find_2 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int * 'elt option * ('key, 'elt) coq_R_find
  | R_find_3 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int * 'elt option * ('key, 'elt) coq_R_find
  
  (** val coq_R_find_rect :
      'a1 coq_OrderedType -> 'a1 -> (('a1, 'a2) tree -> __ -> 'a3) -> (('a1,
      'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int ->
      __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2
      -> ('a1, 'a2) tree -> int -> __ -> __ -> 'a2 option -> ('a1, 'a2)
      coq_R_find -> 'a3 -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1
      -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> 'a2 option -> ('a1,
      'a2) coq_R_find -> 'a3 -> 'a3) -> ('a1, 'a2) tree -> 'a2 option ->
      ('a1, 'a2) coq_R_find -> 'a3 **)
  
  let rec coq_R_find_rect key_OT x f f0 f1 f2 m o = function
  | R_find_0 m0 -> f m0 __
  | R_find_1 (m0, l, y, d, r0, _x) -> f0 m0 l y d r0 _x __ __
  | R_find_2 (m0, l, y, d, r0, _x, res, r1) ->
    f1 m0 l y d r0 _x __ __ res r1
      (coq_R_find_rect key_OT x f f0 f1 f2 l res r1)
  | R_find_3 (m0, l, y, d, r0, _x, res, r1) ->
    f2 m0 l y d r0 _x __ __ res r1
      (coq_R_find_rect key_OT x f f0 f1 f2 r0 res r1)
  
  (** val coq_R_find_rec :
      'a1 coq_OrderedType -> 'a1 -> (('a1, 'a2) tree -> __ -> 'a3) -> (('a1,
      'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int ->
      __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2
      -> ('a1, 'a2) tree -> int -> __ -> __ -> 'a2 option -> ('a1, 'a2)
      coq_R_find -> 'a3 -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1
      -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> 'a2 option -> ('a1,
      'a2) coq_R_find -> 'a3 -> 'a3) -> ('a1, 'a2) tree -> 'a2 option ->
      ('a1, 'a2) coq_R_find -> 'a3 **)
  
  let rec coq_R_find_rec key_OT x f f0 f1 f2 m o = function
  | R_find_0 m0 -> f m0 __
  | R_find_1 (m0, l, y, d, r0, _x) -> f0 m0 l y d r0 _x __ __
  | R_find_2 (m0, l, y, d, r0, _x, res, r1) ->
    f1 m0 l y d r0 _x __ __ res r1
      (coq_R_find_rec key_OT x f f0 f1 f2 l res r1)
  | R_find_3 (m0, l, y, d, r0, _x, res, r1) ->
    f2 m0 l y d r0 _x __ __ res r1
      (coq_R_find_rec key_OT x f f0 f1 f2 r0 res r1)
  
  (** val create :
      ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> ('a1, 'a2) tree **)
  
  let create l x e r =
    Node (l, x, e, r, (Z.add (Z.max (height l) (height r)) 1))
  
  (** val assert_false :
      ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> ('a1, 'a2) tree **)
  
  let assert_false =
    create
  
  (** val bal :
      ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> ('a1, 'a2) tree **)
  
  let bal l x d r =
    let hl = height l in
    let hr = height r in
    if coq_Z_gt_le_dec hl (Z.add hr ((fun p->2*p) 1))
    then (match l with
          | Leaf -> assert_false l x d r
          | Node (ll, lx, ld, lr, z) ->
            if coq_Z_ge_lt_dec (height ll) (height lr)
            then create ll lx ld (create lr x d r)
            else (match lr with
                  | Leaf -> assert_false l x d r
                  | Node (lrl, lrx, lrd, lrr, z0) ->
                    create (create ll lx ld lrl) lrx lrd (create lrr x d r)))
    else if coq_Z_gt_le_dec hr (Z.add hl ((fun p->2*p) 1))
         then (match r with
               | Leaf -> assert_false l x d r
               | Node (rl, rx, rd, rr, z) ->
                 if coq_Z_ge_lt_dec (height rr) (height rl)
                 then create (create l x d rl) rx rd rr
                 else (match rl with
                       | Leaf -> assert_false l x d r
                       | Node (rll, rlx, rld, rlr, z0) ->
                         create (create l x d rll) rlx rld
                           (create rlr rx rd rr)))
         else create l x d r
  
  type ('key, 'elt) coq_R_bal =
  | R_bal_0 of ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree
  | R_bal_1 of ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree
     * ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree * int
  | R_bal_2 of ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree
     * ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree * int
  | R_bal_3 of ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree
     * ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree * int
     * ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree * int
  | R_bal_4 of ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree
  | R_bal_5 of ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree
     * ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree * int
  | R_bal_6 of ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree
     * ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree * int
  | R_bal_7 of ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree
     * ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree * int
     * ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree * int
  | R_bal_8 of ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree
  
  (** val coq_R_bal_rect :
      (('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> __ -> __ -> __ ->
      'a3) -> (('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> __ -> __
      -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __
      -> __ -> 'a3) -> (('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree ->
      __ -> __ -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int ->
      __ -> __ -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1,
      'a2) tree -> __ -> __ -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2)
      tree -> int -> __ -> __ -> __ -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1,
      'a2) tree -> int -> __ -> 'a3) -> (('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> __ -> __ -> __ -> __ -> __ -> 'a3) -> (('a1, 'a2)
      tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> __ -> __ -> __ -> __ -> ('a1,
      'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> __ ->
      'a3) -> (('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> __ -> __
      -> __ -> __ -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int
      -> __ -> __ -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> __ -> __ -> __ -> __ -> ('a1, 'a2) tree -> 'a1 ->
      'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> __ -> ('a1, 'a2) tree ->
      'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> 'a3) -> (('a1, 'a2) tree
      -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> __ -> __ -> __ -> __ -> 'a3) ->
      ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> ('a1, 'a2) tree ->
      ('a1, 'a2) coq_R_bal -> 'a3 **)
  
  let coq_R_bal_rect f f0 f1 f2 f3 f4 f5 f6 f7 l x d r t0 = function
  | R_bal_0 (x0, x1, x2, x3) -> f x0 x1 x2 x3 __ __ __
  | R_bal_1 (x0, x1, x2, x3, x4, x5, x6, x7, x8) ->
    f0 x0 x1 x2 x3 __ __ x4 x5 x6 x7 x8 __ __ __
  | R_bal_2 (x0, x1, x2, x3, x4, x5, x6, x7, x8) ->
    f1 x0 x1 x2 x3 __ __ x4 x5 x6 x7 x8 __ __ __ __
  | R_bal_3 (x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13) ->
    f2 x0 x1 x2 x3 __ __ x4 x5 x6 x7 x8 __ __ __ x9 x10 x11 x12 x13 __
  | R_bal_4 (x0, x1, x2, x3) -> f3 x0 x1 x2 x3 __ __ __ __ __
  | R_bal_5 (x0, x1, x2, x3, x4, x5, x6, x7, x8) ->
    f4 x0 x1 x2 x3 __ __ __ __ x4 x5 x6 x7 x8 __ __ __
  | R_bal_6 (x0, x1, x2, x3, x4, x5, x6, x7, x8) ->
    f5 x0 x1 x2 x3 __ __ __ __ x4 x5 x6 x7 x8 __ __ __ __
  | R_bal_7 (x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13) ->
    f6 x0 x1 x2 x3 __ __ __ __ x4 x5 x6 x7 x8 __ __ __ x9 x10 x11 x12 x13 __
  | R_bal_8 (x0, x1, x2, x3) -> f7 x0 x1 x2 x3 __ __ __ __
  
  (** val coq_R_bal_rec :
      (('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> __ -> __ -> __ ->
      'a3) -> (('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> __ -> __
      -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __
      -> __ -> 'a3) -> (('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree ->
      __ -> __ -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int ->
      __ -> __ -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1,
      'a2) tree -> __ -> __ -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2)
      tree -> int -> __ -> __ -> __ -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1,
      'a2) tree -> int -> __ -> 'a3) -> (('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> __ -> __ -> __ -> __ -> __ -> 'a3) -> (('a1, 'a2)
      tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> __ -> __ -> __ -> __ -> ('a1,
      'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> __ ->
      'a3) -> (('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> __ -> __
      -> __ -> __ -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int
      -> __ -> __ -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> __ -> __ -> __ -> __ -> ('a1, 'a2) tree -> 'a1 ->
      'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> __ -> ('a1, 'a2) tree ->
      'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> 'a3) -> (('a1, 'a2) tree
      -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> __ -> __ -> __ -> __ -> 'a3) ->
      ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> ('a1, 'a2) tree ->
      ('a1, 'a2) coq_R_bal -> 'a3 **)
  
  let coq_R_bal_rec f f0 f1 f2 f3 f4 f5 f6 f7 l x d r t0 = function
  | R_bal_0 (x0, x1, x2, x3) -> f x0 x1 x2 x3 __ __ __
  | R_bal_1 (x0, x1, x2, x3, x4, x5, x6, x7, x8) ->
    f0 x0 x1 x2 x3 __ __ x4 x5 x6 x7 x8 __ __ __
  | R_bal_2 (x0, x1, x2, x3, x4, x5, x6, x7, x8) ->
    f1 x0 x1 x2 x3 __ __ x4 x5 x6 x7 x8 __ __ __ __
  | R_bal_3 (x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13) ->
    f2 x0 x1 x2 x3 __ __ x4 x5 x6 x7 x8 __ __ __ x9 x10 x11 x12 x13 __
  | R_bal_4 (x0, x1, x2, x3) -> f3 x0 x1 x2 x3 __ __ __ __ __
  | R_bal_5 (x0, x1, x2, x3, x4, x5, x6, x7, x8) ->
    f4 x0 x1 x2 x3 __ __ __ __ x4 x5 x6 x7 x8 __ __ __
  | R_bal_6 (x0, x1, x2, x3, x4, x5, x6, x7, x8) ->
    f5 x0 x1 x2 x3 __ __ __ __ x4 x5 x6 x7 x8 __ __ __ __
  | R_bal_7 (x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13) ->
    f6 x0 x1 x2 x3 __ __ __ __ x4 x5 x6 x7 x8 __ __ __ x9 x10 x11 x12 x13 __
  | R_bal_8 (x0, x1, x2, x3) -> f7 x0 x1 x2 x3 __ __ __ __
  
  (** val add :
      'a1 coq_OrderedType -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> ('a1, 'a2) tree **)
  
  let rec add key_OT x d = function
  | Leaf -> Node (Leaf, x, d, Leaf, 1)
  | Node (l, y, d', r, h) ->
    (match compare key_OT x y with
     | Eq -> Node (l, y, d, r, h)
     | Lt -> bal (add key_OT x d l) y d' r
     | Gt -> bal l y d' (add key_OT x d r))
  
  type ('key, 'elt) coq_R_add =
  | R_add_0 of ('key, 'elt) tree
  | R_add_1 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int
  | R_add_2 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int * ('key, 'elt) tree * ('key, 'elt) coq_R_add
  | R_add_3 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int * ('key, 'elt) tree * ('key, 'elt) coq_R_add
  
  (** val coq_R_add_rect :
      'a1 coq_OrderedType -> 'a1 -> 'a2 -> (('a1, 'a2) tree -> __ -> 'a3) ->
      (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree ->
      int -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1
      -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a2) tree ->
      ('a1, 'a2) coq_R_add -> 'a3 -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2)
      tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a2)
      tree -> ('a1, 'a2) coq_R_add -> 'a3 -> 'a3) -> ('a1, 'a2) tree -> ('a1,
      'a2) tree -> ('a1, 'a2) coq_R_add -> 'a3 **)
  
  let rec coq_R_add_rect key_OT x d f f0 f1 f2 m t0 = function
  | R_add_0 m0 -> f m0 __
  | R_add_1 (m0, l, y, d', r0, h) -> f0 m0 l y d' r0 h __ __
  | R_add_2 (m0, l, y, d', r0, h, res, r1) ->
    f1 m0 l y d' r0 h __ __ res r1
      (coq_R_add_rect key_OT x d f f0 f1 f2 l res r1)
  | R_add_3 (m0, l, y, d', r0, h, res, r1) ->
    f2 m0 l y d' r0 h __ __ res r1
      (coq_R_add_rect key_OT x d f f0 f1 f2 r0 res r1)
  
  (** val coq_R_add_rec :
      'a1 coq_OrderedType -> 'a1 -> 'a2 -> (('a1, 'a2) tree -> __ -> 'a3) ->
      (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree ->
      int -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1
      -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a2) tree ->
      ('a1, 'a2) coq_R_add -> 'a3 -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2)
      tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a2)
      tree -> ('a1, 'a2) coq_R_add -> 'a3 -> 'a3) -> ('a1, 'a2) tree -> ('a1,
      'a2) tree -> ('a1, 'a2) coq_R_add -> 'a3 **)
  
  let rec coq_R_add_rec key_OT x d f f0 f1 f2 m t0 = function
  | R_add_0 m0 -> f m0 __
  | R_add_1 (m0, l, y, d', r0, h) -> f0 m0 l y d' r0 h __ __
  | R_add_2 (m0, l, y, d', r0, h, res, r1) ->
    f1 m0 l y d' r0 h __ __ res r1
      (coq_R_add_rec key_OT x d f f0 f1 f2 l res r1)
  | R_add_3 (m0, l, y, d', r0, h, res, r1) ->
    f2 m0 l y d' r0 h __ __ res r1
      (coq_R_add_rec key_OT x d f f0 f1 f2 r0 res r1)
  
  (** val insert0 :
      'a1 coq_OrderedType -> 'a1 -> 'a2 -> ('a2 -> 'a2) -> ('a1, 'a2) tree ->
      ('a1, 'a2) tree **)
  
  let rec insert0 key_OT x d f = function
  | Leaf -> Node (Leaf, x, d, Leaf, 1)
  | Node (l, y, d', r, h) ->
    (match compare key_OT x y with
     | Eq -> Node (l, y, (f d'), r, h)
     | Lt -> bal (insert0 key_OT x d f l) y d' r
     | Gt -> bal l y d' (insert0 key_OT x d f r))
  
  type ('key, 'elt) coq_R_insert0 =
  | R_insert0_0 of ('key, 'elt) tree
  | R_insert0_1 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int
  | R_insert0_2 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int * ('key, 'elt) tree
     * ('key, 'elt) coq_R_insert0
  | R_insert0_3 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int * ('key, 'elt) tree
     * ('key, 'elt) coq_R_insert0
  
  (** val coq_R_insert0_rect :
      'a1 coq_OrderedType -> 'a1 -> 'a2 -> ('a2 -> 'a2) -> (('a1, 'a2) tree
      -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> int -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1,
      'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1,
      'a2) tree -> ('a1, 'a2) coq_R_insert0 -> 'a3 -> 'a3) -> (('a1, 'a2)
      tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __
      -> __ -> ('a1, 'a2) tree -> ('a1, 'a2) coq_R_insert0 -> 'a3 -> 'a3) ->
      ('a1, 'a2) tree -> ('a1, 'a2) tree -> ('a1, 'a2) coq_R_insert0 -> 'a3 **)
  
  let rec coq_R_insert0_rect key_OT x d f f0 f1 f2 f3 m t0 = function
  | R_insert0_0 m0 -> f0 m0 __
  | R_insert0_1 (m0, l, y, d', r0, h) -> f1 m0 l y d' r0 h __ __
  | R_insert0_2 (m0, l, y, d', r0, h, res, r1) ->
    f2 m0 l y d' r0 h __ __ res r1
      (coq_R_insert0_rect key_OT x d f f0 f1 f2 f3 l res r1)
  | R_insert0_3 (m0, l, y, d', r0, h, res, r1) ->
    f3 m0 l y d' r0 h __ __ res r1
      (coq_R_insert0_rect key_OT x d f f0 f1 f2 f3 r0 res r1)
  
  (** val coq_R_insert0_rec :
      'a1 coq_OrderedType -> 'a1 -> 'a2 -> ('a2 -> 'a2) -> (('a1, 'a2) tree
      -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> int -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1,
      'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1,
      'a2) tree -> ('a1, 'a2) coq_R_insert0 -> 'a3 -> 'a3) -> (('a1, 'a2)
      tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __
      -> __ -> ('a1, 'a2) tree -> ('a1, 'a2) coq_R_insert0 -> 'a3 -> 'a3) ->
      ('a1, 'a2) tree -> ('a1, 'a2) tree -> ('a1, 'a2) coq_R_insert0 -> 'a3 **)
  
  let rec coq_R_insert0_rec key_OT x d f f0 f1 f2 f3 m t0 = function
  | R_insert0_0 m0 -> f0 m0 __
  | R_insert0_1 (m0, l, y, d', r0, h) -> f1 m0 l y d' r0 h __ __
  | R_insert0_2 (m0, l, y, d', r0, h, res, r1) ->
    f2 m0 l y d' r0 h __ __ res r1
      (coq_R_insert0_rec key_OT x d f f0 f1 f2 f3 l res r1)
  | R_insert0_3 (m0, l, y, d', r0, h, res, r1) ->
    f3 m0 l y d' r0 h __ __ res r1
      (coq_R_insert0_rec key_OT x d f f0 f1 f2 f3 r0 res r1)
  
  (** val insert_aux :
      'a1 coq_OrderedType -> 'a1 -> 'a2 -> ('a2 -> 'a2) -> ('a1, 'a2) tree ->
      ('a1, 'a2) tree * bool **)
  
  let rec insert_aux key_OT x d f = function
  | Leaf -> ((Node (Leaf, x, d, Leaf, 1)), true)
  | Node (l, y, d', r, h) ->
    (match compare key_OT x y with
     | Eq -> ((Node (l, y, (f d'), r, h)), false)
     | Lt ->
       let (newl, b) = insert_aux key_OT x d f l in
       if b
       then ((bal newl y d' r), true)
       else ((Node (newl, y, d', r, h)), false)
     | Gt ->
       let (newr, b) = insert_aux key_OT x d f r in
       if b
       then ((bal l y d' newr), true)
       else ((Node (l, y, d', newr, h)), false))
  
  (** val insert :
      'a1 coq_OrderedType -> 'a1 -> 'a2 -> ('a2 -> 'a2) -> ('a1, 'a2) tree ->
      ('a1, 'a2) tree **)
  
  let insert key_OT x d f m =
    fst (insert_aux key_OT x d f m)
  
  type ('key, 'elt) coq_R_insert_aux =
  | R_insert_aux_0 of ('key, 'elt) tree
  | R_insert_aux_1 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 
     'elt * ('key, 'elt) tree * int
  | R_insert_aux_2 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 
     'elt * ('key, 'elt) tree * int * (('key, 'elt) tree * bool)
     * ('key, 'elt) coq_R_insert_aux * ('key, 'elt) tree * bool
  | R_insert_aux_3 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 
     'elt * ('key, 'elt) tree * int * (('key, 'elt) tree * bool)
     * ('key, 'elt) coq_R_insert_aux * ('key, 'elt) tree * bool
  | R_insert_aux_4 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 
     'elt * ('key, 'elt) tree * int * (('key, 'elt) tree * bool)
     * ('key, 'elt) coq_R_insert_aux * ('key, 'elt) tree * bool
  | R_insert_aux_5 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 
     'elt * ('key, 'elt) tree * int * (('key, 'elt) tree * bool)
     * ('key, 'elt) coq_R_insert_aux * ('key, 'elt) tree * bool
  
  (** val coq_R_insert_aux_rect :
      'a1 coq_OrderedType -> 'a1 -> 'a2 -> ('a2 -> 'a2) -> (('a1, 'a2) tree
      -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> int -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1,
      'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> (('a1,
      'a2) tree * bool) -> ('a1, 'a2) coq_R_insert_aux -> 'a3 -> ('a1, 'a2)
      tree -> bool -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree
      -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> (('a1, 'a2)
      tree * bool) -> ('a1, 'a2) coq_R_insert_aux -> 'a3 -> ('a1, 'a2) tree
      -> bool -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree ->
      'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> (('a1, 'a2)
      tree * bool) -> ('a1, 'a2) coq_R_insert_aux -> 'a3 -> ('a1, 'a2) tree
      -> bool -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree ->
      'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> (('a1, 'a2)
      tree * bool) -> ('a1, 'a2) coq_R_insert_aux -> 'a3 -> ('a1, 'a2) tree
      -> bool -> __ -> __ -> 'a3) -> ('a1, 'a2) tree -> (('a1, 'a2)
      tree * bool) -> ('a1, 'a2) coq_R_insert_aux -> 'a3 **)
  
  let rec coq_R_insert_aux_rect key_OT x d f f0 f1 f2 f3 f4 f5 m p = function
  | R_insert_aux_0 m0 -> f0 m0 __
  | R_insert_aux_1 (m0, l, y, d', r0, h) -> f1 m0 l y d' r0 h __ __
  | R_insert_aux_2 (m0, l, y, d', r0, h, res, r1, newl, b) ->
    f2 m0 l y d' r0 h __ __ res r1
      (coq_R_insert_aux_rect key_OT x d f f0 f1 f2 f3 f4 f5 l res r1) newl b
      __ __
  | R_insert_aux_3 (m0, l, y, d', r0, h, res, r1, newl, b) ->
    f3 m0 l y d' r0 h __ __ res r1
      (coq_R_insert_aux_rect key_OT x d f f0 f1 f2 f3 f4 f5 l res r1) newl b
      __ __
  | R_insert_aux_4 (m0, l, y, d', r0, h, res, r1, newr, b) ->
    f4 m0 l y d' r0 h __ __ res r1
      (coq_R_insert_aux_rect key_OT x d f f0 f1 f2 f3 f4 f5 r0 res r1) newr b
      __ __
  | R_insert_aux_5 (m0, l, y, d', r0, h, res, r1, newr, b) ->
    f5 m0 l y d' r0 h __ __ res r1
      (coq_R_insert_aux_rect key_OT x d f f0 f1 f2 f3 f4 f5 r0 res r1) newr b
      __ __
  
  (** val coq_R_insert_aux_rec :
      'a1 coq_OrderedType -> 'a1 -> 'a2 -> ('a2 -> 'a2) -> (('a1, 'a2) tree
      -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> int -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1,
      'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> (('a1,
      'a2) tree * bool) -> ('a1, 'a2) coq_R_insert_aux -> 'a3 -> ('a1, 'a2)
      tree -> bool -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree
      -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> (('a1, 'a2)
      tree * bool) -> ('a1, 'a2) coq_R_insert_aux -> 'a3 -> ('a1, 'a2) tree
      -> bool -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree ->
      'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> (('a1, 'a2)
      tree * bool) -> ('a1, 'a2) coq_R_insert_aux -> 'a3 -> ('a1, 'a2) tree
      -> bool -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree ->
      'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> (('a1, 'a2)
      tree * bool) -> ('a1, 'a2) coq_R_insert_aux -> 'a3 -> ('a1, 'a2) tree
      -> bool -> __ -> __ -> 'a3) -> ('a1, 'a2) tree -> (('a1, 'a2)
      tree * bool) -> ('a1, 'a2) coq_R_insert_aux -> 'a3 **)
  
  let rec coq_R_insert_aux_rec key_OT x d f f0 f1 f2 f3 f4 f5 m p = function
  | R_insert_aux_0 m0 -> f0 m0 __
  | R_insert_aux_1 (m0, l, y, d', r0, h) -> f1 m0 l y d' r0 h __ __
  | R_insert_aux_2 (m0, l, y, d', r0, h, res, r1, newl, b) ->
    f2 m0 l y d' r0 h __ __ res r1
      (coq_R_insert_aux_rec key_OT x d f f0 f1 f2 f3 f4 f5 l res r1) newl b
      __ __
  | R_insert_aux_3 (m0, l, y, d', r0, h, res, r1, newl, b) ->
    f3 m0 l y d' r0 h __ __ res r1
      (coq_R_insert_aux_rec key_OT x d f f0 f1 f2 f3 f4 f5 l res r1) newl b
      __ __
  | R_insert_aux_4 (m0, l, y, d', r0, h, res, r1, newr, b) ->
    f4 m0 l y d' r0 h __ __ res r1
      (coq_R_insert_aux_rec key_OT x d f f0 f1 f2 f3 f4 f5 r0 res r1) newr b
      __ __
  | R_insert_aux_5 (m0, l, y, d', r0, h, res, r1, newr, b) ->
    f5 m0 l y d' r0 h __ __ res r1
      (coq_R_insert_aux_rec key_OT x d f f0 f1 f2 f3 f4 f5 r0 res r1) newr b
      __ __
  
  (** val adjust :
      'a1 coq_OrderedType -> 'a1 -> ('a2 -> 'a2) -> ('a1, 'a2) tree -> ('a1,
      'a2) tree **)
  
  let rec adjust key_OT x f = function
  | Leaf -> Leaf
  | Node (l, y, d', r, h) ->
    (match compare key_OT x y with
     | Eq -> Node (l, y, (f d'), r, h)
     | Lt -> Node ((adjust key_OT x f l), y, d', r, h)
     | Gt -> Node (l, y, d', (adjust key_OT x f r), h))
  
  type ('key, 'elt) coq_R_adjust =
  | R_adjust_0 of ('key, 'elt) tree
  | R_adjust_1 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int
  | R_adjust_2 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int * ('key, 'elt) tree
     * ('key, 'elt) coq_R_adjust
  | R_adjust_3 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int * ('key, 'elt) tree
     * ('key, 'elt) coq_R_adjust
  
  (** val coq_R_adjust_rect :
      'a1 coq_OrderedType -> 'a1 -> ('a2 -> 'a2) -> (('a1, 'a2) tree -> __ ->
      'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2)
      tree -> int -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree
      -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a2) tree
      -> ('a1, 'a2) coq_R_adjust -> 'a3 -> 'a3) -> (('a1, 'a2) tree -> ('a1,
      'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1,
      'a2) tree -> ('a1, 'a2) coq_R_adjust -> 'a3 -> 'a3) -> ('a1, 'a2) tree
      -> ('a1, 'a2) tree -> ('a1, 'a2) coq_R_adjust -> 'a3 **)
  
  let rec coq_R_adjust_rect key_OT x f f0 f1 f2 f3 m t0 = function
  | R_adjust_0 m0 -> f0 m0 __
  | R_adjust_1 (m0, l, y, d', r0, h) -> f1 m0 l y d' r0 h __ __
  | R_adjust_2 (m0, l, y, d', r0, h, res, r1) ->
    f2 m0 l y d' r0 h __ __ res r1
      (coq_R_adjust_rect key_OT x f f0 f1 f2 f3 l res r1)
  | R_adjust_3 (m0, l, y, d', r0, h, res, r1) ->
    f3 m0 l y d' r0 h __ __ res r1
      (coq_R_adjust_rect key_OT x f f0 f1 f2 f3 r0 res r1)
  
  (** val coq_R_adjust_rec :
      'a1 coq_OrderedType -> 'a1 -> ('a2 -> 'a2) -> (('a1, 'a2) tree -> __ ->
      'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2)
      tree -> int -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree
      -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a2) tree
      -> ('a1, 'a2) coq_R_adjust -> 'a3 -> 'a3) -> (('a1, 'a2) tree -> ('a1,
      'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1,
      'a2) tree -> ('a1, 'a2) coq_R_adjust -> 'a3 -> 'a3) -> ('a1, 'a2) tree
      -> ('a1, 'a2) tree -> ('a1, 'a2) coq_R_adjust -> 'a3 **)
  
  let rec coq_R_adjust_rec key_OT x f f0 f1 f2 f3 m t0 = function
  | R_adjust_0 m0 -> f0 m0 __
  | R_adjust_1 (m0, l, y, d', r0, h) -> f1 m0 l y d' r0 h __ __
  | R_adjust_2 (m0, l, y, d', r0, h, res, r1) ->
    f2 m0 l y d' r0 h __ __ res r1
      (coq_R_adjust_rec key_OT x f f0 f1 f2 f3 l res r1)
  | R_adjust_3 (m0, l, y, d', r0, h, res, r1) ->
    f3 m0 l y d' r0 h __ __ res r1
      (coq_R_adjust_rec key_OT x f f0 f1 f2 f3 r0 res r1)
  
  (** val remove_min :
      ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> ('a1, 'a2)
      tree * ('a1 * 'a2) **)
  
  let rec remove_min l x d r =
    match l with
    | Leaf -> (r, (x, d))
    | Node (ll, lx, ld, lr, lh) ->
      let (l', m) = remove_min ll lx ld lr in ((bal l' x d r), m)
  
  type ('key, 'elt) coq_R_remove_min =
  | R_remove_min_0 of ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree
  | R_remove_min_1 of ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree
     * ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree * int
     * (('key, 'elt) tree * ('key * 'elt)) * ('key, 'elt) coq_R_remove_min
     * ('key, 'elt) tree * ('key * 'elt)
  
  (** val coq_R_remove_min_rect :
      (('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> __ -> 'a3) ->
      (('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> ('a1, 'a2) tree ->
      'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> (('a1, 'a2)
      tree * ('a1 * 'a2)) -> ('a1, 'a2) coq_R_remove_min -> 'a3 -> ('a1, 'a2)
      tree -> ('a1 * 'a2) -> __ -> 'a3) -> ('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> (('a1, 'a2) tree * ('a1 * 'a2)) -> ('a1, 'a2)
      coq_R_remove_min -> 'a3 **)
  
  let rec coq_R_remove_min_rect f f0 l x d r p = function
  | R_remove_min_0 (l0, x0, d0, r1) -> f l0 x0 d0 r1 __
  | R_remove_min_1 (l0, x0, d0, r1, ll, lx, ld, lr, _x, res, r2, l', m) ->
    f0 l0 x0 d0 r1 ll lx ld lr _x __ res r2
      (coq_R_remove_min_rect f f0 ll lx ld lr res r2) l' m __
  
  (** val coq_R_remove_min_rec :
      (('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> __ -> 'a3) ->
      (('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> ('a1, 'a2) tree ->
      'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> (('a1, 'a2)
      tree * ('a1 * 'a2)) -> ('a1, 'a2) coq_R_remove_min -> 'a3 -> ('a1, 'a2)
      tree -> ('a1 * 'a2) -> __ -> 'a3) -> ('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> (('a1, 'a2) tree * ('a1 * 'a2)) -> ('a1, 'a2)
      coq_R_remove_min -> 'a3 **)
  
  let rec coq_R_remove_min_rec f f0 l x d r p = function
  | R_remove_min_0 (l0, x0, d0, r1) -> f l0 x0 d0 r1 __
  | R_remove_min_1 (l0, x0, d0, r1, ll, lx, ld, lr, _x, res, r2, l', m) ->
    f0 l0 x0 d0 r1 ll lx ld lr _x __ res r2
      (coq_R_remove_min_rec f f0 ll lx ld lr res r2) l' m __
  
  (** val merge : ('a1, 'a2) tree -> ('a1, 'a2) tree -> ('a1, 'a2) tree **)
  
  let merge s1 s2 =
    match s1 with
    | Leaf -> s2
    | Node (t0, k, e, t1, z) ->
      (match s2 with
       | Leaf -> s1
       | Node (l2, x2, d2, r2, h2) ->
         let (s2', p) = remove_min l2 x2 d2 r2 in
         let (x, d) = p in bal s1 x d s2')
  
  type ('key, 'elt) coq_R_merge =
  | R_merge_0 of ('key, 'elt) tree * ('key, 'elt) tree
  | R_merge_1 of ('key, 'elt) tree * ('key, 'elt) tree * ('key, 'elt) tree
     * 'key * 'elt * ('key, 'elt) tree * int
  | R_merge_2 of ('key, 'elt) tree * ('key, 'elt) tree * ('key, 'elt) tree
     * 'key * 'elt * ('key, 'elt) tree * int * ('key, 'elt) tree * 'key
     * 'elt * ('key, 'elt) tree * int * ('key, 'elt) tree * ('key * 'elt)
     * 'key * 'elt
  
  (** val coq_R_merge_rect :
      (('a1, 'a2) tree -> ('a1, 'a2) tree -> __ -> 'a3) -> (('a1, 'a2) tree
      -> ('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree
      -> int -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree ->
      ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> ('a1,
      'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> ('a1, 'a2)
      tree -> ('a1 * 'a2) -> __ -> 'a1 -> 'a2 -> __ -> 'a3) -> ('a1, 'a2)
      tree -> ('a1, 'a2) tree -> ('a1, 'a2) tree -> ('a1, 'a2) coq_R_merge ->
      'a3 **)
  
  let coq_R_merge_rect f f0 f1 s1 s2 t0 = function
  | R_merge_0 (x, x0) -> f x x0 __
  | R_merge_1 (x, x0, x1, x2, x3, x4, x5) -> f0 x x0 x1 x2 x3 x4 x5 __ __
  | R_merge_2 (x, x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13,
               x14) ->
    f1 x x0 x1 x2 x3 x4 x5 __ x6 x7 x8 x9 x10 __ x11 x12 __ x13 x14 __
  
  (** val coq_R_merge_rec :
      (('a1, 'a2) tree -> ('a1, 'a2) tree -> __ -> 'a3) -> (('a1, 'a2) tree
      -> ('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree
      -> int -> __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree ->
      ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> ('a1,
      'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> ('a1, 'a2)
      tree -> ('a1 * 'a2) -> __ -> 'a1 -> 'a2 -> __ -> 'a3) -> ('a1, 'a2)
      tree -> ('a1, 'a2) tree -> ('a1, 'a2) tree -> ('a1, 'a2) coq_R_merge ->
      'a3 **)
  
  let coq_R_merge_rec f f0 f1 s1 s2 t0 = function
  | R_merge_0 (x, x0) -> f x x0 __
  | R_merge_1 (x, x0, x1, x2, x3, x4, x5) -> f0 x x0 x1 x2 x3 x4 x5 __ __
  | R_merge_2 (x, x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13,
               x14) ->
    f1 x x0 x1 x2 x3 x4 x5 __ x6 x7 x8 x9 x10 __ x11 x12 __ x13 x14 __
  
  (** val remove :
      'a1 coq_OrderedType -> 'a1 -> ('a1, 'a2) tree -> ('a1, 'a2) tree **)
  
  let rec remove key_OT x = function
  | Leaf -> Leaf
  | Node (l, y, d, r, h) ->
    (match compare key_OT x y with
     | Eq -> merge l r
     | Lt -> bal (remove key_OT x l) y d r
     | Gt -> bal l y d (remove key_OT x r))
  
  type ('key, 'elt) coq_R_remove =
  | R_remove_0 of ('key, 'elt) tree
  | R_remove_1 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int
  | R_remove_2 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int * ('key, 'elt) tree
     * ('key, 'elt) coq_R_remove
  | R_remove_3 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int * ('key, 'elt) tree
     * ('key, 'elt) coq_R_remove
  
  (** val coq_R_remove_rect :
      'a1 coq_OrderedType -> 'a1 -> (('a1, 'a2) tree -> __ -> 'a3) -> (('a1,
      'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int ->
      __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2
      -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a2) tree -> ('a1, 'a2)
      coq_R_remove -> 'a3 -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree ->
      'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a2) tree ->
      ('a1, 'a2) coq_R_remove -> 'a3 -> 'a3) -> ('a1, 'a2) tree -> ('a1, 'a2)
      tree -> ('a1, 'a2) coq_R_remove -> 'a3 **)
  
  let rec coq_R_remove_rect key_OT x f f0 f1 f2 m t0 = function
  | R_remove_0 m0 -> f m0 __
  | R_remove_1 (m0, l, y, d, r0, _x) -> f0 m0 l y d r0 _x __ __
  | R_remove_2 (m0, l, y, d, r0, _x, res, r1) ->
    f1 m0 l y d r0 _x __ __ res r1
      (coq_R_remove_rect key_OT x f f0 f1 f2 l res r1)
  | R_remove_3 (m0, l, y, d, r0, _x, res, r1) ->
    f2 m0 l y d r0 _x __ __ res r1
      (coq_R_remove_rect key_OT x f f0 f1 f2 r0 res r1)
  
  (** val coq_R_remove_rec :
      'a1 coq_OrderedType -> 'a1 -> (('a1, 'a2) tree -> __ -> 'a3) -> (('a1,
      'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int ->
      __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2
      -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a2) tree -> ('a1, 'a2)
      coq_R_remove -> 'a3 -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree ->
      'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a2) tree ->
      ('a1, 'a2) coq_R_remove -> 'a3 -> 'a3) -> ('a1, 'a2) tree -> ('a1, 'a2)
      tree -> ('a1, 'a2) coq_R_remove -> 'a3 **)
  
  let rec coq_R_remove_rec key_OT x f f0 f1 f2 m t0 = function
  | R_remove_0 m0 -> f m0 __
  | R_remove_1 (m0, l, y, d, r0, _x) -> f0 m0 l y d r0 _x __ __
  | R_remove_2 (m0, l, y, d, r0, _x, res, r1) ->
    f1 m0 l y d r0 _x __ __ res r1
      (coq_R_remove_rec key_OT x f f0 f1 f2 l res r1)
  | R_remove_3 (m0, l, y, d, r0, _x, res, r1) ->
    f2 m0 l y d r0 _x __ __ res r1
      (coq_R_remove_rec key_OT x f f0 f1 f2 r0 res r1)
  
  (** val join :
      'a1 coq_OrderedType -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree
      -> ('a1, 'a2) tree **)
  
  let rec join key_OT l = match l with
  | Leaf -> add key_OT
  | Node (ll, lx, ld, lr, lh) ->
    (fun x d ->
      let rec join_aux r = match r with
      | Leaf -> add key_OT x d l
      | Node (rl, rx, rd, rr, rh) ->
        if coq_Z_gt_le_dec lh (Z.add rh ((fun p->2*p) 1))
        then bal ll lx ld (join key_OT lr x d r)
        else if coq_Z_gt_le_dec rh (Z.add lh ((fun p->2*p) 1))
             then bal (join_aux rl) rx rd rr
             else create l x d r
      in join_aux)
  
  type ('key, 'elt) triple = { t_left : ('key, 'elt) tree;
                               t_opt : 'elt option;
                               t_right : ('key, 'elt) tree }
  
  (** val triple_rect :
      (('a1, 'a2) tree -> 'a2 option -> ('a1, 'a2) tree -> 'a3) -> ('a1, 'a2)
      triple -> 'a3 **)
  
  let triple_rect f t0 =
    let { t_left = x; t_opt = x0; t_right = x1 } = t0 in f x x0 x1
  
  (** val triple_rec :
      (('a1, 'a2) tree -> 'a2 option -> ('a1, 'a2) tree -> 'a3) -> ('a1, 'a2)
      triple -> 'a3 **)
  
  let triple_rec f t0 =
    let { t_left = x; t_opt = x0; t_right = x1 } = t0 in f x x0 x1
  
  (** val t_left : ('a1, 'a2) triple -> ('a1, 'a2) tree **)
  
  let t_left x = x.t_left
  
  (** val t_opt : ('a1, 'a2) triple -> 'a2 option **)
  
  let t_opt x = x.t_opt
  
  (** val t_right : ('a1, 'a2) triple -> ('a1, 'a2) tree **)
  
  let t_right x = x.t_right
  
  (** val split :
      'a1 coq_OrderedType -> 'a1 -> ('a1, 'a2) tree -> ('a1, 'a2) triple **)
  
  let rec split key_OT x = function
  | Leaf -> { t_left = Leaf; t_opt = None; t_right = Leaf }
  | Node (l, y, d, r, h) ->
    (match compare key_OT x y with
     | Eq -> { t_left = l; t_opt = (Some d); t_right = r }
     | Lt ->
       let { t_left = ll; t_opt = o; t_right = rl } = split key_OT x l in
       { t_left = ll; t_opt = o; t_right = (join key_OT rl y d r) }
     | Gt ->
       let { t_left = rl; t_opt = o; t_right = rr } = split key_OT x r in
       { t_left = (join key_OT l y d rl); t_opt = o; t_right = rr })
  
  type ('key, 'elt) coq_R_split =
  | R_split_0 of ('key, 'elt) tree
  | R_split_1 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int
  | R_split_2 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int * ('key, 'elt) triple
     * ('key, 'elt) coq_R_split * ('key, 'elt) tree * 'elt option
     * ('key, 'elt) tree
  | R_split_3 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 'elt
     * ('key, 'elt) tree * int * ('key, 'elt) triple
     * ('key, 'elt) coq_R_split * ('key, 'elt) tree * 'elt option
     * ('key, 'elt) tree
  
  (** val coq_R_split_rect :
      'a1 coq_OrderedType -> 'a1 -> (('a1, 'a2) tree -> __ -> 'a3) -> (('a1,
      'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int ->
      __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2
      -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a2) triple -> ('a1,
      'a2) coq_R_split -> 'a3 -> ('a1, 'a2) tree -> 'a2 option -> ('a1, 'a2)
      tree -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2
      -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a2) triple -> ('a1,
      'a2) coq_R_split -> 'a3 -> ('a1, 'a2) tree -> 'a2 option -> ('a1, 'a2)
      tree -> __ -> 'a3) -> ('a1, 'a2) tree -> ('a1, 'a2) triple -> ('a1,
      'a2) coq_R_split -> 'a3 **)
  
  let rec coq_R_split_rect key_OT x f f0 f1 f2 m t0 = function
  | R_split_0 m0 -> f m0 __
  | R_split_1 (m0, l, y, d, r0, _x) -> f0 m0 l y d r0 _x __ __
  | R_split_2 (m0, l, y, d, r0, _x, res, r1, ll, o, rl) ->
    f1 m0 l y d r0 _x __ __ res r1
      (coq_R_split_rect key_OT x f f0 f1 f2 l res r1) ll o rl __
  | R_split_3 (m0, l, y, d, r0, _x, res, r1, rl, o, rr) ->
    f2 m0 l y d r0 _x __ __ res r1
      (coq_R_split_rect key_OT x f f0 f1 f2 r0 res r1) rl o rr __
  
  (** val coq_R_split_rec :
      'a1 coq_OrderedType -> 'a1 -> (('a1, 'a2) tree -> __ -> 'a3) -> (('a1,
      'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int ->
      __ -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2
      -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a2) triple -> ('a1,
      'a2) coq_R_split -> 'a3 -> ('a1, 'a2) tree -> 'a2 option -> ('a1, 'a2)
      tree -> __ -> 'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2
      -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a2) triple -> ('a1,
      'a2) coq_R_split -> 'a3 -> ('a1, 'a2) tree -> 'a2 option -> ('a1, 'a2)
      tree -> __ -> 'a3) -> ('a1, 'a2) tree -> ('a1, 'a2) triple -> ('a1,
      'a2) coq_R_split -> 'a3 **)
  
  let rec coq_R_split_rec key_OT x f f0 f1 f2 m t0 = function
  | R_split_0 m0 -> f m0 __
  | R_split_1 (m0, l, y, d, r0, _x) -> f0 m0 l y d r0 _x __ __
  | R_split_2 (m0, l, y, d, r0, _x, res, r1, ll, o, rl) ->
    f1 m0 l y d r0 _x __ __ res r1
      (coq_R_split_rec key_OT x f f0 f1 f2 l res r1) ll o rl __
  | R_split_3 (m0, l, y, d, r0, _x, res, r1, rl, o, rr) ->
    f2 m0 l y d r0 _x __ __ res r1
      (coq_R_split_rec key_OT x f f0 f1 f2 r0 res r1) rl o rr __
  
  (** val concat :
      'a1 coq_OrderedType -> ('a1, 'a2) tree -> ('a1, 'a2) tree -> ('a1, 'a2)
      tree **)
  
  let concat key_OT m1 m2 =
    match m1 with
    | Leaf -> m2
    | Node (t0, k, e, t1, z) ->
      (match m2 with
       | Leaf -> m1
       | Node (l2, x2, d2, r2, z0) ->
         let (m2', xd) = remove_min l2 x2 d2 r2 in
         join key_OT m1 (fst xd) (snd xd) m2')
  
  type ('key, 'elt) coq_R_concat =
  | R_concat_0 of ('key, 'elt) tree * ('key, 'elt) tree
  | R_concat_1 of ('key, 'elt) tree * ('key, 'elt) tree * ('key, 'elt) tree
     * 'key * 'elt * ('key, 'elt) tree * int
  | R_concat_2 of ('key, 'elt) tree * ('key, 'elt) tree * ('key, 'elt) tree
     * 'key * 'elt * ('key, 'elt) tree * int * ('key, 'elt) tree * 'key
     * 'elt * ('key, 'elt) tree * int * ('key, 'elt) tree * ('key * 'elt)
  
  (** val coq_R_concat_rect :
      'a1 coq_OrderedType -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> __ ->
      'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1
      -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> 'a3) -> (('a1, 'a2)
      tree -> ('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2)
      tree -> int -> __ -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree
      -> int -> __ -> ('a1, 'a2) tree -> ('a1 * 'a2) -> __ -> 'a3) -> ('a1,
      'a2) tree -> ('a1, 'a2) tree -> ('a1, 'a2) tree -> ('a1, 'a2)
      coq_R_concat -> 'a3 **)
  
  let coq_R_concat_rect key_OT f f0 f1 m1 m2 t0 = function
  | R_concat_0 (x, x0) -> f x x0 __
  | R_concat_1 (x, x0, x1, x2, x3, x4, x5) -> f0 x x0 x1 x2 x3 x4 x5 __ __
  | R_concat_2 (x, x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12) ->
    f1 x x0 x1 x2 x3 x4 x5 __ x6 x7 x8 x9 x10 __ x11 x12 __
  
  (** val coq_R_concat_rec :
      'a1 coq_OrderedType -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> __ ->
      'a3) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1
      -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> 'a3) -> (('a1, 'a2)
      tree -> ('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2)
      tree -> int -> __ -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree
      -> int -> __ -> ('a1, 'a2) tree -> ('a1 * 'a2) -> __ -> 'a3) -> ('a1,
      'a2) tree -> ('a1, 'a2) tree -> ('a1, 'a2) tree -> ('a1, 'a2)
      coq_R_concat -> 'a3 **)
  
  let coq_R_concat_rec key_OT f f0 f1 m1 m2 t0 = function
  | R_concat_0 (x, x0) -> f x x0 __
  | R_concat_1 (x, x0, x1, x2, x3, x4, x5) -> f0 x x0 x1 x2 x3 x4 x5 __ __
  | R_concat_2 (x, x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12) ->
    f1 x x0 x1 x2 x3 x4 x5 __ x6 x7 x8 x9 x10 __ x11 x12 __
  
  (** val elements_aux :
      ('a1 * 'a2) list -> ('a1, 'a2) tree -> ('a1 * 'a2) list **)
  
  let rec elements_aux acc = function
  | Leaf -> acc
  | Node (l, x, d, r, z) -> elements_aux ((x, d) :: (elements_aux acc r)) l
  
  (** val elements : ('a1, 'a2) tree -> ('a1 * 'a2) list **)
  
  let elements m =
    elements_aux [] m
  
  (** val fold :
      ('a1 -> 'a2 -> 'a3 -> 'a3) -> ('a1, 'a2) tree -> 'a3 -> 'a3 **)
  
  let rec fold f m a =
    match m with
    | Leaf -> a
    | Node (l, x, d, r, z) -> fold f r (f x d (fold f l a))
  
  type ('key, 'elt) enumeration =
  | End
  | More of 'key * 'elt * ('key, 'elt) tree * ('key, 'elt) enumeration
  
  (** val enumeration_rect :
      'a3 -> ('a1 -> 'a2 -> ('a1, 'a2) tree -> ('a1, 'a2) enumeration -> 'a3
      -> 'a3) -> ('a1, 'a2) enumeration -> 'a3 **)
  
  let rec enumeration_rect f f0 = function
  | End -> f
  | More (k, e0, t0, e1) -> f0 k e0 t0 e1 (enumeration_rect f f0 e1)
  
  (** val enumeration_rec :
      'a3 -> ('a1 -> 'a2 -> ('a1, 'a2) tree -> ('a1, 'a2) enumeration -> 'a3
      -> 'a3) -> ('a1, 'a2) enumeration -> 'a3 **)
  
  let rec enumeration_rec f f0 = function
  | End -> f
  | More (k, e0, t0, e1) -> f0 k e0 t0 e1 (enumeration_rec f f0 e1)
  
  (** val cons :
      ('a1, 'a2) tree -> ('a1, 'a2) enumeration -> ('a1, 'a2) enumeration **)
  
  let rec cons m e =
    match m with
    | Leaf -> e
    | Node (l, x, d, r, h) -> cons l (More (x, d, r, e))
  
  (** val equal_more :
      'a1 coq_OrderedType -> ('a2 -> 'a2 -> bool) -> 'a1 -> 'a2 -> (('a1,
      'a2) enumeration -> bool) -> ('a1, 'a2) enumeration -> bool **)
  
  let equal_more key_OT cmp x1 d1 cont = function
  | End -> false
  | More (x2, d2, r2, e3) ->
    (match compare key_OT x1 x2 with
     | Eq -> if cmp d1 d2 then cont (cons r2 e3) else false
     | _ -> false)
  
  (** val equal_cont :
      'a1 coq_OrderedType -> ('a2 -> 'a2 -> bool) -> ('a1, 'a2) tree ->
      (('a1, 'a2) enumeration -> bool) -> ('a1, 'a2) enumeration -> bool **)
  
  let rec equal_cont key_OT cmp m1 cont e2 =
    match m1 with
    | Leaf -> cont e2
    | Node (l1, x1, d1, r1, z) ->
      equal_cont key_OT cmp l1
        (equal_more key_OT cmp x1 d1 (equal_cont key_OT cmp r1 cont)) e2
  
  (** val equal_end : ('a1, 'a2) enumeration -> bool **)
  
  let equal_end = function
  | End -> true
  | More (k, e, t0, e0) -> false
  
  (** val equal :
      'a1 coq_OrderedType -> ('a2 -> 'a2 -> bool) -> ('a1, 'a2) tree -> ('a1,
      'a2) tree -> bool **)
  
  let equal key_OT cmp m1 m2 =
    equal_cont key_OT cmp m1 equal_end (cons m2 End)
  
  (** val map : ('a2 -> 'a3) -> ('a1, 'a2) tree -> ('a1, 'a3) tree **)
  
  let rec map f = function
  | Leaf -> Leaf
  | Node (l, x, d, r, h) -> Node ((map f l), x, (f d), (map f r), h)
  
  (** val mapi :
      ('a1 -> 'a2 -> 'a3) -> ('a1, 'a2) tree -> ('a1, 'a3) tree **)
  
  let rec mapi f = function
  | Leaf -> Leaf
  | Node (l, x, d, r, h) -> Node ((mapi f l), x, (f x d), (mapi f r), h)
  
  (** val map_option :
      'a1 coq_OrderedType -> ('a1 -> 'a2 -> 'a3 option) -> ('a1, 'a2) tree ->
      ('a1, 'a3) tree **)
  
  let rec map_option key_OT f = function
  | Leaf -> Leaf
  | Node (l, x, d, r, h) ->
    (match f x d with
     | Some d' ->
       join key_OT (map_option key_OT f l) x d' (map_option key_OT f r)
     | None -> concat key_OT (map_option key_OT f l) (map_option key_OT f r))
  
  (** val map2_opt :
      'a1 coq_OrderedType -> ('a1 -> 'a2 -> 'a3 option -> 'a4 option) ->
      (('a1, 'a2) tree -> ('a1, 'a4) tree) -> (('a1, 'a3) tree -> ('a1, 'a4)
      tree) -> ('a1, 'a2) tree -> ('a1, 'a3) tree -> ('a1, 'a4) tree **)
  
  let rec map2_opt key_OT f mapl mapr m1 m2 =
    match m1 with
    | Leaf -> mapr m2
    | Node (l1, x1, d1, r1, h1) ->
      (match m2 with
       | Leaf -> mapl m1
       | Node (t0, k, y, t1, z) ->
         let { t_left = l2'; t_opt = o2; t_right = r2' } = split key_OT x1 m2
         in
         (match f x1 d1 o2 with
          | Some e ->
            join key_OT (map2_opt key_OT f mapl mapr l1 l2') x1 e
              (map2_opt key_OT f mapl mapr r1 r2')
          | None ->
            concat key_OT (map2_opt key_OT f mapl mapr l1 l2')
              (map2_opt key_OT f mapl mapr r1 r2')))
  
  type ('key, 'elt, 'elt') coq_R_map_option =
  | R_map_option_0 of ('key, 'elt) tree
  | R_map_option_1 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 
     'elt * ('key, 'elt) tree * int * 'elt' * ('key, 'elt') tree
     * ('key, 'elt, 'elt') coq_R_map_option * ('key, 'elt') tree
     * ('key, 'elt, 'elt') coq_R_map_option
  | R_map_option_2 of ('key, 'elt) tree * ('key, 'elt) tree * 'key * 
     'elt * ('key, 'elt) tree * int * ('key, 'elt') tree
     * ('key, 'elt, 'elt') coq_R_map_option * ('key, 'elt') tree
     * ('key, 'elt, 'elt') coq_R_map_option
  
  (** val coq_R_map_option_rect :
      'a1 coq_OrderedType -> ('a1 -> 'a2 -> 'a3 option) -> (('a1, 'a2) tree
      -> __ -> 'a4) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> int -> __ -> 'a3 -> __ -> ('a1, 'a3) tree -> ('a1,
      'a2, 'a3) coq_R_map_option -> 'a4 -> ('a1, 'a3) tree -> ('a1, 'a2, 'a3)
      coq_R_map_option -> 'a4 -> 'a4) -> (('a1, 'a2) tree -> ('a1, 'a2) tree
      -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a3) tree
      -> ('a1, 'a2, 'a3) coq_R_map_option -> 'a4 -> ('a1, 'a3) tree -> ('a1,
      'a2, 'a3) coq_R_map_option -> 'a4 -> 'a4) -> ('a1, 'a2) tree -> ('a1,
      'a3) tree -> ('a1, 'a2, 'a3) coq_R_map_option -> 'a4 **)
  
  let rec coq_R_map_option_rect key_OT f f0 f1 f2 m t0 = function
  | R_map_option_0 m0 -> f0 m0 __
  | R_map_option_1 (m0, l, x, d, r0, _x, d', res0, r1, res, r2) ->
    f1 m0 l x d r0 _x __ d' __ res0 r1
      (coq_R_map_option_rect key_OT f f0 f1 f2 l res0 r1) res r2
      (coq_R_map_option_rect key_OT f f0 f1 f2 r0 res r2)
  | R_map_option_2 (m0, l, x, d, r0, _x, res0, r1, res, r2) ->
    f2 m0 l x d r0 _x __ __ res0 r1
      (coq_R_map_option_rect key_OT f f0 f1 f2 l res0 r1) res r2
      (coq_R_map_option_rect key_OT f f0 f1 f2 r0 res r2)
  
  (** val coq_R_map_option_rec :
      'a1 coq_OrderedType -> ('a1 -> 'a2 -> 'a3 option) -> (('a1, 'a2) tree
      -> __ -> 'a4) -> (('a1, 'a2) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> int -> __ -> 'a3 -> __ -> ('a1, 'a3) tree -> ('a1,
      'a2, 'a3) coq_R_map_option -> 'a4 -> ('a1, 'a3) tree -> ('a1, 'a2, 'a3)
      coq_R_map_option -> 'a4 -> 'a4) -> (('a1, 'a2) tree -> ('a1, 'a2) tree
      -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __ -> __ -> ('a1, 'a3) tree
      -> ('a1, 'a2, 'a3) coq_R_map_option -> 'a4 -> ('a1, 'a3) tree -> ('a1,
      'a2, 'a3) coq_R_map_option -> 'a4 -> 'a4) -> ('a1, 'a2) tree -> ('a1,
      'a3) tree -> ('a1, 'a2, 'a3) coq_R_map_option -> 'a4 **)
  
  let rec coq_R_map_option_rec key_OT f f0 f1 f2 m t0 = function
  | R_map_option_0 m0 -> f0 m0 __
  | R_map_option_1 (m0, l, x, d, r0, _x, d', res0, r1, res, r2) ->
    f1 m0 l x d r0 _x __ d' __ res0 r1
      (coq_R_map_option_rec key_OT f f0 f1 f2 l res0 r1) res r2
      (coq_R_map_option_rec key_OT f f0 f1 f2 r0 res r2)
  | R_map_option_2 (m0, l, x, d, r0, _x, res0, r1, res, r2) ->
    f2 m0 l x d r0 _x __ __ res0 r1
      (coq_R_map_option_rec key_OT f f0 f1 f2 l res0 r1) res r2
      (coq_R_map_option_rec key_OT f f0 f1 f2 r0 res r2)
  
  type ('key, 'elt, 'elt', 'elt'') coq_R_map2_opt =
  | R_map2_opt_0 of ('key, 'elt) tree * ('key, 'elt') tree
  | R_map2_opt_1 of ('key, 'elt) tree * ('key, 'elt') tree
     * ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree * int
  | R_map2_opt_2 of ('key, 'elt) tree * ('key, 'elt') tree
     * ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree * int
     * ('key, 'elt') tree * 'key * 'elt' * ('key, 'elt') tree * int
     * ('key, 'elt') tree * 'elt' option * ('key, 'elt') tree * 'elt''
     * ('key, 'elt'') tree * ('key, 'elt, 'elt', 'elt'') coq_R_map2_opt
     * ('key, 'elt'') tree * ('key, 'elt, 'elt', 'elt'') coq_R_map2_opt
  | R_map2_opt_3 of ('key, 'elt) tree * ('key, 'elt') tree
     * ('key, 'elt) tree * 'key * 'elt * ('key, 'elt) tree * int
     * ('key, 'elt') tree * 'key * 'elt' * ('key, 'elt') tree * int
     * ('key, 'elt') tree * 'elt' option * ('key, 'elt') tree
     * ('key, 'elt'') tree * ('key, 'elt, 'elt', 'elt'') coq_R_map2_opt
     * ('key, 'elt'') tree * ('key, 'elt, 'elt', 'elt'') coq_R_map2_opt
  
  (** val coq_R_map2_opt_rect :
      'a1 coq_OrderedType -> ('a1 -> 'a2 -> 'a3 option -> 'a4 option) ->
      (('a1, 'a2) tree -> ('a1, 'a4) tree) -> (('a1, 'a3) tree -> ('a1, 'a4)
      tree) -> (('a1, 'a2) tree -> ('a1, 'a3) tree -> __ -> 'a5) -> (('a1,
      'a2) tree -> ('a1, 'a3) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1,
      'a2) tree -> int -> __ -> __ -> 'a5) -> (('a1, 'a2) tree -> ('a1, 'a3)
      tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __
      -> ('a1, 'a3) tree -> 'a1 -> 'a3 -> ('a1, 'a3) tree -> int -> __ ->
      ('a1, 'a3) tree -> 'a3 option -> ('a1, 'a3) tree -> __ -> 'a4 -> __ ->
      ('a1, 'a4) tree -> ('a1, 'a2, 'a3, 'a4) coq_R_map2_opt -> 'a5 -> ('a1,
      'a4) tree -> ('a1, 'a2, 'a3, 'a4) coq_R_map2_opt -> 'a5 -> 'a5) ->
      (('a1, 'a2) tree -> ('a1, 'a3) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> int -> __ -> ('a1, 'a3) tree -> 'a1 -> 'a3 -> ('a1,
      'a3) tree -> int -> __ -> ('a1, 'a3) tree -> 'a3 option -> ('a1, 'a3)
      tree -> __ -> __ -> ('a1, 'a4) tree -> ('a1, 'a2, 'a3, 'a4)
      coq_R_map2_opt -> 'a5 -> ('a1, 'a4) tree -> ('a1, 'a2, 'a3, 'a4)
      coq_R_map2_opt -> 'a5 -> 'a5) -> ('a1, 'a2) tree -> ('a1, 'a3) tree ->
      ('a1, 'a4) tree -> ('a1, 'a2, 'a3, 'a4) coq_R_map2_opt -> 'a5 **)
  
  let rec coq_R_map2_opt_rect key_OT f mapl mapr f0 f1 f2 f3 m1 m2 t0 = function
  | R_map2_opt_0 (m3, m4) -> f0 m3 m4 __
  | R_map2_opt_1 (m3, m4, l1, x1, d1, r1, _x) ->
    f1 m3 m4 l1 x1 d1 r1 _x __ __
  | R_map2_opt_2 (m3, m4, l1, x1, d1, r1, _x, _x0, _x1, _x2, _x3, _x4, l2',
                  o2, r2', e, res0, r0, res, r2) ->
    f2 m3 m4 l1 x1 d1 r1 _x __ _x0 _x1 _x2 _x3 _x4 __ l2' o2 r2' __ e __ res0
      r0 (coq_R_map2_opt_rect key_OT f mapl mapr f0 f1 f2 f3 l1 l2' res0 r0)
      res r2
      (coq_R_map2_opt_rect key_OT f mapl mapr f0 f1 f2 f3 r1 r2' res r2)
  | R_map2_opt_3 (m3, m4, l1, x1, d1, r1, _x, _x0, _x1, _x2, _x3, _x4, l2',
                  o2, r2', res0, r0, res, r2) ->
    f3 m3 m4 l1 x1 d1 r1 _x __ _x0 _x1 _x2 _x3 _x4 __ l2' o2 r2' __ __ res0
      r0 (coq_R_map2_opt_rect key_OT f mapl mapr f0 f1 f2 f3 l1 l2' res0 r0)
      res r2
      (coq_R_map2_opt_rect key_OT f mapl mapr f0 f1 f2 f3 r1 r2' res r2)
  
  (** val coq_R_map2_opt_rec :
      'a1 coq_OrderedType -> ('a1 -> 'a2 -> 'a3 option -> 'a4 option) ->
      (('a1, 'a2) tree -> ('a1, 'a4) tree) -> (('a1, 'a3) tree -> ('a1, 'a4)
      tree) -> (('a1, 'a2) tree -> ('a1, 'a3) tree -> __ -> 'a5) -> (('a1,
      'a2) tree -> ('a1, 'a3) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1,
      'a2) tree -> int -> __ -> __ -> 'a5) -> (('a1, 'a2) tree -> ('a1, 'a3)
      tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 -> ('a1, 'a2) tree -> int -> __
      -> ('a1, 'a3) tree -> 'a1 -> 'a3 -> ('a1, 'a3) tree -> int -> __ ->
      ('a1, 'a3) tree -> 'a3 option -> ('a1, 'a3) tree -> __ -> 'a4 -> __ ->
      ('a1, 'a4) tree -> ('a1, 'a2, 'a3, 'a4) coq_R_map2_opt -> 'a5 -> ('a1,
      'a4) tree -> ('a1, 'a2, 'a3, 'a4) coq_R_map2_opt -> 'a5 -> 'a5) ->
      (('a1, 'a2) tree -> ('a1, 'a3) tree -> ('a1, 'a2) tree -> 'a1 -> 'a2 ->
      ('a1, 'a2) tree -> int -> __ -> ('a1, 'a3) tree -> 'a1 -> 'a3 -> ('a1,
      'a3) tree -> int -> __ -> ('a1, 'a3) tree -> 'a3 option -> ('a1, 'a3)
      tree -> __ -> __ -> ('a1, 'a4) tree -> ('a1, 'a2, 'a3, 'a4)
      coq_R_map2_opt -> 'a5 -> ('a1, 'a4) tree -> ('a1, 'a2, 'a3, 'a4)
      coq_R_map2_opt -> 'a5 -> 'a5) -> ('a1, 'a2) tree -> ('a1, 'a3) tree ->
      ('a1, 'a4) tree -> ('a1, 'a2, 'a3, 'a4) coq_R_map2_opt -> 'a5 **)
  
  let rec coq_R_map2_opt_rec key_OT f mapl mapr f0 f1 f2 f3 m1 m2 t0 = function
  | R_map2_opt_0 (m3, m4) -> f0 m3 m4 __
  | R_map2_opt_1 (m3, m4, l1, x1, d1, r1, _x) ->
    f1 m3 m4 l1 x1 d1 r1 _x __ __
  | R_map2_opt_2 (m3, m4, l1, x1, d1, r1, _x, _x0, _x1, _x2, _x3, _x4, l2',
                  o2, r2', e, res0, r0, res, r2) ->
    f2 m3 m4 l1 x1 d1 r1 _x __ _x0 _x1 _x2 _x3 _x4 __ l2' o2 r2' __ e __ res0
      r0 (coq_R_map2_opt_rec key_OT f mapl mapr f0 f1 f2 f3 l1 l2' res0 r0)
      res r2
      (coq_R_map2_opt_rec key_OT f mapl mapr f0 f1 f2 f3 r1 r2' res r2)
  | R_map2_opt_3 (m3, m4, l1, x1, d1, r1, _x, _x0, _x1, _x2, _x3, _x4, l2',
                  o2, r2', res0, r0, res, r2) ->
    f3 m3 m4 l1 x1 d1 r1 _x __ _x0 _x1 _x2 _x3 _x4 __ l2' o2 r2' __ __ res0
      r0 (coq_R_map2_opt_rec key_OT f mapl mapr f0 f1 f2 f3 l1 l2' res0 r0)
      res r2
      (coq_R_map2_opt_rec key_OT f mapl mapr f0 f1 f2 f3 r1 r2' res r2)
  
  (** val map2 :
      'a1 coq_OrderedType -> ('a2 option -> 'a3 option -> 'a4 option) ->
      ('a1, 'a2) tree -> ('a1, 'a3) tree -> ('a1, 'a4) tree **)
  
  let map2 key_OT f =
    map2_opt key_OT (fun x d o -> f (Some d) o)
      (map_option key_OT (fun x d -> f (Some d) None))
      (map_option key_OT (fun x d' -> f None (Some d')))
  
  module Proofs = 
   struct 
    module PX = KeyOrderedType
    
    module L = MapList.MapList
    
    (** val fold' :
        'a1 coq_OrderedType -> ('a1 -> 'a2 -> 'a3 -> 'a3) -> ('a1, 'a2) tree
        -> 'a3 -> 'a3 **)
    
    let fold' key_OT f s =
      L.fold key_OT f (elements s)
    
    (** val flatten_e : ('a1, 'a2) enumeration -> ('a1 * 'a2) list **)
    
    let rec flatten_e = function
    | End -> []
    | More (x, e0, t0, r) -> (x, e0) :: (app (elements t0) (flatten_e r))
   end
 end

module Raw = MapAVL

type ('key, 'elt) dict =
  ('key, 'elt) Raw.tree
  (* singleton inductive, whose constructor was Bst *)

(** val this :
    'a1 coq_OrderedType -> ('a1, 'a2) dict -> ('a1, 'a2) Raw.tree **)

let this key_OT d =
  d

type ('key, 'elt) t = ('key, 'elt) dict

(** val empty : 'a1 coq_OrderedType -> ('a1, 'a2) t **)

let empty key_OT =
  Raw.empty

(** val is_empty : 'a1 coq_OrderedType -> ('a1, 'a2) t -> bool **)

let is_empty key_OT m =
  Raw.is_empty (this key_OT m)

(** val add :
    'a1 coq_OrderedType -> 'a1 -> 'a2 -> ('a1, 'a2) t -> ('a1, 'a2) t **)

let add key_OT x e m =
  Raw.add key_OT x e (this key_OT m)

(** val insert :
    'a1 coq_OrderedType -> 'a1 -> 'a2 -> ('a2 -> 'a2) -> ('a1, 'a2) t ->
    ('a1, 'a2) t **)

let insert key_OT x d f m =
  Raw.insert key_OT x d f (this key_OT m)

(** val adjust :
    'a1 coq_OrderedType -> 'a1 -> ('a2 -> 'a2) -> ('a1, 'a2) t -> ('a1, 'a2)
    t **)

let adjust key_OT x f m =
  Raw.adjust key_OT x f (this key_OT m)

(** val remove :
    'a1 coq_OrderedType -> 'a1 -> ('a1, 'a2) t -> ('a1, 'a2) t **)

let remove key_OT x m =
  Raw.remove key_OT x (this key_OT m)

(** val mem : 'a1 coq_OrderedType -> 'a1 -> ('a1, 'a2) t -> bool **)

let mem key_OT x m =
  Raw.mem key_OT x (this key_OT m)

(** val find : 'a1 coq_OrderedType -> 'a1 -> ('a1, 'a2) t -> 'a2 option **)

let find key_OT x m =
  Raw.find key_OT x (this key_OT m)

(** val map :
    'a1 coq_OrderedType -> ('a2 -> 'a3) -> ('a1, 'a2) t -> ('a1, 'a3) t **)

let map key_OT f m =
  Raw.map f (this key_OT m)

(** val mapi :
    'a1 coq_OrderedType -> ('a1 -> 'a2 -> 'a3) -> ('a1, 'a2) t -> ('a1, 'a3)
    t **)

let mapi key_OT f m =
  Raw.mapi f (this key_OT m)

(** val map2 :
    'a1 coq_OrderedType -> ('a2 option -> 'a3 option -> 'a4 option) -> ('a1,
    'a2) t -> ('a1, 'a3) t -> ('a1, 'a4) t **)

let map2 key_OT f m m' =
  Raw.map2 key_OT f (this key_OT m) (this key_OT m')

(** val elements :
    'a1 coq_OrderedType -> ('a1, 'a2) t -> ('a1 * 'a2) list **)

let elements key_OT m =
  Raw.elements (this key_OT m)

(** val cardinal : 'a1 coq_OrderedType -> ('a1, 'a2) t -> int **)

let cardinal key_OT m =
  Raw.cardinal (this key_OT m)

(** val fold :
    'a1 coq_OrderedType -> ('a1 -> 'a2 -> 'a3 -> 'a3) -> ('a1, 'a2) t -> 'a3
    -> 'a3 **)

let fold key_OT f m i =
  Raw.fold f (this key_OT m) i

(** val equal :
    'a1 coq_OrderedType -> ('a2 -> 'a2 -> bool) -> ('a1, 'a2) t -> ('a1, 'a2)
    t -> bool **)

let equal key_OT cmp m m' =
  Raw.equal key_OT cmp (this key_OT m) (this key_OT m')

(** val compare_more :
    'a1 coq_OrderedType -> 'a2 coq_OrderedType -> 'a1 -> 'a2 -> (('a1, 'a2)
    Raw.enumeration -> comparison) -> ('a1, 'a2) Raw.enumeration ->
    comparison **)

let compare_more key_OT elt_OT x1 d1 cont = function
| Raw.End -> Gt
| Raw.More (x2, d2, r2, e3) ->
  (match compare key_OT x1 x2 with
   | Eq ->
     (match compare elt_OT d1 d2 with
      | Eq -> cont (Raw.cons r2 e3)
      | x -> x)
   | x -> x)

(** val compare_cont :
    'a1 coq_OrderedType -> 'a2 coq_OrderedType -> ('a1, 'a2) Raw.tree ->
    (('a1, 'a2) Raw.enumeration -> comparison) -> ('a1, 'a2) Raw.enumeration
    -> comparison **)

let rec compare_cont key_OT elt_OT s1 cont e2 =
  match s1 with
  | Raw.Leaf -> cont e2
  | Raw.Node (l1, x1, d1, r1, z) ->
    compare_cont key_OT elt_OT l1
      (compare_more key_OT elt_OT x1 d1 (compare_cont key_OT elt_OT r1 cont))
      e2

(** val compare_end : ('a1, 'a2) Raw.enumeration -> comparison **)

let compare_end = function
| Raw.End -> Eq
| Raw.More (k, e, t0, e0) -> Lt

(** val compare_pure :
    'a1 coq_OrderedType -> 'a2 coq_OrderedType -> ('a1, 'a2) Raw.tree ->
    ('a1, 'a2) Raw.tree -> comparison **)

let compare_pure key_OT elt_OT s1 s2 =
  compare_cont key_OT elt_OT s1 compare_end (Raw.cons s2 Raw.End)

(** val map_OrderedType :
    'a1 coq_OrderedType -> 'a2 coq_OrderedType -> ('a1, 'a2) dict
    coq_OrderedType **)

let map_OrderedType key_OT elt_OT s1 s2 =
  compare_pure key_OT elt_OT (this key_OT s1) (this key_OT s2)

