open BinInt
open Datatypes
open OrderedType
open Peano
open ZArith_dec

type __ = Obj.t
let __ = let rec f _ = Obj.repr f in Obj.repr f

module SetAVL = 
 struct 
  type 'elt tree =
  | Leaf
  | Node of 'elt tree * 'elt * 'elt tree * int
  
  (** val tree_rect :
      'a1 coq_OrderedType -> 'a2 -> ('a1 tree -> 'a2 -> 'a1 -> 'a1 tree ->
      'a2 -> int -> 'a2) -> 'a1 tree -> 'a2 **)
  
  let rec tree_rect h f f0 = function
  | Leaf -> f
  | Node (t0, y, t1, z) ->
    f0 t0 (tree_rect h f f0 t0) y t1 (tree_rect h f f0 t1) z
  
  (** val tree_rec :
      'a1 coq_OrderedType -> 'a2 -> ('a1 tree -> 'a2 -> 'a1 -> 'a1 tree ->
      'a2 -> int -> 'a2) -> 'a1 tree -> 'a2 **)
  
  let rec tree_rec h f f0 = function
  | Leaf -> f
  | Node (t0, y, t1, z) ->
    f0 t0 (tree_rec h f f0 t0) y t1 (tree_rec h f f0 t1) z
  
  (** val height : 'a1 coq_OrderedType -> 'a1 tree -> int **)
  
  let height h = function
  | Leaf -> 0
  | Node (t, e, t0, h0) -> h0
  
  (** val cardinal : 'a1 coq_OrderedType -> 'a1 tree -> int **)
  
  let rec cardinal h = function
  | Leaf -> 0
  | Node (l, e, r, z) -> Pervasives.succ (plus (cardinal h l) (cardinal h r))
  
  (** val empty : 'a1 coq_OrderedType -> 'a1 tree **)
  
  let empty h =
    Leaf
  
  (** val is_empty : 'a1 coq_OrderedType -> 'a1 tree -> bool **)
  
  let is_empty h = function
  | Leaf -> true
  | Node (t, y, t0, z) -> false
  
  (** val mem : 'a1 coq_OrderedType -> 'a1 -> 'a1 tree -> bool **)
  
  let rec mem h x = function
  | Leaf -> false
  | Node (l, y, r, z) ->
    (match compare h x y with
     | Eq -> true
     | Lt -> mem h x l
     | Gt -> mem h x r)
  
  (** val singleton : 'a1 coq_OrderedType -> 'a1 -> 'a1 tree **)
  
  let singleton h x =
    Node (Leaf, x, Leaf, 1)
  
  (** val create :
      'a1 coq_OrderedType -> 'a1 tree -> 'a1 -> 'a1 tree -> 'a1 tree **)
  
  let create h l x r =
    Node (l, x, r, (Z.add (Z.max (height h l) (height h r)) 1))
  
  (** val assert_false :
      'a1 coq_OrderedType -> 'a1 tree -> 'a1 -> 'a1 tree -> 'a1 tree **)
  
  let assert_false h =
    create h
  
  (** val bal :
      'a1 coq_OrderedType -> 'a1 tree -> 'a1 -> 'a1 tree -> 'a1 tree **)
  
  let bal h l x r =
    let hl = height h l in
    let hr = height h r in
    if coq_Z_gt_le_dec hl (Z.add hr ((fun p->2*p) 1))
    then (match l with
          | Leaf -> assert_false h l x r
          | Node (ll, lx, lr, z) ->
            if coq_Z_ge_lt_dec (height h ll) (height h lr)
            then create h ll lx (create h lr x r)
            else (match lr with
                  | Leaf -> assert_false h l x r
                  | Node (lrl, lrx, lrr, z0) ->
                    create h (create h ll lx lrl) lrx (create h lrr x r)))
    else if coq_Z_gt_le_dec hr (Z.add hl ((fun p->2*p) 1))
         then (match r with
               | Leaf -> assert_false h l x r
               | Node (rl, rx, rr, z) ->
                 if coq_Z_ge_lt_dec (height h rr) (height h rl)
                 then create h (create h l x rl) rx rr
                 else (match rl with
                       | Leaf -> assert_false h l x r
                       | Node (rll, rlx, rlr, z0) ->
                         create h (create h l x rll) rlx (create h rlr rx rr)))
         else create h l x r
  
  (** val add : 'a1 coq_OrderedType -> 'a1 -> 'a1 tree -> 'a1 tree **)
  
  let rec add h x = function
  | Leaf -> Node (Leaf, x, Leaf, 1)
  | Node (l, y, r, h0) ->
    (match compare h x y with
     | Eq -> Node (l, y, r, h0)
     | Lt -> bal h (add h x l) y r
     | Gt -> bal h l y (add h x r))
  
  (** val join :
      'a1 coq_OrderedType -> 'a1 tree -> 'a1 -> 'a1 tree -> 'a1 tree **)
  
  let rec join h l = match l with
  | Leaf -> add h
  | Node (ll, lx, lr, lh) ->
    (fun x ->
      let rec join_aux r = match r with
      | Leaf -> add h x l
      | Node (rl, rx, rr, rh) ->
        if coq_Z_gt_le_dec lh (Z.add rh ((fun p->2*p) 1))
        then bal h ll lx (join h lr x r)
        else if coq_Z_gt_le_dec rh (Z.add lh ((fun p->2*p) 1))
             then bal h (join_aux rl) rx rr
             else create h l x r
      in join_aux)
  
  (** val remove_min :
      'a1 coq_OrderedType -> 'a1 tree -> 'a1 -> 'a1 tree -> 'a1 tree * 'a1 **)
  
  let rec remove_min h l x r =
    match l with
    | Leaf -> (r, x)
    | Node (ll, lx, lr, lh) ->
      let (l', m) = remove_min h ll lx lr in ((bal h l' x r), m)
  
  (** val merge : 'a1 coq_OrderedType -> 'a1 tree -> 'a1 tree -> 'a1 tree **)
  
  let merge h s1 s2 =
    match s1 with
    | Leaf -> s2
    | Node (t, y, t0, z) ->
      (match s2 with
       | Leaf -> s1
       | Node (l2, x2, r2, h2) ->
         let (s2', m) = remove_min h l2 x2 r2 in bal h s1 m s2')
  
  (** val remove : 'a1 coq_OrderedType -> 'a1 -> 'a1 tree -> 'a1 tree **)
  
  let rec remove h x = function
  | Leaf -> Leaf
  | Node (l, y, r, h0) ->
    (match compare h x y with
     | Eq -> merge h l r
     | Lt -> bal h (remove h x l) y r
     | Gt -> bal h l y (remove h x r))
  
  (** val min_elt : 'a1 coq_OrderedType -> 'a1 tree -> 'a1 option **)
  
  let rec min_elt h = function
  | Leaf -> None
  | Node (l, y, t, z) ->
    (match l with
     | Leaf -> Some y
     | Node (t0, y0, t1, z0) -> min_elt h l)
  
  (** val max_elt : 'a1 coq_OrderedType -> 'a1 tree -> 'a1 option **)
  
  let rec max_elt h = function
  | Leaf -> None
  | Node (t, y, r, z) ->
    (match r with
     | Leaf -> Some y
     | Node (t0, y0, t1, z0) -> max_elt h r)
  
  (** val choose : 'a1 coq_OrderedType -> 'a1 tree -> 'a1 option **)
  
  let choose h =
    min_elt h
  
  (** val concat :
      'a1 coq_OrderedType -> 'a1 tree -> 'a1 tree -> 'a1 tree **)
  
  let concat h s1 s2 =
    match s1 with
    | Leaf -> s2
    | Node (t, y, t0, z) ->
      (match s2 with
       | Leaf -> s1
       | Node (l2, x2, r2, z0) ->
         let (s2', m) = remove_min h l2 x2 r2 in join h s1 m s2')
  
  type 'elt triple = { t_left : 'elt tree; t_in : bool; t_right : 'elt tree }
  
  (** val triple_rect :
      'a1 coq_OrderedType -> ('a1 tree -> bool -> 'a1 tree -> 'a2) -> 'a1
      triple -> 'a2 **)
  
  let triple_rect h f t =
    let { t_left = x; t_in = x0; t_right = x1 } = t in f x x0 x1
  
  (** val triple_rec :
      'a1 coq_OrderedType -> ('a1 tree -> bool -> 'a1 tree -> 'a2) -> 'a1
      triple -> 'a2 **)
  
  let triple_rec h f t =
    let { t_left = x; t_in = x0; t_right = x1 } = t in f x x0 x1
  
  (** val t_left : 'a1 coq_OrderedType -> 'a1 triple -> 'a1 tree **)
  
  let t_left _ x = x.t_left
  
  (** val t_in : 'a1 coq_OrderedType -> 'a1 triple -> bool **)
  
  let t_in _ x = x.t_in
  
  (** val t_right : 'a1 coq_OrderedType -> 'a1 triple -> 'a1 tree **)
  
  let t_right _ x = x.t_right
  
  (** val split : 'a1 coq_OrderedType -> 'a1 -> 'a1 tree -> 'a1 triple **)
  
  let rec split h x = function
  | Leaf -> { t_left = Leaf; t_in = false; t_right = Leaf }
  | Node (l, y, r, h0) ->
    (match compare h x y with
     | Eq -> { t_left = l; t_in = true; t_right = r }
     | Lt ->
       let { t_left = ll; t_in = b; t_right = rl } = split h x l in
       { t_left = ll; t_in = b; t_right = (join h rl y r) }
     | Gt ->
       let { t_left = rl; t_in = b; t_right = rr } = split h x r in
       { t_left = (join h l y rl); t_in = b; t_right = rr })
  
  (** val inter : 'a1 coq_OrderedType -> 'a1 tree -> 'a1 tree -> 'a1 tree **)
  
  let rec inter h s1 s2 =
    match s1 with
    | Leaf -> Leaf
    | Node (l1, x1, r1, h1) ->
      (match s2 with
       | Leaf -> Leaf
       | Node (t, y, t0, z) ->
         let { t_left = l2'; t_in = pres; t_right = r2' } = split h x1 s2 in
         if pres
         then join h (inter h l1 l2') x1 (inter h r1 r2')
         else concat h (inter h l1 l2') (inter h r1 r2'))
  
  (** val diff : 'a1 coq_OrderedType -> 'a1 tree -> 'a1 tree -> 'a1 tree **)
  
  let rec diff h s1 s2 =
    match s1 with
    | Leaf -> Leaf
    | Node (l1, x1, r1, h1) ->
      (match s2 with
       | Leaf -> s1
       | Node (t, y, t0, z) ->
         let { t_left = l2'; t_in = pres; t_right = r2' } = split h x1 s2 in
         if pres
         then concat h (diff h l1 l2') (diff h r1 r2')
         else join h (diff h l1 l2') x1 (diff h r1 r2'))
  
  (** val union : 'a1 coq_OrderedType -> 'a1 tree -> 'a1 tree -> 'a1 tree **)
  
  let rec union h s1 s2 =
    match s1 with
    | Leaf -> s2
    | Node (l1, x1, r1, h1) ->
      (match s2 with
       | Leaf -> s1
       | Node (t, y, t0, z) ->
         let { t_left = l2'; t_in = x; t_right = r2' } = split h x1 s2 in
         join h (union h l1 l2') x1 (union h r1 r2'))
  
  (** val elements_aux :
      'a1 coq_OrderedType -> 'a1 list -> 'a1 tree -> 'a1 list **)
  
  let rec elements_aux h acc = function
  | Leaf -> acc
  | Node (l, x, r, z) -> elements_aux h (x :: (elements_aux h acc r)) l
  
  (** val elements : 'a1 coq_OrderedType -> 'a1 tree -> 'a1 list **)
  
  let elements h =
    elements_aux h []
  
  (** val filter_acc :
      'a1 coq_OrderedType -> ('a1 -> bool) -> 'a1 tree -> 'a1 tree -> 'a1
      tree **)
  
  let rec filter_acc h f acc = function
  | Leaf -> acc
  | Node (l, x, r, h0) ->
    filter_acc h f (filter_acc h f (if f x then add h x acc else acc) l) r
  
  (** val filter :
      'a1 coq_OrderedType -> ('a1 -> bool) -> 'a1 tree -> 'a1 tree **)
  
  let filter h f =
    filter_acc h f Leaf
  
  (** val partition_acc :
      'a1 coq_OrderedType -> ('a1 -> bool) -> ('a1 tree * 'a1 tree) -> 'a1
      tree -> 'a1 tree * 'a1 tree **)
  
  let rec partition_acc h f acc = function
  | Leaf -> acc
  | Node (l, x, r, z) ->
    let (acct, accf) = acc in
    partition_acc h f
      (partition_acc h f
        (if f x then ((add h x acct), accf) else (acct, (add h x accf))) l) r
  
  (** val partition :
      'a1 coq_OrderedType -> ('a1 -> bool) -> 'a1 tree -> 'a1 tree * 'a1 tree **)
  
  let partition h f =
    partition_acc h f (Leaf, Leaf)
  
  (** val for_all :
      'a1 coq_OrderedType -> ('a1 -> bool) -> 'a1 tree -> bool **)
  
  let rec for_all h f = function
  | Leaf -> true
  | Node (l, x, r, z) ->
    if if f x then for_all h f l else false then for_all h f r else false
  
  (** val exists_ :
      'a1 coq_OrderedType -> ('a1 -> bool) -> 'a1 tree -> bool **)
  
  let rec exists_ h f = function
  | Leaf -> false
  | Node (l, x, r, z) ->
    if if f x then true else exists_ h f l then true else exists_ h f r
  
  (** val map_monotonic :
      'a1 coq_OrderedType -> 'a2 coq_OrderedType -> ('a1 -> 'a2) -> 'a1 tree
      -> 'a2 tree **)
  
  let rec map_monotonic h hB f = function
  | Leaf -> Leaf
  | Node (l, x, r, h0) ->
    Node ((map_monotonic h hB f l), (f x), (map_monotonic h hB f r), h0)
  
  (** val fold :
      'a1 coq_OrderedType -> ('a1 -> 'a2 -> 'a2) -> 'a1 tree -> 'a2 -> 'a2 **)
  
  let rec fold h f s x =
    match s with
    | Leaf -> x
    | Node (l, x0, r, z) -> fold h f r (f x0 (fold h f l x))
  
  (** val subsetl :
      'a1 coq_OrderedType -> ('a1 tree -> bool) -> 'a1 -> 'a1 tree -> bool **)
  
  let rec subsetl h subset_l1 x1 s2 = match s2 with
  | Leaf -> false
  | Node (l2, x2, r2, h2) ->
    (match compare h x1 x2 with
     | Eq -> subset_l1 l2
     | Lt -> subsetl h subset_l1 x1 l2
     | Gt -> if mem h x1 r2 then subset_l1 s2 else false)
  
  (** val subsetr :
      'a1 coq_OrderedType -> ('a1 tree -> bool) -> 'a1 -> 'a1 tree -> bool **)
  
  let rec subsetr h subset_r1 x1 s2 = match s2 with
  | Leaf -> false
  | Node (l2, x2, r2, h2) ->
    (match compare h x1 x2 with
     | Eq -> subset_r1 r2
     | Lt -> if mem h x1 l2 then subset_r1 s2 else false
     | Gt -> subsetr h subset_r1 x1 r2)
  
  (** val subset : 'a1 coq_OrderedType -> 'a1 tree -> 'a1 tree -> bool **)
  
  let rec subset h s1 s2 =
    match s1 with
    | Leaf -> true
    | Node (l1, x1, r1, h1) ->
      (match s2 with
       | Leaf -> false
       | Node (l2, x2, r2, h2) ->
         (match compare h x1 x2 with
          | Eq -> if subset h l1 l2 then subset h r1 r2 else false
          | Lt ->
            if subsetl h (subset h l1) x1 l2 then subset h r1 s2 else false
          | Gt ->
            if subsetr h (subset h r1) x1 r2 then subset h l1 s2 else false))
  
  type 'elt enumeration =
  | End
  | More of 'elt * 'elt tree * 'elt enumeration
  
  (** val enumeration_rect :
      'a1 coq_OrderedType -> 'a2 -> ('a1 -> 'a1 tree -> 'a1 enumeration ->
      'a2 -> 'a2) -> 'a1 enumeration -> 'a2 **)
  
  let rec enumeration_rect h f f0 = function
  | End -> f
  | More (e0, t, e1) -> f0 e0 t e1 (enumeration_rect h f f0 e1)
  
  (** val enumeration_rec :
      'a1 coq_OrderedType -> 'a2 -> ('a1 -> 'a1 tree -> 'a1 enumeration ->
      'a2 -> 'a2) -> 'a1 enumeration -> 'a2 **)
  
  let rec enumeration_rec h f f0 = function
  | End -> f
  | More (e0, t, e1) -> f0 e0 t e1 (enumeration_rec h f f0 e1)
  
  (** val cons :
      'a1 coq_OrderedType -> 'a1 tree -> 'a1 enumeration -> 'a1 enumeration **)
  
  let rec cons h s e =
    match s with
    | Leaf -> e
    | Node (l, x, r, h0) -> cons h l (More (x, r, e))
  
  (** val compare_more :
      'a1 coq_OrderedType -> 'a1 -> ('a1 enumeration -> comparison) -> 'a1
      enumeration -> comparison **)
  
  let compare_more h x1 cont = function
  | End -> Gt
  | More (x2, r2, e3) ->
    (match compare h x1 x2 with
     | Eq -> cont (cons h r2 e3)
     | x -> x)
  
  (** val compare_cont :
      'a1 coq_OrderedType -> 'a1 tree -> ('a1 enumeration -> comparison) ->
      'a1 enumeration -> comparison **)
  
  let rec compare_cont h s1 cont e2 =
    match s1 with
    | Leaf -> cont e2
    | Node (l1, x1, r1, z) ->
      compare_cont h l1 (compare_more h x1 (compare_cont h r1 cont)) e2
  
  (** val compare_end :
      'a1 coq_OrderedType -> 'a1 enumeration -> comparison **)
  
  let compare_end h = function
  | End -> Eq
  | More (e, t, e0) -> Lt
  
  (** val compare :
      'a1 coq_OrderedType -> 'a1 tree -> 'a1 tree -> comparison **)
  
  let compare h s1 s2 =
    compare_cont h s1 (compare_end h) (cons h s2 End)
  
  (** val equal : 'a1 coq_OrderedType -> 'a1 tree -> 'a1 tree -> bool **)
  
  let equal h s1 s2 =
    match compare h s1 s2 with
    | Eq -> true
    | _ -> false
  
  type 'elt coq_R_mem =
  | R_mem_0 of 'elt tree
  | R_mem_1 of 'elt tree * 'elt tree * 'elt * 'elt tree * int
  | R_mem_2 of 'elt tree * 'elt tree * 'elt * 'elt tree * int * bool
     * 'elt coq_R_mem
  | R_mem_3 of 'elt tree * 'elt tree * 'elt * 'elt tree * int * bool
     * 'elt coq_R_mem
  
  (** val coq_R_mem_rect :
      'a1 coq_OrderedType -> 'a1 -> ('a1 tree -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> bool -> 'a1 coq_R_mem
      -> 'a2 -> 'a2) -> ('a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __
      -> __ -> bool -> 'a1 coq_R_mem -> 'a2 -> 'a2) -> 'a1 tree -> bool ->
      'a1 coq_R_mem -> 'a2 **)
  
  let rec coq_R_mem_rect h x f f0 f1 f2 s b = function
  | R_mem_0 s0 -> f s0 __
  | R_mem_1 (s0, l, y, r0, _x) -> f0 s0 l y r0 _x __ __
  | R_mem_2 (s0, l, y, r0, _x, res, r1) ->
    f1 s0 l y r0 _x __ __ res r1 (coq_R_mem_rect h x f f0 f1 f2 l res r1)
  | R_mem_3 (s0, l, y, r0, _x, res, r1) ->
    f2 s0 l y r0 _x __ __ res r1 (coq_R_mem_rect h x f f0 f1 f2 r0 res r1)
  
  (** val coq_R_mem_rec :
      'a1 coq_OrderedType -> 'a1 -> ('a1 tree -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> bool -> 'a1 coq_R_mem
      -> 'a2 -> 'a2) -> ('a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __
      -> __ -> bool -> 'a1 coq_R_mem -> 'a2 -> 'a2) -> 'a1 tree -> bool ->
      'a1 coq_R_mem -> 'a2 **)
  
  let rec coq_R_mem_rec h x f f0 f1 f2 s b = function
  | R_mem_0 s0 -> f s0 __
  | R_mem_1 (s0, l, y, r0, _x) -> f0 s0 l y r0 _x __ __
  | R_mem_2 (s0, l, y, r0, _x, res, r1) ->
    f1 s0 l y r0 _x __ __ res r1 (coq_R_mem_rec h x f f0 f1 f2 l res r1)
  | R_mem_3 (s0, l, y, r0, _x, res, r1) ->
    f2 s0 l y r0 _x __ __ res r1 (coq_R_mem_rec h x f f0 f1 f2 r0 res r1)
  
  type 'elt coq_R_bal =
  | R_bal_0 of 'elt tree * 'elt * 'elt tree
  | R_bal_1 of 'elt tree * 'elt * 'elt tree * 'elt tree * 'elt * 'elt tree
     * int
  | R_bal_2 of 'elt tree * 'elt * 'elt tree * 'elt tree * 'elt * 'elt tree
     * int
  | R_bal_3 of 'elt tree * 'elt * 'elt tree * 'elt tree * 'elt * 'elt tree
     * int * 'elt tree * 'elt * 'elt tree * int
  | R_bal_4 of 'elt tree * 'elt * 'elt tree
  | R_bal_5 of 'elt tree * 'elt * 'elt tree * 'elt tree * 'elt * 'elt tree
     * int
  | R_bal_6 of 'elt tree * 'elt * 'elt tree * 'elt tree * 'elt * 'elt tree
     * int
  | R_bal_7 of 'elt tree * 'elt * 'elt tree * 'elt tree * 'elt * 'elt tree
     * int * 'elt tree * 'elt * 'elt tree * int
  | R_bal_8 of 'elt tree * 'elt * 'elt tree
  
  (** val coq_R_bal_rect :
      'a1 coq_OrderedType -> ('a1 tree -> 'a1 -> 'a1 tree -> __ -> __ -> __
      -> 'a2) -> ('a1 tree -> 'a1 -> 'a1 tree -> __ -> __ -> 'a1 tree -> 'a1
      -> 'a1 tree -> int -> __ -> __ -> __ -> 'a2) -> ('a1 tree -> 'a1 -> 'a1
      tree -> __ -> __ -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ ->
      __ -> __ -> 'a2) -> ('a1 tree -> 'a1 -> 'a1 tree -> __ -> __ -> 'a1
      tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> __ -> 'a1 tree -> 'a1 ->
      'a1 tree -> int -> __ -> 'a2) -> ('a1 tree -> 'a1 -> 'a1 tree -> __ ->
      __ -> __ -> __ -> __ -> 'a2) -> ('a1 tree -> 'a1 -> 'a1 tree -> __ ->
      __ -> __ -> __ -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> __
      -> 'a2) -> ('a1 tree -> 'a1 -> 'a1 tree -> __ -> __ -> __ -> __ -> 'a1
      tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> __ -> __ -> 'a2) -> ('a1
      tree -> 'a1 -> 'a1 tree -> __ -> __ -> __ -> __ -> 'a1 tree -> 'a1 ->
      'a1 tree -> int -> __ -> __ -> __ -> 'a1 tree -> 'a1 -> 'a1 tree -> int
      -> __ -> 'a2) -> ('a1 tree -> 'a1 -> 'a1 tree -> __ -> __ -> __ -> __
      -> 'a2) -> 'a1 tree -> 'a1 -> 'a1 tree -> 'a1 tree -> 'a1 coq_R_bal ->
      'a2 **)
  
  let coq_R_bal_rect h f f0 f1 f2 f3 f4 f5 f6 f7 l x r t = function
  | R_bal_0 (x0, x1, x2) -> f x0 x1 x2 __ __ __
  | R_bal_1 (x0, x1, x2, x3, x4, x5, x6) ->
    f0 x0 x1 x2 __ __ x3 x4 x5 x6 __ __ __
  | R_bal_2 (x0, x1, x2, x3, x4, x5, x6) ->
    f1 x0 x1 x2 __ __ x3 x4 x5 x6 __ __ __ __
  | R_bal_3 (x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10) ->
    f2 x0 x1 x2 __ __ x3 x4 x5 x6 __ __ __ x7 x8 x9 x10 __
  | R_bal_4 (x0, x1, x2) -> f3 x0 x1 x2 __ __ __ __ __
  | R_bal_5 (x0, x1, x2, x3, x4, x5, x6) ->
    f4 x0 x1 x2 __ __ __ __ x3 x4 x5 x6 __ __ __
  | R_bal_6 (x0, x1, x2, x3, x4, x5, x6) ->
    f5 x0 x1 x2 __ __ __ __ x3 x4 x5 x6 __ __ __ __
  | R_bal_7 (x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10) ->
    f6 x0 x1 x2 __ __ __ __ x3 x4 x5 x6 __ __ __ x7 x8 x9 x10 __
  | R_bal_8 (x0, x1, x2) -> f7 x0 x1 x2 __ __ __ __
  
  (** val coq_R_bal_rec :
      'a1 coq_OrderedType -> ('a1 tree -> 'a1 -> 'a1 tree -> __ -> __ -> __
      -> 'a2) -> ('a1 tree -> 'a1 -> 'a1 tree -> __ -> __ -> 'a1 tree -> 'a1
      -> 'a1 tree -> int -> __ -> __ -> __ -> 'a2) -> ('a1 tree -> 'a1 -> 'a1
      tree -> __ -> __ -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ ->
      __ -> __ -> 'a2) -> ('a1 tree -> 'a1 -> 'a1 tree -> __ -> __ -> 'a1
      tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> __ -> 'a1 tree -> 'a1 ->
      'a1 tree -> int -> __ -> 'a2) -> ('a1 tree -> 'a1 -> 'a1 tree -> __ ->
      __ -> __ -> __ -> __ -> 'a2) -> ('a1 tree -> 'a1 -> 'a1 tree -> __ ->
      __ -> __ -> __ -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> __
      -> 'a2) -> ('a1 tree -> 'a1 -> 'a1 tree -> __ -> __ -> __ -> __ -> 'a1
      tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> __ -> __ -> 'a2) -> ('a1
      tree -> 'a1 -> 'a1 tree -> __ -> __ -> __ -> __ -> 'a1 tree -> 'a1 ->
      'a1 tree -> int -> __ -> __ -> __ -> 'a1 tree -> 'a1 -> 'a1 tree -> int
      -> __ -> 'a2) -> ('a1 tree -> 'a1 -> 'a1 tree -> __ -> __ -> __ -> __
      -> 'a2) -> 'a1 tree -> 'a1 -> 'a1 tree -> 'a1 tree -> 'a1 coq_R_bal ->
      'a2 **)
  
  let coq_R_bal_rec h f f0 f1 f2 f3 f4 f5 f6 f7 l x r t = function
  | R_bal_0 (x0, x1, x2) -> f x0 x1 x2 __ __ __
  | R_bal_1 (x0, x1, x2, x3, x4, x5, x6) ->
    f0 x0 x1 x2 __ __ x3 x4 x5 x6 __ __ __
  | R_bal_2 (x0, x1, x2, x3, x4, x5, x6) ->
    f1 x0 x1 x2 __ __ x3 x4 x5 x6 __ __ __ __
  | R_bal_3 (x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10) ->
    f2 x0 x1 x2 __ __ x3 x4 x5 x6 __ __ __ x7 x8 x9 x10 __
  | R_bal_4 (x0, x1, x2) -> f3 x0 x1 x2 __ __ __ __ __
  | R_bal_5 (x0, x1, x2, x3, x4, x5, x6) ->
    f4 x0 x1 x2 __ __ __ __ x3 x4 x5 x6 __ __ __
  | R_bal_6 (x0, x1, x2, x3, x4, x5, x6) ->
    f5 x0 x1 x2 __ __ __ __ x3 x4 x5 x6 __ __ __ __
  | R_bal_7 (x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10) ->
    f6 x0 x1 x2 __ __ __ __ x3 x4 x5 x6 __ __ __ x7 x8 x9 x10 __
  | R_bal_8 (x0, x1, x2) -> f7 x0 x1 x2 __ __ __ __
  
  type 'elt coq_R_add =
  | R_add_0 of 'elt tree
  | R_add_1 of 'elt tree * 'elt tree * 'elt * 'elt tree * int
  | R_add_2 of 'elt tree * 'elt tree * 'elt * 'elt tree * int * 'elt tree
     * 'elt coq_R_add
  | R_add_3 of 'elt tree * 'elt tree * 'elt * 'elt tree * int * 'elt tree
     * 'elt coq_R_add
  
  (** val coq_R_add_rect :
      'a1 coq_OrderedType -> 'a1 -> ('a1 tree -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a1 tree -> 'a1
      coq_R_add -> 'a2 -> 'a2) -> ('a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree ->
      int -> __ -> __ -> 'a1 tree -> 'a1 coq_R_add -> 'a2 -> 'a2) -> 'a1 tree
      -> 'a1 tree -> 'a1 coq_R_add -> 'a2 **)
  
  let rec coq_R_add_rect h x f f0 f1 f2 s t = function
  | R_add_0 s0 -> f s0 __
  | R_add_1 (s0, l, y, r0, h0) -> f0 s0 l y r0 h0 __ __
  | R_add_2 (s0, l, y, r0, h0, res, r1) ->
    f1 s0 l y r0 h0 __ __ res r1 (coq_R_add_rect h x f f0 f1 f2 l res r1)
  | R_add_3 (s0, l, y, r0, h0, res, r1) ->
    f2 s0 l y r0 h0 __ __ res r1 (coq_R_add_rect h x f f0 f1 f2 r0 res r1)
  
  (** val coq_R_add_rec :
      'a1 coq_OrderedType -> 'a1 -> ('a1 tree -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a1 tree -> 'a1
      coq_R_add -> 'a2 -> 'a2) -> ('a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree ->
      int -> __ -> __ -> 'a1 tree -> 'a1 coq_R_add -> 'a2 -> 'a2) -> 'a1 tree
      -> 'a1 tree -> 'a1 coq_R_add -> 'a2 **)
  
  let rec coq_R_add_rec h x f f0 f1 f2 s t = function
  | R_add_0 s0 -> f s0 __
  | R_add_1 (s0, l, y, r0, h0) -> f0 s0 l y r0 h0 __ __
  | R_add_2 (s0, l, y, r0, h0, res, r1) ->
    f1 s0 l y r0 h0 __ __ res r1 (coq_R_add_rec h x f f0 f1 f2 l res r1)
  | R_add_3 (s0, l, y, r0, h0, res, r1) ->
    f2 s0 l y r0 h0 __ __ res r1 (coq_R_add_rec h x f f0 f1 f2 r0 res r1)
  
  type 'elt coq_R_remove_min =
  | R_remove_min_0 of 'elt tree * 'elt * 'elt tree
  | R_remove_min_1 of 'elt tree * 'elt * 'elt tree * 'elt tree * 'elt
     * 'elt tree * int * ('elt tree * 'elt) * 'elt coq_R_remove_min
     * 'elt tree * 'elt
  
  (** val coq_R_remove_min_rect :
      'a1 coq_OrderedType -> ('a1 tree -> 'a1 -> 'a1 tree -> __ -> 'a2) ->
      ('a1 tree -> 'a1 -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int ->
      __ -> ('a1 tree * 'a1) -> 'a1 coq_R_remove_min -> 'a2 -> 'a1 tree ->
      'a1 -> __ -> 'a2) -> 'a1 tree -> 'a1 -> 'a1 tree -> ('a1 tree * 'a1) ->
      'a1 coq_R_remove_min -> 'a2 **)
  
  let rec coq_R_remove_min_rect h f f0 l x r p = function
  | R_remove_min_0 (l0, x0, r1) -> f l0 x0 r1 __
  | R_remove_min_1 (l0, x0, r1, ll, lx, lr, _x, res, r2, l', m) ->
    f0 l0 x0 r1 ll lx lr _x __ res r2
      (coq_R_remove_min_rect h f f0 ll lx lr res r2) l' m __
  
  (** val coq_R_remove_min_rec :
      'a1 coq_OrderedType -> ('a1 tree -> 'a1 -> 'a1 tree -> __ -> 'a2) ->
      ('a1 tree -> 'a1 -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int ->
      __ -> ('a1 tree * 'a1) -> 'a1 coq_R_remove_min -> 'a2 -> 'a1 tree ->
      'a1 -> __ -> 'a2) -> 'a1 tree -> 'a1 -> 'a1 tree -> ('a1 tree * 'a1) ->
      'a1 coq_R_remove_min -> 'a2 **)
  
  let rec coq_R_remove_min_rec h f f0 l x r p = function
  | R_remove_min_0 (l0, x0, r1) -> f l0 x0 r1 __
  | R_remove_min_1 (l0, x0, r1, ll, lx, lr, _x, res, r2, l', m) ->
    f0 l0 x0 r1 ll lx lr _x __ res r2
      (coq_R_remove_min_rec h f f0 ll lx lr res r2) l' m __
  
  type 'elt coq_R_merge =
  | R_merge_0 of 'elt tree * 'elt tree
  | R_merge_1 of 'elt tree * 'elt tree * 'elt tree * 'elt * 'elt tree * int
  | R_merge_2 of 'elt tree * 'elt tree * 'elt tree * 'elt * 'elt tree * 
     int * 'elt tree * 'elt * 'elt tree * int * 'elt tree * 'elt
  
  (** val coq_R_merge_rect :
      'a1 coq_OrderedType -> ('a1 tree -> 'a1 tree -> __ -> 'a2) -> ('a1 tree
      -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2)
      -> ('a1 tree -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> 'a1 -> __ ->
      'a2) -> 'a1 tree -> 'a1 tree -> 'a1 tree -> 'a1 coq_R_merge -> 'a2 **)
  
  let coq_R_merge_rect h f f0 f1 s1 s2 t = function
  | R_merge_0 (x, x0) -> f x x0 __
  | R_merge_1 (x, x0, x1, x2, x3, x4) -> f0 x x0 x1 x2 x3 x4 __ __
  | R_merge_2 (x, x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10) ->
    f1 x x0 x1 x2 x3 x4 __ x5 x6 x7 x8 __ x9 x10 __
  
  (** val coq_R_merge_rec :
      'a1 coq_OrderedType -> ('a1 tree -> 'a1 tree -> __ -> 'a2) -> ('a1 tree
      -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2)
      -> ('a1 tree -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> 'a1 -> __ ->
      'a2) -> 'a1 tree -> 'a1 tree -> 'a1 tree -> 'a1 coq_R_merge -> 'a2 **)
  
  let coq_R_merge_rec h f f0 f1 s1 s2 t = function
  | R_merge_0 (x, x0) -> f x x0 __
  | R_merge_1 (x, x0, x1, x2, x3, x4) -> f0 x x0 x1 x2 x3 x4 __ __
  | R_merge_2 (x, x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10) ->
    f1 x x0 x1 x2 x3 x4 __ x5 x6 x7 x8 __ x9 x10 __
  
  type 'elt coq_R_remove =
  | R_remove_0 of 'elt tree
  | R_remove_1 of 'elt tree * 'elt tree * 'elt * 'elt tree * int
  | R_remove_2 of 'elt tree * 'elt tree * 'elt * 'elt tree * int * 'elt tree
     * 'elt coq_R_remove
  | R_remove_3 of 'elt tree * 'elt tree * 'elt * 'elt tree * int * 'elt tree
     * 'elt coq_R_remove
  
  (** val coq_R_remove_rect :
      'a1 coq_OrderedType -> 'a1 -> ('a1 tree -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a1 tree -> 'a1
      coq_R_remove -> 'a2 -> 'a2) -> ('a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree
      -> int -> __ -> __ -> 'a1 tree -> 'a1 coq_R_remove -> 'a2 -> 'a2) ->
      'a1 tree -> 'a1 tree -> 'a1 coq_R_remove -> 'a2 **)
  
  let rec coq_R_remove_rect h x f f0 f1 f2 s t = function
  | R_remove_0 s0 -> f s0 __
  | R_remove_1 (s0, l, y, r0, _x) -> f0 s0 l y r0 _x __ __
  | R_remove_2 (s0, l, y, r0, _x, res, r1) ->
    f1 s0 l y r0 _x __ __ res r1 (coq_R_remove_rect h x f f0 f1 f2 l res r1)
  | R_remove_3 (s0, l, y, r0, _x, res, r1) ->
    f2 s0 l y r0 _x __ __ res r1 (coq_R_remove_rect h x f f0 f1 f2 r0 res r1)
  
  (** val coq_R_remove_rec :
      'a1 coq_OrderedType -> 'a1 -> ('a1 tree -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a1 tree -> 'a1
      coq_R_remove -> 'a2 -> 'a2) -> ('a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree
      -> int -> __ -> __ -> 'a1 tree -> 'a1 coq_R_remove -> 'a2 -> 'a2) ->
      'a1 tree -> 'a1 tree -> 'a1 coq_R_remove -> 'a2 **)
  
  let rec coq_R_remove_rec h x f f0 f1 f2 s t = function
  | R_remove_0 s0 -> f s0 __
  | R_remove_1 (s0, l, y, r0, _x) -> f0 s0 l y r0 _x __ __
  | R_remove_2 (s0, l, y, r0, _x, res, r1) ->
    f1 s0 l y r0 _x __ __ res r1 (coq_R_remove_rec h x f f0 f1 f2 l res r1)
  | R_remove_3 (s0, l, y, r0, _x, res, r1) ->
    f2 s0 l y r0 _x __ __ res r1 (coq_R_remove_rec h x f f0 f1 f2 r0 res r1)
  
  type 'elt coq_R_min_elt =
  | R_min_elt_0 of 'elt tree
  | R_min_elt_1 of 'elt tree * 'elt tree * 'elt * 'elt tree * int
  | R_min_elt_2 of 'elt tree * 'elt tree * 'elt * 'elt tree * int * 'elt tree
     * 'elt * 'elt tree * int * 'elt option * 'elt coq_R_min_elt
  
  (** val coq_R_min_elt_rect :
      'a1 coq_OrderedType -> ('a1 tree -> __ -> 'a2) -> ('a1 tree -> 'a1 tree
      -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2) -> ('a1 tree -> 'a1 tree
      -> 'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> 'a1 -> 'a1 tree -> int
      -> __ -> 'a1 option -> 'a1 coq_R_min_elt -> 'a2 -> 'a2) -> 'a1 tree ->
      'a1 option -> 'a1 coq_R_min_elt -> 'a2 **)
  
  let rec coq_R_min_elt_rect h f f0 f1 s o = function
  | R_min_elt_0 s0 -> f s0 __
  | R_min_elt_1 (s0, l, y, _x, _x0) -> f0 s0 l y _x _x0 __ __
  | R_min_elt_2 (s0, l, y, _x, _x0, _x1, _x2, _x3, _x4, res, r0) ->
    f1 s0 l y _x _x0 __ _x1 _x2 _x3 _x4 __ res r0
      (coq_R_min_elt_rect h f f0 f1 l res r0)
  
  (** val coq_R_min_elt_rec :
      'a1 coq_OrderedType -> ('a1 tree -> __ -> 'a2) -> ('a1 tree -> 'a1 tree
      -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2) -> ('a1 tree -> 'a1 tree
      -> 'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> 'a1 -> 'a1 tree -> int
      -> __ -> 'a1 option -> 'a1 coq_R_min_elt -> 'a2 -> 'a2) -> 'a1 tree ->
      'a1 option -> 'a1 coq_R_min_elt -> 'a2 **)
  
  let rec coq_R_min_elt_rec h f f0 f1 s o = function
  | R_min_elt_0 s0 -> f s0 __
  | R_min_elt_1 (s0, l, y, _x, _x0) -> f0 s0 l y _x _x0 __ __
  | R_min_elt_2 (s0, l, y, _x, _x0, _x1, _x2, _x3, _x4, res, r0) ->
    f1 s0 l y _x _x0 __ _x1 _x2 _x3 _x4 __ res r0
      (coq_R_min_elt_rec h f f0 f1 l res r0)
  
  type 'elt coq_R_max_elt =
  | R_max_elt_0 of 'elt tree
  | R_max_elt_1 of 'elt tree * 'elt tree * 'elt * 'elt tree * int
  | R_max_elt_2 of 'elt tree * 'elt tree * 'elt * 'elt tree * int * 'elt tree
     * 'elt * 'elt tree * int * 'elt option * 'elt coq_R_max_elt
  
  (** val coq_R_max_elt_rect :
      'a1 coq_OrderedType -> ('a1 tree -> __ -> 'a2) -> ('a1 tree -> 'a1 tree
      -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2) -> ('a1 tree -> 'a1 tree
      -> 'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> 'a1 -> 'a1 tree -> int
      -> __ -> 'a1 option -> 'a1 coq_R_max_elt -> 'a2 -> 'a2) -> 'a1 tree ->
      'a1 option -> 'a1 coq_R_max_elt -> 'a2 **)
  
  let rec coq_R_max_elt_rect h f f0 f1 s o = function
  | R_max_elt_0 s0 -> f s0 __
  | R_max_elt_1 (s0, _x, y, r0, _x0) -> f0 s0 _x y r0 _x0 __ __
  | R_max_elt_2 (s0, _x, y, r0, _x0, _x1, _x2, _x3, _x4, res, r1) ->
    f1 s0 _x y r0 _x0 __ _x1 _x2 _x3 _x4 __ res r1
      (coq_R_max_elt_rect h f f0 f1 r0 res r1)
  
  (** val coq_R_max_elt_rec :
      'a1 coq_OrderedType -> ('a1 tree -> __ -> 'a2) -> ('a1 tree -> 'a1 tree
      -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2) -> ('a1 tree -> 'a1 tree
      -> 'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> 'a1 -> 'a1 tree -> int
      -> __ -> 'a1 option -> 'a1 coq_R_max_elt -> 'a2 -> 'a2) -> 'a1 tree ->
      'a1 option -> 'a1 coq_R_max_elt -> 'a2 **)
  
  let rec coq_R_max_elt_rec h f f0 f1 s o = function
  | R_max_elt_0 s0 -> f s0 __
  | R_max_elt_1 (s0, _x, y, r0, _x0) -> f0 s0 _x y r0 _x0 __ __
  | R_max_elt_2 (s0, _x, y, r0, _x0, _x1, _x2, _x3, _x4, res, r1) ->
    f1 s0 _x y r0 _x0 __ _x1 _x2 _x3 _x4 __ res r1
      (coq_R_max_elt_rec h f f0 f1 r0 res r1)
  
  type 'elt coq_R_concat =
  | R_concat_0 of 'elt tree * 'elt tree
  | R_concat_1 of 'elt tree * 'elt tree * 'elt tree * 'elt * 'elt tree * int
  | R_concat_2 of 'elt tree * 'elt tree * 'elt tree * 'elt * 'elt tree * 
     int * 'elt tree * 'elt * 'elt tree * int * 'elt tree * 'elt
  
  (** val coq_R_concat_rect :
      'a1 coq_OrderedType -> ('a1 tree -> 'a1 tree -> __ -> 'a2) -> ('a1 tree
      -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2)
      -> ('a1 tree -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> 'a1 -> __ ->
      'a2) -> 'a1 tree -> 'a1 tree -> 'a1 tree -> 'a1 coq_R_concat -> 'a2 **)
  
  let coq_R_concat_rect h f f0 f1 s1 s2 t = function
  | R_concat_0 (x, x0) -> f x x0 __
  | R_concat_1 (x, x0, x1, x2, x3, x4) -> f0 x x0 x1 x2 x3 x4 __ __
  | R_concat_2 (x, x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10) ->
    f1 x x0 x1 x2 x3 x4 __ x5 x6 x7 x8 __ x9 x10 __
  
  (** val coq_R_concat_rec :
      'a1 coq_OrderedType -> ('a1 tree -> 'a1 tree -> __ -> 'a2) -> ('a1 tree
      -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2)
      -> ('a1 tree -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> 'a1 -> __ ->
      'a2) -> 'a1 tree -> 'a1 tree -> 'a1 tree -> 'a1 coq_R_concat -> 'a2 **)
  
  let coq_R_concat_rec h f f0 f1 s1 s2 t = function
  | R_concat_0 (x, x0) -> f x x0 __
  | R_concat_1 (x, x0, x1, x2, x3, x4) -> f0 x x0 x1 x2 x3 x4 __ __
  | R_concat_2 (x, x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10) ->
    f1 x x0 x1 x2 x3 x4 __ x5 x6 x7 x8 __ x9 x10 __
  
  type 'elt coq_R_split =
  | R_split_0 of 'elt tree
  | R_split_1 of 'elt tree * 'elt tree * 'elt * 'elt tree * int
  | R_split_2 of 'elt tree * 'elt tree * 'elt * 'elt tree * int * 'elt triple
     * 'elt coq_R_split * 'elt tree * bool * 'elt tree
  | R_split_3 of 'elt tree * 'elt tree * 'elt * 'elt tree * int * 'elt triple
     * 'elt coq_R_split * 'elt tree * bool * 'elt tree
  
  (** val coq_R_split_rect :
      'a1 coq_OrderedType -> 'a1 -> ('a1 tree -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a1 triple -> 'a1
      coq_R_split -> 'a2 -> 'a1 tree -> bool -> 'a1 tree -> __ -> 'a2) ->
      ('a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a1
      triple -> 'a1 coq_R_split -> 'a2 -> 'a1 tree -> bool -> 'a1 tree -> __
      -> 'a2) -> 'a1 tree -> 'a1 triple -> 'a1 coq_R_split -> 'a2 **)
  
  let rec coq_R_split_rect h x f f0 f1 f2 s t = function
  | R_split_0 s0 -> f s0 __
  | R_split_1 (s0, l, y, r0, _x) -> f0 s0 l y r0 _x __ __
  | R_split_2 (s0, l, y, r0, _x, res, r1, ll, b, rl) ->
    f1 s0 l y r0 _x __ __ res r1 (coq_R_split_rect h x f f0 f1 f2 l res r1)
      ll b rl __
  | R_split_3 (s0, l, y, r0, _x, res, r1, rl, b, rr) ->
    f2 s0 l y r0 _x __ __ res r1 (coq_R_split_rect h x f f0 f1 f2 r0 res r1)
      rl b rr __
  
  (** val coq_R_split_rec :
      'a1 coq_OrderedType -> 'a1 -> ('a1 tree -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2) -> ('a1 tree ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a1 triple -> 'a1
      coq_R_split -> 'a2 -> 'a1 tree -> bool -> 'a1 tree -> __ -> 'a2) ->
      ('a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a1
      triple -> 'a1 coq_R_split -> 'a2 -> 'a1 tree -> bool -> 'a1 tree -> __
      -> 'a2) -> 'a1 tree -> 'a1 triple -> 'a1 coq_R_split -> 'a2 **)
  
  let rec coq_R_split_rec h x f f0 f1 f2 s t = function
  | R_split_0 s0 -> f s0 __
  | R_split_1 (s0, l, y, r0, _x) -> f0 s0 l y r0 _x __ __
  | R_split_2 (s0, l, y, r0, _x, res, r1, ll, b, rl) ->
    f1 s0 l y r0 _x __ __ res r1 (coq_R_split_rec h x f f0 f1 f2 l res r1) ll
      b rl __
  | R_split_3 (s0, l, y, r0, _x, res, r1, rl, b, rr) ->
    f2 s0 l y r0 _x __ __ res r1 (coq_R_split_rec h x f f0 f1 f2 r0 res r1)
      rl b rr __
  
  type 'elt coq_R_inter =
  | R_inter_0 of 'elt tree * 'elt tree
  | R_inter_1 of 'elt tree * 'elt tree * 'elt tree * 'elt * 'elt tree * int
  | R_inter_2 of 'elt tree * 'elt tree * 'elt tree * 'elt * 'elt tree * 
     int * 'elt tree * 'elt * 'elt tree * int * 'elt tree * bool * 'elt tree
     * 'elt tree * 'elt coq_R_inter * 'elt tree * 'elt coq_R_inter
  | R_inter_3 of 'elt tree * 'elt tree * 'elt tree * 'elt * 'elt tree * 
     int * 'elt tree * 'elt * 'elt tree * int * 'elt tree * bool * 'elt tree
     * 'elt tree * 'elt coq_R_inter * 'elt tree * 'elt coq_R_inter
  
  (** val coq_R_inter_rect :
      'a1 coq_OrderedType -> ('a1 tree -> 'a1 tree -> __ -> 'a2) -> ('a1 tree
      -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2)
      -> ('a1 tree -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> bool -> 'a1
      tree -> __ -> __ -> 'a1 tree -> 'a1 coq_R_inter -> 'a2 -> 'a1 tree ->
      'a1 coq_R_inter -> 'a2 -> 'a2) -> ('a1 tree -> 'a1 tree -> 'a1 tree ->
      'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> 'a1 -> 'a1 tree -> int ->
      __ -> 'a1 tree -> bool -> 'a1 tree -> __ -> __ -> 'a1 tree -> 'a1
      coq_R_inter -> 'a2 -> 'a1 tree -> 'a1 coq_R_inter -> 'a2 -> 'a2) -> 'a1
      tree -> 'a1 tree -> 'a1 tree -> 'a1 coq_R_inter -> 'a2 **)
  
  let rec coq_R_inter_rect h f f0 f1 f2 s1 s2 t = function
  | R_inter_0 (s3, s4) -> f s3 s4 __
  | R_inter_1 (s3, s4, l1, x1, r1, _x) -> f0 s3 s4 l1 x1 r1 _x __ __
  | R_inter_2 (s3, s4, l1, x1, r1, _x, _x0, _x1, _x2, _x3, l2', pres, r2',
               res0, r0, res, r2) ->
    f1 s3 s4 l1 x1 r1 _x __ _x0 _x1 _x2 _x3 __ l2' pres r2' __ __ res0 r0
      (coq_R_inter_rect h f f0 f1 f2 l1 l2' res0 r0) res r2
      (coq_R_inter_rect h f f0 f1 f2 r1 r2' res r2)
  | R_inter_3 (s3, s4, l1, x1, r1, _x, _x0, _x1, _x2, _x3, l2', pres, r2',
               res0, r0, res, r2) ->
    f2 s3 s4 l1 x1 r1 _x __ _x0 _x1 _x2 _x3 __ l2' pres r2' __ __ res0 r0
      (coq_R_inter_rect h f f0 f1 f2 l1 l2' res0 r0) res r2
      (coq_R_inter_rect h f f0 f1 f2 r1 r2' res r2)
  
  (** val coq_R_inter_rec :
      'a1 coq_OrderedType -> ('a1 tree -> 'a1 tree -> __ -> 'a2) -> ('a1 tree
      -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2)
      -> ('a1 tree -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> bool -> 'a1
      tree -> __ -> __ -> 'a1 tree -> 'a1 coq_R_inter -> 'a2 -> 'a1 tree ->
      'a1 coq_R_inter -> 'a2 -> 'a2) -> ('a1 tree -> 'a1 tree -> 'a1 tree ->
      'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> 'a1 -> 'a1 tree -> int ->
      __ -> 'a1 tree -> bool -> 'a1 tree -> __ -> __ -> 'a1 tree -> 'a1
      coq_R_inter -> 'a2 -> 'a1 tree -> 'a1 coq_R_inter -> 'a2 -> 'a2) -> 'a1
      tree -> 'a1 tree -> 'a1 tree -> 'a1 coq_R_inter -> 'a2 **)
  
  let rec coq_R_inter_rec h f f0 f1 f2 s1 s2 t = function
  | R_inter_0 (s3, s4) -> f s3 s4 __
  | R_inter_1 (s3, s4, l1, x1, r1, _x) -> f0 s3 s4 l1 x1 r1 _x __ __
  | R_inter_2 (s3, s4, l1, x1, r1, _x, _x0, _x1, _x2, _x3, l2', pres, r2',
               res0, r0, res, r2) ->
    f1 s3 s4 l1 x1 r1 _x __ _x0 _x1 _x2 _x3 __ l2' pres r2' __ __ res0 r0
      (coq_R_inter_rec h f f0 f1 f2 l1 l2' res0 r0) res r2
      (coq_R_inter_rec h f f0 f1 f2 r1 r2' res r2)
  | R_inter_3 (s3, s4, l1, x1, r1, _x, _x0, _x1, _x2, _x3, l2', pres, r2',
               res0, r0, res, r2) ->
    f2 s3 s4 l1 x1 r1 _x __ _x0 _x1 _x2 _x3 __ l2' pres r2' __ __ res0 r0
      (coq_R_inter_rec h f f0 f1 f2 l1 l2' res0 r0) res r2
      (coq_R_inter_rec h f f0 f1 f2 r1 r2' res r2)
  
  type 'elt coq_R_diff =
  | R_diff_0 of 'elt tree * 'elt tree
  | R_diff_1 of 'elt tree * 'elt tree * 'elt tree * 'elt * 'elt tree * int
  | R_diff_2 of 'elt tree * 'elt tree * 'elt tree * 'elt * 'elt tree * 
     int * 'elt tree * 'elt * 'elt tree * int * 'elt tree * bool * 'elt tree
     * 'elt tree * 'elt coq_R_diff * 'elt tree * 'elt coq_R_diff
  | R_diff_3 of 'elt tree * 'elt tree * 'elt tree * 'elt * 'elt tree * 
     int * 'elt tree * 'elt * 'elt tree * int * 'elt tree * bool * 'elt tree
     * 'elt tree * 'elt coq_R_diff * 'elt tree * 'elt coq_R_diff
  
  (** val coq_R_diff_rect :
      'a1 coq_OrderedType -> ('a1 tree -> 'a1 tree -> __ -> 'a2) -> ('a1 tree
      -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2)
      -> ('a1 tree -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> bool -> 'a1
      tree -> __ -> __ -> 'a1 tree -> 'a1 coq_R_diff -> 'a2 -> 'a1 tree ->
      'a1 coq_R_diff -> 'a2 -> 'a2) -> ('a1 tree -> 'a1 tree -> 'a1 tree ->
      'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> 'a1 -> 'a1 tree -> int ->
      __ -> 'a1 tree -> bool -> 'a1 tree -> __ -> __ -> 'a1 tree -> 'a1
      coq_R_diff -> 'a2 -> 'a1 tree -> 'a1 coq_R_diff -> 'a2 -> 'a2) -> 'a1
      tree -> 'a1 tree -> 'a1 tree -> 'a1 coq_R_diff -> 'a2 **)
  
  let rec coq_R_diff_rect h f f0 f1 f2 s1 s2 t = function
  | R_diff_0 (s3, s4) -> f s3 s4 __
  | R_diff_1 (s3, s4, l1, x1, r1, _x) -> f0 s3 s4 l1 x1 r1 _x __ __
  | R_diff_2 (s3, s4, l1, x1, r1, _x, _x0, _x1, _x2, _x3, l2', pres, r2',
              res0, r0, res, r2) ->
    f1 s3 s4 l1 x1 r1 _x __ _x0 _x1 _x2 _x3 __ l2' pres r2' __ __ res0 r0
      (coq_R_diff_rect h f f0 f1 f2 l1 l2' res0 r0) res r2
      (coq_R_diff_rect h f f0 f1 f2 r1 r2' res r2)
  | R_diff_3 (s3, s4, l1, x1, r1, _x, _x0, _x1, _x2, _x3, l2', pres, r2',
              res0, r0, res, r2) ->
    f2 s3 s4 l1 x1 r1 _x __ _x0 _x1 _x2 _x3 __ l2' pres r2' __ __ res0 r0
      (coq_R_diff_rect h f f0 f1 f2 l1 l2' res0 r0) res r2
      (coq_R_diff_rect h f f0 f1 f2 r1 r2' res r2)
  
  (** val coq_R_diff_rec :
      'a1 coq_OrderedType -> ('a1 tree -> 'a1 tree -> __ -> 'a2) -> ('a1 tree
      -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2)
      -> ('a1 tree -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> bool -> 'a1
      tree -> __ -> __ -> 'a1 tree -> 'a1 coq_R_diff -> 'a2 -> 'a1 tree ->
      'a1 coq_R_diff -> 'a2 -> 'a2) -> ('a1 tree -> 'a1 tree -> 'a1 tree ->
      'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> 'a1 -> 'a1 tree -> int ->
      __ -> 'a1 tree -> bool -> 'a1 tree -> __ -> __ -> 'a1 tree -> 'a1
      coq_R_diff -> 'a2 -> 'a1 tree -> 'a1 coq_R_diff -> 'a2 -> 'a2) -> 'a1
      tree -> 'a1 tree -> 'a1 tree -> 'a1 coq_R_diff -> 'a2 **)
  
  let rec coq_R_diff_rec h f f0 f1 f2 s1 s2 t = function
  | R_diff_0 (s3, s4) -> f s3 s4 __
  | R_diff_1 (s3, s4, l1, x1, r1, _x) -> f0 s3 s4 l1 x1 r1 _x __ __
  | R_diff_2 (s3, s4, l1, x1, r1, _x, _x0, _x1, _x2, _x3, l2', pres, r2',
              res0, r0, res, r2) ->
    f1 s3 s4 l1 x1 r1 _x __ _x0 _x1 _x2 _x3 __ l2' pres r2' __ __ res0 r0
      (coq_R_diff_rec h f f0 f1 f2 l1 l2' res0 r0) res r2
      (coq_R_diff_rec h f f0 f1 f2 r1 r2' res r2)
  | R_diff_3 (s3, s4, l1, x1, r1, _x, _x0, _x1, _x2, _x3, l2', pres, r2',
              res0, r0, res, r2) ->
    f2 s3 s4 l1 x1 r1 _x __ _x0 _x1 _x2 _x3 __ l2' pres r2' __ __ res0 r0
      (coq_R_diff_rec h f f0 f1 f2 l1 l2' res0 r0) res r2
      (coq_R_diff_rec h f f0 f1 f2 r1 r2' res r2)
  
  type 'elt coq_R_union =
  | R_union_0 of 'elt tree * 'elt tree
  | R_union_1 of 'elt tree * 'elt tree * 'elt tree * 'elt * 'elt tree * int
  | R_union_2 of 'elt tree * 'elt tree * 'elt tree * 'elt * 'elt tree * 
     int * 'elt tree * 'elt * 'elt tree * int * 'elt tree * bool * 'elt tree
     * 'elt tree * 'elt coq_R_union * 'elt tree * 'elt coq_R_union
  
  (** val coq_R_union_rect :
      'a1 coq_OrderedType -> ('a1 tree -> 'a1 tree -> __ -> 'a2) -> ('a1 tree
      -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2)
      -> ('a1 tree -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> bool -> 'a1
      tree -> __ -> 'a1 tree -> 'a1 coq_R_union -> 'a2 -> 'a1 tree -> 'a1
      coq_R_union -> 'a2 -> 'a2) -> 'a1 tree -> 'a1 tree -> 'a1 tree -> 'a1
      coq_R_union -> 'a2 **)
  
  let rec coq_R_union_rect h f f0 f1 s1 s2 t = function
  | R_union_0 (s3, s4) -> f s3 s4 __
  | R_union_1 (s3, s4, l1, x1, r1, _x) -> f0 s3 s4 l1 x1 r1 _x __ __
  | R_union_2 (s3, s4, l1, x1, r1, _x, _x0, _x1, _x2, _x3, l2', _x4, r2',
               res0, r0, res, r2) ->
    f1 s3 s4 l1 x1 r1 _x __ _x0 _x1 _x2 _x3 __ l2' _x4 r2' __ res0 r0
      (coq_R_union_rect h f f0 f1 l1 l2' res0 r0) res r2
      (coq_R_union_rect h f f0 f1 r1 r2' res r2)
  
  (** val coq_R_union_rec :
      'a1 coq_OrderedType -> ('a1 tree -> 'a1 tree -> __ -> 'a2) -> ('a1 tree
      -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> __ -> 'a2)
      -> ('a1 tree -> 'a1 tree -> 'a1 tree -> 'a1 -> 'a1 tree -> int -> __ ->
      'a1 tree -> 'a1 -> 'a1 tree -> int -> __ -> 'a1 tree -> bool -> 'a1
      tree -> __ -> 'a1 tree -> 'a1 coq_R_union -> 'a2 -> 'a1 tree -> 'a1
      coq_R_union -> 'a2 -> 'a2) -> 'a1 tree -> 'a1 tree -> 'a1 tree -> 'a1
      coq_R_union -> 'a2 **)
  
  let rec coq_R_union_rec h f f0 f1 s1 s2 t = function
  | R_union_0 (s3, s4) -> f s3 s4 __
  | R_union_1 (s3, s4, l1, x1, r1, _x) -> f0 s3 s4 l1 x1 r1 _x __ __
  | R_union_2 (s3, s4, l1, x1, r1, _x, _x0, _x1, _x2, _x3, l2', _x4, r2',
               res0, r0, res, r2) ->
    f1 s3 s4 l1 x1 r1 _x __ _x0 _x1 _x2 _x3 __ l2' _x4 r2' __ res0 r0
      (coq_R_union_rec h f f0 f1 l1 l2' res0 r0) res r2
      (coq_R_union_rec h f f0 f1 r1 r2' res r2)
  
  (** val fold' :
      'a1 coq_OrderedType -> ('a1 -> 'a2 -> 'a2) -> 'a1 tree -> 'a2 -> 'a2 **)
  
  let fold' helt f s =
    SetList.SetList.fold helt f (elements helt s)
  
  (** val flatten_e : 'a1 coq_OrderedType -> 'a1 enumeration -> 'a1 list **)
  
  let rec flatten_e helt = function
  | End -> []
  | More (x, t, r) -> x :: (app (elements helt t) (flatten_e helt r))
 end

module S = SetAVL

type 'elt set =
  'elt S.tree
  (* singleton inductive, whose constructor was Bst *)

(** val this : 'a1 coq_OrderedType -> 'a1 set -> 'a1 S.tree **)

let this helt s =
  s

(** val mem : 'a1 coq_OrderedType -> 'a1 -> 'a1 set -> bool **)

let mem helt x s =
  S.mem helt x (this helt s)

(** val empty : 'a1 coq_OrderedType -> 'a1 set **)

let empty helt =
  S.empty helt

(** val is_empty : 'a1 coq_OrderedType -> 'a1 set -> bool **)

let is_empty helt s =
  S.is_empty helt (this helt s)

(** val singleton : 'a1 coq_OrderedType -> 'a1 -> 'a1 set **)

let singleton helt x =
  S.singleton helt x

(** val add : 'a1 coq_OrderedType -> 'a1 -> 'a1 set -> 'a1 set **)

let add helt x s =
  S.add helt x (this helt s)

(** val remove : 'a1 coq_OrderedType -> 'a1 -> 'a1 set -> 'a1 set **)

let remove helt x s =
  S.remove helt x (this helt s)

(** val inter : 'a1 coq_OrderedType -> 'a1 set -> 'a1 set -> 'a1 set **)

let inter helt s s' =
  S.inter helt (this helt s) (this helt s')

(** val union : 'a1 coq_OrderedType -> 'a1 set -> 'a1 set -> 'a1 set **)

let union helt s s' =
  S.union helt (this helt s) (this helt s')

(** val diff : 'a1 coq_OrderedType -> 'a1 set -> 'a1 set -> 'a1 set **)

let diff helt s s' =
  S.diff helt (this helt s) (this helt s')

(** val elements : 'a1 coq_OrderedType -> 'a1 set -> 'a1 list **)

let elements helt s =
  S.elements helt (this helt s)

(** val min_elt : 'a1 coq_OrderedType -> 'a1 set -> 'a1 option **)

let min_elt helt s =
  S.min_elt helt (this helt s)

(** val max_elt : 'a1 coq_OrderedType -> 'a1 set -> 'a1 option **)

let max_elt helt s =
  S.max_elt helt (this helt s)

(** val choose : 'a1 coq_OrderedType -> 'a1 set -> 'a1 option **)

let choose helt s =
  S.choose helt (this helt s)

(** val fold :
    'a1 coq_OrderedType -> ('a1 -> 'a2 -> 'a2) -> 'a1 set -> 'a2 -> 'a2 **)

let fold helt f s x =
  S.fold helt f (this helt s) x

(** val cardinal : 'a1 coq_OrderedType -> 'a1 set -> int **)

let cardinal helt s =
  S.cardinal helt (this helt s)

(** val filter :
    'a1 coq_OrderedType -> ('a1 -> bool) -> 'a1 set -> 'a1 set **)

let filter helt f s =
  S.filter helt f (this helt s)

(** val for_all : 'a1 coq_OrderedType -> ('a1 -> bool) -> 'a1 set -> bool **)

let for_all helt f s =
  S.for_all helt f (this helt s)

(** val exists_ : 'a1 coq_OrderedType -> ('a1 -> bool) -> 'a1 set -> bool **)

let exists_ helt f s =
  S.exists_ helt f (this helt s)

(** val partition :
    'a1 coq_OrderedType -> ('a1 -> bool) -> 'a1 set -> 'a1 set * 'a1 set **)

let partition helt f s =
  let p = S.partition helt f (this helt s) in ((fst p), (snd p))

(** val equal : 'a1 coq_OrderedType -> 'a1 set -> 'a1 set -> bool **)

let equal helt s s' =
  S.equal helt (this helt s) (this helt s')

(** val subset : 'a1 coq_OrderedType -> 'a1 set -> 'a1 set -> bool **)

let subset helt s s' =
  S.subset helt (this helt s) (this helt s')

(** val set_compare :
    'a1 coq_OrderedType -> 'a1 set -> 'a1 set -> comparison **)

let set_compare h s1 s2 =
  S.compare h (this h s1) (this h s2)

(** val set_OrderedType :
    'a1 coq_OrderedType -> 'a1 set coq_SpecificOrderedType **)

let set_OrderedType h =
  set_compare h

