open Datatypes
open DomAbsValue
open DomBasic
open MapInterface
open SetInterface

val local : idT -> varT

type literal =
| Coq_lit_int of int
| Coq_lit_int_top
| Coq_lit_top
| Coq_lit_bot

type ctype =
| Coq_typ_uint
| Coq_typ_uchar
| Coq_typ_ushort
| Coq_typ_ulong
| Coq_typ_ulonglong
| Coq_typ_int
| Coq_typ_char
| Coq_typ_short
| Coq_typ_long
| Coq_typ_longlong
| Coq_typ_struct of varT
| Coq_typ_union of varT
| Coq_typ_pointer of ctype
| Coq_typ_array of ctype * int
| Coq_typ_function of ctype
| Coq_typ_void
| Coq_typ_va_list
| Coq_typ_unknown

type exprT =
| Coq_expr_const of literal
| Coq_expr_op2 of op2 * exprT * exprT
| Coq_expr_op1 of op1 * exprT
| Coq_expr_value_of of lv
| Coq_expr_addr_of of lv
| Coq_expr_cast of ctype * exprT
| Coq_expr_size_of of exprT
and lv =
| Coq_lv_id of varT
| Coq_lv_deref of exprT
| Coq_lv_array of exprT * exprT
| Coq_lv_struct of lv * idT

type alloc =
| Coq_alloc_arr of exprT * exprT

type lcall =
| Coq_airac_print
| Coq_airac_dump
| Coq_rand
| Coq_strlen
| Coq_strcpy
| Coq_scanf
| Coq_errno_location
| Coq_builtin_alloca
| Coq_realloc
| Coq_builtin_choose_expr
| Coq_unknown

type instrT =
| Coq_instr_assign of lv * exprT
| Coq_instr_alloc of lv * alloc * allocT
| Coq_instr_assume of exprT
| Coq_instr_libcall of lv option * lcall * exprT list * allocT
| Coq_instr_forget of locT
| Coq_instr_skip

type tmnT =
| Coq_tmn_jump of bidT set
| Coq_tmn_call of bidT * fidT set * exprT list * varT option
| Coq_tmn_exit of exprT

type cmdT = instrT list * tmnT

val cinstrs : cmdT -> instrT list

val ctmn : cmdT -> tmnT

type funcT = (bidT * (bidT, cmdT) dict) * idT list

val fentry : funcT -> bidT

val fcmds : funcT -> (bidT, cmdT) dict

val fparams : funcT -> idT list

type progT = fidT * (fidT, funcT) dict

val pfuncs : progT -> (fidT, funcT) dict

