Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import List.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomSet.
Require Import DomMap.

Require Import Syntax.

Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.

Require Import SemAbsBlock.
Require Import SemAccSen.
Require Import SemAccInsen.

Require Import SemSenLocal.
Require Import SemInsenLocal.

Require Import TrsSen.

Set Implicit Arguments.

Section Accs.
  Variables (prog:progT) (acc:accT) (state:stateT).

  Definition tstate' := tstate prog state.

  Lemma tacc_correct_acclocs_expr expr fid mem
        (Hexpr : acc_correct_acclocs_expr acc fid expr mem) :
    SemAccSen.acc_correct_acclocs_expr acc fid expr mem

  with tacc_correct_acclocs_lv l fid mem
        (Hlv : acc_correct_acclocs_lv acc fid l mem) :
    SemAccSen.acc_correct_acclocs_lv acc fid l mem.
  Proof.
    rewrite/SemAccSen.acc_correct_acclocs_expr.
    elim: expr Hexpr=> [l|op l Hl r Hr|op l Hl|lv|lv|ct l Hl|l Hl] He mem'.
    { reflexivity. }
    { rewrite/= in He *.
      move/andP: He=> [Hle Hre].
      rewrite Hl; last done.
      rewrite Hr; last done.
      reflexivity.
    }
    { rewrite/= in He *.
      rewrite Hl; last done.
      reflexivity.
    }
    { rewrite/= in He.
      move/andP: He=> [He Hlv].
      exploit tacc_correct_acclocs_lv; [by eauto|move=> {He} He].
      rewrite/=.
      erewrite He.
      dom_tac.
      move: (sem_expr_lv _ _ _) {He} Hlv=> [locs|]; last done.
      dom_tac=> Hlocs.
      move: mem=> [mem|]; last reflexivity.
      dom_tac; rewrite/memT_hasJoin/withtop_hasJoin/withtop_join.
      apply: le_antisym; apply: memory_lookup_lub=> loc Hloc.
      { rewrite <- memory_lookup_ub; last by eauto.
        rewrite ? [mlookup _ (Elt _)]/mlookup/memT_maplike/withtop_maplike.
        rewrite map_join_spec.
        dom_tac.
        apply SetFacts.subset_iff in Hlocs.
        move: (Hlocs loc)=> {Hlocs} Hlocs.
        exploit Hlocs; first by apply SetFacts.mem_iff. 
        move=> {Hloc Hlocs} Hloc.
        apply SetFacts.mem_iff in Hloc.
        rewrite/smem/set_setlike Hloc /negb.
        rewrite join_bot.
        exact/DomType.eq_le/eq_refl.
      }
      { rewrite <- memory_lookup_ub; last by eauto.
        rewrite ? [mlookup _ (Elt _)]/mlookup/memT_maplike/withtop_maplike.
        rewrite map_join_spec.
        rewrite <- join_lhs.
        dom_tac.
        apply SetFacts.subset_iff in Hlocs.
        move: (Hlocs loc)=> {Hlocs} Hlocs.
        exploit Hlocs; first by apply SetFacts.mem_iff. 
        move=> {Hloc Hlocs} Hloc.
        apply SetFacts.mem_iff in Hloc.
        rewrite/smem/set_setlike Hloc.
        exact/DomType.eq_le/eq_refl.
      }
    }
    { rewrite/= in He.
      exploit tacc_correct_acclocs_lv; [by eauto|move=> {He} He].
      rewrite/=.
      do ? split; rewrite/fst/snd; try done.
      erewrite He.
      reflexivity.
    }
    { rewrite/=.
      rewrite Hl; last done.
      reflexivity.
    }
    { reflexivity. }

    rewrite/SemAccSen.acc_correct_acclocs_lv.
    elim: l Hlv=> [x|e|arr idx|e pos] Hlv mem'.
    { reflexivity. }
    { rewrite/= in Hlv *.
      exploit tacc_correct_acclocs_expr; [by eauto|move=> He].
      instantiate (1 := mem') in He.
      move: (sem_expr _ _ _) He=> v He.
      apply join_mor.
      by apply locs_of_mor.
      apply locs_from_arrvT_mor.
      by apply arrv_of_mor.
    }
    { rewrite/= in Hlv *.
      move/andP: Hlv=> [Harr _].
      exploit tacc_correct_acclocs_expr; [by eauto|move=> {Harr} Harr].
      instantiate (1 := mem') in Harr.
      move: (sem_expr _ _ _) Harr=> v Harr.
      apply locs_from_arrvT_mor.
      by apply arrv_of_mor.
    }
    { rewrite/=.
      assert (append_map_mor :
                forall pos,
                  Proper (eq ==> eq)
                         (fun locs =>
                            match locs with
                              | Elt locs' => Elt (DomSet.map (append_field pos) locs')
                              | Top => top
                            end)).
      { move=> pos' [x|] [y|] Hxy; rewrite/eq/= in Hxy *; try done.
        apply DomSet.map_mor; last done.
        reflexivity.
      }
      intro. apply append_map_mor.
      apply pos. apply mem'.
    }
  Qed.

  Lemma tacc_correct_acclocs_lv_opt opt_l fid mem
        (Hlv : acc_correct_acclocs_lv_opt acc fid opt_l mem) :
    SemAccSen.acc_correct_acclocs_lv_opt acc fid opt_l mem.
  Proof.
    case: opt_l Hlv => [opt|] Hlv /=;
    rewrite/acc_correct_acclocs_lv_opt in Hlv.
    { right; exists opt; split; first by done.
      by apply tacc_correct_acclocs_lv.
    }
    { left; done. }
  Qed.  

  Lemma tacc_correct_acclocs_exprs :
    forall fid exprs mem
           (Hexpr : acc_correct_acclocs_exprs
                      acc fid exprs mem),
      SemAccSen.acc_correct_acclocs_exprs
        acc fid exprs mem.
  Proof.
    intros.
    move: Hexpr.
    elim: exprs => [|expr exprs];
    rewrite/SemAccSen.acc_correct_acclocs_exprs
           /acc_correct_acclocs_exprs//.
    move=> IH Hexpr mem'.
    move/andP in Hexpr.
    inversion Hexpr as [Hexpr' Hexprs']; clear Hexpr.
    rewrite/sem_exprs/eq/eq_/list_setoid/DomBasic.list_eq.
    split.
    { by apply tacc_correct_acclocs_expr. }
    { by apply (IH Hexprs'). }
  Qed.
End Accs.
