Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomBridge.
Require Import DomSet.
Require Import DomMap.

Require Import Syntax.

Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomPidxSen.

Require Import SemAbsBlock.
Require Import SemAccSen.

Require Import SemSenLocal.
Require Import SemSen.
Require Import TrsLocal.
Require Import CorAcc.

Set Implicit Arguments.

Obligation Tactic := Simpless.tactic.

Section Postfixs.
  Variables (prog:progT) (acc:accT) (state:SemAccSen.stateT) (Hacc:acc_correct prog acc state).

  Lemma tstate_instrs
        fid func (Hfunc : MapsTo fid func (pfuncs prog))
        bid instrs tmn (Hcmd : MapsTo bid (instrs, tmn) (fcmds func))
        stack (Hc : pidx_consistent prog fid stack)
        (Hpostfix : SemSenLocal.postfix_instrs
                      acc fid instrs
                      state[(fid, bid, false, stack)]!
                      state[(fid, bid, true, stack)]!) :
    postfix_instrs
      acc fid instrs
      (tstate prog acc state)[(fid, bid, false, stack)]!
      (tstate prog acc state)[(fid, bid, true, stack)]!.
  Proof.
    rewrite ? tstate_lookup Hc.
    rewrite ? simpl_tmem.
    move: (match stack with | nil => _ | (_,_)::_ => _ end)=> prevmem.
    elim: {Hcmd} instrs (state[_]!) (state[_]!) Hpostfix=> [|istr istrs IH] mnew mold Hpostfix.
    { inv Hpostfix; constructor.
      apply join_le_mor; last exact/DomType.eq_le/eq_refl.
      by apply mrestrict_mor.
    }
    { inv Hpostfix; econstructor; last by eauto.
      exploit tacc_correct_acclocs_instr; eauto.
      move=> Hsem'.
      instantiate (1 := prevmem) in Hsem'.
      rewrite -> Hsem'.
      apply join_le_mor; last exact/DomType.eq_le/eq_refl.
      by apply mrestrict_mor.
    }
  Qed.

  Lemma tstate_jump
        fid func (Hfunc : MapsTo fid func (pfuncs prog))
        bid stack nbids
        (Hc : pidx_consistent prog fid stack = true)
        (Hpostfix : SemSenLocal.postfix_jump state fid func bid stack nbids) :
    postfix_jump (tstate prog acc state) fid func bid stack nbids.
  Proof.
    move: Hpostfix.
    refine (forall_implies _ _ _ _) => nbid.
    refine (forall_implies _ _ _ _) => Hnbid.
    rewrite ? tstate_lookup Hc ? simpl_tmem.
    move=> Htmn; rewrite <- Htmn.
    exact/DomType.eq_le/eq_refl.
  Qed.

  Lemma tstate_call
        fid func bid stack instrs nbid (nfid:fidT) (nfids:Set![fidT]) args retto
        (Hc : pidx_consistent prog fid stack)
        (Hfunc : MapsTo fid func (pfuncs prog))
        (Hcmd : MapsTo bid (instrs, tmn_call nbid nfids args retto) (fcmds func))
        (Hnfid: nfid !\in nfids)
        (Hpostfix : SemSenLocal.postfix_call prog acc state fid func bid stack nbid nfid args) :
    postfix_call prog acc (tstate prog acc state) fid func bid stack nbid nfid args.
  Proof.
    move: Hpostfix.
    refine (forall_implies _ _ _ _) => nfunc.
    refine (forall_implies _ _ _ _) => Hnfunc.
    move: (Hacc)=> [_ _ Hal _].
    move: {Hal} (Hal _ _ Hfunc)=> [Hal Halp].
    move: {Hal} (Hal _ _ stack Hcmd Hc)=> [Htmn Htmn'].
    move=> [Hout Hin].
    rewrite/= Hc.
    apply MF.find_mapsto_iff in Hfunc; rewrite Hfunc.
    apply MF.find_mapsto_iff in Hcmd; rewrite Hcmd.
    rewrite/= in Hnfid.
    rewrite/ctmn/snd Hnfid /andb.
    dom_tac.
    rewrite/acc_correct_acclocs_exprs in Htmn.
    rewrite {1}simpl_tmem.
    case Hpremem:
      (match stack with
         | nil => bot
         | (fid0, bid0) :: stack0 =>
           tmem acc state fid0 bid0 true stack0
       end)=> [premem|]; last first.
    { dom_tac.
      eapply le_trans; last by apply join_rhs.
      rewrite simpl_tmem /mrestrict_join Hpremem.
      dom_tac.
      rewrite Top_top => /=.
      rewrite Top_top.
      exact: top_prop.
    }
    rewrite -> Htmn.
    eapply le_trans; last first.
    { eapply mrestrict_join_mor; eauto.
      { by apply eq_refl. }
      { by apply DomType.eq_le; apply eq_refl. }
    }
    set params := (fparams nfunc).
    set argvals := ((sem_exprs fid args 
                               (state) [(fid, bid, true, stack) ]!)).
    assert (Hcaller :
              (∀ i : idT, List.In i params →
                           var_loc nfid (local i) nil
                                   !\in
                             set_to_powerset (acclocs acc) [fid ]!));
      first by intros i Hiparams; eapply acc_imply_params_in_caller; eauto.
    assert (Hcallee :
              (∀ i : idT, List.In i params →
                           var_loc nfid (local i) nil
                                   !\in
                             set_to_powerset (acclocs acc) [nfid ]!));
      first by intros i Hiparams; eapply acc_imply_params_in_callee; eauto.
    apply DomType.eq_le.
    rewrite bind_tmem_commutative; last done.
    unfold tmem_mem.
    rewrite <- (@mrestrict_join_subset _ _ _ _ _ _ _ _ _ _ memT_maplike _ _ _ (set_to_powerset (acclocs acc) [fid ]!) (set_to_powerset (acclocs acc) [nfid ]!));
      last by eapply acc_imply_call_subset_acclocs; eauto.
    rewrite (bind_mrestrict_join_commutative acc
               (set_to_powerset (acclocs acc) [fid ]!)); last done.
    rewrite mrestrict_join_bind_merge; last done.
    rewrite <- simpl_tmem.
    rewrite (mrestrict_join_imply_mrestrict (K:=locT) (E:=valueT) (M:=memT)).
    rewrite bind_mrestrict_commutative; last done.
    reflexivity.
  Qed.

  Lemma mem_le_spec' (m n:memT) (Hmn: le m n) (loc:locT) :
    le (A:=valueT) m[loc]! n[loc]!.
  Proof.
    move: n m Hmn=> [n|] [m|] Hmn.
    { rewrite/le/= in Hmn.
      rewrite/= ? map_mlookup_spec.
        by apply Hmn.
    }
    { inv Hmn. }
    { exact: top_prop. }
    { done. }
  Qed.
  
  Lemma tstate_exit
        ofid ofunc (Hofunc : MapsTo ofid ofunc (pfuncs prog))
        obid oinstrs nbid (fid:fidT) (fids:Set![fidT]) args retto
        (Hocmd : MapsTo obid (oinstrs, tmn_call nbid fids args retto) (fcmds ofunc))
        (Hfid: fid !\in fids)
        func (Hfunc : MapsTo fid func (pfuncs prog))
        bid instrs ret (Hcmd : MapsTo bid (instrs, tmn_exit ret) (fcmds func))
        ostack (Hc : pidx_consistent prog fid ((ofid,obid)::ostack))
        (Hcall : SemSenLocal.postfix_call prog acc state ofid ofunc obid ostack nbid fid args)
        (Hpostfix : SemSenLocal.postfix_exit prog acc state fid func bid ((ofid,obid)::ostack) ret) :
    postfix_exit prog acc (tstate prog acc state) fid func bid ((ofid,obid)::ostack) ret.
  Proof.
    move: {Hcall} (Hcall _ Hfunc)=> [_ Hout].
    move: {Hpostfix} (Hpostfix _ Hofunc _ Hocmd)=> Hpostfix.

    move: (Hc)=> Hco.
    rewrite simpl_pidx_consistent in Hco.
    move/andP: Hco=> [_ Hco].

    move: (Hacc)=> [_ _ Hal _].
    move: {Hal} (Hal) (Hal _ _ Hofunc)=> Hal' [Hal _].
    move: {Hal} (Hal _ _ _ Hocmd Hco)=> [_ Hret]; simpl in Hret.
    move: {Hal'} (Hal' _ _ Hfunc)=> [Hal' _].
    move: {Hal'} (Hal' _ _ _ Hcmd Hc)=> Hretval.
    rewrite/acc_correct_acclocs_blk/acc_correct_acclocs_tmn/ctmn/snd in Hretval.

    move=> ofunc' Hofunc'.
    assert (J := (MF.MapsTo_fun Hofunc' Hofunc)); subst.
    move=> ocmd' Hocmd'.
    assert (J := (MF.MapsTo_fun Hocmd' Hocmd)); subst.
    rewrite/ctmn/snd in Hpostfix *.

    move: Hretval Hout Hpostfix.
    rewrite ? tstate_lookup Hc Hco.
    rewrite ? simpl_tmem.
    rewrite/sem_bind_ret.

    move: (state[_]!) (state[_]!) (state[_]!)
                      (match ostack with
                         | nil => bot
                         | (fid0, bid0) :: stack =>
                           tmem acc state fid0 bid0 true stack
                       end) Hret=> mem1 mem2 mem3 mem4 Hret Hretval.

    rewrite/acc_correct_acclocs_expr in Hretval.
    
    assert (J : forall (loc':locT), ((loc' !\in (set_to_powerset (acclocs acc)[fid]!)) -> (loc' !\in (set_to_powerset (acclocs acc)[ofid]!)))).
    { move=> loc.
      rewrite ? set_to_powerset_spec.
      move=> Hin.
      move: Hacc=> [Hcf _ _ Hac].
      move: {Hcf} (Hcf _ _ Hofunc _ _ Hocmd)=> Hcf.
      rewrite/acc_correct_callfroms_blk/ctmn/snd in Hcf.
      move/set_forallP in Hcf.
      exploit Hcf; first by eauto.
      move=> {Hcf} Hcf.
      move: {Hcf} (Hac _ _ Hcf)=> Hcf.
      rewrite/fst in Hcf.
      exploit Hcf.
      { by apply SetFacts.mem_iff; eauto. }
      move=> Hin''.
      by apply SetFacts.mem_iff in Hin''.
    }
    move: Hret; dom_tac=> Hret.
    move: ((acclocs _)[_]!) ((acclocs _)[_]!) J Hret Hretval=> ofid_alocs fid_locs Hlocs Hret Hretval.

    move: mem1=> [mem1|]; last first.
    { move=> _ _.
      rewrite [X in le _ X]/mrestrict_join.
      eapply le_trans; last by apply join_lhs.
      exact: top_prop.
    }

    move: mem2 Hretval=> [mem2|] Hretval; last first.
    { move=> ? Hcont.
      rewrite/mrestrict/memT_restricting_maplike/withtop_restricting_maplike in Hcont.
      move: retto Hocmd Hret Hocmd' Hcont=> [l|] Hocmd Hret Hocmd' Hcont; last by case: (not_recursive _ _) in Hcont *.
      by case: (not_recursive _ _) in Hcont.
    }

    move: mem3 Hretval=> [mem3|] Hretval; last first.
    { move=> ? Hcont.
      move: retto Hocmd Hret Hocmd' Hcont=> [l|] Hocmd Hret Hocmd' Hcont; last by case: (not_recursive _ _) in Hcont *.
      by case: (not_recursive _ _) in Hcont.
    }

    move: mem4=> [mem4|]; last first.
    { move=> _ _.
      rewrite [X in le _ X]/mrestrict_join.
      eapply le_trans; last by apply join_rhs.
      exact: top_prop.
    }

    destruct retto as [retto|]; last first.
    { rewrite/=; dom_tac.
      rewrite/memT_hasJoin/withtop_hasJoin/withtop_join.
      rewrite ? if_Elt.
      rewrite/memT_poset/withtop_poset/withtop_le.
      move=> Hmem2 Hmem3.

      apply mem_le_spec; [done|done|move=> loc].
      move: (@mem_le_spec' _ _ Hmem3 loc); clear Hmem3.
      move: (Hmem2 loc); clear Hmem2.
      rewrite ? memory_update_spec.
      rewrite ? map_mlookup_spec ? map_mrestrict_spec.
      rewrite ? powerset_neg_spec.
      rewrite ?[(Elt _)[_]!]/mlookup/memT_maplike/withtop_maplike.

      case Hin: (_ !\in _); rewrite/negb.
      { move=> _.
        case Hrec: (not_recursive _ _).
        { rewrite map_join_spec ? map_mrestrict_spec.
          case Hin'': (_ !\in _); last by move=> _; exact: bot_prop.
          rewrite map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec Hin /negb.
          rewrite powerset_neg_spec Hlocs /negb; last done.
          by rewrite ? join_bot.
        }
        { rewrite map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec Hin /negb.
          rewrite ? map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec Hlocs /negb; last done.
          by rewrite ? join_bot.
        }
      }
      { move=> Hmem2.
        case Hrec: (not_recursive _ _).
        { rewrite map_join_spec ? map_mrestrict_spec.
          case Hin'': (_ !\in _); last by move=> _; exact: bot_prop.
          rewrite map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec Hin /negb.
          rewrite map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec.
          case Hin''': (_ !\in _); rewrite/negb.
          { by rewrite ?[join bot _]join_commutative ? join_bot. }
          { rewrite ?[join bot _]join_commutative ? join_bot.
              by move=> _; apply/DomType.eq_le/eq_refl.
          }
        }
        { rewrite map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec Hin /negb.
          rewrite ? map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec.
          case Hin'': (_ !\in _); rewrite/negb.
          { by rewrite ?[join bot _]join_commutative ? join_bot. }
          { rewrite ?[join bot _]join_commutative ? join_bot.
              by move=> _; apply/DomType.eq_le/eq_refl.
          }
        }
      }
    }

    rewrite/=; dom_tac.
    rewrite/memT_hasJoin/withtop_hasJoin/withtop_join.
    rewrite ? if_Elt.
    rewrite/memT_poset/withtop_poset/withtop_le.
    move=> Hmem2 Hmem3.
    apply mem_le_spec; [done|done|move=> loc].
    move: (@mem_le_spec' _ _ Hmem3 loc); clear Hmem3.
    move: (Hmem2 loc); clear Hmem2.
    rewrite ? memory_update_spec.
    rewrite ? map_mlookup_spec ? map_mrestrict_spec.
    rewrite ? powerset_neg_spec.
    rewrite [_ !\in SetInterface.singleton _]/smem/set_setlike SetFacts.singleton_b is_compare_eqb.
    rewrite ? SetProperties.singleton_cardinal eqb_refl; rewrite/andb.
    rewrite ?[(Elt _)[_]!]/mlookup/memT_maplike/withtop_maplike.

    set le := sem_expr _ _ _.
    set re := sem_expr _ _ _.
    assert (He : eq le re).
    { rewrite/le/re.
      rewrite/acc_correct_acclocs_expr in Hretval.
      rewrite <- Hretval.
      apply sem_expr_mor; try reflexivity.
      apply mem_eq_spec; [done|done|move=> loc'].
      rewrite/mrestrict_join.
      rewrite mem_join_spec.
      rewrite ? mem_mrestrict_spec.
      rewrite/mlookup/memT_maplike/withtop_maplike.
      rewrite powerset_neg_spec.
      rewrite map_join_spec map_mrestrict_spec.
      case Hin: (_ !\in _); rewrite/negb.
      { apply join_mor; first reflexivity.
        rewrite map_mrestrict_spec powerset_neg_spec Hin /negb.
        reflexivity.
      }
      { reflexivity. }
    }
    move: le re He=> le re He.
    
    case Hin: (_ !\in _); rewrite/negb.
    { move=> _.
      case Hin': (eqb _ _).
      { move/eqbP in Hin'; usual_tac Hin'.
        case Hstrong: (is_strong _ _).
        { rewrite map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec Hlocs; rewrite/negb; last done.
          by rewrite He join_bot.
        }
        { case Hrec: (not_recursive _ _).
          { rewrite ? map_mrestrict_spec.
            case Hin'': (_ !\in _).
            { rewrite map_join_spec ? map_mrestrict_spec.
              rewrite powerset_neg_spec Hin; rewrite/negb.
              rewrite map_join_spec ? map_mrestrict_spec.
              rewrite powerset_neg_spec Hlocs; rewrite/negb; last done.
              by rewrite He ? join_bot.
            }
            { rewrite map_join_spec ? map_mrestrict_spec.
              rewrite powerset_neg_spec Hlocs; rewrite/negb; last done.
              by rewrite He ?[join bot _]join_commutative ? join_bot.
            }
          }
          { rewrite map_join_spec ? map_mrestrict_spec.
            rewrite powerset_neg_spec Hin; rewrite/negb.
            rewrite map_join_spec ? map_mrestrict_spec.
            rewrite powerset_neg_spec Hlocs; rewrite/negb; last done.
            by rewrite He ? join_bot.
          }
        }
      }
      { case Hrec: (not_recursive _ _).
        { rewrite ? map_join_spec ? map_mrestrict_spec.
          case Hin'': (_ !\in _); last by move=> _; exact: bot_prop.
          rewrite map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec Hin /negb.
          rewrite powerset_neg_spec Hlocs; rewrite/negb; last done.
          by rewrite ? join_bot.
        }
        { rewrite map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec Hin /negb.
          rewrite map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec Hlocs; rewrite/negb; last done.
          by rewrite ? join_bot.
        }
      }
    }
    { move=> Hmem2.
      rewrite set_smem_spec in Hret.
      case Hin': (eqb _ _).
      { move/eqbP in Hin'; usual_tac Hin'.
        case Hstrong: (is_strong _ _).
        { rewrite map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec set_to_powerset_spec Hret /negb.
          by rewrite He join_bot.
        }
        { case Hrec: (not_recursive _ _).
          { rewrite map_join_spec ? map_mrestrict_spec.
            case Hin'': (_ !\in _).
            { rewrite map_join_spec ? map_mrestrict_spec.
              rewrite powerset_neg_spec Hin /negb.
              rewrite powerset_neg_spec set_to_powerset_spec Hret /negb.
              rewrite map_join_spec ? map_mrestrict_spec.
              rewrite powerset_neg_spec set_to_powerset_spec Hret /negb.
              move=> Hle.
              rewrite ?[join bot _]join_commutative ? join_bot.
              apply join_lub; first done.
              eapply le_trans; last by apply Hle.
              rewrite <- He.
              by apply join_rhs.
            }
            { rewrite powerset_neg_spec set_to_powerset_spec Hret /negb.
              rewrite ?[join bot _]join_commutative ? join_bot.
              by rewrite He.
            }
          }
          { rewrite map_join_spec ? map_mrestrict_spec.
            rewrite powerset_neg_spec Hin /negb.
            rewrite map_join_spec ? map_mrestrict_spec.
            rewrite powerset_neg_spec set_to_powerset_spec Hret /negb.
            rewrite map_join_spec ? map_mrestrict_spec.
            rewrite powerset_neg_spec set_to_powerset_spec Hret /negb.
            rewrite ?[join bot _]join_commutative ? join_bot.
            move=> Hle.
            rewrite ?[join bot _]join_commutative ? join_bot.
            apply join_lub; first done.
            eapply le_trans; last by apply Hle.
            rewrite <- He.
            by apply join_rhs.
          }
        }
      }
      { case Hrec: (not_recursive _ _).
        { rewrite map_join_spec ? map_mrestrict_spec.
          case Hin'': (_ !\in _); last by move=> _; exact: bot_prop.
          rewrite map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec Hin /negb.
          rewrite map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec.
          case Hin''': (_ !\in _); rewrite/negb.
          { by rewrite ?[join bot _]join_commutative ? join_bot. }
          { rewrite ?[join bot _]join_commutative ? join_bot.
            by move=> _; apply/DomType.eq_le/eq_refl.
          }
        }
        { rewrite map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec Hin /negb.
          rewrite ? map_join_spec ? map_mrestrict_spec.
          rewrite powerset_neg_spec.
          case Hin'': (_ !\in _); rewrite/negb.
          { by rewrite ?[join bot _]join_commutative ? join_bot. }
          { rewrite ?[join bot _]join_commutative ? join_bot.
            by move=> _; apply/DomType.eq_le/eq_refl.
          }
        }
      }
    }
  Qed.

  Lemma tstate_correct (Hpostfix : SemSenLocal.postfix prog acc state) :
    postfix prog acc (tstate prog acc state).
  Proof.
    move: (Hpostfix).
    refine (forall_implies _ _ _ _) => fid.
    refine (forall_implies _ _ _ _) => func.
    refine (forall_implies _ _ _ _) => Hfunc.
    refine (forall_implies _ _ _ _) => bid.
    refine (forall_implies _ _ _ _) => stack.
    refine (forall_implies _ _ _ _); move=> [instrs tmn].
    refine (forall_implies _ _ _ _) => Hcmd.
    case Hc: (pidx_consistent prog fid stack);
      last by rewrite/tstate/= Hc; split; apply: eq_refl.
    move=> [Hinstrs Htmn]; split.
    { by eapply tstate_instrs; eauto. }
    { case: tmn Hcmd Htmn=> [nbids|nbid nfids args retto|ret] Hcmd {Hinstrs}.
      { by eapply tstate_jump; eauto. }
      { refine (forall_implies _ _ _ _) => nfid.
        refine (forall_implies _ _ _ _) => Hnfid.
        by eapply tstate_call; eauto.
      }
      { case: stack Hc=> [|[ofid obid] ostack] Hc; first done.
        move: (Hc).
        rewrite simpl_pidx_consistent.
        move/andP=> [Hco Hcostack].
        case Hofunc: ((pfuncs prog)[ofid])=> [ofunc|]; rewrite Hofunc in Hco; last done.
        apply MF.find_mapsto_iff in Hofunc.
        case Hocmd: ((fcmds ofunc)[obid])=> [ocmd|]; rewrite Hocmd in Hco; last done.
        apply MF.find_mapsto_iff in Hocmd.
        move: ocmd Hocmd Hco=> [oinstrs [|nbid nfid args retto|]] Hocmd Hco; try done.
        rewrite/ctmn/snd in Hco.
        eapply tstate_exit; eauto.
        move: {Hpostfix} (Hpostfix _ _ Hofunc)=> Hpostfix.
        move: {Hpostfix} (Hpostfix _ ostack _ Hocmd)=> Hpostfix.
        rewrite Hcostack in Hpostfix.
        move: Hpostfix=> [_ Hpostfix].
        exact: Hpostfix.
      }
    }
  Qed.
End Postfixs.
